﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using SimpleGenerate.Properties;
using System.Xml;

namespace SimpleGenerate
{
    public partial class generateForm : Form
    {
        Dictionary<string, Int64[]> maps = new Dictionary<string, Int64[]>();

        class Enenmy
        {
            public Enenmy(string iname, string ix, string iy, string itype, string[] icoordinates, int iminDistance, int ienemySpeed, int ibulletSpeed, int ibulletLifeTime, int ibulletShotCounterMax)
            {
                name = iname;
                x = ix;
                y = iy;
                type = itype;
                coordinates = icoordinates;
                minDistance = iminDistance;
                enemySpeed = ienemySpeed;
                bulletSpeed = ibulletSpeed;
                bulletLifeTime = ibulletLifeTime;
                bulletShotCounterMax = ibulletShotCounterMax;
            }

            public List<way> ways = null;
            public string name;
            public string x;
            public string y;
            public string type;
            public string []coordinates;
            public int minDistance;
            public int enemySpeed;
            public int bulletSpeed;
            public int bulletLifeTime;
            public int bulletShotCounterMax;

        }

        int w16 = 20;
        int h16 = 19;

        List<Enenmy> enemys = new List<Enenmy>();

        public generateForm()
        {
            InitializeComponent();
        }

        

        void print(string txt)
        {
            generateBox.AppendText(txt);
        }

        private void clearToolStripMenuItem_Click(object sender, EventArgs e)
        {
            generateBox.Clear();
        }

        private void sourceFileSelect_Click(object sender, EventArgs e)
        {
            try
            {
                OpenFileDialog openFileDialog = new OpenFileDialog();

                if (TmxSourceFile.Text.Length != 0)
                {
                    openFileDialog.FileName = TmxSourceFile.Text;
                    openFileDialog.InitialDirectory = Path.GetDirectoryName(TmxSourceFile.Text);
                }
                else
                {
                    openFileDialog.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                }

                openFileDialog.Filter = "TMX file (*.tmx;*.TMX)|*.tmx;*.TMX|All files (*.*)|*.*";
                openFileDialog.FilterIndex = 2;
                openFileDialog.RestoreDirectory = true;

                if (openFileDialog.ShowDialog() == DialogResult.OK)
                {
                    string fileName = openFileDialog.FileName;
                    TmxSourceFile.Text = fileName;

                }
            }
            catch (Exception)
            {
                print("Error while selecting/open file or dir\n");
            }
        }

        private void destinationSelectButton_Click(object sender, EventArgs e)
        {
            try
            {
                
                FolderBrowserDialog saveFileDialog = new FolderBrowserDialog();

                if (DestFilesPath.Text.Length != 0)
                {
                    saveFileDialog.SelectedPath = Path.GetDirectoryName(DestFilesPath.Text);
                }
                else
                {
                    saveFileDialog.SelectedPath = System.IO.Directory.GetCurrentDirectory();
                }

            /*    saveFileDialog.Filter = "All files (*.*)|*.*";
                saveFileDialog.FilterIndex = 1;
                saveFileDialog.RestoreDirectory = true;*/

                if (saveFileDialog.ShowDialog() == DialogResult.OK)
                {
                    string fileName = saveFileDialog.SelectedPath;
                    DestFilesPath.Text = fileName;
                }
            }
            catch (Exception)
            {
                print("Error while selecting/open file or dir\n");
            }
        }

        private void generateForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            Settings.Default.Save();
        }

        private void generateForm_Load(object sender, EventArgs e)
        {
            Settings.Default.Reload();
        }



        private void buttonGenerate_Click(object sender, EventArgs e)
        {
            maps.Clear();
            enemys.Clear();

            // READ XML

            XmlTextReader xmlReader = new XmlTextReader(TmxSourceFile.Text);
            XmlDocument xmlDocument = new XmlDocument();
           
            try
            {
                xmlDocument.Load(xmlReader);
                XmlElement root = xmlDocument["map"];

                print("Reading TMX ....\n");

                if (root != null)
                {
                    foreach (XmlElement elem in root)
                    {
                        if (elem.Name == "tileset")
                        {
                            XmlAttributeCollection elemAttr = elem.Attributes;
                            if (elemAttr != null)
                            {
                                string tilesetName = elemAttr["name"].Value;
                                if (tilesetName == "tileset")   // tileset for the normal map
                                {
                                    foreach (XmlElement subElem in elem)
                                    {                                        
                                        if (subElem.Name == "image")    // has image ?
                                        {
                                            XmlAttributeCollection subElemAttr = subElem.Attributes;
                                            string swidth = subElemAttr["width"].Value;
                                            string sheight = subElemAttr["height"].Value;

                                            int width = Convert.ToInt32(swidth);
                                            int height = Convert.ToInt32(sheight);

                                            w16 = (width - 1) / 17;
                                            h16 = (height - 1) / 17;

                                            print("Normal Tileset has a size of "+width+"x"+height+" -> "+w16+"x"+h16+" Tiles\n");
                                        }
                                    }
                                }
                            }
                        }
                        else if (elem.Name == "layer")       // layer
                        {
                            XmlAttributeCollection elemAttr = elem.Attributes;
                            if (elemAttr != null)
                            {

                                string layerName = elemAttr["name"].Value;
                                print("Layer name = " + layerName + "");

                                XmlElement data = elem["data"];
                                XmlAttributeCollection dataAttr = data.Attributes;
                                string encoding = dataAttr["encoding"].Value;
                                print("  Data encoding = " + encoding + "\n");
                                if (encoding == "csv")
                                {
                                    string[] words = data.InnerText.Split(',');
                                    Int64[] wordsInt = new Int64[words.Length];
                                    print(" -> Read/Convert " + wordsInt.Length + " Data \n");
                                    for (int i = 0; i < words.Length; i++)
                                    {
                                        wordsInt[i] = Convert.ToInt64(words[i]);
                                    }

                                    maps[layerName] = wordsInt;
                                }
                            }
                        }
                        else if (elem.Name == "objectgroup")    // objectgroup
                        { 
                            XmlAttributeCollection elemAttr = elem.Attributes;
                            if (elemAttr != null)
                            {
                                string objectGroupName = elemAttr["name"].Value;

                                print("Object group name = " + objectGroupName + "\n");

                                foreach (XmlElement subElem in elem)
                                {
                                    if(subElem.Name == "object")
                                    {
                                        XmlAttributeCollection subElemAttr = subElem.Attributes;

                                        string pathName = subElemAttr["name"].Value;
                                        string pathX = subElemAttr["x"].Value;
                                        string pathY = subElemAttr["y"].Value;                                       
                                        string prTypeValue = "";

                                        // defaults

                                        int minDistance = 120;
                                        int enemySpeed = 1;
                                        int bulletSpeed = 2;
                                        int bulletLifeTime = 30 * bulletSpeed;
                                        int bulletShotCounterMax = 30;
                                        string[] coords = null;

                                        print(" -> Sub object '" + pathName + "' at (" + pathX + ";" + pathY + ")\n");

                                        foreach (XmlElement sub2Elem in subElem)
                                        {
                                            XmlAttributeCollection sub2ElemAttr = sub2Elem.Attributes;

                                            if (sub2Elem.Name == "properties")
                                            {
                                                foreach (XmlElement sub3Elem in sub2Elem)
                                                {
                                                    if (sub3Elem.Name == "property")
                                                    {
                                                        XmlAttributeCollection sub3ElemAttr = sub3Elem.Attributes;

                                                        string prType = sub3ElemAttr["name"].Value;
                                                        string prValue = sub3ElemAttr["value"].Value;

                                                        if(prType == "type")
                                                        {
                                                            prTypeValue = prValue;
                                                        }
                                                        if(prType == "minDistance")
                                                        {
                                                            minDistance = Convert.ToInt32(prValue);
                                                        }
                                                        if (prType == "enemySpeed")
                                                        {
                                                            enemySpeed = Convert.ToInt32(prValue);
                                                        }
                                                        if (prType == "bulletSpeed")
                                                        {
                                                            bulletSpeed = Convert.ToInt32(prValue);
                                                        }
                                                        if (prType == "bulletLifeTime")
                                                        {
                                                            bulletLifeTime = Convert.ToInt32(prValue);
                                                        }
                                                        if (prType == "bulletShotCounterMax")
                                                        {
                                                            bulletShotCounterMax = Convert.ToInt32(prValue);
                                                        }
                                                       
                                                        print("  -> " + prType + " = " + prValue + "\n");
                                                    }
                                                }
                                                
                                            }
                                            else if(sub2Elem.Name == "polyline")
                                            {
                                                coords = sub2ElemAttr["points"].InnerText.Split(' ');                                             
                                                print(" -> Read/Convert " + coords.Length + " Coordinates \n");
                                            }
                                        }

                                        Enenmy enemy = new Enenmy(pathName, pathX, pathY, prTypeValue, coords, minDistance, enemySpeed, bulletSpeed, bulletLifeTime, bulletShotCounterMax);
                                        enemys.Add(enemy);
                                    }
                                   
 
                                }

                              //  XmlElement xobject = elem["object"];
                              //  XmlAttributeCollection xobjectAttr = xobject.Attributes; ;

                            }
                        }

                    }

                }

                print("TMX Reading finish\n");
            }
            catch (Exception ex)
            {
                print("Error: " + ex.Message + "\n");
            }

            // store

            storeMapBin("Background", 1);
            storeMapBin("Ctrl", (20 * 19) + 1);
            storeAnims("Background", "Anim2", "Anim3", "Anim4", "Anim5", "Anim6", "Anim7", "Anim8", "Anim9");
            storeTileLut();
            storeEnemys();

            xmlReader.Close();
        }

        void storeEnemys()
        {
            print("Store enemys\n");

            StringBuilder sb = new StringBuilder();
            // run one calculate ways

            foreach (Enenmy e in enemys)
            {
                string[] coordinatesString = e.coordinates;
                int x = Convert.ToInt32(e.x);
                int y = Convert.ToInt32(e.y);

                Point[] points = new Point[coordinatesString.Length];

                for (int i = 0; i < coordinatesString.Length; i++)
                {
                    string []coordinateXYString = coordinatesString[i].Split(',');
                    int px = Convert.ToInt32(coordinateXYString[0]);
                    int py = Convert.ToInt32(coordinateXYString[1]);

                    points[i] = new Point(x + px, y + py);
                }

                e.ways = calculateWays(points);
            }

            // generate offsets
            int wCount = 0;     
            sb.Append("\tword\t" + enemys.Count + "\t;enemy count\n");wCount += 1;

            int offset = 0;
            int waysToDo = enemys.Count;

            foreach (Enenmy e in enemys)
            {
                int wayCount = e.ways.Count;

                sb.Append("\tword\t" + (offset + waysToDo + 1) + "\t;enemy offset\n"); wCount += 1;

                offset += (wayCount * 6) +20;
                //waysToDo--;
            }

                 
            foreach (Enenmy e in enemys)
            {
                sb.Append("\n\t;enemy data (" + wCount + ")\n\n");
                List<way> eWays = e.ways;

                sb.Append("\tword\t0\t;x\n"); wCount += 1;
                sb.Append("\tword\t0\t;y\n"); wCount += 1;
                sb.Append("\tword\t0\t;sourl\n"); wCount += 1;
                sb.Append("\tword\t0\t;direction\n"); wCount += 1;
                sb.Append("\tword\t0\t;moving\n"); wCount += 1;
                sb.Append("\tword\t0\t;movecounter\n"); wCount += 1;
                sb.Append("\tword\t0\t;anim delay\n"); wCount += 1;
                sb.Append("\tword\t0\t;shot delay\n"); wCount += 1;
                sb.Append("\tword\t0\t;state\n"); wCount += 1;
                sb.Append("\tword\t0\t;current way\n"); wCount += 1;
                sb.Append("\tword\t"+e.ways.Count+"\t;max ways\n"); wCount += 1;
                sb.Append("\tword\t0\t;pos in way\n"); wCount += 1;
                sb.Append("\tword\t" + e.type + "\t;enemy type\n"); wCount += 1;
                sb.Append("\tword\t0\t;direction to player\n"); wCount += 1;
                sb.Append("\tword\t0\t;distance to player\n"); wCount += 1;
                sb.Append("\tword\t" + e.minDistance + "\t;min distance to player\n"); wCount += 1;
                sb.Append("\tword\t" + e.enemySpeed + "\t;enemy speed\n"); wCount += 1;
                sb.Append("\tword\t" + e.bulletSpeed + "\t;bullet speed\n"); wCount += 1;
                sb.Append("\tword\t" + e.bulletLifeTime + "\t;bullet lifetime\n"); wCount += 1;
                sb.Append("\tword\t" + e.bulletShotCounterMax + "\t;bullet shoot counter max 60/n\n"); wCount += 1;
                
                foreach (way w in eWays)
                {
                    sb.Append(string.Format("\t.data\t{0},\t{1},\t{2},\t${3:X4},${4:X4},\t{5}\t;x,y,length,sx,sy,direction\n", w.x, w.y, (int)w.length, w.sx, w.sy, w.direction)); wCount += 6;
                }

                sb.Append("\n");
            }

            int dataCount = wCount;

            int fixedMaxLength = 1024;
            sb.Append("\t.space\t" + (fixedMaxLength - dataCount) + "\t;valid data is " + dataCount + " of " + fixedMaxLength + " \n");

            TextWriter tw = new StreamWriter(DestFilesPath.Text + "/enemys.s");
            tw.Write(sb.ToString());
            tw.Close();

        }

        private List<way> calculateWays(Point[] points)
        {
            // calc

            List<way> ways = new List<way>();

            int lastPoints = points.Length - 1;

          //  print(string.Format("\t.data\t0,\t{0}\t;current, last way\n", lastPoints));

            for (int i = 0; i <= lastPoints; i += 1)
            {

                int p1 = i;
                int p2 = (i == lastPoints) ? 0 : (i + 1);

                Point from = points[p1];
                Point to = points[p2];

                float dx = ((float)to.X) - ((float)from.X);
                float dy = ((float)to.Y) - ((float)from.Y);

                float length = (float)Math.Sqrt((double)(dx * dx + dy * dy));

              /*  if (length <=2.5f)
                {
                    continue;
                }*/

                float sx = (dx) / length;
                float sy = (dy) / length;

                Int16 isx = (Int16)(sx * 256);
                Int16 isy = (Int16)(sy * 256);

                // calc direction




                double angle = Math.Atan2(dy, dx) * 180.0 / Math.PI;
                double ding = Math.Floor((angle + 90.0 + 22.5) / 45.0);

                int iding = (int)(ding);

                if (iding < 0) iding = 8 + iding;
               // if (iding == -1) iding = 7;
                // print(angle + " ===" + (ding) + "===\n"); ;



                ways.Add(new way(from.X, from.Y, length, isx, isy, iding));


               // print(string.Format("\t.data\t{0},\t{1},\t{2},\t${3:X4},${4:X4},0,\t{5}\t;x,y,length,sx,sy,n,direction\n", from.X, from.Y, (int)length, isx, isy, iding));
            }

            return ways;
        }

        class way
        {
            public way(float ix, float iy, float ilength, Int16 isx, Int16 isy, int idirection)
            {
                x = ix;
                y = iy;
                length = ilength;
                sx = isx;
                sy = isy;
                n = 0;
                direction = idirection;
            }

            public float x;
            public float y;
            public float length;
            public Int16 sx;
            public Int16 sy;
            public int direction;

            // tmp

            public float n;
        }

       

        void storeMap(string name,Int64 subtract)
        {
            print("Store map = '" + name +"'\n");
            StringBuilder sb = new StringBuilder();

            sb.Append("\t;map name = " + name + "\n");

            if (maps.ContainsKey(name))
            {
                Int64 []mapData = maps[name];

                int div = 256;
                for (int i = 0; i < mapData.Length; i++)
                {
                    Int64 d = (uint)mapData[i];

                    if ((i % div) == 0)
                    {
                        sb.Append("\n\t.data\t");
                    }

                    if(d != 0)
                    {
                        d = d - subtract;
                    }                   

                    d = (d & 0xffff) | (d >> 16);

                    sb.Append(string.Format("${0:X4}{1}", d, (i % div) == (div - 1) ? " " : ","));
                }
            }

            TextWriter tw = new StreamWriter(DestFilesPath.Text + "/" + name + ".map.s");
            tw.Write(sb.ToString());
            tw.Close();
        }

        void storeMapBin(string name, Int64 subtract)
        {
            FileStream fs = File.Create(DestFilesPath.Text + "/" + name + ".map.raw");
            BinaryWriter bw = new BinaryWriter(fs);

            if (maps.ContainsKey(name))
            {
                Int64[] mapData = maps[name];

                for (int i = 0; i < mapData.Length; i++)
                {
                    Int64 d = (uint)mapData[i];

                    if (d != 0)
                    {
                        d = d - subtract;
                    }

                    d = (d & 0xffff) | (d >> 16);

                    UInt16 du16 = (UInt16) ((d<<8) | ((d >> 8) & 0xff));  // bigendian verdammt nochma

                    bw.Write(du16);                    
                }
            }

            bw.Close();
            fs.Close();
        }

        private void storeAnims(string referenceMap,params string []animLayers)
        {
            print("Store anim\n");
            StringBuilder sb = new StringBuilder();

            Int64[] stdMap = maps[referenceMap];

            int animCount = animLayers.Length;
            int mapLength = 256 * 256;

            uint[,] animations = new uint[animCount, mapLength];

            for (int x = 0; x < animCount; x++)
            {
                Int64 []od = maps[animLayers[x]];
                for (int i = 0; i < mapLength; i++) // simple copy
                {
                    animations[x, i] = (uint)od[i];
                }

            }
        
            int fixedMaxLength = 1024;                      
            int dataCount = 0;

            sb.Append("\t;map anim\n");

            for (int i = 0; i < mapLength; i++)
            {
                uint[] sequence = new uint[animCount];

                int sequenceLength = 0;
                for (int anim = 0; anim < animCount; anim++)
                {
                    uint w = animations[anim, i];

                    if (w != 0)
                    {
                        sequenceLength++;
                    }
                    sequence[anim] = w;
                }

                String line = "";

                if (sequenceLength > 0)
                {
                    line += ("\t.data\t" + (sequenceLength + 1) + ",\t" + i + ",\t");
                    dataCount += 2;

                    for (int anim = -1; anim < animCount; anim++)
                    {
                        uint w = 0;

                        if (anim == -1)
                        {
                            w = (uint)stdMap[i];
                        }
                        else
                        {
                            w = sequence[anim];

                        }

                        if (w != 0)
                        {
                            w--;
                            w = (w & 0xffff) | (w >> 16);
                            dataCount++;
                            line += "" + w + ",";
                        }
                    }
                    line = line.Remove(line.Length - 1);
                    sb.Append(line + "\t\t; cnt length    position    data\n");
                }
            }

            sb.Append("\t.data\t$0,$0\t;stop\n"); dataCount += 2;
            sb.Append("\t.space\t" + (fixedMaxLength - dataCount) + "\t;valid data is " + dataCount + " of " + fixedMaxLength + " \n");

            TextWriter tw = new StreamWriter(DestFilesPath.Text + "/anims.s");
            tw.Write(sb.ToString());
            tw.Close();
        }

        private void storeTileLut()
        {
            print("Generate tile lut\n");
            StringBuilder sb = new StringBuilder();

            int w = w16;
            int h = h16;
            int ex = 1;
            int ey = 1;
            int ox = 1;
            int oy = 1;
            int tw = 16;
            int th = 16;

            print(" lut support "+w+"x"+h+" tiles\n");

            int yAdrMul = w * (tw + ex) + ox;

            sb.Append("\t;tile lut size = " + (w * h * 2) + " words\n");

            for (int y = 0; y < h; y++)
            {
                sb.Append("\t.data\t");
                for (int x = 0; x < w; x++)
                {
                    int rx = x * (tw + ex) + ox;
                    int ry = y * (th + ey) + oy;
                    int adr = rx + (ry * yAdrMul);

                    sb.Append(string.Format("${0:X4},${1:X4}{2}", adr >> 16, adr & 0xffff, ((x + 1) == w) ? "" : ","));
                }
                sb.Append("\n");
            }

            TextWriter two = new StreamWriter(DestFilesPath.Text + "/tile_lut.s");
            two.Write(sb.ToString());
            two.Close();
        }
    }
}
