 
 	.ifndef	HCORETOOLSI
	.def	HCORETOOLSI
	
;************************************************************************************************************************
;*
;* udiv32 
;*
;* needs 47 cycles	
;*
;* return
;*
;* qoutient  = divident / divider
;* remainder = divident % divider
;*
	
	.macro	udiv32(remainder,qoutient,divident,divider)
	udivinit	divident,divider,32
	move	remainder,0

udiv32loop@	
	udivstep	last
	udivstep	last
	udivstep	last
	udivstep	last
	cmpeqi	remainder,4
	add	remainder,remainder,1	
	br.tc	udiv32loop@
	udivstep	last		;delay slot
	udivstep	last		;delay slot
	udivstep	last		;delay slot
	udivstep	qoutient	;delay slot
	udivgrem	remainder
	.endmacro
	
;************************************************************************************************************************
;*
;* doRake (jump parallel without sync after)
;*
;* handOverReg	 	is given to all cores (gpa 0)
;* maxCoresReg 		contains maximum to prallel (gpa 2 will return cores that work on this function)
;* functionAddress 	is address to jump parallel
;* r126			is used as link register
;*
;* need 11 cycles
;*
;************************************************************************************************************************
	
	.macro	doRake(handOverReg,maxCoresReg,functionAddress)	
	
	spa	handOverReg,0		;hand_over_address = handOverReg
	move	hold,functionAddress
	moveh	hold,last,>>functionAddress
	spa	last,1			;jump_address	   = functionAddress
	spa	maxCoresReg,2	;;max_cores 	   = maxCoresReg

	gpc	r126,2			;return address
	rake	hold			;ignore cores work result
	nop	;delay slot
	nop	;delay slot
	nop	;delay slot
	nop	;delay slot
		
	.endmacro
	
;************************************************************************************************************************
;*
;* doRakeSync (jump parallel with sync after retrun)
;*
;* handOverReg	 	is given to all cores (gpa 0)
;* maxCoresReg 		contains maximum to prallel (gpa 2 will return cores that work on this function)
;* functionAddress 	is address to jump parallel
;* r126			is used as link register
;* r127			is stack pointer
;*
;*
;************************************************************************************************************************	
	
	.macro	doRakeSync(handOverReg,maxCoresReg,functionAddress)	
	
	spa	handOverReg,0		;hand_over_address = handOverReg
	move	hold,functionAddress
	moveh	hold,last,>>functionAddress
	spa	last,1			;jump_address	   = functionAddress
	spa	maxCoresReg,2	;;max_cores 	   = maxCoresReg
	
	;make sync space
	
	sub	r127,r127,8
	move	hold,0
	st	last,r127,0	;clear 0/1
	st	last,r127,2	;clear 2/3
	st	last,r127,4	;clear 4/5
	st	last,r127,6	;clear 6/7
	rqld	r127,0		;sync to memory
	ld	hold		;ignore read value (only sync)
	
	;
	
	gpc	r126,2			;return address
	rake	hold			;get cores work
	st	last,r127,8		;and store on stack	delay slot
	nop				;			delay slot
	nop				;			delay slot
	nop				;			delay slot
	
	;wait for other cores
	
	rqld	r127,8	;total cnt
	rqld	r127,0	;0/1 preload
	rqld	r127,2	;2/3 preload
	rqld	r127,4	;4/5 preload
	rqld	r127,6	;6/7 preload	
	ld	r4
	
syncLoop@		

	ld	r0
	ld	r1
	ld	r2
	ld	r3
	
	add	hold,r1,r0		;  	000X000Y = r1 + r0
	add	hold,last,r2		;   	000X000Y + r2
	add	r0,last,r3		;r0 	000X000Y + r3
	lsr	hold,last,16		;last	0000000X
	add	hold,last,r0		;last   000X000Z
	moveh	hold,last,0		;last   0000000Z
	cmpeq	last,r4			;last = total count ?
	br.tc	syncLoop@		;no poll again
	rqld.tc	r127,0			;0/1	delay slot
	rqld.tc	r127,2			;2/3	delay slot
	rqld.tc	r127,4			;4/5	delay slot
	rqld.tc	r127,6			;6/7	delay slot
	
	add	r127,r127,8	;restore stack	
		
	.endmacro
	
;************************************************************************************************************************
;*
;* returnSync (set sync word should called at end of thread)
;*	

	.macro	returnSync(handOverSp,coreNumber,tmpReg)
	
	jmp	r126,0				;return	
	sub	hold,handOverSp,8		;sp - 8		
	add	tmpReg,last,coreNumber		
	move	hold,1				;		delay slot
	stw	last,tmpReg,0			;		delay slot
	
	.endmacro
	
;************************************************************************************************************************
;*
;* returnUnSync (set sync word should called at end of thread)
;*	

	.macro	returnUnSync
	
	jmp	r126,0		;return
	nop			;delay slot
	nop			;delay slot
	nop			;delay slot
	nop			;delay slot
	
	.endmacro	
	
	.endif
