﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Drawing.Drawing2D;
using System.IO;
using GfxConvert.Properties;
using System.Reflection;
using System.Drawing.Imaging;
using gfx_convert_library;
using gfx_convert_library.converter;

namespace GfxConvert
{
    public partial class GfxConvertForm : Form, ISelectionAreaChange, IGFXConvertPrint
    {
        private SelectionArea m_selectionArea;
        private float m_scaleFactor;
        bool m_moveMode;
        bool m_mouseDown;
        int m_stepX, m_stepY;
        UInt64 m_downCount;
        bool m_selectionEventEnable;
        Bitmap m_image;
        DataConvert m_dataConvert = null;
        bool m_enableCloseDiaglog = true;

        public GfxConvertForm()
        {
            InitializeComponent();
            m_selectionArea = new SelectionArea(this);
            m_selectionEventEnable = true;

            comboBoxOutputFormat.Items.Add(Outputformats_e.Linear_16Bit_4X4R4G4B_ColorFormat.ToString());
            comboBoxOutputFormat.Items.Add(Outputformats_e.Linear_16Bit_1A3C4R4G4B_ColorFormat.ToString());
            comboBoxOutputFormat.Items.Add(Outputformats_e.Linear_16Bit_1A5R5G5B_ColorFormat.ToString());
            comboBoxOutputFormat.Items.Add(Outputformats_e.Linear_16Bit_5R6G5B_ColorFormat.ToString());
            comboBoxOutputFormat.Items.Add(Outputformats_e.Linear_32Bit_8X8R8G8B_ColorFormat.ToString());
            comboBoxOutputFormat.Items.Add(Outputformats_e.Tiled_8x8PixelBlock_4BitColor.ToString());
            comboBoxOutputFormat.Items.Add(Outputformats_e.Tiled_8x8PixelBlock_8BitColor.ToString());           
            comboBoxOutputFormat.Items.Add(Outputformats_e.Tiled_16x16PixelBlock_4BitColor.ToString());
            comboBoxOutputFormat.Items.Add(Outputformats_e.Tiled_16x16PixelBlock_8BitColor.ToString());
            comboBoxOutputFormat.Items.Add(Outputformats_e.Linear_1BitColor.ToString());
            comboBoxOutputFormat.Items.Add(Outputformats_e.linear_4BitColor.ToString());
            comboBoxOutputFormat.Items.Add(Outputformats_e.linear_8BitColor.ToString());
           

            comboBoxOutputFormat.SelectedIndex = 0;

            comboBoxDataFormat.Items.Add("RAW");
            comboBoxDataFormat.Items.Add("MIF");
            comboBoxDataFormat.Items.Add("ASM");
            comboBoxDataFormat.Items.Add("C");
            comboBoxDataFormat.SelectedIndex = 0;

            comboBoxDataWidth.Items.Add("8");
            comboBoxDataWidth.Items.Add("16");
            comboBoxDataWidth.Items.Add("32");
            comboBoxDataWidth.SelectedIndex = 2;

            numericUpDownStepX.Value = 16;
            numericUpDownStepY.Value = 16;
            m_stepX = 16;
            m_stepY = 16;
            m_downCount = 0;

            buttonDoSave.Enabled = false;
            m_mouseDown = false;
        }

        public bool enableCloseDialog
        {
            get { return m_enableCloseDiaglog; }
            set { m_enableCloseDiaglog = value; }
        }

        #region ISelectionAreaChange Member

        void ISelectionAreaChange.changed()
        {
            if (m_selectionArea != null && m_image != null && m_selectionEventEnable == true)
            {
                try
                {
                    numericUpDownSX.Value = m_selectionArea.x;
                    numericUpDownSY.Value = m_selectionArea.y;
                    numericUpDownSelectWidth.Value = m_selectionArea.w;
                    numericUpDownSelectHight.Value = m_selectionArea.h;
                }
                catch (Exception ex)
                {
                    print(ex.Message);
                }
            }
        }

        #endregion

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private DialogResult showCloseDialog()
        {
            DialogResult dr = DialogResult.Yes;
            if (enableCloseDialog)
            {
                MessageBox.Show("Quit GfxConvert?", "Quit?", MessageBoxButtons.YesNo);
            }
            return dr;
        }

        private void BitSwapForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (showCloseDialog() == DialogResult.No)
            {
                e.Cancel = true;
            }
            else 
            {
                Settings.Default.Save();
            }
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AboutBox ab = new AboutBox();
            ab.Show();
        }

        private void buttonLoad_Click(object sender, EventArgs e)
        {
            try
            {
                OpenFileDialog openFileDialog1 = new OpenFileDialog();

                if (textBoxSourceFile.Text != "")
                {
                    openFileDialog1.FileName = textBoxSourceFile.Text;
                    openFileDialog1.InitialDirectory = Path.GetDirectoryName(textBoxSourceFile.Text);
                }
                else
                {
                    openFileDialog1.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                }

                openFileDialog1.Filter = "All files (*.*)|*.*";
                openFileDialog1.FilterIndex = 2;
                openFileDialog1.RestoreDirectory = true;

                if (openFileDialog1.ShowDialog() == DialogResult.OK)
                {
                    string fileName = openFileDialog1.FileName;
                    textBoxSourceFile.Text = fileName; 
                }
            }
            catch (Exception)
            {
                print("Error while selecting/open file or dir\n");
            }
        }

        private void buttonSave_Click(object sender, EventArgs e)
        {
            try
            {
                SaveFileDialog saveFileDialog1 = new SaveFileDialog();

                if (textBoxDestinationFile.Text.Length != 0)
                {
                    saveFileDialog1.FileName = textBoxDestinationFile.Text;
                    saveFileDialog1.InitialDirectory = Path.GetDirectoryName(textBoxDestinationFile.Text);
                }
                else
                {
                    saveFileDialog1.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                }

                saveFileDialog1.Filter = "All files (*.*)|*.*";
                saveFileDialog1.FilterIndex = 1;
                saveFileDialog1.RestoreDirectory = true;

                if (saveFileDialog1.ShowDialog() == DialogResult.OK)
                {
                    string fileName = saveFileDialog1.FileName;
                    textBoxDestinationFile.Text = fileName;
                }
            }
            catch (Exception)
            {
                print("Error while selecting/open file or dir\n");
            }

         
        }


        private void buttonDoLoad_Click(object sender, EventArgs e)
        {
            loadPictureFromFile(textBoxSourceFile.Text);
        }

        private void buttonDoSave_Click(object sender, EventArgs e)
        {
            Outputformats_e oFormat = Outputformats_e.UNKNOWN;

            if (comboBoxOutputFormat.Text == Outputformats_e.Tiled_16x16PixelBlock_4BitColor.ToString())
            {
                oFormat = Outputformats_e.Tiled_16x16PixelBlock_4BitColor;
            }
            else if (comboBoxOutputFormat.Text == Outputformats_e.Tiled_8x8PixelBlock_4BitColor.ToString())
            {
                oFormat = Outputformats_e.Tiled_8x8PixelBlock_4BitColor;
            }
            else if (comboBoxOutputFormat.Text == Outputformats_e.Tiled_8x8PixelBlock_8BitColor.ToString())
            {
                oFormat = Outputformats_e.Tiled_8x8PixelBlock_8BitColor;
            }
            else if (comboBoxOutputFormat.Text == Outputformats_e.Tiled_16x16PixelBlock_8BitColor.ToString())
            {
                oFormat = Outputformats_e.Tiled_16x16PixelBlock_8BitColor;
            }
            else if (comboBoxOutputFormat.Text == Outputformats_e.linear_4BitColor.ToString())
            {
                oFormat = Outputformats_e.linear_4BitColor;
            }
            else if (comboBoxOutputFormat.Text == Outputformats_e.linear_8BitColor.ToString())
            {
                oFormat = Outputformats_e.linear_8BitColor;
            }
            else if (comboBoxOutputFormat.Text == Outputformats_e.Linear_16Bit_4X4R4G4B_ColorFormat.ToString())
            {
                oFormat = Outputformats_e.Linear_16Bit_4X4R4G4B_ColorFormat;
            }

            else if (comboBoxOutputFormat.Text == Outputformats_e.Linear_1BitColor.ToString())
            {
                oFormat = Outputformats_e.Linear_1BitColor;
            }            
            else if (comboBoxOutputFormat.Text == Outputformats_e.Linear_16Bit_5R6G5B_ColorFormat.ToString())
            {
                oFormat = Outputformats_e.Linear_16Bit_5R6G5B_ColorFormat;
            }
            else if (comboBoxOutputFormat.Text == Outputformats_e.Linear_16Bit_1A3C4R4G4B_ColorFormat.ToString())
            {
                oFormat = Outputformats_e.Linear_16Bit_1A3C4R4G4B_ColorFormat;
            }
            else if (comboBoxOutputFormat.Text == Outputformats_e.Linear_16Bit_1A5R5G5B_ColorFormat.ToString())
            {
                oFormat = Outputformats_e.Linear_16Bit_1A5R5G5B_ColorFormat;
            }
            else if (comboBoxOutputFormat.Text == Outputformats_e.Linear_32Bit_8X8R8G8B_ColorFormat.ToString())
            {
                oFormat = Outputformats_e.Linear_32Bit_8X8R8G8B_ColorFormat;
            }
            else
            { 
                print("Error: unknwon output format\n"); 
            }

            if (oFormat != Outputformats_e.UNKNOWN)
            {;
                m_dataConvert = new DataConvert(textBoxDestinationFile.Text, oFormat,this,textBoxAsmOpcode.Text);

                try
                {
                    int ix, iy, iw, ih;
                    m_selectionArea.get(out ix, out iy, out iw, out ih);

                    if (ix < 0)
                    {
                        iw = iw + ix; 
                        ix = 0;
                    }

                    if (iy < 0)
                    {
                        ih = ih + iy; 
                        iy = 0;
                    }

                    if (ix + iw > m_image.Width)
                    {
                        iw = m_image.Width - ix;
                    }

                    if (iy + ih > m_image.Height)
                    {
                        ih = m_image.Height - iy;
                    }

                    if (iw <= 0 || ih <= 0)
                    {
                        print("Can't save a image with no width or depth\n"); 
                    }
                    else
                    {
 

                        print("Try to convert from selection (" + ix + ";" + iy + ")-(" + iw + ";" + ih + ")\n");

                        writeMode_t wm = writeMode_t.RAW;
                        if (comboBoxDataFormat.Text == "MIF")
                        {
                            wm = writeMode_t.MIF;
                        }
                        else if (comboBoxDataFormat.Text == "ASM")
                        {
                            wm = writeMode_t.ASM;
                        }
                        else if (comboBoxDataFormat.Text == "C")
                        {
                            wm = writeMode_t.C;
                        }


                        int wlength = numericUpDownForcedLength.Enabled ? (int)numericUpDownForcedLength.Value : 0; // 0 is write all

                        int wbitwidth = 32;
                        switch( comboBoxDataWidth.Text )
                        {
                            case "8": wbitwidth = 8; break;
                            case "16": wbitwidth = 16; break;
                            case "32": wbitwidth = 32; break;
                        }

                        m_dataConvert.code = (UInt32) numericUpDownDataCode.Value;
                        m_dataConvert.setWriteBitWidth(wbitwidth);       
                        m_dataConvert.setWriteForcedLength(wlength);
                        m_dataConvert.setWriteMode(wm);
                        m_dataConvert.convert(m_image, (uint)ix, (uint)iy, (uint)iw, (uint)ih);
                        m_dataConvert.store();

                        print("Convert and store complete\n");
                    }
                }
                catch (Exception ex)
                {
                    print(ex.Message);
                }
            }
        }

        public void print(string txt)
        {
            debugBox.AppendText(txt);
        }

        public void printd(string txt)
        {
            if (checkBoxDebugoutput.Checked == true)
            {
                print(txt);
            }
        }

        void loadPictureFromFile(string fileName)
        {
             try
             {
                m_image = new Bitmap(fileName);

                m_selectionArea.set(0, 0, m_image.Width, m_image.Height);

                if (m_image.PixelFormat == PixelFormat.Format4bppIndexed ||
                    m_image.PixelFormat == PixelFormat.Format8bppIndexed)
                {
                    numericUpDownColors.Value = m_image.Palette.Entries.Length; // auto refresh
                }
                else
                {
                    numericUpDownColors.Value = 2 ^ 24;
                }

                zoomBar.Value = 200;
                numericUpDownWidth.Value = m_image.Width;
                numericUpDownHeight.Value = m_image.Height;

                print("Load complete.\n");

                buttonDoSave.Enabled = true;

                updateZoom(zoomBar.Value / 100f); 

             }
             catch (Exception e)
             {
                 print("Error " + e.Message + "\n");
             }

             
        }
  
        private void picBox_MouseMove(object sender, MouseEventArgs e)
        {
            int mx = e.X;
            int my = e.Y;

            float offsetX = clamp(hScrollBar.Value, 0);
            float offsetY = clamp(vScrollBar.Value, 0);

            mx += (int)offsetX;
            my += (int)offsetY;

            // descale

            float fdx = mx / m_scaleFactor;
            float fdy = my / m_scaleFactor;

            // step raster

            if (m_stepX != 0 && m_stepY != 0)
            {
                fdx = fdx - (fdx % m_stepX);
                fdy = fdy - (fdy % m_stepX);
            }

            if (m_moveMode == true)
            {
                if (m_mouseDown == true)
                {
                    m_selectionArea.x = (int)fdx;
                    m_selectionArea.y = (int)fdy;
                }
            }
            else 
            {
                if (m_mouseDown == true)
                {
                    if(m_downCount == 0)
                    {
                        m_selectionArea.x = (int)fdx;
                        m_selectionArea.y = (int)fdy;
                    }
                    else
                    {
                        int w = (int)fdx;
                        int h = (int)fdy;

                        w = w - m_selectionArea.x + m_stepX;
                        h = h - m_selectionArea.y + m_stepY;

                        if (w < m_stepX)
                        {
                            w = m_stepX;
                        }

                        if (h < m_stepY)
                        {
                            h = m_stepY;
                        }

                        m_selectionArea.w = (int)w;
                        m_selectionArea.h = (int)h;
                    }

                    m_downCount ++;
                }
                else
                { 
                    m_downCount = 0;
                }

            }

           

            picBox.Refresh();
        }

        int clamp(int v, int min = 0, int max = 2147483647)
        {
            v = v < min ? min : v;
            v = v > max ? max : v;
            return v;
        }

        private void picBox_Paint(object sender, PaintEventArgs e)
        {         
            Graphics g = e.Graphics;

            g.PixelOffsetMode = PixelOffsetMode.Half;
            g.SmoothingMode = SmoothingMode.None;
            g.InterpolationMode = InterpolationMode.NearestNeighbor;

            if(m_image != null)
            {

                // scrollbars

                float iWidth = m_image.Width;
                float iHeight = m_image.Height;

                iWidth *= m_scaleFactor;
                iHeight *= m_scaleFactor;

                hScrollBar.Maximum = clamp((int)(iWidth - picBox.Width),0);
                vScrollBar.Maximum = clamp((int)(iHeight - picBox.Height),0);

                float offsetX = -clamp(hScrollBar.Value, 0);
                float offsetY = -clamp(vScrollBar.Value, 0);

                // draw selection

                Pen pn = new Pen(Color.Blue);
                pn.DashStyle = DashStyle.Dash;
                pn.Width = m_scaleFactor;

                float sx,sy,sw,sh, csw, csh;
                m_selectionArea.get(m_scaleFactor, out sx, out sy, out sw, out sh);

                sx += offsetX;
                sy += offsetY;

                csw = sw;
                csh = sh;

                if (sx + sw > iWidth)
                {
                    csw = iWidth - sx;
                }

                if (sy + sh > iHeight)
                {
                    csh = iHeight - sy;
                }
                
                g.FillRectangle(Brushes.LightCoral, sx, sy, sw, sh);

                try
                {
                    g.DrawImage(m_image, offsetX, offsetY, iWidth, iHeight);
                }
                catch (Exception)
                {
                }

                pn.Color = Color.Blue;
                g.DrawRectangle(pn, sx, sy, csw, csh);
            

                if (checkBoxDrawCollCircle.Checked == true)
                {
                    Pen pnc = new Pen(Color.White);
                    pnc.Width = m_scaleFactor;

                    //float sx,sy,sw,sh, csw, csh;
                    m_selectionArea.get(m_scaleFactor, out sx, out sy, out sw, out sh);

                    float cR = (float)numericUpDownCSR.Value * m_scaleFactor;

                    float cX = ((float)numericUpDownCSX.Value) + (sx + sw/2.0f) - cR;
                    float cY = ((float)numericUpDownCSY.Value) + (sy + sh/2.0f) - cR;
                    

                    g.DrawArc(pnc, cX, cY, cR*2, cR*2, 0, 360);

                }
            }
        }

        private void numericUpDownStepX_ValueChanged(object sender, EventArgs e)
        {
            m_stepX = (int)numericUpDownStepX.Value;
        }

        private void numericUpDownStepY_ValueChanged(object sender, EventArgs e)
        {
            m_stepY = (int)numericUpDownStepY.Value;
        }

        private void numericUpDownSelectWidth_ValueChanged(object sender, EventArgs e)
        {
            m_selectionEventEnable = false;
            m_selectionArea.w = (int)numericUpDownSelectWidth.Value;
            m_selectionEventEnable = true;
            picBox.Refresh();
        }

        private void numericUpDownSelectHight_ValueChanged(object sender, EventArgs e)
        {
            m_selectionEventEnable = false;
            m_selectionArea.h = (int)numericUpDownSelectHight.Value;
            m_selectionEventEnable = true;
            picBox.Refresh();
        }

        private void numericUpDownSX_ValueChanged(object sender, EventArgs e)
        {
            m_selectionEventEnable = false;
            m_selectionArea.x = (int)numericUpDownSX.Value;
            m_selectionEventEnable = true;
            picBox.Refresh();
        }

        private void numericUpDownSY_ValueChanged(object sender, EventArgs e)
        {
            m_selectionEventEnable = false;
            m_selectionArea.y = (int)numericUpDownSY.Value;
            m_selectionEventEnable = true;
            picBox.Refresh();
        }

      

       

        private void picBox_MouseUp(object sender, MouseEventArgs e)
        {
            m_mouseDown = false;
        }

        private void picBox_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                m_moveMode = false;
            }
            else
            {
                m_moveMode = true;
            }

            m_mouseDown = true;

        }

        private void BitSwapForm_Load(object sender, EventArgs e)
        {
            Settings.Default.Reload();

            textBoxSourceFile.Text = Settings.Default.sourceFilename;
            textBoxDestinationFile.Text = Settings.Default.destinationFilename;

           
        }

        private void clearToolStripMenuItem_Click(object sender, EventArgs e)
        {
            debugBox.Clear();
        }

        private void checkBoxForceLength_CheckedChanged(object sender, EventArgs e)
        {
            numericUpDownForcedLength.Enabled = checkBoxForceLength.Checked;

            if (m_dataConvert != null)
            {
                int wlength = numericUpDownForcedLength.Enabled ? (int)numericUpDownForcedLength.Value : 0; // 0 is write all
                m_dataConvert.setWriteForcedLength(wlength);
            }
        }

        private void numericUpDownForcedLength_ValueChanged(object sender, EventArgs e)
        {
            if (m_dataConvert != null)
            {
                int wlength = numericUpDownForcedLength.Enabled ? (int)numericUpDownForcedLength.Value : 0; // 0 is write all
                m_dataConvert.setWriteForcedLength(wlength);
            }
        }

        private void comboBoxDataWidth_VisibleChanged(object sender, EventArgs e)
        {
            if (m_dataConvert != null)
            {
                int wbitwidth = 32;
                switch (comboBoxDataWidth.Text)
                {
                    case "8": wbitwidth = 8; break;
                    case "16": wbitwidth = 16; break;
                    case "32": wbitwidth = 32; break;
                }

                m_dataConvert.setWriteBitWidth(wbitwidth);
            }
        }

        #region FileJoin
        string m_FJsourceFile = "";

        private void buttonFJAddInputFiles_Click(object sender, EventArgs e)
        {
            try
            {
                OpenFileDialog openFileDialog1 = new OpenFileDialog();

                if (textBoxSourceFile.Text != "")
                {
                    openFileDialog1.FileName = m_FJsourceFile;
                }

                openFileDialog1.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                openFileDialog1.Filter = "All files (*.*)|*.*";
                openFileDialog1.FilterIndex = 2;
                openFileDialog1.RestoreDirectory = true;
                openFileDialog1.Multiselect = true;

                if (openFileDialog1.ShowDialog() == DialogResult.OK)
                {
                    string fileName = openFileDialog1.FileName;
                    m_FJsourceFile = fileName;
                }

                foreach( string fname in openFileDialog1.FileNames)
                {
                    listBoxFJ.Items.Add(fname);
                }
            }
            catch (Exception)
            {
                print("Error while selecting/open file or dir\n");
            }
        }
       

        private void buttonFJMoveUp_Click(object sender, EventArgs e)
        {
            int selected = listBoxFJ.SelectedIndex;
            if (selected >= 1)
            {
                string before = listBoxFJ.Items[selected - 1].ToString();
                string current = listBoxFJ.Items[selected].ToString();
                listBoxFJ.Items[selected] = before;
                listBoxFJ.Items[selected - 1] = current;
                listBoxFJ.SelectedIndex = selected - 1;
            }
        }

        private void buttonFJMoveDown_Click(object sender, EventArgs e)
        {
            int selected = listBoxFJ.SelectedIndex;
            if (selected < listBoxFJ.Items.Count-1)
            {
                string next = listBoxFJ.Items[selected + 1].ToString();
                string current = listBoxFJ.Items[selected].ToString();
                listBoxFJ.Items[selected] = next;
                listBoxFJ.Items[selected + 1] = current;
                listBoxFJ.SelectedIndex = selected + 1;
            }
        }

        private void buttonFJremove_Click(object sender, EventArgs e)
        {
            int selected = listBoxFJ.SelectedIndex;
            if (selected >= 0)
            {
                listBoxFJ.Items.RemoveAt(selected);
            }
        }

        private void buttonFJclear_Click(object sender, EventArgs e)
        {
            listBoxFJ.Items.Clear();
        }

        string m_FJdestFile = "";
        private void buttonFJStore_Click(object sender, EventArgs e)
        {
            // get store filename

            try
            {
                SaveFileDialog saveFileDialog1 = new SaveFileDialog();

                if (textBoxDestinationFile.Text.Length != 0)
                {
                    saveFileDialog1.FileName = m_FJdestFile;
                }

                saveFileDialog1.InitialDirectory = System.IO.Directory.GetCurrentDirectory();
                saveFileDialog1.Filter = "All files (*.*)|*.*";
                saveFileDialog1.FilterIndex = 1;
                saveFileDialog1.RestoreDirectory = true;

                if (saveFileDialog1.ShowDialog() == DialogResult.OK)
                {
                    string fileName = saveFileDialog1.FileName;
                    m_FJdestFile = fileName;
                }
            }
            catch (Exception)
            {
                print("Error while selecting/open file or dir\n");
            }

            //

            List<byte> dataArray = new List<byte>();
            try
            {
                int fileCount = listBoxFJ.Items.Count;
                for (int i = 0; i < fileCount; i++)
                {
                    FileStream fs = File.OpenRead(listBoxFJ.Items[i].ToString());
                    BinaryReader br = new BinaryReader(fs);

                    int length = (int)fs.Length;

                    for (uint n = 0; n < length; n++)
                    {
                        dataArray.Add(br.ReadByte());
                    }
                    br.Close();
                    fs.Close();
                }

                // store

                FileStream fsw = File.Create(m_FJdestFile);
                BinaryWriter bw = new BinaryWriter(fsw);

                foreach(byte b in dataArray)    // it slow but doesnt care me currently ;-)
                {
                    bw.Write(b);
                }

                bw.Close();
                fsw.Close();
            }
            catch (Exception ex)
            {
                print(ex.Message);
            }
        }
        #endregion

        private void numericUpDownCSX_ValueChanged(object sender, EventArgs e)
        {
            picBox.Refresh();
        }

        private void numericUpDownCSY_ValueChanged(object sender, EventArgs e)
        {
            picBox.Refresh();
        }

        private void numericUpDownCSR_ValueChanged(object sender, EventArgs e)
        {
            picBox.Refresh();
        }

        private void checkBoxDrawCollCircle_CheckedChanged(object sender, EventArgs e)
        {
            picBox.Refresh();
        }

        private void GfxConvertForm_DragDrop(object sender, DragEventArgs e)
        {
            print("lol");
        }

        private void vScrollBar_ValueChanged(object sender, EventArgs e)
        {
            picBox.Refresh();
        }

        private void hScrollBar_Scroll(object sender, ScrollEventArgs e)
        {
            picBox.Refresh();
        }

        private void trackBar1_Scroll(object sender, EventArgs e)
        {
            updateZoom(zoomBar.Value / 100f); 
           
        }

        private void updateZoom(float zoomValue)
        {
             m_scaleFactor = zoomValue;

            zoomLabel.Text = string.Format("Zoom X{0:0.00}", m_scaleFactor);
            picBox.Refresh();
        }
    }   
}
