﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using gfx_convert_library;
using System.Drawing;
using gfx_convert_library.converter;

namespace gfx_convert_con
{
    class Program
    {
        static void Main(string[] args)
        {
            string sourceFileName = "";
            string destinationFileName = "";
            string formatName = "";
            string outputTypeName = "RAW";
            string assemblerOpcode = "word";
            int dataWidth = 32;
            UInt32 code = 0;
            int forceLength = 0;
            Outputformats_e outputFormat = Outputformats_e.UNKNOWN;
                  
            int argCount = args.Length;

            if (argCount == 0)
            {
                print("Usage: gfx_convert_con.exe -s sourceFile -o destFile -f format [-w outputType] [-bw data width] [-fl force length] [-c code]\n");
                print("\n");
                print("Description:   -s          source file\n");
                print("               -o          output file\n");
                print("               -f          format to write (see below)\n");               
                print("                            Linear_32_8X8R8G8B\n");  
                print("                            Linear_16_4X4R4G4B\n");
                print("                            Linear_16_1A5R5G5B\n");
                print("                            Linear_16_1A3C4R4G4B\n");  
                print("                            Linear_16_5R6G5B\n");
                print("                            Linear_1\n");  
                print("                            Linear_4\n");  
                print("                            Linear_8\n");  
                print("                            Tile_16x16x4\n");  
                print("                            Tile_8x8x4\n");  
                print("                            Tile_16x16x8\n");
                print("                            Tile_8x8x8\n");  
                print("               -w          output type (see below), default is RAW\n");     
                print("                            MIF\n"); 
                print("                            RAW\n"); 
                print("                            ASM\n");
                print("                            C\n");
                print("               -c          code (will set to every pixel see C in format)\n");
                print("               -ao         assembler opcode\n");
                print("               -bw          datatype width in bits (only for ASM/C/MIF, default is 32)");
                print("               -fl          force output length (only for ASM/C/MIF,0 is full length, default is 0)");
                print("\n");

                return;
            }

            for (uint i = 0; i < argCount; i++)
            {
                switch (args[i])
                {
                    case "-bw":

                        if (i + 1 >= argCount)
                        {
                            print("error: no data width after -bw\n");
                            return;
                        }

                        try
                        {
                            dataWidth = Convert.ToInt32(args[i + 1]);
                        }
                        catch (Exception ex)
                        {
                            print("error: invalid data width (" + ex.Message + ")\n");
                            return;
                        }

                        break;

                    case "-ao":

                        if (i + 1 >= argCount)
                        {
                            print("error: no data width after -ao\n");
                            return;
                        }

                        assemblerOpcode = args[i + 1];                       
                        break;


                    case "-fl":

                        if (i + 1 >= argCount)
                        {
                            print("error: no length after -fl\n");
                            return;
                        }

                        try
                        {
                            forceLength = Convert.ToInt32(args[i + 1]);
                        }
                        catch (Exception ex)
                        {
                            print("error: invalid length (" + ex.Message + ")\n");
                            return;
                        }

                        break;

                    case "-c":

                        if (i + 1 >= argCount)
                        {
                            print("error: no value after -c\n");
                            return;
                        }

                        try
                        {
                            code = Convert.ToUInt32(args[i + 1]);
                        }
                        catch (Exception ex)
                        {
                            print("error: invalid value (" + ex.Message + ")\n");
                            return;
                        }

                        break;

                    case "-w":

                        if (i + 1 >= argCount)
                        {
                            print("error: no type -w\n");
                            return;
                        }

                        outputTypeName = args[i + 1];

                        break;

                    case "-f":

                        if (i + 1 >= argCount)
                        {
                            print("error: no format after -f\n");
                            return;
                        }

                        formatName = args[i + 1];

                        break;

                    case "-s":

                        if (i + 1 >= argCount)
                        {
                            print("error: no source file after -s\n");
                            return;
                        }

                        sourceFileName = args[i + 1];

                        break;
              
                    case "-o":

                        if (i + 1 >= argCount)
                        {
                            print("error: no destination file after -o\n");
                            return;
                        }

                        destinationFileName = args[i + 1];

                        break;

                    default:

                        break;
                }
            }

            // check

            if (destinationFileName == "")
            {
                print("error: no destination file!\n");
                return;
            }

            if (sourceFileName == "")
            {
                print("error: no source file!\n");
                return;
            }

            if (formatName == "")
            {
                print("error: no format!\n");
                return;
            }
            else
            {
                // check format

                outputFormat = Outputformats_e.UNKNOWN;

                switch (formatName)
                {

                    case "Linear_32_8X8R8G8B": outputFormat = Outputformats_e.Linear_32Bit_8X8R8G8B_ColorFormat; break;
                    case "Linear_16_4X4R4G4B": outputFormat = Outputformats_e.Linear_16Bit_4X4R4G4B_ColorFormat; break;
                    case "Linear_16_1A5R5G5B": outputFormat = Outputformats_e.Linear_16Bit_1A5R5G5B_ColorFormat; break;
                    case "Linear_16_1A3C4R4G4B": outputFormat = Outputformats_e.Linear_16Bit_1A3C4R4G4B_ColorFormat; break;
                    case "Linear_16_5R6G5B": outputFormat = Outputformats_e.Linear_16Bit_5R6G5B_ColorFormat; break;
                    case "Linear_1": outputFormat = Outputformats_e.Linear_1BitColor; break;
                    case "Linear_4": outputFormat = Outputformats_e.linear_4BitColor; break;
                    case "Linear_8": outputFormat = Outputformats_e.linear_8BitColor; break;
                    case "Tile_16x16x4": outputFormat = Outputformats_e.Tiled_16x16PixelBlock_4BitColor; break;
                    case "Tile_8x8x4": outputFormat = Outputformats_e.Tiled_8x8PixelBlock_4BitColor; break;
                    case "Tile_16x16x8": outputFormat = Outputformats_e.Tiled_16x16PixelBlock_8BitColor; break;
                    case "Tile_8x8x8": outputFormat = Outputformats_e.Tiled_8x8PixelBlock_8BitColor; break;           
                }

                if (outputFormat == Outputformats_e.UNKNOWN)
                {
                    print("error: unknown format!\n");
                    return;
                }

            }

            // check output type

            writeMode_t writeMode = writeMode_t.RAW;

            switch (outputTypeName)
            {
                case "MIF": writeMode = writeMode_t.MIF;break;
                case "RAW": writeMode = writeMode_t.RAW;break;
                case "ASM": writeMode = writeMode_t.ASM;break;
                case "C": writeMode = writeMode_t.C; break;
                default: 
                    
                    print("error: unknown output type!\n");
                    return;
            }

            // do

            try
            {
                Bitmap image = new Bitmap(sourceFileName);
                DataConvert dataConvert = new DataConvert(destinationFileName, outputFormat, new printOutputDirector(), assemblerOpcode);

                dataConvert.code = code;
                dataConvert.setWriteBitWidth(dataWidth);
                dataConvert.setWriteForcedLength(forceLength);
                dataConvert.setWriteMode(writeMode);
                dataConvert.convert(image, (uint)0, (uint)0, (uint)image.Width, (uint)image.Height); // ever full size
                dataConvert.store();

                print("finish successfully\n");
            }
            catch (Exception ex)
            {
                print("error: " + ex.Message + "\n");
            }
        }

        static void print(string txt)
        {
            Console.Write(txt);
        }

        public class printOutputDirector : IGFXConvertPrint
        {
            public printOutputDirector()
            {
               
            }

            public void print(string txt)
            {
                Program.print(txt);
            }

            public void printd(string txt)
            {
               
            }
        }
    }
}
