﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing.Drawing2D;
using gfx_convert_library.Store;
using System.Drawing;
using System.Drawing.Imaging;
using gfx_convert_library.converter;

namespace gfx_convert_library
{
    public enum Outputformats_e
    {
        Linear_32Bit_8X8R8G8B_ColorFormat,
        Linear_16Bit_4X4R4G4B_ColorFormat,
        Linear_16Bit_1A3C4R4G4B_ColorFormat,
        Linear_16Bit_1A5R5G5B_ColorFormat,
        Linear_16Bit_5R6G5B_ColorFormat,
        Linear_1BitColor,
        linear_4BitColor,
        linear_8BitColor,
        Tiled_16x16PixelBlock_4BitColor,
        Tiled_8x8PixelBlock_4BitColor,
        Tiled_16x16PixelBlock_8BitColor,
        Tiled_8x8PixelBlock_8BitColor,
        UNKNOWN
    };

    public enum writeMode_t
    {
        MIF,
        RAW,
        ASM,
        C
    }

    public class DataConvert : IGFXConvertPrint
    {
        private string m_outfileName;
        private Outputformats_e m_oFormat;
        private IGFXConvertPrint m_printInterface;
        writeMode_t m_writeMode;
        int m_forcedWriteLength;
        int m_writeBitWidth;
        string m_extra;
        UInt32 m_code = 0;

        List<uint> m_gfx = new List<uint>();
        List<Color> m_pal = new List<Color>();

        public DataConvert(string outfileName, Outputformats_e oFormat, IGFXConvertPrint printInterface, string extra)
        {
            m_outfileName = outfileName;
            m_oFormat = oFormat;
            m_printInterface = printInterface;
            m_forcedWriteLength = 0;
            m_writeMode = writeMode_t.RAW;
            m_code = 0;
            m_extra = extra != null ? extra : "";
        }

        public UInt32 code
        {
            get { return m_code; }
            set { m_code = value; }
        }

        void convertGfxData(DataStream bw)
        {
            if (m_oFormat == Outputformats_e.Tiled_16x16PixelBlock_4BitColor || m_oFormat == Outputformats_e.Tiled_8x8PixelBlock_4BitColor || m_oFormat == Outputformats_e.linear_4BitColor)
            {
                for (int i = 0; i < m_gfx.Count; i += 2)
                {
                    byte b = (byte)(((m_gfx[i] & 0xf) << 4) | (m_gfx[i + 1] & 0xf));

                    bw.Write(b);
                }
            }
            else if (m_oFormat == Outputformats_e.Tiled_16x16PixelBlock_8BitColor || m_oFormat == Outputformats_e.Tiled_8x8PixelBlock_8BitColor || m_oFormat == Outputformats_e.linear_8BitColor)
            {
                for (int i = 0; i < m_gfx.Count; i++)
                {
                    bw.Write((byte)m_gfx[i]);
                }
            }
            else if (m_oFormat == Outputformats_e.Linear_16Bit_4X4R4G4B_ColorFormat)
            {
                for (int i = 0; i < m_gfx.Count; i++)
                {
                    UInt32 orgb = m_gfx[i];     // aarrggbb
                    UInt16 srgb = (UInt16)((orgb & 0xf0) >> 4);  // 000b
                    srgb |= (UInt16)((orgb & 0x00f000) >> 8);  // 00gb
                    srgb |= (UInt16)((orgb & 0xf00000) >> 12);  // argb
                    srgb |= (UInt16)((orgb >> (32 - 4)) << (16 - 4));    // a

                    //    srgb = (UInt16)((srgb >> 8) | (srgb << 8)); // endian shit

                    bw.Write(srgb);
                }
            }
            else if (m_oFormat == Outputformats_e.Linear_16Bit_1A3C4R4G4B_ColorFormat)
            {
                for (int i = 0; i < m_gfx.Count; i++)
                {
                    UInt32 orgb = m_gfx[i];     // accc rrrr gggg bbbb
                    UInt16 srgb = (UInt16)((orgb & 0xf0) >> 4);  // 000b
                    srgb |= (UInt16)((orgb & 0x00f000) >> 8);  // 00gb
                    srgb |= (UInt16)((orgb & 0xf00000) >> 12);  // argb
                    srgb |= (UInt16)((orgb >> 31) << 15);    // a
                    srgb |= (UInt16)((m_code & 7) << 12);   // code

                    bw.Write(srgb);
                }
            }
            else if (m_oFormat == Outputformats_e.Linear_1BitColor)
            {
                UInt32[] pixelBuffer = new UInt32[8];
                int pixelFlushCount = 0;

                for (int i = 0; i < m_gfx.Count; i++)
                {
                    UInt32 orgb = m_gfx[i];     // aarrggbb

                    pixelBuffer[pixelFlushCount++] = orgb;
                    if (pixelFlushCount == 8)
                    {
                        pixelFlushCount = 0;
                        Byte d = 0;

                        for (int j = 0; j < 8; j++)
                        {
                            bool isValue = (pixelBuffer[j] >> 24) != 0xff;

                            printd("" + (isValue ? 0 : 1));
                            d = (byte)(d | ((isValue ? 0 : (1 << (7 - j)))));
                        }

                        printd(string.Format(" -> {0:X}\n", d));

                        bw.Write(d);
                    }
                }
            }
            else if (m_oFormat == Outputformats_e.Linear_16Bit_1A5R5G5B_ColorFormat)
            {
                for (int i = 0; i < m_gfx.Count; i++)
                {
                    UInt32 orgb = m_gfx[i];     // arrr rrrgg gggb bbbb
                    UInt16 srgb = (UInt16)((orgb & 0xf8) >> 3);  // 000b
                    srgb |= (UInt16)((orgb & 0x00f800) >> 6);  // 00gb
                    srgb |= (UInt16)((orgb & 0xf80000) >> 9);  // 0rgb
                    srgb |= (UInt16)((orgb >> 31) << 15);    // a

                    //    srgb = (UInt16)((srgb >> 8) | (srgb << 8)); // endian shit

                    bw.Write(srgb);
                }
            }
            else if (m_oFormat == Outputformats_e.Linear_16Bit_5R6G5B_ColorFormat)
            {
                for (int i = 0; i < m_gfx.Count; i++)
                {
                    UInt32 orgb = m_gfx[i];     // 00rrggbb
                    UInt16 srgb = (UInt16)((orgb & 0xf8) >> 3);  // 000b
                    srgb |= (UInt16)((orgb & 0x00fc00) >> 5);  // 00gb
                    srgb |= (UInt16)((orgb & 0xf80000) >> 8);  // 0rgb

                    //    srgb = (UInt16)((srgb >> 8) | (srgb << 8)); // endian shit

                    bw.Write(srgb);
                }
            }
            else if (m_oFormat == Outputformats_e.Linear_32Bit_8X8R8G8B_ColorFormat)
            {
                for (int i = 0; i < m_gfx.Count; i++)
                {
                    bw.Write(m_gfx[i]);
                }
            }
            else
            {
                throw new Exception("Convert implements no gfx data convertion for this mode\n");
            }
        }

        public void store()
        {
            // gfx

            DataStream ds = new DataStream();

            convertGfxData(ds);
            Save.write(m_outfileName + ".dat", m_writeMode, ds, m_forcedWriteLength, m_writeBitWidth,m_extra);     // picture datas

            // pal

            if (m_pal.Count != 0)
            {
                DataStream dsp = new DataStream();

                uint colorIndex = 0;
                foreach (Color dc in m_pal)
                {

                    byte bh = (byte)(dc.R / 16);
                    byte bl = (byte)((((int)dc.G / 16) << 4) | ((int)dc.B / 16));

                    m_printInterface.printd(String.Format("Color {0:D3} ${1:X2}{2:X2}\n", (colorIndex++), (uint)bh, (uint)bl));

                    dsp.Write(bh);
                    dsp.Write(bl);
                }

                Save.write(m_outfileName + ".pal", m_writeMode, dsp, 0,m_writeBitWidth, m_extra);     // palette datas

            }

        }

        unsafe uint getPixelIndex(void* data, uint x, uint y, uint w, uint h, PixelFormat pixelFormat)
        {
            byte* bdata = (byte*)data;
            UInt32* ldata = (UInt32*)data;
            uint rValue = 0;

            if (pixelFormat == PixelFormat.Format4bppIndexed)
            {
                uint pos = x / 2 + (y * w / 2);
                if (pos >= (w * h) / 2)
                {
                    throw new Exception("getPixelIndex is out of bounds\n");
                }

                rValue = (uint)bdata[pos];

                if ((x & 1) == 0)
                {
                    rValue = rValue >> 4;
                }
                else
                {
                    rValue = rValue & 0xf;
                }
            }
            else if (pixelFormat == PixelFormat.Format8bppIndexed)
            {
                uint pos = x + (y * w);
                if (pos >= w * h)
                {
                    throw new Exception("getPixelIndex is out of bounds\n");
                }

                rValue = (uint)bdata[pos];
            }
            else if (pixelFormat == PixelFormat.Format24bppRgb)
            {
                uint pos = x + (y * w);
                if (pos >= w * h)
                {
                    throw new Exception("getPixelIndex is out of bounds\n");
                }

                UInt32 xrgb = 0;
                xrgb |= (UInt32)bdata[pos * 3 + 2] << 16;
                xrgb |= (UInt32)bdata[pos * 3 + 1] << 8;
                xrgb |= (UInt32)bdata[pos * 3];
                rValue = xrgb;
            }
            else if (pixelFormat == PixelFormat.Format32bppArgb)
            {
                uint pos = x + (y * w);
                if (pos >= w * h)
                {
                    throw new Exception("getPixelIndex is out of bounds\n");
                }

                UInt32 xrgb = 0;
                xrgb |= (UInt32)bdata[pos * 4 + 3] << 24;
                xrgb |= (UInt32)bdata[pos * 4 + 2] << 16;
                xrgb |= (UInt32)bdata[pos * 4 + 1] << 8;
                xrgb |= (UInt32)bdata[pos * 4];
                rValue = xrgb;
            }
            else
            {
                throw new Exception("illegal Pixelformat (only 16 or 256 indexed colors supported)\n");
            }

            return rValue;
        }

        unsafe public void convert(Bitmap m_image, uint x, uint y, uint w, uint h)
        {
            m_image.PixelFormat.ToString();



            m_gfx.Clear();
            m_pal.Clear();

            // data o

            uint tWidth = 1;
            uint tHeight = 1;
            uint tColors = 1;
            PixelFormat tPixelFormat = PixelFormat.Undefined;

            if (m_oFormat == Outputformats_e.Tiled_16x16PixelBlock_4BitColor)
            {
                tWidth = 16;
                tHeight = 16;
                tColors = 16;
                tPixelFormat = PixelFormat.Format4bppIndexed;

            }
            else if (m_oFormat == Outputformats_e.Tiled_8x8PixelBlock_4BitColor)
            {
                tWidth = 8;
                tHeight = 8;
                tColors = 16;
                tPixelFormat = PixelFormat.Format4bppIndexed;
            }
            else if (m_oFormat == Outputformats_e.Tiled_16x16PixelBlock_8BitColor)
            {
                tWidth = 16;
                tHeight = 16;
                tColors = 256;
                tPixelFormat = PixelFormat.Format8bppIndexed;
            }
            else if (m_oFormat == Outputformats_e.Tiled_8x8PixelBlock_8BitColor)
            {
                tWidth = 8;
                tHeight = 8;
                tColors = 256;
                tPixelFormat = PixelFormat.Format8bppIndexed;
            }
            else if (m_oFormat == Outputformats_e.linear_4BitColor)
            {
                tWidth = w;
                tHeight = h;
                tColors = 16;
                tPixelFormat = PixelFormat.Format8bppIndexed;
            }
            else if (m_oFormat == Outputformats_e.linear_8BitColor)
            {
                tWidth = w;
                tHeight = h;
                tColors = 256;
                tPixelFormat = PixelFormat.Format8bppIndexed;
            }
            else if (
                m_oFormat == Outputformats_e.Linear_1BitColor
                || m_oFormat == Outputformats_e.Linear_16Bit_4X4R4G4B_ColorFormat
                || m_oFormat == Outputformats_e.Linear_16Bit_1A5R5G5B_ColorFormat
                || m_oFormat == Outputformats_e.Linear_16Bit_1A3C4R4G4B_ColorFormat
                || m_oFormat == Outputformats_e.Linear_16Bit_5R6G5B_ColorFormat
                || m_oFormat == Outputformats_e.Linear_32Bit_8X8R8G8B_ColorFormat)
            {
                tWidth = w;
                tHeight = h;
                tColors = 0xffffffff;
                tPixelFormat = PixelFormat.Format32bppArgb; // deal as full argb
            }
            BitmapData bitmapData;

            try
            {
                bitmapData = m_image.LockBits(new Rectangle((int)0, (int)0, (int)m_image.Width, (int)m_image.Height), ImageLockMode.ReadOnly, tPixelFormat);
            }
            catch (Exception)
            {
                throw new Exception("Input and Outputformat is not compatible (check Colordepth)\n");
            }

            void* rawData = bitmapData.Scan0.ToPointer();


            try
            {

                for (uint j = y; j < (y + h); j += tHeight) 
                {
                    for (uint i = x; i < (x + w); i += tWidth)
                    {         
                        print(string.Format("Convert at ({0};{1})-({2};{3})\n", i, j, tWidth, tHeight));
                        converTile(rawData, i, j, tWidth, tHeight, tColors, (uint)m_image.Width, (uint)m_image.Height, tPixelFormat);                
                    }
                }
            }
            catch (Exception ex)
            {
                m_printInterface.print("Error: " + ex.Message);
            }

            // lut

            foreach (Color dc in m_image.Palette.Entries)
            {
                m_pal.Add(dc);
            }

            m_image.UnlockBits(bitmapData);

        }

        private int getColorIndex(Bitmap m_image, Color s)
        {
            int clutPos = 0;
            foreach (Color dc in m_image.Palette.Entries)
            {
                if (s == dc)
                {
                    return clutPos;
                }

                clutPos++;
            }

            return -1;
        }

        unsafe private void converTile(void* rawData, uint x, uint y, uint w, uint h, uint c, uint bitmapWidth, uint bitMapHeight, PixelFormat pixelFormat)
        {
            for (uint j = y; j < (y + h); j++)
            {
                for (uint i = x; i < (x + w); i++)
                {
                    uint ci = getPixelIndex(rawData, i, j, bitmapWidth, bitMapHeight, pixelFormat);

                    if (ci < 0 || ci > c)
                    {
                        throw new Exception("Error: color index (" + ci + " out of range (" + c + ")\n");
                    }

                    m_gfx.Add((uint)ci);
                }
            }
        }

        public void setWriteMode(writeMode_t wm)
        {
            m_writeMode = wm;
        }

        public void setWriteForcedLength(int wlength)
        {
            m_forcedWriteLength = wlength;
        }

        public void setWriteBitWidth(int wbitwidth)
        {
            m_writeBitWidth = wbitwidth;
        }

        public void printd(string txt)
        {
            m_printInterface.printd(txt);
        }

        public void print(string txt)
        {
            m_printInterface.print(txt);
        }
    }
}
