﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Text.RegularExpressions;
using System.IO;
using hcore_assembler_library.Opcode;
using xcore_tools.Common;
using xcore_tools;
using xcore_tools.Data;

namespace hcore_assembler_library
{
    public class HCoreAssembler : bitExtraction, IPrint
    {
        IPrint m_iprint;
        LineTextReader m_reader;
        Programm m_program;

        public HCoreAssembler(IPrint iprint)
        {
            m_iprint = iprint;
        }

        public string sourceFile
        {
            get;
            set;
        }

        public List<Line> lines
        {
            get { return m_reader.lines; }
        }


        public void parse(Dictionary<string, string> preDefines, bool trace)
        {
            //init
            m_program = new Programm(this);

            // read source
            m_reader = new LineTextReader(sourceFile,this);           

            // preprocess
            Dictionary<string, string> defines = preprocess(preDefines,m_reader.lines);

            // debug

          /*  foreach (Line l in m_reader.lines)
            {
                print("("+l.line + ") " +l.value);
            }*/

            // analyse

            print("Start HCore assembler @" + DateTime.Now.ToString("HH:mm:ss") + "\n");
            MyParser analyse = new MyParser(m_program,this);

            analyse.Setup();
            analyse.Parse(m_reader);

            // resolve

            m_program.addDefines(defines);
            m_program.preresolve();
            m_program.resolve();
            m_program.check(); 

            // debug

            if (trace)
            {
                print("HCore Trace Long Style\n");
                m_program.trace();
            }
        }

        String txtParseTree = "";

        private void DrawReductionTree(GOLD.Reduction Root)
        {
            //This procedure starts the recursion that draws the parse tree.
            StringBuilder tree = new StringBuilder();

            tree.AppendLine("+-" + Root.Parent.Text(false));
            DrawReduction(tree, Root, 1);

            txtParseTree = tree.ToString();
        }

        private void DrawReduction(StringBuilder tree, GOLD.Reduction reduction, int indent)
        {
            //This is a simple recursive procedure that draws an ASCII version of the parse
            //tree

            int n;
            string indentText = "";

            for (n = 1; n <= indent; n++)
            {
                indentText += "| ";
            }

            //=== Display the children of the reduction
            for (n = 0; n < reduction.Count(); n++)
            {
                switch (reduction[n].Type())
                {
                    case GOLD.SymbolType.Nonterminal:
                        GOLD.Reduction branch = (GOLD.Reduction)reduction[n].Data;

                        tree.AppendLine(indentText + "+- P" + branch.Parent.Text(false));
                        DrawReduction(tree, branch, indent + 1);
                        break;

                    default:
                        string leaf = (string)reduction[n].Data;

                        tree.AppendLine(indentText + "+- L" + leaf);
                        break;
                }
            }
        }

        public List<uint> assembleAndGetData()
        {
            return m_program.assembleAndGetData();
        }

        Dictionary<string, string> preprocess(Dictionary<string, string> preDefines,List<Line> lines)
        {
            HcorePreprocessor pre = new HcorePreprocessor(this);
            pre.process(preDefines,lines);
            return pre.defines;
        }

        public void print(string txt)
        {
            m_iprint.print(txt);
        }

        public void dprint(string txt)
        {
            m_iprint.dprint(txt);
        }
    }
}
