﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using xcore_tools.Common;

namespace hcore_assembler_library.Opcode
{
    public enum HOpcodeExecutionPath_e
    {
        inOrder,            /* std */
        memoryOutOfOrder,   /* to/from memory */
        uctrlOutOfOrder,    /* to/from uctrl */
        unknownOutOfOrder
    };

    public class HOpcodeDescription
    {
        public HOpcodeDescription(string _name, uint _id,bool _writeBack = false, bool _hasCC = true,HOpcodeExecutionPath_e _path = HOpcodeExecutionPath_e.inOrder)
        {
            name = _name;
            id = _id;
            hasCC = _hasCC;
            path = _path;
            writeBack = _writeBack;
        }

        public string name { get; set; }
        public uint id { get; set; }
        public bool hasCC { get; set; }
        public HOpcodeExecutionPath_e path { get; set; }
        public bool writeBack { get; set; }
    };

    public class HOpcodeDescriptions
    {
        static bool m_init = false;

        static SortedDictionary<string, HOpcodeDescription> m_nameBased;
        static SortedDictionary<uint, SortedDictionary<bool, HOpcodeDescription>> m_idBased;

        public HOpcodeDescriptions()
        {
            if (m_init == false)
            {
                init();
            }
        }

        public HOpcodeDescription getOpcode(string name)
        {
            if (m_nameBased.ContainsKey(name) == true)
            {
                return m_nameBased[name];
            }

            return null;
        }

        public HOpcodeDescription getOpcode(uint id, bool isCC)
        {
            if (m_idBased.ContainsKey(id))
            {
                if (m_idBased[id].ContainsKey(isCC))
                {
                    return m_idBased[id][isCC];
                }
            }

            return null;
        }


        void init()
        {
            List<HOpcodeDescription> opcodes = new List<HOpcodeDescription>();

            m_init = true;

            opcodes.Add(new HOpcodeDescription("move"       , 0, true));
            opcodes.Add(new HOpcodeDescription("moveh"      , 1, true));
            opcodes.Add(new HOpcodeDescription("add"        , 2, true));
            opcodes.Add(new HOpcodeDescription("sub"        , 3, true));
            opcodes.Add(new HOpcodeDescription("cmplo"      , 4));
            opcodes.Add(new HOpcodeDescription("cmplos"     , 5));
            opcodes.Add(new HOpcodeDescription("cmple"      , 6));
            opcodes.Add(new HOpcodeDescription("cmples"     , 7));
            opcodes.Add(new HOpcodeDescription("cmpeq"      , 8));
            opcodes.Add(new HOpcodeDescription("jmp"        , 9));
            opcodes.Add(new HOpcodeDescription("br"         , 10));
            opcodes.Add(new HOpcodeDescription("gpc"        , 11, true));
            opcodes.Add(new HOpcodeDescription("and"        , 12, true));
            opcodes.Add(new HOpcodeDescription("or"         , 13, true));
            opcodes.Add(new HOpcodeDescription("xor"        , 14, true));
            opcodes.Add(new HOpcodeDescription("bic"        , 15, true));
            opcodes.Add(new HOpcodeDescription("mul"        , 16, true));
            opcodes.Add(new HOpcodeDescription("mul_fead"   , 16, true));
            opcodes.Add(new HOpcodeDescription("mulh"       , 17, true));
            opcodes.Add(new HOpcodeDescription("mulh_fead"  , 17, true));
            opcodes.Add(new HOpcodeDescription("lsr"        , 18, true));
            opcodes.Add(new HOpcodeDescription("lsl"        , 19, true));
            opcodes.Add(new HOpcodeDescription("asr"        , 20, true));
            opcodes.Add(new HOpcodeDescription("st"         , 21, false,true, HOpcodeExecutionPath_e.memoryOutOfOrder));
            opcodes.Add(new HOpcodeDescription("stw"        , 22, false,true, HOpcodeExecutionPath_e.memoryOutOfOrder));
            opcodes.Add(new HOpcodeDescription("uexec"      , 23, false,true, HOpcodeExecutionPath_e.uctrlOutOfOrder));
            opcodes.Add(new HOpcodeDescription("rsub"       , 24, true));
            opcodes.Add(new HOpcodeDescription("stc"        , 25, false, true, HOpcodeExecutionPath_e.memoryOutOfOrder));
            opcodes.Add(new HOpcodeDescription("roe"        , 26));
            opcodes.Add(new HOpcodeDescription("idle"       , 25));
            opcodes.Add(new HOpcodeDescription("dump"       , 46));
            opcodes.Add(new HOpcodeDescription("fead"       , 47));
            opcodes.Add(new HOpcodeDescription("nop"        , 47));


            opcodes.Add(new HOpcodeDescription("ld"         , 0, true, false, HOpcodeExecutionPath_e.memoryOutOfOrder)); // not realy a rcwe
            opcodes.Add(new HOpcodeDescription("ldw"        , 1, true, false, HOpcodeExecutionPath_e.memoryOutOfOrder)); // not realy a rcwe
            opcodes.Add(new HOpcodeDescription("ldws"       , 2, true, false, HOpcodeExecutionPath_e.memoryOutOfOrder)); // not realy a rcwe
            opcodes.Add(new HOpcodeDescription("gssr"       , 3, true, false));
            opcodes.Add(new HOpcodeDescription("sssr"       , 4, false, false));
            opcodes.Add(new HOpcodeDescription("uexecs"     , 5, true, false, HOpcodeExecutionPath_e.uctrlOutOfOrder));
            opcodes.Add(new HOpcodeDescription("ldc"        , 6, true, false, HOpcodeExecutionPath_e.memoryOutOfOrder)); // not realy a rcwe

            // build name/id table

            m_nameBased = new SortedDictionary<string, HOpcodeDescription>();
            m_idBased = new SortedDictionary<uint, SortedDictionary<bool, HOpcodeDescription>>();

            foreach (HOpcodeDescription o in opcodes)
            {
                m_nameBased[o.name] = o;
                if (m_idBased.ContainsKey(o.id) == false)
                {
                    m_idBased[o.id] = new SortedDictionary<bool, HOpcodeDescription>();
                }
                m_idBased[o.id][o.hasCC] = o;
            }

        }

    };

    public class HOpcode : bitExtraction
    {

        const uint delaySlotLength = 5;

        HOpcodeDescriptions m_hDesc = new HOpcodeDescriptions();

        //

        HOpcodeExecutionPath_e m_executionPath;
        uint m_opcode = 0;
        bool m_withCondition = true;
        uint m_cc = 0;
        uint m_aibico = 0;
        int m_rc = 0;
        int m_rb = 0;
        int m_ra = 0;
        Int64 m_imm = 0;
        string m_immString = "";
        bool m_immIsValue = true;
        bool m_immIsPcRelative = false;
        bool m_isCommand = false;
        string m_commandName = "";
        int m_immShr = 0;


        public List<string> labelNames { get; set; }
        uint m_pc = 0;
        Labels m_labels;
        bool m_skip = false;
        public HOpcodeInfo opcodeInfo { get; set; }

        //

        UInt32 m_data;

        public HOpcode()
        {
            labelNames = new List<string>();            
        }

        public HOpcode(Labels labels,HOpcodeInfo iopcodeInfo)
        {
            labelNames = new List<string>();
            m_labels = labels;
            opcodeInfo = iopcodeInfo; // store for debug
            init(iopcodeInfo);
        }

        public bool isCommand   { get { return m_isCommand; } }
        public bool useCAsReg   { get { return (m_aibico & 1) == 0; } }
        public bool useBAsReg   { get { return (m_aibico & 2) == 0; } }
        public bool useAAsReg   { get { return (m_aibico & 4) == 0; } }
        public int RC           { get { return m_rc; } }
        public int RB           { get { return m_rb; } }
        public int RA           { get { return m_ra; } }
        public bool isRCWE
        {
            get 
            {
                HOpcodeDescription hd = m_hDesc.getOpcode(m_opcode, m_withCondition);
                if (hd != null)
                {
                    return hd.writeBack;
                }            
                throw new Exception("HOpcode error, can't detect RCWE of given opcode");
            }
        }

        public bool canLockReg
        {
            get
            {
                HOpcodeDescription hd = m_hDesc.getOpcode(m_opcode, m_withCondition);
                if (hd != null)
                {
                    return hd.writeBack == true && hd.path != HOpcodeExecutionPath_e.inOrder;
                }
                throw new Exception("HOpcode error, can't detect canLockReg of given opcode");
            }
        }

        public UInt32 getData(out bool hasData)
        {
            hasData = !m_skip;
            return m_data;
        }

        public bool hasData()
        {
            return !m_skip;
        }

        public uint preresolve(uint pc)
        {
            uint nextPc = pc + 2;
            if (m_isCommand)
            {
                if (m_commandName == "org")
                {
                    if(m_immString != "")
                    {
                        m_imm = m_labels.getLabel(m_immString);
                    }
                    nextPc = (uint)m_imm;
                    m_skip = true; // dont generate data from derective
                }
            }

            m_pc = pc;
            foreach (string labelName in labelNames)
            { 
                if (labelName != "")
                {
                    m_labels.addLabel(labelName, pc);
                }
            }

            return nextPc;
        }

        public void resolve()
        {
            if (m_immIsValue == false)
            {
                m_imm = m_labels.getLabel(m_immString); // parse evtl.
                m_imm = m_imm >> m_immShr;
                if (m_immIsPcRelative)
                {
                    m_imm = m_imm - m_pc - (delaySlotLength * 2);
                }            
            }
        }

        public uint getPc()
        {
            return m_pc;
        }

        public void assemble()
        {
            int param = 0;
            m_skip = false;


            if (m_isCommand)
            {
                if (m_commandName == "word")
                {
                    m_data = (uint)m_imm;
                }
                else if (m_commandName == "org")
                {
                    m_pc = (uint)m_imm;
                    m_skip = true; // dont generate data from derective
                }
                else
                {
                    throw new Exception("illegal assembler command '" + m_commandName  + "'");
                }
            }
            else
            {

                switch (m_aibico)
                {
                    case 0:
                        {
                            param = m_rc << 14 | m_rb << 7 | m_ra;
                        }
                        break;
                    case 1:
                        {
                            Int32 i = Immediate.translate((Int32)m_imm, 7, true, false);
                            param = m_rb << 14 | m_ra << 7 | i;
                        }
                        break;
                    case 2:
                        {
                            Int32 i = Immediate.translate((Int32)m_imm, 7, true, false);
                            param = m_rc << 14 | m_ra << 7 | i;
                        }
                        break;
                    case 3:
                        {
                            Int32 i = Immediate.translate((Int32)m_imm, 13, false, false);
                            param = m_ra << 14 | i;

                        }
                        break;
                    case 4:
                        {
                            Int32 i = Immediate.translate((Int32)m_imm, 7, false, false);
                            param = m_rc << 14 | m_rb << 7 | i;
                        }
                        break;
                    case 5:
                        {
                            Int32 i = Immediate.translate((Int32)m_imm, 13, false, false);
                            param = m_rb << 14 | i;
                        }
                        break;
                    case 6:
                        {
                            Int32 i = Immediate.translate((Int32)m_imm, 13, false, false);
                            param = m_rc << 14 | i;
                        }
                        break;
                    case 7:
                        {
                            Int32 i;
                            if (m_immIsPcRelative)
                            {
                                i = Immediate.translate((Int32)m_imm, 21, true, false);
                            }
                            else
                            {
                                if (m_immIsValue)
                                {
                                    i = Immediate.translate((Int32)m_imm, 21, false, false);
                                }
                                else
                                {
                                    i = Immediate.translate((Int32)m_imm, 21, false, true);
                                }
                            }
                            param = i & ((1 << 21) - 1);
                        }
                        break;
                    default:
                        throw new Exception("Opcode has internal error??");
                }

                // set data

                m_data = 0;

                if (m_withCondition == true)
                {
                    m_data |= m_opcode << 26;
                    m_data |= m_cc << 24;
                }
                else 
                {
                    m_data |= 0xc0000000;
                    m_data |= m_opcode << 24;
                }

                m_data |= m_aibico << 21;
                m_data |= (uint)param;
            }
        }

        void init(HOpcodeInfo iop)
        {
         /*   labelName = iop.getLabelName();
            if (labelName != "")
            {
                labelName = labelName;
            }*/

            m_immShr = iop.getImmediateShr();

            // command

            switch (iop.getOpcodeName())
            {
                case "word": m_isCommand = true; m_commandName = "word"; break;
                case "org": m_isCommand = true; m_commandName = "org"; break;
                default: m_isCommand = false; m_commandName = ""; break;
            }

            // real opcode

            m_withCondition = true;
            m_executionPath = HOpcodeExecutionPath_e.inOrder;

            string opcodeName = iop.getOpcodeName();

           HOpcodeDescription hd = m_hDesc.getOpcode(opcodeName);

           if (hd == null)
           {
               m_opcode = 47;   // nop       
           }
           else
           {

               m_opcode = hd.id;
               m_withCondition = hd.hasCC;
               m_executionPath = hd.path;
           }
          

            if (m_withCondition == true)
            {
                // in order opcodes have condition

                switch (iop.getCondition())
                {
                    case HOpcodeInfo.condition_e.ever: m_cc = 0; break;
                    case HOpcodeInfo.condition_e.ts: m_cc = 1; break;
                    case HOpcodeInfo.condition_e.tc: m_cc = 2; break;
                    case HOpcodeInfo.condition_e.never: m_cc = 3; break;
                    default: m_cc = 3; break;
                }
            }
            else
            {
                // out of order opcodes have no condition
                m_cc = 0; // ever
            }


            m_immIsPcRelative = false;
            switch (iop.getOpcodeName())
            {
                case "br": m_immIsPcRelative = true; break;
                case "gpc": m_immIsPcRelative = true; break;
            }

           

            m_rc = iop.getDestinationRegister();
            m_rb = iop.getSourceBRegister();
            m_ra = iop.getSourceARegister();

            m_immIsValue = iop.isImmidiatValue();
            m_imm = m_immIsValue ? iop.getImmediate() : 0;
            m_immString = m_immIsValue ? "" : iop.getImmediateString();

            // test register

            if (iop.isDestinationRegister() && (m_rc < 0 || m_rc > 127))
            {
                throw new Exception("Destination register is out of range (r0..r127 != r" + m_rc + ")");                
            }

            if (iop.isSourceBRegister() && (m_rb < 0 || m_rb > 127))
            {
                throw new Exception("Source register (B) is out of range (r0..r127 != r" + m_rb + ")");
            }

            if (iop.isSourceARegister() && (m_ra < 0 || m_ra > 127))
            {
                throw new Exception("Source register (A) is out of range (r0..r127 != r" + m_ra + ")");
            }

            //

            switch (iop.getCodeType())
            {
                case HOpcodeInfo.codeType_e.CBA:
                    {
                        if (iop.isDestinationRegister())
                        {
                            if (iop.isSourceBRegister())
                            {
                                if (iop.isSourceARegister())    // C B A
                                {
                                    m_aibico = 0;
                                }
                                else                            // C B I6
                                {
                                    m_aibico = 4;
                                }
                            }
                            else
                            {
                                if (iop.isSourceARegister())    // C A
                                {
                                    m_aibico = 2;
                                }
                                else                            // C 13
                                {
                                    m_aibico = 6;
                                }
                            }
                        }
                        else
                        {
                            if (iop.isSourceBRegister())
                            {
                                if (iop.isSourceARegister())    // B A
                                {
                                    m_aibico = 1;
                                }
                                else                            // B I13
                                {
                                    m_aibico = 5;
                                }
                            }
                            else
                            {
                                if (iop.isSourceARegister())    // A
                                {
                                    m_aibico = 3;
                                }
                                else                            // I21
                                {
                                    m_aibico = 7;
                                }
                            }
                        }
                    }
                    break;

                case HOpcodeInfo.codeType_e.BA:
                     {
                         if (iop.isSourceBRegister())   
                         {
                             if (iop.isSourceARegister())   // B A
                             {
                                 m_aibico = 1;
                             }
                             else                           // B I
                             {
                                 m_aibico = 5;
                             }
                         }
                         else 
                         {
                             if (iop.isSourceARegister())   // A
                             {
                                 m_aibico = 3;
                             }
                             else                           // I21
                             {
                                 m_aibico = 7;
                             }
                         }
                    }
                    break;

                case HOpcodeInfo.codeType_e.BAS:
                    if (iop.isSourceBRegister())
                    {
                        m_aibico = 1; // B A I7
                    }
                    else 
                    {
                        m_aibico = 3; // A I14
                    }
                    break;

                case HOpcodeInfo.codeType_e.CA:
                    {
                        if (iop.isDestinationRegister())
                        {
                            if (iop.isSourceARegister())    // C A
                            {
                                m_aibico = 2;
                            }
                            else                            // C I13
                            {
                                m_aibico = 6;
                            }
                        }
                        else
                        {
                            if (iop.isSourceARegister())    // A
                            {
                                m_aibico = 3;
                            }
                            else                            // I21
                            {
                                m_aibico = 7;
                            }
                        }
                    }
                    break;

                case HOpcodeInfo.codeType_e.CB:
                     {
                         if (iop.isDestinationRegister())
                         {
                             if (iop.isSourceBRegister())   // C B
                             {
                                 m_aibico = 4;
                             }
                             else                           // C I13
                             {
                                 m_aibico = 5;
                             }
                         }
                         else
                         {
                             if (iop.isSourceBRegister())   // B
                             {
                                 m_aibico = 5;
                             }
                             else                           // I21
                             {
                                 m_aibico = 7;
                             }
                         }
                    }
                    break;

                case HOpcodeInfo.codeType_e.A:
                    {
                        if (iop.isSourceARegister())
                        {

                            m_aibico = 3;
                        }
                        else
                        {
                            m_aibico = 7;
                        }
                    }
                    break;

                case HOpcodeInfo.codeType_e.C:
                    if (iop.isDestinationRegister())
                    {
                        m_aibico = 6; // C
                    }
                    else
                    {
                        m_aibico = 7; 
                    }
                    break;

                case HOpcodeInfo.codeType_e.NOTHING: 
                    m_aibico = 7;
                    break;

                default: m_aibico = 7; break;
            }
        }

        public void decode(uint din)
        {
            if (extr(din, 31, 30) == 3)
            {
                // oo

                m_executionPath = HOpcodeExecutionPath_e.unknownOutOfOrder;
                m_withCondition = false;
                m_opcode = extr(din, 29, 24);
                m_cc = 0;

                switch (m_opcode)
                {
                    case 0:
                    case 1:
                    case 2:
                    case 3:
                    case 4: m_executionPath = HOpcodeExecutionPath_e.memoryOutOfOrder;
                        break;

                    case 5:
                    case 6: m_executionPath = HOpcodeExecutionPath_e.uctrlOutOfOrder;
                        break;
                }
            }
            else
            {
                // io
                m_withCondition = true;
                m_executionPath = HOpcodeExecutionPath_e.inOrder;
                m_opcode = extr(din, 31, 26);
                m_cc = extr(din, 25, 24);
            }
                                       
            m_aibico = extr(din, 23, 21);

            switch (m_aibico)
            {
                case 0: m_rc = (int)extr(din, 20, 14); m_rb = (int)extr(din, 13, 7); m_ra = (int)extr(din, 6, 0); m_imm = 0; break;
                case 1: m_rc = 0; m_rb = (int)extr(din, 20, 14); m_ra = (int)extr(din, 13, 7); m_imm = extr(din, 6, 0); break;
                case 2: m_rc = (int)extr(din, 20, 14); m_rb = 0; m_ra = (int)extr(din, 13, 7); m_imm = extr(din, 6, 0); break;
                case 3: m_rc = 0; m_rb = 0; m_ra = (int)extr(din, 20, 14); m_imm = extr(din, 13, 0); break;
                case 4: m_rc = (int)extr(din, 20, 14); m_rb = (int)extr(din, 13, 7); m_imm = extr(din, 6, 0); break;
                case 5: m_rc = 0; m_rb = (int)extr(din, 20, 14); m_imm = extr(din, 13, 0); break;
                case 6: m_rc = (int)extr(din, 20, 14); m_rb = 0; m_imm = extr(din, 13, 0); break;
                case 7: m_rc = 0; m_rb = 0; m_imm = (uint)exts(extr(din, 20, 0), 20); break;
                default: break;
            }
        }

        public string trace()
        {
            string rstr = "";
            bool writeBack = false;

            HOpcodeDescription hd = m_hDesc.getOpcode(m_opcode, m_withCondition);
            if (hd != null)
            {
                m_executionPath = hd.path;
                writeBack = hd.writeBack;
                rstr = hd.name;
            }         
            else
            {
                rstr = "error";
            }
         
            if (m_withCondition)
            {
                switch (m_cc)
                {
                    // case 0: rstr+=".ev";break; // ever is standard so dont display
                    case 1: rstr += ".ts"; break;
                    case 2: rstr += ".tc"; break;
                    case 3: rstr += ".nv"; break;
                }
            }

           rstr = string.Format("{0,-16}",rstr);

           if (writeBack == true)
           {
               switch (m_aibico)
               {
                   case 0: rstr += string.Format("r{0},r{1},r{2}", m_rc, m_rb, m_ra); break;
                   case 1: rstr += string.Format("hold,r{0},r{1}[,${2:X3}]", m_rb, m_ra, (uint)(m_imm & 0x7f)); break;
                   case 2: rstr += string.Format("r{0},last,r{1}[,${2:X3}]", m_rc, m_ra, (uint)(m_imm & 0x7f)); break;
                   case 3: rstr += string.Format("hold,last,r{0}[,${1:X4}]", m_ra, (uint)(m_imm & 0x3fff)); break;
                   case 4: rstr += string.Format("r{0},r{1},${2:X3}", m_rc, m_rb, (uint)(m_imm & 0x7f)); break;
                   case 5: rstr += string.Format("hold,r{0},${1:X4}", m_rb, (uint)(m_imm & 0x3fff)); break;
                   case 6: rstr += string.Format("r{0},last,${1:X4}", m_rc, (uint)(m_imm & 0x3fff)); break;
                   case 7: rstr += string.Format("[hold,][last,]${0:X6}", (uint)(m_imm & 0x1fffff)); break;
                   default: break;
               }
           }
           else
           {
               switch (m_aibico)
               {
                   case 0: rstr += string.Format("r{1},r{2}", m_rc, m_rb, m_ra); break;
                   case 1: rstr += string.Format("r{0},r{1}[,${2:X3}]", m_rb, m_ra, (uint)(m_imm & 0x7f)); break;
                   case 2: rstr += string.Format("last,r{1}[,${2:X3}]", m_rc, m_ra, (uint)(m_imm & 0x7f)); break;
                   case 3: rstr += string.Format("last,r{0}[,${1:X4}]", m_ra, (uint)(m_imm & 0x3fff)); break;
                   case 4: rstr += string.Format("r{1},${2:X3}", m_rc, m_rb, (uint)(m_imm & 0x7f)); break;
                   case 5: rstr += string.Format("r{0},${1:X4}", m_rb, (uint)(m_imm & 0x3fff)); break;
                   case 6: rstr += string.Format("last,${1:X4}", m_rc, (uint)(m_imm & 0x3fff)); break;
                   case 7: rstr += string.Format("[last,]${0:X6}", (uint)(m_imm & 0x1fffff)); break;
                   default: break;
               }
           }

            return rstr;
        }

        public bool isInOrder()
        {
            return m_withCondition;
        }

        public HOpcodeExecutionPath_e getExecutionPath()
        {
            return m_executionPath;
        }

        public string getExecutionPathAsString()
        {
            switch (m_executionPath)
            {
                case HOpcodeExecutionPath_e.inOrder: return "in order";
                case HOpcodeExecutionPath_e.memoryOutOfOrder: return "memory";
                case HOpcodeExecutionPath_e.uctrlOutOfOrder: return "uctrl";
                case HOpcodeExecutionPath_e.unknownOutOfOrder: return "unknown";
                
            }
            return "unknown";
        }
    }
}
