﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using xcore_tools.Data;

namespace hcore_assembler_library.Opcode
{
    public class HOpcodeInfo
    {
        public enum condition_e
        {
            ever,
            ts,
            tc,
            never
        }

        public enum codeType_e
        {
            CBA,
            BA,
            BAS,
            CA,
            CB,
            A,
            C,
            NOTHING
        }

        public void clear()
        {
            line = null;
            setLabelName();
            setOpcodeName();
            setCodeType();
            setCondition();
            setDestination(true);
            setSourceB(true);
            setSourceA(true);
            setImmediate(0);

        }

        public Line line{ get;set; }

        public HOpcodeInfo()
        {
        }

        public HOpcodeInfo(Line iline)
        {
            line = iline;
        }
       
        public void setLabelName(string name = "")
        {
            m_labelName = name;
        }

        public string getLabelName()
        {
            return m_labelName;
        }

        public void setOpcodeName(string name = "")
        {
            m_opcodeName = name;
        }

        public string getOpcodeName()
        {
            return m_opcodeName;
        }


        public void setCodeType(codeType_e codeType = codeType_e.NOTHING)
        {
            m_codeType = codeType;
        }

        public codeType_e getCodeType()
        {
            return m_codeType;
        }

        public void setCondition(string cc = "")
        {
            switch (cc)
            {
                case "ev": m_condition = condition_e.ever; break;
                case "ts": m_condition = condition_e.ts; break;
                case "tc": m_condition = condition_e.tc; break;
                case "ne": m_condition = condition_e.never; break;
                default  : m_condition = condition_e.ever; break;
            }

        }

        public condition_e getCondition()
        {
            return m_condition;
        }

        public void setDestination(bool isRegister, int register = 0)
        {
            m_destIsRegister = isRegister;
            m_destRegister = isRegister ? register : 0;
        }

        public int getDestinationRegister()
        {
            return m_destRegister;
        }

        public bool isDestinationRegister()
        {
            return m_destIsRegister;
        }

        public void setSourceB(bool isRegister, int register = 0)
        {
            m_sourceBIsRegister = isRegister;
            m_sourceBRegister = isRegister ? register : 0;
        }

        public int getSourceBRegister()
        {
            return m_sourceBRegister;
        }

        public bool isSourceBRegister()
        {
            return m_sourceBIsRegister;
        }

        public void setSourceA(bool isRegister, int register = 0)
        {
            m_sourceAIsRegister = isRegister;
            m_sourceARegister = isRegister ? register : 0;
        }

        public int getSourceARegister()
        {
            return m_sourceARegister;
        }

        public bool isSourceARegister()
        {
            return m_sourceAIsRegister;
        }

        public void setImmediate(Int64 imm,int shrCount = 0)
        {
            m_immValue = true;
            m_imm = imm;
            m_immShrCount = shrCount;
        }

        public int getImmediateShr()
        {
            return m_immShrCount;
        }

        public void setImmediate(string imm,int shrCount = 0)
        {
            m_immValue = false;
            m_immString = imm;
            m_immShrCount = shrCount;
        }

        public bool isImmidiatValue()
        {
            return m_immValue; 
        }

        public Int64 getImmediate()
        {
            return m_imm;
        }

        public string getImmediateString()
        {
            return m_immString;
        }

        public string printable()
        {
            return String.Format("LABEL = {0} OPCODE = {1} CC = {2} DRR = {3} DR = {4} DBR = {5} DB = {6} DAR = {7} DR = {8} IMV = {9} IMM = {10} IMMSTR = {11} SHRC = {12}",m_labelName,m_opcodeName,m_condition.ToString(),m_destIsRegister,m_destRegister,m_sourceBIsRegister,m_sourceBRegister,m_sourceAIsRegister,m_sourceARegister,m_immValue,m_imm,m_immString,m_immShrCount);
        }

        int m_destRegister = 0;
        bool m_destIsRegister = false;

        int m_sourceBRegister = 0;
        bool m_sourceBIsRegister = false;

        int m_sourceARegister = 0;
        bool m_sourceAIsRegister = false;

        string m_labelName = "";
        condition_e m_condition = condition_e.ever;
        codeType_e m_codeType = codeType_e.NOTHING;
        public string m_cc = "";
        string m_opcodeName = "";

        int m_immShrCount = 0;
        bool m_immValue = true;
        Int64 m_imm = 0;
        string m_immString = "";
    }
}
