﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_assembler_library.Common;
using System.Text.RegularExpressions;

namespace ucore_assembler_library.Data
{
    public partial class FOpcode : bitExtraction, IOpcode<UInt128>
    {

        const int m_maxMux = 31;
       
        enum condition
        {
               ts,tc,ne,ev
        };

        enum calc_src
        {
            x,y,imm
        }

        enum shift_src
        {
            x, imm
        }

        public enum component_type
        {
            alu,buf,los
        }

        component_type m_componentType;

      
        // std

        bool m_use_y;
        bool m_use_x;
        UInt64 m_data_y_sel;
        UInt64 m_data_x_sel;
        UInt32 m_data_valid_count;
        UInt64 m_event_out;
        const int std_offset = 46;

        // los/buf/alu

        UInt64 m_t_in_sel = 0; // FIXME: implement me

        UInt64 m_immval = 0;
        UInt64 m_mainOpcode = 0;
        UInt64 m_sftOpcode = 0;
        condition m_cc;
        calc_src m_calcSrc;
        shift_src m_shiftSrc;
        UInt64 m_ccSel = 0;
     
        
        bool m_en_internal;
        bool m_use_internal;
        UInt64 m_sel_internal;
        UInt64 m_epc, m_event;
        bool m_stop;
       

        public FOpcode()
        {
            initPatterns();
        }

        public string decodeString(UInt128 data_in,bool detailed = true)
        {
            string rString = "";

            // std block

            UInt128 dctrl = 0;

            UInt32 validCount = extr(data_in, 31, 0);
            bool useX = extr(data_in, 32) != 0;
            bool useY = extr(data_in, 33) != 0;
            UInt32 sourceX = extr(data_in, 38, 34);
            UInt32 sourceY = extr(data_in, 43, 39);
            bool eventOut = extr(data_in, 44) != 0;
            bool stop = extr(data_in, 45) != 0;

            rString = String.Format("{0:X8} {1}{2} {3}{4} {5} {6}", validCount, useX ? "X" : "x", sourceX, useY ? "Y" : "y", sourceY, eventOut?"!":" ",stop?"STOP ":"     ");
            dctrl = data_in >> std_offset;

            //

            switch (m_componentType)
            {                
                case component_type.alu:

                    // opcode

                    switch ((UInt32)extr(dctrl, 36, 32))
                    {
                        case 0:  rString += "skip"; break;      
                        case 1:  rString += "move"; break;      
                        case 2:  rString += "cmpeq"; break;      
                        case 3:  rString += "extr"; break;      
                        case 4:  rString += "or"; break;      
                        case 5:  rString += "and"; break;      
                        case 6:  rString += "xor"; break;      
                        case 7:  rString += "bic"; break;      
                        case 8:  rString += "cmplo"; break;      
                        case 9:  rString += "cmplos"; break;      
                        case 10: rString += "add"; break;      
                        case 11: rString += "sub"; break;      
                        case 12: rString += "addt"; break;      
                        case 13: rString += "subt"; break;      
                        case 14: rString += "mul18"; break;      
                        case 15: rString += "muls18"; break;      
                        case 16: rString += "ffom"; break;      
                        case 17: rString += "min"; break;      
                        case 18: rString += "max"; break;      
                        case 19: rString += "mins"; break;      
                        case 20: rString += "maxs"; break;
                        default: rString += "illegal"; break;      	 
                    }
                                                       
                    // cc

                    switch ((UInt32)extr(dctrl, 41, 40))
                    {
                        case 0: rString += ".ts"; break;
                        case 1: rString += ".tc"; break;
                        case 2: rString += ".ev"; break;
                        case 3: rString += ".ne"; break;
                    }
                    rString += (" t" + (UInt32)extr(dctrl, 57) + " ");
                  
                    // opcode shift

                    if (detailed)
                    {
                        rString += " shift opcode ='";
                    }
                    else
                    {
                        rString += " so=";
                    }

                    switch ((UInt32)extr(dctrl, 39, 37))
                    {
                        case 0: rString += "nop"; break;
                        case 1: rString += "lsl"; break;
                        case 2: rString += "lsr"; break;
                        case 3: rString += "asr"; break;
                        case 4: rString += "sxt"; break;
                        case 5: rString += "ext"; break;
                    }

                    if (detailed)
                    {
                        rString += "' calc src ='";
                    }
                    else
                    {
                        rString += " cs=";
                    }

                    switch ((UInt32)extr(dctrl, 43, 42))
                    {
                        case 0: rString += "const"; break;
                        case 1: rString += "x"; break;
                        case 2: rString += "y"; break;
                    }

                    if (detailed)
                    {
                        rString += "' shift src ='";
                    }
                    else
                    {
                        rString += " ss=";
                    }

                    switch ((UInt32)extr(dctrl, 44, 44))
                    {
                        case 0: rString += "const"; break;
                        case 1: rString += "x"; break;
                    }

                    if (detailed)
                    {
                        rString += string.Format("' IMM[{0:X8}]", (UInt32)extr(dctrl, 31, 0));
                    }
                    else
                    {
                        rString += string.Format(" ${0:X8}", (UInt32)extr(dctrl, 31, 0));
                    }

                    break;
                case component_type.los:                  
                case component_type.buf:


                    // opcode

                    switch ((UInt32)extr(dctrl, 3, 0))
                    {
                        case 0: rString += "skip        "; break;
                        case 1: rString += "ld          "; break;
                        case 2: rString += "st          "; break;
                        case 3: rString += "ldw         "; break;
                        case 4: rString += "stw         "; break;
                        case 5: rString += "set_adrh    "; break;
                        case 6: rString += "set_adrl_msk"; break;
                        case 7: rString += "set_add     "; break;
                        case 8: rString += "set_base    "; break;
                        default: rString+= "illegal     "; break;
                    }

                    rString += string.Format("' IMM[{0:X8}]", (UInt32)extr(dctrl, 40, 9));

                    rString += string.Format(" setup internal counter = {0} use internal '{1}' enable internal calc '{2}'", (UInt32)extr(dctrl, 7, 6), (extr(dctrl, 5) == 1) ? "yes" : "no", ((UInt32)extr(dctrl, 8) == 1) ? "yes" : "no");

                    break;
            }

            return rString;
        }

        public component_type componentType
        {
            get { return m_componentType; }
            set { m_componentType = value; }
        }

        public UInt128 generate()
        {
            UInt128 rv = new UInt128();
            UInt128 dctrl = new UInt128();

            if (m_componentType == component_type.buf || m_componentType == component_type.los)
            {
                dctrl |= m_mainOpcode;
                dctrl |= (m_t_in_sel & 1) << 4;
                dctrl |= m_use_internal ? 1UL << 5 : 0;
                dctrl |= m_sel_internal << 6;
                dctrl |= m_en_internal ? 1UL << 8 : 0;
                dctrl |= (m_immval & 0xffffffffUL) << 9;

            }           
            else if (m_componentType == component_type.alu)
            {

                dctrl |= m_immval & 0xffffffff;
                dctrl |= m_mainOpcode << 32;
                dctrl |= m_sftOpcode << 37;

                switch (m_cc)
                {
                    case condition.ts: break;
                    case condition.tc: dctrl |= 1UL << 40; break;
                    case condition.ev: dctrl |= 2UL << 40; break;
                    case condition.ne: dctrl |= 3UL << 40; break;
                }

                switch (m_calcSrc)
                {
                    case calc_src.imm: break;
                    case calc_src.x: dctrl |= 1UL << 42; break;
                    case calc_src.y: dctrl |= 2UL << 42; break;

                }

                switch (m_shiftSrc)
                {
                    case shift_src.imm: break;
                    case shift_src.x: dctrl |= 1UL << 44; break;
                }

                dctrl |= (m_t_in_sel & 1) << 57;    // t sel

            }

            // std code

            rv |= m_data_valid_count & 0xffffffff;
            rv |= m_use_y ? 1UL << 32 : 0;
            rv |= m_use_x ? 1UL << 33 : 0;
            rv |= m_data_y_sel << 34;
            rv |= m_data_x_sel << 39;
            rv |= m_event_out << 44;
            rv |= m_stop? 1UL << 45:0;
            rv |= dctrl << std_offset;

            return rv;
        }

        public string trace(bool longFormat = true)
        {
            throw new NotImplementedException();
        }

        public Labels labelDictonary
        {
            get
            {
                throw new NotImplementedException();
            }
            set
            {
                throw new NotImplementedException();
            }
        }

        public bool noData()
        {
            throw new NotImplementedException();
        }

        public bool checkResolveImm()
        {
            throw new NotImplementedException();
        }

        public uint pc
        {
            get
            {
                throw new NotImplementedException();
            }
            set
            {
                throw new NotImplementedException();
            }
        }
      
        const string palu_p1    = @"^[\s]*alu.(?<OP1>[A-Za-z]+[A-Za-z0-9_]+)" +                  
                                  @"(" + 
                                  @"\.(?<CC>(ts|tc|ne|ev))\[(?<CCS>[%$A-Fa-f0-9]+)\]" +            
                                  @")?" +
                                  @"(?<EVENT>\!)?" +
                                  @"[\s]+" +
                                  @"(?<Y>[yY])\[(?<YSEL>[%$A-Fa-f0-9]+)\][\s]*,[\s]*((y)|(?<IMM>[\+\-%$A-Fa-f0-9]+)|((?<XP>[xX])\[(?<XPSEL>[%$A-Fa-f0-9]+)\]))" +
                                  @"([\s]+(?<OP2>[A-Za-z]+[A-Za-z0-9_]+)[\s]+((?<IMMSS>[\+\-%$A-Fa-f0-9]+)|(?<X>[xX])(\[(?<XSEL>[%$A-Fa-f0-9]+)\])?))?" +
                                  @"[\s]*([\s]*\@(?<COUNT>[\+\-%$A-Fa-f0-9]+))?" +
                                  @"([\s]+(?<STOP>(stop|STOP)))?" +
                                  @"[\s]*(?<comment>;[^\n]*)?$";

        const string plosbuf_p1 = @"^[\s]*(?<TYPE>(buf|los)).(?<OP1>[A-Za-z]+[A-Za-z0-9_]+)(.(?<CC>(ue|u|e)))?" +
                                    @"(?<EVENT>\!)?" +
                                    @"[\s]*([\s]*\@(?<COUNT>[\+\-%$A-Fa-f0-9]+))?" +
                                    @"([\s]+(?<STOP>(stop|STOP)))?" +
                                    @"[\s]*(?<comment>;[^\n]*)?$";  // op empty -> internal address gene

        const string plosbuf_p2 = @"^[\s]*(?<TYPE>(buf|los)).(?<OP1>[A-Za-z]+[A-Za-z0-9_]+)(.(?<CC>(ue|u|e)))?" +
                                  @"(?<EVENT>\!)?" +
                                  @"[\s]+((?<IN1>[dD])\[(?<IN1SEL>[%$A-Fa-f0-9]+)\])" +
                                  @"[\s]*([\s]*\@(?<COUNT>[\+\-%$A-Fa-f0-9]+))?" +
                                  @"([\s]+(?<STOP>(stop|STOP)))?" +
                                  @"[\s]*(?<comment>;[^\n]*)?$";  // op d[n] -> internal address gene

        const string plosbuf_p3 = @"^[\s]*(?<TYPE>(buf|los)).(?<OP1>[A-Za-z]+[A-Za-z0-9_]+)(.(?<CC>(ue|u|e)))?" +
                                  @"(?<EVENT>\!)?" +
                                  @"[\s]+((?<OUT>[aA])\[(?<OUTSEL>[%$A-Fa-f0-9]+)\])[\s]*,[\s]*" +
                                  @"((?<IN1>[dD])\[(?<IN1SEL>[%$A-Fa-f0-9]+)\])" +
                                  @"[\s]*([\s]*\@(?<COUNT>[\+\-%$A-Fa-f0-9]+))?" +
                                  @"([\s]+(?<STOP>(stop|STOP)))?" +
                                  @"[\s]*(?<comment>;[^\n]*)?$";  // op a[m],d[n] 

        const string plosbuf_p4 = @"^[\s]*(?<TYPE>(buf|los)).(?<OP1>[A-Za-z]+[A-Za-z0-9_]+)(.(?<CC>(ue|u|e)))?" +
                                  @"(?<EVENT>\!)?" +
                                  @"[\s]+((?<OUT>[aA])\[(?<OUTSEL>[%$A-Fa-f0-9]+)\])" +
                                  @"[\s]*([\s]*\@(?<COUNT>[\+\-%$A-Fa-f0-9]+))?" +
                                  @"([\s]+(?<STOP>(stop|STOP)))?" +
                                  @"[\s]*(?<comment>;[^\n]*)?$";  // op a[n] 

        const string plosbuf_p5 = @"^[\s]*(?<TYPE>(buf|los)).(?<OP1>[A-Za-z]+[A-Za-z0-9_]+)(.(?<CC>(ue|u|e)))?" +
                                  @"(?<EVENT>\!)?" +
                                  @"[\s]+(?<OUT>[xy])[\s]*,[\s]*" +
                                  @"(?<IN1IMM>[\+\-%$A-Fa-f0-9]+)" +
                                  @"[\s]*([\s]*\@(?<COUNT>[\+\-%$A-Fa-f0-9]+))?" +
                                  @"([\s]+(?<STOP>(stop|STOP)))?" +
                                  @"[\s]*(?<comment>;[^\n]*)?$";  // op x,$12345678

        const string plosbuf_p6 = @"^[\s]*(?<TYPE>(buf|los)).(?<OP1>[A-Za-z]+[A-Za-z0-9_]+)(.(?<CC>(ue|u|e)))?" +
                                  @"(?<EVENT>\!)?" +
                                  @"[\s]+" +
                                  @"(?<IN1IMM>[\+\-%$A-Fa-f0-9]+)" +
                                  @"[\s]*([\s]*\@(?<COUNT>[\+\-%$A-Fa-f0-9]+))?" +
                                  @"([\s]+(?<STOP>(stop|STOP)))?" +
                                  @"[\s]*(?<comment>;[^\n]*)?$";  // op $12345678

        enum fcorePatternType_t
        {
            alu,
            los_buf_op_empty,
            los_buf_op_d,
            los_buf_op_a_d,
            los_buf_op_a,
            los_buf_op_xy_imm32,
            los_buf_op_imm32
        }

        static List<parsePattern<fcorePatternType_t>> m_parsePatterns = null;

        static void initPatterns()
        {
            if (m_parsePatterns == null)
            {
                m_parsePatterns = new List<parsePattern<fcorePatternType_t>>();
                m_parsePatterns.Add(new parsePattern<fcorePatternType_t>(new Regex(palu_p1, RegexOptions.Compiled), fcorePatternType_t.alu));
                m_parsePatterns.Add(new parsePattern<fcorePatternType_t>(new Regex(plosbuf_p1, RegexOptions.Compiled), fcorePatternType_t.los_buf_op_empty));
                m_parsePatterns.Add(new parsePattern<fcorePatternType_t>(new Regex(plosbuf_p2, RegexOptions.Compiled), fcorePatternType_t.los_buf_op_d));
                m_parsePatterns.Add(new parsePattern<fcorePatternType_t>(new Regex(plosbuf_p3, RegexOptions.Compiled), fcorePatternType_t.los_buf_op_a_d));
                m_parsePatterns.Add(new parsePattern<fcorePatternType_t>(new Regex(plosbuf_p4, RegexOptions.Compiled), fcorePatternType_t.los_buf_op_a));
                m_parsePatterns.Add(new parsePattern<fcorePatternType_t>(new Regex(plosbuf_p5, RegexOptions.Compiled), fcorePatternType_t.los_buf_op_xy_imm32));
                m_parsePatterns.Add(new parsePattern<fcorePatternType_t>(new Regex(plosbuf_p6, RegexOptions.Compiled), fcorePatternType_t.los_buf_op_imm32));
            }
        }

        public bool parse_check(Line currentLine)
        {
             bool isOk = false;
             Match match;

             m_data_valid_count = 0;
             m_stop = false;

             foreach (parsePattern<fcorePatternType_t> pp in m_parsePatterns)
             {
                 match = pp.regex.Match(currentLine.value);
                 fcorePatternType_t m_patternType = pp.patternType;

                 // STD (c0,c1)


                 string validCount = match.Groups["COUNT"].Value;

                 if (validCount != "")
                 {
                     m_data_valid_count = (UInt32)Tools.extractInteger(validCount);
                 }

                 string stop = match.Groups["STOP"].Value;

                 if (stop.ToLower() == "stop")
                 {
                     m_stop = true;
                 }


                 // LOS | BUF

                 if (match.Success == true
                         && (  m_patternType == fcorePatternType_t.los_buf_op_empty
                            || m_patternType == fcorePatternType_t.los_buf_op_d
                            || m_patternType == fcorePatternType_t.los_buf_op_a_d
                            || m_patternType == fcorePatternType_t.los_buf_op_a
                            || m_patternType == fcorePatternType_t.los_buf_op_xy_imm32
                            || m_patternType == fcorePatternType_t.los_buf_op_imm32     ))
                 {
                      m_sel_internal = 0;
                      m_event_out = 0;
                    
                     string ctype = match.Groups["TYPE"].Value;
                     string op1 = match.Groups["OP1"].Value;
                     string sevent = match.Groups["EVENT"].Value;
                     string enabler = match.Groups["CC"].Value;
                    

                     if (ctype == "los")
                     {
                         m_componentType = component_type.los;

                         m_mainOpcode = checkLosOpcode(op1, out isOk);
                         if (!isOk)
                         {
                             throw new Exception("unknown main opcode '" + op1 + "'");
                         }

                     }
                     else if (ctype == "buf")
                     {
                         m_componentType = component_type.buf;

                         m_mainOpcode = checkBufOpcode(op1, out isOk);
                         if (!isOk)
                         {
                             throw new Exception("unknown main opcode '" + op1 + "'");
                         }

                     }
                     else
                     {
                         throw new Exception("unkonwon component type");
                     }

                       // event ena

                     if (sevent == "!")
                     {
                         m_event_out = 1;
                     }
                  
                      // enabler

                      m_use_internal = false;
                      m_en_internal = false;

                      switch(enabler)
                      {
                          case ""   : break;
                          case "ue" : m_use_internal = true; m_en_internal = true; break;
                          case "u"  : m_use_internal = true; break;
                          case "e"  : m_en_internal = true; break;
                          default: throw new Exception("unknwon enabler (ue?)");
                      }

                 }


                 if (match.Success == true
                     && m_patternType == fcorePatternType_t.los_buf_op_empty)
                 {
                     isOk = true;
                 }

                 if (match.Success == true
                    && m_patternType == fcorePatternType_t.los_buf_op_d)
                 {
                     losBufCheckDataSel(match.Groups["IN1"].Value, match.Groups["IN1SEL"].Value, out m_data_x_sel, out m_use_x);                   
                     isOk = true;
                 }

                 if (match.Success == true
                  && m_patternType == fcorePatternType_t.los_buf_op_a_d)
                 {
                     losBufCheckAdrSel(match.Groups["OUT"].Value, match.Groups["OUTSEL"].Value, out m_data_y_sel, out m_use_y);
                     losBufCheckDataSel(match.Groups["IN1"].Value, match.Groups["IN1SEL"].Value, out m_data_x_sel, out m_use_x);      
                     isOk = true;
                 }

                 if (match.Success == true
                 && m_patternType == fcorePatternType_t.los_buf_op_a)
                 {
                     losBufCheckAdrSel(match.Groups["OUT"].Value, match.Groups["OUTSEL"].Value, out m_data_y_sel, out m_use_y);
                     isOk = true;
                 }

                 if (match.Success == true
                && m_patternType == fcorePatternType_t.los_buf_op_xy_imm32)
                 {
                     string dout = match.Groups["OUT"].Value;
                     string din1imm = match.Groups["IN1IMM"].Value;

                     if (dout == "x")
                     {
                         m_sel_internal = 0;
                     }
                     else if (dout == "y")
                     {
                         m_sel_internal = 1;
                     }
                     else
                     {
                         throw new Exception("error in out selector (xy?)");
                     }
                     m_immval = (UInt64)Tools.extractInteger(din1imm);
                     isOk = true;
                 }

                 if (match.Success == true
               && m_patternType == fcorePatternType_t.los_buf_op_imm32)
                 {
                     string din1imm = match.Groups["IN1IMM"].Value;
                   
                     m_immval = (UInt64)Tools.extractInteger(din1imm);
                     isOk = true;
                 }
               
                 // ALU

                 if (match.Success == true && m_patternType == fcorePatternType_t.alu)
                 {
                     string mainopcode  = match.Groups["OP1"].Value;
                     string cc          = match.Groups["CC"].Value;
                     string ccs         = match.Groups["CCS"].Value;
                     string sevent = match.Groups["EVENT"].Value;
                     string y           = match.Groups["Y"].Value;
                     string ysel        = match.Groups["YSEL"].Value;

                     string xp          = match.Groups["XP"].Value;
                     string xpsel       = match.Groups["XPSEL"].Value;

                     m_data_x_sel = 0;
                     m_data_y_sel = 0;
                     m_use_x = false;
                     m_use_y = false;

                     // mainopcode

                     m_mainOpcode = checkMainOpcode(mainopcode, out isOk);
                     if (!isOk)
                     {
                         throw new Exception("unknown main opcode '" + mainopcode + "'");
                     }

                     // cc

                     m_cc = condition.ev;
                     if (cc != "")
                     {
                         switch (cc)
                         {
                             case "ts": m_cc = condition.ts; break;
                             case "tc": m_cc = condition.tc; break;
                             case "ev": m_cc = condition.ev; break;
                             case "ne": m_cc = condition.ne; break;
                             default:
                                 throw new Exception("unknown condition");
                         }

                         m_ccSel = (UInt64)Tools.extractInteger(ccs);

                     }

                     // event ena

                     if (sevent == "!")
                     {
                         m_event_out = 1;
                     }

                     // y check

                    aluCheckYSel(y, ysel, out m_data_y_sel, out m_use_y);                    
                     
                    m_sftOpcode = 0;
                    m_immval = 0;
                    m_calcSrc = calc_src.y;
                    m_shiftSrc = shift_src.imm;

                    string xpv = match.Groups["XP"].Value;
                    if (xpv != "")
                    {
                        string xsel = match.Groups["XPSEL"].Value;
                        aluCheckXSel(xpv, xsel, out m_data_x_sel, out m_use_x);   

                        m_shiftSrc = shift_src.x;
                    }

                    string imm = match.Groups["IMM"].Value;
                    if (imm != "")
                    {
                        m_calcSrc = calc_src.imm;
                        m_immval = (UInt64)Tools.extractInteger(imm);
                    }

                    // shift opcode ?

                    string shiftopcode = match.Groups["OP2"].Value;
                    if (shiftopcode != "")
                    {
                        m_sftOpcode = checkShiftOpcode(shiftopcode, out isOk);
                        if (!isOk)
                        {
                            throw new Exception("unknown shift opcode '" + m_sftOpcode + "'");
                        }
                                                         
                        string x = match.Groups["X"].Value;
                        if (x != "")
                        {
                            string xsel = match.Groups["XSEL"].Value;
                            if (xsel != "")
                            {
                                if (xpv != "")
                                {
                                    throw new Exception("error in x selector (no one defined)");
                                }
                                aluCheckXSel(x, xsel, out m_data_x_sel, out m_use_x);                       
                            }
                            
                            m_shiftSrc = shift_src.x;
                        }

                        string immss = match.Groups["IMMSS"].Value;
                        if (imm != "" && immss != "")
                        {
                            if (imm != immss)
                            {
                                throw new Exception("error if use two immidiate values these values has to be equal");
                            }
                        }

                        if (immss != "")
                        {
                            m_immval = (UInt64)Tools.extractInteger(immss);
                        }
                    }

                    isOk = true;
                 }
             }

             if (isOk == false)
             {
                 throw new Exception("illegal fcore command");
             }

             return isOk;
        }

        void losBufCheckDataSel(string mux,string mxu_sel,out UInt64 mux_sel,out bool mux_used)
        {
            checkSel("D", mux, mxu_sel, out mux_sel, out mux_used);
        }

        void losBufCheckAdrSel(string mux, string mxu_sel, out UInt64 mux_sel, out bool mux_used)
        {
            checkSel("A", mux, mxu_sel, out mux_sel, out mux_used);
        }

        void aluCheckYSel(string mux, string mxu_sel, out UInt64 mux_sel, out bool mux_used)
        {
            checkSel("Y", mux, mxu_sel, out mux_sel, out mux_used);
        }

        void aluCheckXSel(string mux, string mxu_sel, out UInt64 mux_sel, out bool mux_used)
        {
            checkSel("X", mux, mxu_sel, out mux_sel, out mux_used);
        }

        void checkSel(string selChar,string mux, string mxu_sel, out UInt64 mux_sel, out bool mux_used)
        {
            mux_sel = (UInt64)Tools.extractInteger(mxu_sel);
            mux_used = false;

            if (mux == selChar.ToUpper()) { }
            else if (mux == selChar.ToLower()) { mux_used = true; }
            else
            {
                throw new Exception("error in '"+selChar+"' selector");
            }
        }

    }
}
