﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using ucore_emulator.core.components.extMemory;
using ucore_assembler_library;
using xcore_tools;

namespace ucore_emulator.Load
{
    public class LoadRAW : ILoad
    {
        public string name()
        {
            return "LoadRAW";
        }



        public bool load(IPrint printInterface, string filename, core.components.memory.IMemIO mem, out int loadLength, bool byteSwap)
        {
            loadLength = 0;

            try
            {
                FileStream fs = File.OpenRead(filename);
                BinaryReader br = new BinaryReader(fs);

                int length = (int)fs.Length;

                for (uint i = 0; i < length / 2; i++)
                {
                    UInt16 data = br.ReadUInt16();
                    if(byteSwap)
                    {
                        data = (UInt16)((data >> 8) | (data << 8));
                    }

                    mem.write(i, data);
                }
                   
                printInterface.print("Load " + length / 2 + " words (16 Bit)\n");              

                loadLength = length / 2;

                br.Close();
                fs.Close();
            }
            catch (Exception ex)
            {
                printInterface.print(ex.Message);
                return false;
            }

            return true;
        }

        public bool load(IPrint printInterface, string filename, IExtMemory mem, out int loadLength, uint addressOffset)
        {
            loadLength = 0;

            try
            {
                FileStream fs = File.OpenRead(filename);
                BinaryReader br = new BinaryReader(fs);

                int length = (int)fs.Length;

                for (uint i = 0; i < length / 2; i++)
                {
                    UInt16 din = br.ReadUInt16();
                    din = (UInt16) ((din >> 8) | (din << 8));
                    mem.write(i + addressOffset, din);
                }

                printInterface.print("Load " + length / 2 + " words (16 Bit) into external device\n");

                loadLength = length / 2;

                br.Close();
                fs.Close();
            }
            catch (Exception ex)
            {
                printInterface.print(ex.Message);
                return false;
            }

            return true;
        }
    }
}
