﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace ucore_emulator.common
{
    public class ScanCodeTranslator
    {
        public ScanCodeTranslator()
        {
            m_codes = new Queue<byte>();
        }

        public void reset()
        {
            m_codes.Clear();
        }

        public bool hasData
        {
            get { return m_codes.Count != 0; }
        }

        public void addKey(byte b)
        {
          //  if (m_codes.Count <= 16)    // limit to 16 scan codes fifo like hardware
          //  {
                m_codes.Enqueue(b);
          //  }
        }

        public void addKeys(List<byte> lb)
        {
            foreach (byte b in lb)
            {
                addKey(b);
            }
        }

        public byte lookScan()
        {
            if (hasData)
            {
                return m_codes.First();
            }
            else
            {
                return 0xee;
            }
        }

        public byte popScan()
        {
            return m_codes.Dequeue();
        }

        public List<byte> translate(System.Windows.Forms.Keys k, bool pressed)
        {
            List<byte> rBytes = new List<byte>();
            switch (k)
            {

                case System.Windows.Forms.Keys.Back: if (pressed) { rBytes.Add(0x66); } else { rBytes.Add(0xf0); rBytes.Add(0x66); } break;
                case System.Windows.Forms.Keys.Tab: if (pressed) { rBytes.Add(0x0d); } else { rBytes.Add(0xf0); rBytes.Add(0x0d); } break;
                case System.Windows.Forms.Keys.Enter: if (pressed) { rBytes.Add(0x5A); } else { rBytes.Add(0xf0); rBytes.Add(0x5A); } break;
                case System.Windows.Forms.Keys.ShiftKey: if (pressed) { rBytes.Add(0x12); } else { rBytes.Add(0xf0); rBytes.Add(0x12); } break;
                case System.Windows.Forms.Keys.ControlKey: if (pressed) { rBytes.Add(0x14); } else { rBytes.Add(0xf0); rBytes.Add(0x14); } break;
                case System.Windows.Forms.Keys.Menu: /*alt*/if (pressed) { rBytes.Add(0x11); } else { rBytes.Add(0xf0); rBytes.Add(0x11); } break;
                case System.Windows.Forms.Keys.Pause: if (pressed) { rBytes.Add(0xe1); rBytes.Add(0x14); rBytes.Add(0x77); rBytes.Add(0xe1); rBytes.Add(0xf0); rBytes.Add(0x14); rBytes.Add(0xf0); rBytes.Add(0x77); } break;
                case System.Windows.Forms.Keys.CapsLock: if (pressed) { rBytes.Add(0x58); } else { rBytes.Add(0xf0); rBytes.Add(0x58); } break;
                case System.Windows.Forms.Keys.Escape: if (pressed) { rBytes.Add(0x76); } else { rBytes.Add(0xf0); rBytes.Add(0x76); } break;
                case System.Windows.Forms.Keys.Space: if (pressed) { rBytes.Add(0x29); } else { rBytes.Add(0xf0); rBytes.Add(0x29); } break;
                case System.Windows.Forms.Keys.PageUp: if (pressed) { rBytes.Add(0xe0); rBytes.Add(0x7d); } else { rBytes.Add(0xe0); rBytes.Add(0xf0); rBytes.Add(0x7d); } break;
                case System.Windows.Forms.Keys.PageDown: if (pressed) { rBytes.Add(0xe0); rBytes.Add(0x7a); } else { rBytes.Add(0xe0); rBytes.Add(0xf0); rBytes.Add(0x7a); } break;
                case System.Windows.Forms.Keys.End: if (pressed) { rBytes.Add(0xe0); rBytes.Add(0x69); } else { rBytes.Add(0xe0); rBytes.Add(0xf0); rBytes.Add(0x69); } break;
                case System.Windows.Forms.Keys.Home: if (pressed) { rBytes.Add(0xe0); rBytes.Add(0x6c); } else { rBytes.Add(0xe0); rBytes.Add(0xf0); rBytes.Add(0x6c); } break;
               
                case System.Windows.Forms.Keys.Left: if (pressed) { rBytes.Add(0xe0); rBytes.Add(0x6b); } else { rBytes.Add(0xe0); rBytes.Add(0xf0); rBytes.Add(0x6b); } break;
                case System.Windows.Forms.Keys.Up: if (pressed) { rBytes.Add(0xe0); rBytes.Add(0x75); } else { rBytes.Add(0xe0); rBytes.Add(0xf0); rBytes.Add(0x75); } break;
                case System.Windows.Forms.Keys.Right: if (pressed) { rBytes.Add(0xe0); rBytes.Add(0x74); } else { rBytes.Add(0xe0); rBytes.Add(0xf0); rBytes.Add(0x74); } break;
                case System.Windows.Forms.Keys.Down: if (pressed) { rBytes.Add(0xe0); rBytes.Add(0x72); } else { rBytes.Add(0xe0); rBytes.Add(0xf0); rBytes.Add(0x72); } break;
               
                case System.Windows.Forms.Keys.PrintScreen: if (pressed) { rBytes.Add(0xe0); rBytes.Add(0x12); rBytes.Add(0xe0); rBytes.Add(0x7c); } else { rBytes.Add(0xe0); rBytes.Add(0xf0); rBytes.Add(0x7c); rBytes.Add(0xe0); rBytes.Add(0xf0); rBytes.Add(0x12); } break;
                case System.Windows.Forms.Keys.Insert: if (pressed) { rBytes.Add(0xe0); rBytes.Add(0x70); } else { rBytes.Add(0xe0); rBytes.Add(0xf0); rBytes.Add(0x70); } break;
                case System.Windows.Forms.Keys.Delete: if (pressed) { rBytes.Add(0xe0); rBytes.Add(0x71); } else { rBytes.Add(0xe0); rBytes.Add(0xf0); rBytes.Add(0x71); } break;
               
                case System.Windows.Forms.Keys.D0: if (pressed) { rBytes.Add(0x45); } else { rBytes.Add(0xf0); rBytes.Add(0x45); } break;
                case System.Windows.Forms.Keys.D1: if (pressed) { rBytes.Add(0x16); } else { rBytes.Add(0xf0); rBytes.Add(0x16); } break;
                case System.Windows.Forms.Keys.D2: if (pressed) { rBytes.Add(0x1e); } else { rBytes.Add(0xf0); rBytes.Add(0x1e); } break;
                case System.Windows.Forms.Keys.D3: if (pressed) { rBytes.Add(0x26); } else { rBytes.Add(0xf0); rBytes.Add(0x26); } break;
                case System.Windows.Forms.Keys.D4: if (pressed) { rBytes.Add(0x25); } else { rBytes.Add(0xf0); rBytes.Add(0x25); } break;
                case System.Windows.Forms.Keys.D5: if (pressed) { rBytes.Add(0x2e); } else { rBytes.Add(0xf0); rBytes.Add(0x2e); } break;
                case System.Windows.Forms.Keys.D6: if (pressed) { rBytes.Add(0x36); } else { rBytes.Add(0xf0); rBytes.Add(0x36); } break;
                case System.Windows.Forms.Keys.D7: if (pressed) { rBytes.Add(0x3d); } else { rBytes.Add(0xf0); rBytes.Add(0x3d); } break;
                case System.Windows.Forms.Keys.D8: if (pressed) { rBytes.Add(0x3e); } else { rBytes.Add(0xf0); rBytes.Add(0x3e); } break;
                case System.Windows.Forms.Keys.D9: if (pressed) { rBytes.Add(0x46); } else { rBytes.Add(0xf0); rBytes.Add(0x46); } break;

                case System.Windows.Forms.Keys.A: if (pressed) { rBytes.Add(0x1c); } else { rBytes.Add(0xf0); rBytes.Add(0x1c); } break;
                case System.Windows.Forms.Keys.B: if (pressed) { rBytes.Add(0x32); } else { rBytes.Add(0xf0); rBytes.Add(0x32); } break;
                case System.Windows.Forms.Keys.C: if (pressed) { rBytes.Add(0x21); } else { rBytes.Add(0xf0); rBytes.Add(0x21); } break;
                case System.Windows.Forms.Keys.D: if (pressed) { rBytes.Add(0x23); } else { rBytes.Add(0xf0); rBytes.Add(0x23); } break;
                case System.Windows.Forms.Keys.E: if (pressed) { rBytes.Add(0x24); } else { rBytes.Add(0xf0); rBytes.Add(0x24); } break;
                case System.Windows.Forms.Keys.F: if (pressed) { rBytes.Add(0x2b); } else { rBytes.Add(0xf0); rBytes.Add(0x2b); } break;
                case System.Windows.Forms.Keys.G: if (pressed) { rBytes.Add(0x34); } else { rBytes.Add(0xf0); rBytes.Add(0x34); } break;
                case System.Windows.Forms.Keys.H: if (pressed) { rBytes.Add(0x33); } else { rBytes.Add(0xf0); rBytes.Add(0x33); } break;
                case System.Windows.Forms.Keys.I: if (pressed) { rBytes.Add(0x43); } else { rBytes.Add(0xf0); rBytes.Add(0x43); } break;
                case System.Windows.Forms.Keys.J: if (pressed) { rBytes.Add(0x3b); } else { rBytes.Add(0xf0); rBytes.Add(0x3b); } break;
                case System.Windows.Forms.Keys.K: if (pressed) { rBytes.Add(0x42); } else { rBytes.Add(0xf0); rBytes.Add(0x42); } break;
                case System.Windows.Forms.Keys.L: if (pressed) { rBytes.Add(0x4b); } else { rBytes.Add(0xf0); rBytes.Add(0x4b); } break;
                case System.Windows.Forms.Keys.M: if (pressed) { rBytes.Add(0x3a); } else { rBytes.Add(0xf0); rBytes.Add(0x3a); } break;
                case System.Windows.Forms.Keys.N: if (pressed) { rBytes.Add(0x31); } else { rBytes.Add(0xf0); rBytes.Add(0x31); } break;
                case System.Windows.Forms.Keys.O: if (pressed) { rBytes.Add(0x44); } else { rBytes.Add(0xf0); rBytes.Add(0x44); } break;
                case System.Windows.Forms.Keys.P: if (pressed) { rBytes.Add(0x4d); } else { rBytes.Add(0xf0); rBytes.Add(0x4d); } break;
                case System.Windows.Forms.Keys.Q: if (pressed) { rBytes.Add(0x15); } else { rBytes.Add(0xf0); rBytes.Add(0x15); } break;
                case System.Windows.Forms.Keys.R: if (pressed) { rBytes.Add(0x2d); } else { rBytes.Add(0xf0); rBytes.Add(0x2d); } break;
                case System.Windows.Forms.Keys.S: if (pressed) { rBytes.Add(0x1b); } else { rBytes.Add(0xf0); rBytes.Add(0x1b); } break;
                case System.Windows.Forms.Keys.T: if (pressed) { rBytes.Add(0x2c); } else { rBytes.Add(0xf0); rBytes.Add(0x2c); } break;
                case System.Windows.Forms.Keys.U: if (pressed) { rBytes.Add(0x3c); } else { rBytes.Add(0xf0); rBytes.Add(0x3c); } break;
                case System.Windows.Forms.Keys.V: if (pressed) { rBytes.Add(0x2a); } else { rBytes.Add(0xf0); rBytes.Add(0x2a); } break;
                case System.Windows.Forms.Keys.W: if (pressed) { rBytes.Add(0x1d); } else { rBytes.Add(0xf0); rBytes.Add(0x1d); } break;
                case System.Windows.Forms.Keys.X: if (pressed) { rBytes.Add(0x22); } else { rBytes.Add(0xf0); rBytes.Add(0x22); } break;
                case System.Windows.Forms.Keys.Y: if (pressed) { rBytes.Add(0x35); } else { rBytes.Add(0xf0); rBytes.Add(0x35); } break;
                case System.Windows.Forms.Keys.Z: if (pressed) { rBytes.Add(0x1a); } else { rBytes.Add(0xf0); rBytes.Add(0x1a); } break;

                case System.Windows.Forms.Keys.NumPad0: if (pressed) { rBytes.Add(0x70); } else { rBytes.Add(0xf0); rBytes.Add(0x70); } break;
                case System.Windows.Forms.Keys.NumPad1: if (pressed) { rBytes.Add(0x69); } else { rBytes.Add(0xf0); rBytes.Add(0x69); } break;
                case System.Windows.Forms.Keys.NumPad2: if (pressed) { rBytes.Add(0x72); } else { rBytes.Add(0xf0); rBytes.Add(0x72); } break;
                case System.Windows.Forms.Keys.NumPad3: if (pressed) { rBytes.Add(0x7a); } else { rBytes.Add(0xf0); rBytes.Add(0x7a); } break;
                case System.Windows.Forms.Keys.NumPad4: if (pressed) { rBytes.Add(0x6b); } else { rBytes.Add(0xf0); rBytes.Add(0x6b); } break;
                case System.Windows.Forms.Keys.NumPad5: if (pressed) { rBytes.Add(0x73); } else { rBytes.Add(0xf0); rBytes.Add(0x73); } break;
                case System.Windows.Forms.Keys.NumPad6: if (pressed) { rBytes.Add(0x74); } else { rBytes.Add(0xf0); rBytes.Add(0x74); } break;
                case System.Windows.Forms.Keys.NumPad7: if (pressed) { rBytes.Add(0x6c); } else { rBytes.Add(0xf0); rBytes.Add(0x6c); } break;
                case System.Windows.Forms.Keys.NumPad8: if (pressed) { rBytes.Add(0x75); } else { rBytes.Add(0xf0); rBytes.Add(0x75); } break;
                case System.Windows.Forms.Keys.NumPad9: if (pressed) { rBytes.Add(0x7d); } else { rBytes.Add(0xf0); rBytes.Add(0x7d); } break;


                case System.Windows.Forms.Keys.Multiply: if (pressed) { rBytes.Add(0x7c); } else { rBytes.Add(0xf0); rBytes.Add(0x7c); } break;
                case System.Windows.Forms.Keys.Add: if (pressed) { rBytes.Add(0x79); } else { rBytes.Add(0xf0); rBytes.Add(0x79); } break;
                case System.Windows.Forms.Keys.Separator: if (pressed) { rBytes.Add(0xe0); rBytes.Add(0x4a); } else { rBytes.Add(0xe0); rBytes.Add(0xf0); rBytes.Add(0x4a); } break;
                case System.Windows.Forms.Keys.Subtract: if (pressed) { rBytes.Add(0x7b); } else { rBytes.Add(0xf0); rBytes.Add(0x7b); } break;
                case System.Windows.Forms.Keys.Decimal: if (pressed) { rBytes.Add(0x71); } else { rBytes.Add(0xf0); rBytes.Add(0x71); } break;

                case System.Windows.Forms.Keys.F1: if (pressed) { rBytes.Add(0x05); } else { rBytes.Add(0xf0); rBytes.Add(0x05); } break;
                case System.Windows.Forms.Keys.F2: if (pressed) { rBytes.Add(0x06); } else { rBytes.Add(0xf0); rBytes.Add(0x06); } break;
                case System.Windows.Forms.Keys.F3: if (pressed) { rBytes.Add(0x04); } else { rBytes.Add(0xf0); rBytes.Add(0x04); } break;
                case System.Windows.Forms.Keys.F4: if (pressed) { rBytes.Add(0x0c); } else { rBytes.Add(0xf0); rBytes.Add(0x0c); } break;
                case System.Windows.Forms.Keys.F5: if (pressed) { rBytes.Add(0x03); } else { rBytes.Add(0xf0); rBytes.Add(0x03); } break;
                case System.Windows.Forms.Keys.F6: if (pressed) { rBytes.Add(0x0b); } else { rBytes.Add(0xf0); rBytes.Add(0x0b); } break;
                case System.Windows.Forms.Keys.F7: if (pressed) { rBytes.Add(0x83); } else { rBytes.Add(0xf0); rBytes.Add(0x83); } break;
                case System.Windows.Forms.Keys.F8: if (pressed) { rBytes.Add(0x0a); } else { rBytes.Add(0xf0); rBytes.Add(0x0a); } break;
                case System.Windows.Forms.Keys.F9: if (pressed) { rBytes.Add(0x01); } else { rBytes.Add(0xf0); rBytes.Add(0x01); } break;
                case System.Windows.Forms.Keys.F10: if (pressed) { rBytes.Add(0x09); } else { rBytes.Add(0xf0); rBytes.Add(0x09); } break;
                case System.Windows.Forms.Keys.F11: if (pressed) { rBytes.Add(0x78); } else { rBytes.Add(0xf0); rBytes.Add(0x78); } break;
                case System.Windows.Forms.Keys.F12: if (pressed) { rBytes.Add(0x07); } else { rBytes.Add(0xf0); rBytes.Add(0x07); } break;

                case System.Windows.Forms.Keys.NumLock: if (pressed) { rBytes.Add(0x77); } else { rBytes.Add(0xf0); rBytes.Add(0x77); } break;
                case System.Windows.Forms.Keys.Scroll: if (pressed) { rBytes.Add(0x7e); } else { rBytes.Add(0xf0); rBytes.Add(0x7e); } break;

                case System.Windows.Forms.Keys.Oemcomma: if (pressed) { rBytes.Add(0x41); } else { rBytes.Add(0xf0); rBytes.Add(0x41); } break;
                case System.Windows.Forms.Keys.OemPeriod: if (pressed) { rBytes.Add(0x49); } else { rBytes.Add(0xf0); rBytes.Add(0x49); } break;
                case System.Windows.Forms.Keys.OemMinus: if (pressed) { rBytes.Add(0x4e); } else { rBytes.Add(0xf0); rBytes.Add(0x4e); } break;
                case System.Windows.Forms.Keys.OemQuestion: if (pressed) { rBytes.Add(0x4a); } else { rBytes.Add(0xf0); rBytes.Add(0x4a); } break;

                case System.Windows.Forms.Keys.Oem1: if (pressed) { rBytes.Add(0x4c); } else { rBytes.Add(0xf0); rBytes.Add(0x4c); } break;
                case System.Windows.Forms.Keys.Oem7: if (pressed) { rBytes.Add(0x52); } else { rBytes.Add(0xf0); rBytes.Add(0x52); } break;

                case System.Windows.Forms.Keys.OemOpenBrackets: if (pressed) { rBytes.Add(0x54); } else { rBytes.Add(0xf0); rBytes.Add(0x54); } break;
                case System.Windows.Forms.Keys.Oem6: if (pressed) { rBytes.Add(0x5b); } else { rBytes.Add(0xf0); rBytes.Add(0x5b); } break;
            }
      
            return rBytes;
        }

        Queue<byte> m_codes; 
    }
}
