﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_emulator.core.components.memory;

namespace ucore_emulator.core.components.extMemory
{
    public class SDRAM : IExtMemory, IClocked
    {
        UInt16[] m_mem;

        int m_size = 0;
        int m_sizeMask;

        clearMode_t m_clearMode = clearMode_t.RANDOM;

        public SDRAM(int width = 26) // 32 megaword
        {
            m_size = 1 << width;
            m_sizeMask = m_size - 1;
            m_mem = new UInt16[m_size];

            reset();
        }

        public void setClearMode(clearMode_t mode)
        {
            m_clearMode = mode;
        }

        public void write(uint adr, UInt16 value)
        {
            m_mem[adr & m_sizeMask] = value;
        }

        public void read(uint adr, out UInt16 value)
        {
            value = m_mem[adr & m_sizeMask];
        }


        public void write(UInt32 adr, UInt32 value)
        {
            m_mem[(adr) & m_sizeMask] = (UInt16)(value >> 16);
            m_mem[((adr) + 1) & m_sizeMask] = (UInt16)value;
        }

        public void read(UInt32 adr, out UInt32 value)
        {
            value = (UInt32)m_mem[((adr & 0xfffffffe) + 1) & m_sizeMask];
            value |= ((UInt32)m_mem[((adr & 0xfffffffe)) & m_sizeMask]) << 16;
        }

        public void reset()
        {
            Random rnd = new Random();
            for (int i = 0; i < m_size; i++)
            {
                if (m_clearMode == clearMode_t.RANDOM)
                {
                    m_mem[i] = (UInt16)rnd.Next();
                }
                else 
                {
                    m_mem[i] = 0xffff;
                }
            }
        }

        public void step()
        {

        }

        public int size
        {
            get
            {
                return m_size;
            }
        }
    }
}
