﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_emulator.core.components.extMemory;
using ucore_emulator.core.components.memory;
using ucore_assembler_library;
using xcore_tools;

namespace ucore_emulator.core.components
{
    public class ext_bus : IExtMemory, IClocked
    {
        IPrint m_iprint;

        SRAM m_sram = null;
        SDRAM m_sdram = null;
        Flash m_flash = null;
        sdCard_ctrl m_sdCard = null;


        public ext_bus(IPrint iprint)
        {
            m_iprint = iprint;

            m_sram = new SRAM();
            m_sdram = new SDRAM();
            m_flash = new Flash();
            m_sdCard = new sdCard_ctrl(iprint,this);
        }

        public SRAM sram
        {
            get { return m_sram; }
        }

        public SDRAM dsram
        {
            get { return m_sdram; }
        }

        public sdCard_ctrl sdCard
        {
            get { return m_sdCard; }
        }

        public Flash flash
        {
            get { return m_flash; }
        }

        public void write(UInt32 adr, ushort value)
        {
            if ((adr >> 28) == 0xf) // sram
            {              
                m_sram.write(adr, value);
            }
            else if ((adr >> 28) == 0xd) // sdram
            {
                m_sdram.write(adr, value);
            }
            else if ((adr >> 28) == 0xe) // flash
            {
                m_flash.write(adr, value);
            }
            else if ((adr >> 28) == 0xa) // sdcard
            {
                m_sdCard.write(adr, value);
            }
            else
            {
                throw new Exception(string.Format("ext_bus: illegal write (16bit) access to ${0:X8}\n", adr));
            }
        }

        public void read(UInt32 adr, out UInt16 value)
        {
            if ((adr >> 28) == 0xf) // sram
            {
                m_sram.read(adr, out value);             
            }
            else if ((adr >> 28) == 0xd) // sdram
            {
                m_sdram.read(adr, out value);
            }
            else if ((adr >> 28) == 0xe) // flash
            {
                m_flash.read(adr,out value);
            }
            else if ((adr >> 28) == 0xa) // sdcard
            {
                m_sdCard.read(adr, out value);
            }
            else
            {
                throw new Exception(string.Format("ext_bus: illegal read (16bit) access from ${0:X8}\n",adr));
            }
        }

        public void write(UInt32 adr, UInt32 value)
        {
            if ((adr >> 28) == 0xf) // sram
            {
                m_sram.write(adr, value);
            }
            else if ((adr >> 28) == 0xd) // sdram
            {
                m_sdram.write(adr, value);
            }
            else if ((adr >> 28) == 0xe) // flash
            {
                m_flash.write(adr, value);
            }
            else if ((adr >> 28) == 0xa) // sdcard
            {
                m_sdCard.write(adr, value);
            }
            else
            {
                throw new Exception(string.Format("ext_bus: illegal write (32bit) access to ${0:X8}\n",adr));
            }
        }

        public void read(UInt32 adr, out UInt32 value)
        {
            if ((adr >> 28) == 0xf) // sram
            {
                m_sram.read(adr, out value);
            }
            else if ((adr >> 28) == 0xd) // sdram
            {
                m_sdram.read(adr, out value);
            }
            else if ((adr >> 28) == 0xe) // flash
            {
                m_flash.read(adr, out value);
            }
            else if ((adr >> 28) == 0xa) // sdcard
            {
                m_sdCard.read(adr, out value);
            }
            else
            {
                throw new Exception(string.Format("ext_bus: illegal read (32bit) access from ${0:X8}\n",adr));
            }
        }

        public void reset()
        {
            m_sdram.reset();
            m_sram.reset();
            m_flash.reset();
            m_sdCard.reset();
           
        }

        public void setClearMode(clearMode_t mode)
        {
            // nothing todo here
        }

        public void step()
        {
            m_sdCard.step();
        }

        public int size
        {
            get
            {
                return 0;
            }
        }
    }
}
