﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_emulator.core.components.stages_hcore;
using ucore_emulator.core.stages_hcore;
using xcore_tools;
using hcore_assembler_library.Opcode;

namespace ucore_emulator.core.components
{
   public class hcore : IClocked, IPrint
    {
        IPrint m_iprint;
        ext_bus m_bus;
        uint m_hcoreId;
        bool m_reset;
        public bool idle;
        uint m_total_core_count;

        //

        hcore_thread_ctrl m_thread_ctrl;
        hcore_ctrl m_hcore_ctrl;

        // stages

        
        s0_fetch s0_fetch;
        s1_score s1_score;
        s2_decode s2_decode;
        s2b_decode s2b_decode;
        s3_readReg s3_readReg;
        public s4_execute s4_execute;
        s5_merge s5_merge;
        s6_writeReg s6_writeReg;

        // dcache

        hDCache m_hDCache;

        // regs

        const uint regCount = 128;

        UInt32[] m_regs = new UInt32[128];
        bool[] m_regLocks = new bool[128];

        public UInt32 getRegValue(uint regNum)
        {
            return m_regs[regNum];
        }

        public bool getRegLock(uint regNum)
        {
            return m_regLocks[regNum];
        }

        public bool getRegOffsetEnabled(uint regNum)
        {
            return s2b_decode.m_offsetEnable[regNum];
        }

        // signals

        hc_s0_out_s1_in s1_current = new hc_s0_out_s1_in();
        hc_s1_out_s2_in s2_current = new hc_s1_out_s2_in();
        hc_s2_out_s2b_in s2b_current = new hc_s2_out_s2b_in();
        hc_s2b_out_s3_in s3_current = new hc_s2b_out_s3_in();
        hc_s3_out_s4_in s4_current = new hc_s3_out_s4_in();
        hc_s4_out_s5_in s5_current = new hc_s4_out_s5_in();
        hc_s5_out_s6_in s6_current = new hc_s5_out_s6_in();

        public hc_s0_out_s1_in s1_next = new hc_s0_out_s1_in();
        public hc_s1_out_s2_in s2_next = new hc_s1_out_s2_in();
        public hc_s2_out_s2b_in s2b_next = new hc_s2_out_s2b_in();
        public hc_s2b_out_s3_in s3_next = new hc_s2b_out_s3_in();
        public hc_s3_out_s4_in s4_next = new hc_s3_out_s4_in();
        public hc_s4_out_s5_in s5_next = new hc_s4_out_s5_in();
        public hc_s5_out_s6_in s6_next = new hc_s5_out_s6_in();

        //

        public hcore(ext_bus iext_bus, IPrint iprint, uint hcoreId, hcore_thread_ctrl thread_ctrl, uint total_core_count, hcore_ctrl hcore_ctrl)
        {
            m_iprint = iprint;         
            m_bus = iext_bus;
            m_hcoreId = hcoreId;
            m_reset = true;
            m_thread_ctrl = thread_ctrl;
            m_total_core_count = total_core_count;
            m_hcore_ctrl = hcore_ctrl;
            m_hDCache = new hDCache();

            // init stages

            s0_fetch = new s0_fetch(m_bus);
            s1_score = new s1_score();            
            s2_decode = new s2_decode();
            s2b_decode = new s2b_decode();
            s3_readReg = new s3_readReg(iprint,m_regs, m_regLocks);
            s4_execute = new s4_execute(m_bus, iprint, hcoreId, m_thread_ctrl, m_total_core_count, m_hcore_ctrl,m_hDCache);
            s5_merge = new s5_merge(m_iprint, m_hDCache);
            s6_writeReg = new s6_writeReg(iprint,m_regs, m_regLocks);
            

            //

            for (uint i = 0; i < regCount; i++)
            {
                m_regs[i] = 0;
                m_regLocks[i] = false;
            }
        }

        public void reset(bool reset)
        {
            m_reset = reset;
        }

        public void reset()
        {
            print("hcore[" + m_hcoreId + "] reset\n");
            reset(true);
            m_thread_ctrl.reset();

            s0_fetch.reset();
            s1_score.reset();
            s2_decode.reset();
            s2b_decode.reset();
            s3_readReg.reset();
            s4_execute.reset();
            s5_merge.reset();
            s6_writeReg.reset();
     
            s1_current.reset();
            s2_current.reset();
            s2b_current.reset();
            s3_current.reset();
            s4_current.reset();
            s5_current.reset();
            s6_current.reset();

            s1_next.reset();
            s2_next.reset();
            s2b_next.reset();
            s3_next.reset();
            s4_next.reset();
            s5_next.reset();
            s6_next.reset();

            m_hDCache.reset();

            //

            for (uint i = 0; i < regCount; i++)
            {
                m_regLocks[i] = false;
            }

        }

        public void step()
        {
            bool regLockStall = false;
            bool registerLock = false;
            bool dump = false;

            m_reset = m_hcore_ctrl.isHCoresReset();

            if (!idle)
            {
                s0_fetch.process(s5_current, out s1_next, m_reset); //
                s1_score.process(s1_current, out s2_next);
                s2_decode.process(s2_current, out s2b_next);
                s2b_decode.process(s2b_current, out s3_next, s5_current);
                s6_writeReg.process(s6_current);
                s3_readReg.process(s3_current, out s4_next,out regLockStall);
            }

            s4_execute.process(regLockStall, s4_current, out s5_next, m_reset, out idle, out registerLock, out dump);
            s5_merge.process(s5_current,out s6_next);

            if (dump)
            {
                printDump();

            }

            if (!idle)
            {
                s1_next.copy(ref s1_current, regLockStall);
                s2_next.copy(ref s2_current, regLockStall);
                s2b_next.copy(ref s2b_current, regLockStall);
                s3_next.copy(ref s3_current, regLockStall);
                s4_next.copy(ref s4_current, false);
                s5_next.copy(ref s5_current, regLockStall);
                s6_next.copy(ref s6_current,false);

            }
        }

        private void printDump()
        {
            StringBuilder dumpString = new StringBuilder();

            UInt32 value = 0;
            m_bus.dsram.read(s4_current.debug_pc, out value);

            HOpcode tOpcode = new HOpcode();
            tOpcode.decode(value);   // disassemble

            dumpString.Append(string.Format("dump at pc = {0:X8} '{1}' last = {2:X8} t = {3} ", s4_current.debug_pc, tOpcode.trace(), s4_execute.last, s4_execute.t ? 1 : 0));
            for (uint i = 0; i < 128; i++)
            {
                dumpString.Append(string.Format("r{0} = {1:X8}{2} ", i, m_regs[i], m_regLocks[i] ? "*" : " "));
            }
            dumpString.Append("\n");

            print(dumpString.ToString());
        }

        public void print(string txt)
        {
            m_iprint.print(txt);
        }

        public void dprint(string txt)
        {
            m_iprint.dprint(txt);
        }
    }
}
