﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_emulator.core.components.memory.ctrl;
using ucore_emulator.core.components.video;
using ucore_assembler_library;
using xcore_tools.Common;
using xcore_tools;

namespace ucore_emulator.core.components.memory
{
    public class CtrlMemory : bitExtraction, IMemIO, IClocked, IPrint
    {
        BoardCtrl m_bctrl;
        Timer32 m_timer1, m_timer2;
        ucore m_ucore;
        hcore[] m_hcores;
        emulator m_emulator;
        VideoComponent m_video;
        IPrint m_iPrint = null;

        UInt16  m_irq_msk;
        UInt16 m_irq_memory;

        // 

        bool m_identifyAsEmulator = true;

        // video store

        UInt32 lc_add_store;
        UInt32 MEM_STARTADR_store;
        UInt32 MEM_LINEOFFSET_store;
        UInt16 MEM_MODE_store; 		

        // i2c

        I2C m_i2cPortA;
        I2C m_i2cPortB;

        // audio

        UInt16 m_UcAudioDACLeftHighValue = 0;	
        UInt16 m_UcAudioDACLeftLowValue	= 0;		
        UInt16 m_UcAudioDACRightHighValue= 0;		
        UInt16 m_UcAudioDACRightLowValue= 0;

        Audio m_audio;

        // timer

        UInt16 m_timer1_tmp = 0;
        UInt16 m_timer2_tmp = 0;
        UInt32 m_timer1_stored_value = 0;
        UInt32 m_timer2_stored_value = 0;

        // keyboard

        bool m_keyBoardSwitchToLedSet = false;

        // hcore

        bool m_hcoreReset = true;
        uint m_hcoreSel = 0;
        UInt16 m_hcoreAdr = 0;
        UInt16 m_hcoreData = 0;

        // hcore io

        UInt32 m_hcoreRdIoData = 0;
        UInt32 m_hcoreRdCmdData = 0;
        UInt32 m_hcoreWrIoData = 0;
        UInt32 m_hcoreWrCmdData = 0;

        public CtrlMemory(BoardCtrl bctrl, VideoComponent video, IPrint iprint,emulator emulator)
        {
            m_emulator = emulator;
            m_iPrint = iprint;

            m_timer1 = new Timer32();
            m_timer2 = new Timer32();
            m_video = video;
            m_bctrl = bctrl;
            m_ucore = null;
            m_irq_msk = 0;
            m_irq_memory = 0;

            m_i2cPortA = new I2C(iprint);
            m_i2cPortB = new I2C(iprint);
            m_audio = new Audio(iprint);

            reset();
        }

        public Audio audio
        {
            get { return m_audio; }
        }

        public int size
        {
            get { return 64; }
        }

        public bool identifyAsEmulator
        {
            get { return m_identifyAsEmulator; }
            set { m_identifyAsEmulator = value; }
        }

        public UInt16 irqMsk
        {
            get { return m_irq_msk; }
        }

        public UInt16 irqMemory
        {
            get { return m_irq_memory; }
        }

        public Timer32 timer(int i)
        {
            switch(i)
            {
                case 0: return m_timer1;
                case 1: return m_timer2;
                default: return null;
            }
        }

        public void setUCore(ucore iucore)
        {
            m_ucore = iucore;
        }

        internal void setHCores(hcore[] hcores)
        {
            m_hcores = hcores;
        }

        public void write(uint adr, UInt16 value)
        {

            switch (adr & 0x7f)
            {
                case 0:

                    m_bctrl.set_gled((uint) value , 0x1ff);
                    m_bctrl.set_rled((uint) value >> 9, 0x3);
                    break;
                case 1:
                    m_bctrl.set_rled((uint)value, 0xffff);                                          
                    break;
                case 2:
                    m_bctrl.set_hex(0, (uint)value );
                    m_bctrl.set_hex(1, (uint)value >> 8);
                    break;
                case 3:
                    m_bctrl.set_hex(2, (uint)value );
                    m_bctrl.set_hex(3, (uint)value >> 8);
                    break;
                case 4:
                    m_bctrl.set_hex(4, (uint)value );
                    m_bctrl.set_hex(5, (uint)value >> 8);
                    break;
                case 5:
                    m_bctrl.set_hex(6, (uint)value);
                    m_bctrl.set_hex(7, (uint)value >> 8);
                    break;

                case 8: m_irq_msk = value; break;
                case 0x09: m_irq_memory = (UInt16)(value);
                            break;
                case 0xa: m_timer1_tmp = value;  break;
                case 0xb: m_timer1.setHighCount(value); m_timer1.setLowCount(m_timer1_tmp); break;
                case 0xc:   m_timer1.setTimerStart((value & 1) != 0);
                            m_timer1.setTimerStop((value & 2) != 0);
                            m_timer1.setTimerAutoReload((value & 4) != 0);
                            if ((value & 8) != 0)
                            {
                                m_timer1_stored_value = m_timer1.last_count;
                            }
                            break;

                case 0xd: m_timer2_tmp = value; break;
                case 0xe: m_timer2.setHighCount(value); m_timer2.setLowCount(m_timer2_tmp); break;
                case 0xf: m_timer2.setTimerStart((value & 1) != 0);
                            m_timer2.setTimerStop((value & 2) != 0);
                            m_timer2.setTimerAutoReload((value & 4) != 0);
                            if ((value & 8) != 0)
                            {
                                m_timer2_stored_value = m_timer1.last_count;
                            }
                            break;



                case 0x10 : m_video.enableVideo((value&1)!=0); break; 
				case 0x11 :	m_video.HSYNC = value;break;
				case 0x12 :	m_video.HSTART 					= value;break;
				case 0x13 :	m_video.HMEMSTART 				= value;break;
				case 0x14 :	m_video.HSTOP 					= value;break;
				case 0x15 :	m_video.HTOTAL 					= value;break;
				case 0x16 :	m_video.VSYNC 					= value;break;
				case 0x17 :	m_video.VSTART 					= value;break;
				case 0x18 :	m_video.VSTOP 					= value;break;
				case 0x19 :	m_video.VTOTAL 					= value;break;
				case 0x1a :	lc_add_store = (lc_add_store & 0xffff0000) | ((UInt32)value);break;
				case 0x1b :	lc_add_store = (lc_add_store & ~0xffff0000) | (((UInt32)value) << 16);break;
				case 0x1c :	m_video.lc_add 					= lc_add_store;break;
                case 0x1d : m_video.lc_start = value; break;
                case 0x1e: m_video.PVCOUNT = (UInt16) (value & 0x7ff); break;
						
				case 0x20 :	m_video.HLOADSTART 				= value;break;
				case 0x21 :	m_video.HLOADSTOP				= value;break;
				case 0x22 :	m_video.VLOADSTARTt 			= value;break;
				case 0x23 :	m_video.VLOADSTOP 				= value;break;
				case 0x24 :	m_video.VLOADNEXT				= value;break;
						      
				case 0x25 :	MEM_STARTADR_store = (MEM_STARTADR_store & 0xffff0000) | ((UInt32)value);break;
				case 0x26 :	MEM_STARTADR_store = (MEM_STARTADR_store & ~0xffff0000) | (((UInt32)value) << 16);break;
				case 0x27 :	MEM_LINEOFFSET_store = (MEM_LINEOFFSET_store & 0xffff0000) | ((UInt32)value);break;
				case 0x28 : MEM_LINEOFFSET_store = (MEM_LINEOFFSET_store & ~0xffff0000) | (((UInt32)value) << 16);break;

				case 0x29 :	MEM_MODE_store 					= value;break;
						      
				case 0x2a :	m_video.MEM_STARTADR 	= MEM_STARTADR_store; 
							m_video.MEM_LINEOFFSET  = MEM_LINEOFFSET_store;
							m_video.MEM_MODE 		= MEM_MODE_store;
                            break;

                case 0x30:
                            if (((UInt32)value & 0x0100) != 0)
                            {
                                if (m_bctrl.scanCodeTranslator.hasData == true)
                                {
                                    m_bctrl.scanCodeTranslator.popScan();
                                }
                            }

                            if (((UInt32)value & 0x0200) != 0)
                            {
                                if (m_keyBoardSwitchToLedSet)
                                {
                                    m_bctrl.setKeyBoardStatus(BoardCtrl.keyBoardStatus.KS_SCROLL, (value & 1) !=0 );
                                    m_bctrl.setKeyBoardStatus(BoardCtrl.keyBoardStatus.KS_NUM, (value & 2) != 0);
                                    m_bctrl.setKeyBoardStatus(BoardCtrl.keyBoardStatus.KS_CAPS, (value & 4) != 0);

                                    m_bctrl.scanCodeTranslator.addKey(0xfa);    // send ack

                                    m_keyBoardSwitchToLedSet = false;
                                }
                                else
                                {
                                    if ((value & 0xff) == 0xed)
                                    {
                                        m_keyBoardSwitchToLedSet = true;
                                        m_bctrl.scanCodeTranslator.addKey(0xfa);    // send ack
                                    }
                                    else
                                    {
                                        m_bctrl.scanCodeTranslator.addKey(0xfc);    // send wtf
                                    }
                                }                                
                            }

                            break;

                case 0x32:
                            {
                                string message = "";

                                if ((value & 0x100) != 0)
                                {
                                    message += " ADA" + ((value) & 1);
                                }

                                if ((value & 0x200) != 0)
                                {
                                    message += " ACA" + ((value >> 1) & 1);
                                }

                                if (message != "")
                                {
                                    print(message + "\n");
                                }
                            }
                            break;

                case 0x33:
                            {
                                //string message = "";

                                if ((value & 0x300) != 0)
                                {
                                    m_i2cPortB.write((value & 0x2) != 0,(value & 0x1) != 0);
                                }

                              /*  if ((value & 0x100) != 0)
                                {
                                    message += " BDA" + ((value) & 1);
                                }

                                if ((value & 0x200) != 0)
                                {
                                    message += " BCA" + ((value >> 1) & 1);
                                }

                                if (message != "")
                                {
                                    print(message + "\n");
                                }*/
                            }
                            break;

                // audio 

                case 0x34:  m_UcAudioDACLeftHighValue = value; break;
                case 0x35:  m_UcAudioDACLeftLowValue = value; break;
                case 0x36:  m_UcAudioDACRightHighValue = value; break;
                case 0x37:  m_UcAudioDACRightLowValue = value; break;
                case 0x38:
                    {
                        UInt64 wdata = 0;
                        wdata |= ((UInt64)m_UcAudioDACLeftHighValue) << 48;
                        wdata |= ((UInt64)m_UcAudioDACLeftLowValue) << 40;
                        wdata |= ((UInt64)m_UcAudioDACRightHighValue) << 24;
                        wdata |= ((UInt64)m_UcAudioDACRightLowValue) << 16;
                        wdata |= ((UInt64)value);

                        m_audio.pushDacFifo(wdata);
                    }
                    break;

                /* -----------------------------------------------------------------------------------------------------------
                   -- hcores
                   ----------------------------------------------------------------------------------------------------------- */
				
                case 0x3c:
                    m_hcoreSel = (UInt16) (value);
                   
                    break;

                case 0x3d:
                    m_hcoreAdr = value;
                    break;

                case 0x3e:
                    m_hcoreData = value;
                    break;

                case 0x3f:
                    {
                        ctrl_struct cs = new ctrl_struct(m_hcoreSel, m_hcoreAdr, m_hcoreData);                     
                        m_emulator.hcore_ctrl.push(cs);
                    }
                    break;

                /*  -----------------------------------------------------------------------------------------------------------
                    -- hcores io
                    ----------------------------------------------------------------------------------------------------------- */

                case 0x40: m_hcoreWrIoData = (m_hcoreWrIoData & 0xffff0000U) | ((UInt32) value);break;
                case 0x41: m_hcoreWrIoData = (m_hcoreWrIoData & 0x0000ffffU) | (((UInt32) value) << 16);break;
                case 0x42: m_hcoreWrCmdData = (m_hcoreWrCmdData & 0xffff0000U) | ((UInt32)value); break;
                case 0x43: m_hcoreWrCmdData = (m_hcoreWrCmdData & 0x0000ffffU) | (((UInt32)value) << 16); break;

                case 0x44:
                    {
                        if ((value & 1) != 0) // write data
                        {
                            uint hcoreId = extr(m_hcoreWrCmdData, 30, 23);
                            if (hcoreId >= m_emulator.hcoreCount)
                            {
                                throw new Exception("uctrl: Error: send hcore data to an not existing hcore (" + hcoreId + ")!");
                            }

                            ucore_emulator.core.stages_hcore.s4_execute.ooFromUctrl toHCoreData = new ucore_emulator.core.stages_hcore.s4_execute.ooFromUctrl();
                            toHCoreData.data = m_hcoreWrIoData;

                            m_hcores[hcoreId].s4_execute.pushFifoFromUctrl(toHCoreData);
                           

                        }

                        if ((value & 2) != 0) // request data
                        {
                            hcore hc = m_hcores[0]; // to uctrl fifo is static so access the first hcore                        
                            if (hc.s4_execute.hasFifoToUctrlData == false)
                            {
                                throw new Exception("uctrl: Error: request hcore data from empty fifo!");
                            }

                            ucore_emulator.core.stages_hcore.s4_execute.ooToUctrl hdata = hc.s4_execute.popFifoToUctrlData();

                            m_hcoreRdIoData = hdata.data;
                            m_hcoreRdCmdData = 0;

                            m_hcoreRdCmdData |= hdata.doReturn ? 1U << 31 : 0U;
                            m_hcoreRdCmdData |= (hdata.id & 0xffU) << 23;
                            m_hcoreRdCmdData |= (hdata.address & 0xffU) << 8;
                            m_hcoreRdCmdData |= hdata.code;                          
                        }
                    }
                    break;

                default: break;
            }
        }

        public UInt16 read(uint adr)
        {
            UInt16 rValue = 0;
            switch (adr & 0x7f)
            {
                case 0: {
                            UInt32 swr = 0;
                            m_bctrl.get_sw(out swr);
                            rValue = (UInt16)swr;
                    }

                        break;
                case 1:
                        rValue = (UInt16)(m_identifyAsEmulator ? 1 : 0); 
                        break;

                case 8: rValue = m_irq_msk; break;
                case 9: rValue = m_irq_memory; break;

                case 0xa: rValue = (UInt16)m_timer1_stored_value; break;
                case 0xb: rValue = (UInt16)(m_timer1_stored_value>>16); break;

                case 0xd: rValue = (UInt16)m_timer2_stored_value; break;
                case 0xe: rValue = (UInt16)(m_timer2_stored_value >> 16); break;


                case 0x10: rValue = (UInt16)m_video.hcount; break;
                case 0x11: rValue = (UInt16)m_video.vcount; break;
                case 0x12: rValue = (UInt16)(m_video.pvcountHit == true ? 1 : 0); break;

                case 0x30:
                    rValue = (UInt16)( ((UInt16)m_bctrl.scanCodeTranslator.lookScan()) & 0xff);
                    rValue |= (UInt16) (m_bctrl.scanCodeTranslator.hasData == true ? 1 << 8 : 0);
                    rValue |= 0x0600;   // WR WF ever done on emulator
                    break;

                /*  -----------------------------------------------------------------------------------------------------------
				    -- hcores
					----------------------------------------------------------------------------------------------------------- */
					
	            case 0x3c:  rValue = (UInt16) m_emulator.hcoreCount;
                            break;
                case 0x3f:  rValue = (UInt16)0;
                            break;

                /*  -----------------------------------------------------------------------------------------------------------
                   -- hcores io
                   ----------------------------------------------------------------------------------------------------------- */

                case 0x40: rValue = (UInt16)m_hcoreRdIoData; break;
                case 0x41: rValue = (UInt16)(m_hcoreRdIoData>>16); break;
                case 0x42: rValue = (UInt16)m_hcoreRdCmdData; break;
                case 0x43: rValue = (UInt16)(m_hcoreRdCmdData >> 16); break;
                case 0x44: 
                    {
                        rValue = 0;

                        hcore hc = m_hcores[0]; // to uctrl fifo is static so access the first hcore 
                        if (hc.s4_execute.hasFifoToUctrlData == true)
                        {
                            rValue |= 1; // there are data
                        }                    
                    }
                    break;
                            

                default: break;
            }

            return rValue;
        }

        public void reset()
        {
            m_irq_msk = 0;
            m_irq_memory = 0;
            m_uctrlRequest = 0;
            lc_add_store = videoPackage.lc_add_default;
            MEM_LINEOFFSET_store = videoPackage.MEM_LINEOFFSET_default;
            MEM_MODE_store = videoPackage.MEM_MODE_default;
            MEM_STARTADR_store = videoPackage.MEM_STARTADR_default;
            
        }

        uint m_testUctrlFifoCount = 0;
        const uint mc_testUctrlFifoMax = 3U;

        UInt64 m_uctrlRequest = 0;

        public UInt16 irqCurrent
        {
            get 
            {
                UInt16 current = 0;

                // @todo: FIXME
                //  if (m_testUctrlFifoCount >= mc_testUctrlFifoMax) // dont test every time because lock of fifo will slow down
                {
                   m_testUctrlFifoCount = 0;

                   if (m_hcores != null)
                   {
                       bool uctrlDataAvailable = m_hcores[0].s4_execute.hasFifoToUctrlData;
                       if (uctrlDataAvailable)
                       {
                           m_uctrlRequest++;
                       }
                       current |= (UInt16)((uctrlDataAvailable) ? 16U : 0U);
                   }
                }
                m_testUctrlFifoCount++;

                  /* adc irq never occure on emulator ( 8 : 0 )*/
                current |= (UInt16)((m_audio.dacFifoIrq)    ? 4U : 0U);
                current |= (UInt16)((m_timer2.timerHit)     ? 2U : 0U);
                current |= (UInt16)((m_timer1.timerHit)     ? 1U : 0U);
  
                return current; 
            }
        }

        public void step()
        {
            m_timer1.step();
            m_timer2.step();
            m_audio.step();

            if (m_ucore != null)
            {
                
                if ((irqCurrent & m_irq_msk) != 0)
                {
                     m_irq_memory = (UInt16) (m_irq_memory | irqCurrent);
                     m_ucore.setIrq(true);

                }
            }
        }

        #region IPrint
        public void print(string txt)
        {
            if (m_iPrint != null)
            {
                m_iPrint.print(txt);
            }
        }

        public void dprint(string txt)
        {
            if (m_iPrint != null)
            {
                m_iPrint.dprint(txt);
            }
        }
           #endregion     
    }
}
