﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace ucore_emulator.core.components.stages.execute_extensions
{
    public class l2d_s0 : IStage
    {
        public const int eld_u_l        = 0;
        public const int eld_u_h 	    = 1;
        public const int eld_du_l	    = 2;			
        public const int eld_du_h	    = 3;					
        public const int eld_v_l	    = 4;		
        public const int eld_v_h	    = 5;				
        public const int eld_dv_l	    = 6;	
        public const int eld_dv_h	    = 7;		
        public const int eld_xmask	    = 8;
        public const int eld_ymask      = 9;
        public const int eld_xwidth     = 10;			 
        public const int eld_sadr_l	    = 11;
        public const int eld_sadr_h	    = 12;

        public l2d_s0_out_s1_in dataOut
        {
            get { return m_out; }
        }

        public UInt32 iu;
        public UInt32 iv;
        public UInt32 idu;
        public UInt32 idv;
        public UInt16 ixmask;
        public UInt16 iymask;
        public UInt16 ixwidth;
        public UInt32 isadr;

        l2d_s0_out_s1_in m_out = new l2d_s0_out_s1_in();

        public void setup(bool inwe, int wsel, UInt16 indata, int rsel, out UInt16 outdata)
        {
            // unclocked read read

            switch (wsel)
            {
                case eld_u_l: outdata = (UInt16)iu; break;
                case eld_u_h: outdata = (UInt16)(iu >> 16); break;
                case eld_v_l: outdata = (UInt16)iv; break;
                case eld_v_h: outdata = (UInt16)(iv >> 16); break;
                default: outdata = 0; break;
            }

            if (inwe == true)
            {
                switch (wsel)
                {
                    case eld_u_l: iu = (iu & 0xffff0000) | ((UInt32)indata); break;
                    case eld_u_h: iu = (iu & 0x0000ffff) | (((UInt32)indata) << 16); break;
                    case eld_du_l: idu = (idu & 0xffff0000) | ((UInt32)indata); break;
                    case eld_du_h: idu = (idu & 0x0000ffff) | (((UInt32)indata) << 16); break;
                    case eld_v_l: iv = (iv & 0xffff0000) | ((UInt32)indata); break;
                    case eld_v_h: iv = (iv & 0x0000ffff) | (((UInt32)indata) << 16); break;
                    case eld_dv_l: idv = (idv & 0xffff0000) | ((UInt32)indata); break;
                    case eld_dv_h: idv = (idv & 0x0000ffff) | (((UInt32)indata) << 16); break;
                    case eld_xmask: ixmask = indata; break;
                    case eld_ymask: iymask = indata; break;
                    case eld_xwidth: ixwidth = indata; break;
                    case eld_sadr_l: isadr = (isadr & 0xffff0000) | ((UInt32)indata); break;
                    case eld_sadr_h: isadr = (isadr & 0x0000ffff) | (((UInt32)indata) << 16); break;
                    default: break;
                }
            }
        }

        public void process(out l2d_s0_out_s1_in dout)
        {
            dout = m_out;
           
            //

            UInt32 x_tmp,y_tmp;

			x_tmp  = iu >> 16;
			y_tmp  = iv >> 16;

            x_tmp &= ixmask;
            y_tmp &= iymask;

            iu = iu + idu;
            iv = iv + idv;
                    
            dout.x = x_tmp;
            dout.y = y_tmp;						
			

        }

        public void reset()
        {
           
        }
    }
}
