﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_emulator.common;
using ucore_assembler_library;
using ucore_emulator.core.components.stages.execute_extensions;
using xcore_tools.Common;
using xcore_tools;

namespace ucore_emulator.core.components.stages
{   
    public class s4_execute : bitExtraction, IStage, IPrint
    {
        // 

        IPrint m_iprint;
        bool m_runtimeCheck = true;

        // extensions

        l2d m_l2d;

        public l2d l2d
        {
            get { return m_l2d; }
        }

        int l2d_rsel;
        int l2d_wsel;
        UInt16 l2d_outdata;

        // runtime

        public bool enableRuntimeCheck
        {
            set { m_runtimeCheck = value; }
            get { return m_runtimeCheck; }
        }

        //

        public UInt16 m_pc;
        public UInt16 m_stored_pc;
        public UInt16 m_sp;
        public UInt32 m_espage;
        public UInt16 m_mulHigh;
        public UInt16 m_storeOffset;
        public bool m_t;
        public bool m_v;
        public bool m_stored_t;
        public bool m_stored_v;
        public bool m_i;
        public UInt16 m_irq_start_vector;
        public Queue<UInt16> m_extLoadQueue = new Queue<UInt16>();

        // udiv extendsion

        public UInt32 udiv_qoutient;
        public UInt32 udiv_rest;
        public UInt32 udiv_divident;
        public UInt32 udiv_divisor;
        public UInt32 udiv_counter;

        //

        public bool m_enableCache = false;

        //

        public UInt64 m_irqCount;

        // special

        UInt16 m_emulatorSpecialSelect = 0;
        emulator_special m_emulator_special;

     
        //

        public s4_execute(IPrint iprint)
        {
            m_emulator_special = new emulator_special(iprint);
            m_l2d = new l2d(iprint);
            m_iprint = iprint;
            reset();
        }

        public void reset()
        {
            m_stored_pc = m_pc = 0x0000;
            m_irq_start_vector = 0x0010;
            m_sp = 0x8000;
            m_stored_t = m_t = m_stored_v = m_v = false;
            m_i = false;
            m_mulHigh = 0;
            m_irqCount = 0;
            m_storeOffset = 0;

            udiv_qoutient = 0; 
            udiv_rest = 0; 
            udiv_divident = 0; 
            udiv_divisor = 0; 
            udiv_counter = 0;

            m_extLoadQueue.Clear();
            m_emulator_special.reset();
        }

        public s4_out_s5_in dataOut
        {
            get { return m_out; }
        }

        s4_out_s5_in m_out = new s4_out_s5_in();
        ucore m_ucore;

        public void process(ucore ucore,s3_out_s4_in din, out s4_out_s5_in dout, ext_bus extBus, out bool dump)
        {
            dump = false;
            m_ucore = ucore;
          
            //

            dout = m_out;

            // debug

            dout.opcode = din.opcode;
            dout.debug_pc = din.debug_pc;

            ///////////////// non jump opcodes

            // default

            // reg write
				
			dout.alu_rc_we 	= din.reg_rc_we;
			dout.alu_rc		= din.reg_rc;
				
			// memory
				
			dout.mem_we 	= false;
			dout.mem_data	= din.reg_data_a;		// write data ever reg a	

            switch (din.opcode)
            {
                
                case opcode_t.ope_nop       : break;
                case opcode_t.ope_dump      : // only emulator
                                              dump = true; 
                                              break;
                case opcode_t.ope_dexti     : break;    // emaultor (hw will use nop from decoder)
			    case opcode_t.ope_movei     : dout.alu_data_c = (UInt16) (din.reg_data_a); break;
                case opcode_t.ope_moveih 	: dout.alu_data_c = (UInt16) (((din.reg_data_a & 0xff) << 8) | (din.reg_data_b & 0xff)); break;

                case opcode_t.ope_addi: dout.alu_data_c = (UInt16)add(din.reg_data_b, din.reg_data_a, false, out m_t,out m_v); break;
                case opcode_t.ope_add: dout.alu_data_c = (UInt16)add(din.reg_data_b, din.reg_data_a, false, out m_t, out m_v); break;
                case opcode_t.ope_addqi: dout.alu_data_c = (UInt16)add(din.reg_data_b, din.reg_data_a, false, out m_t, out m_v); break;
                case opcode_t.ope_addtqi: dout.alu_data_c = (UInt16)add(din.reg_data_b, din.reg_data_a, m_t, out m_t, out m_v); break;
                case opcode_t.ope_addt: dout.alu_data_c = (UInt16)add(din.reg_data_b, din.reg_data_a, m_t, out m_t, out m_v); break;

                case opcode_t.ope_subi: dout.alu_data_c = (UInt16)sub(din.reg_data_b, din.reg_data_a, true, out m_t, out m_v); break;
                case opcode_t.ope_sub: dout.alu_data_c = (UInt16)sub(din.reg_data_b, din.reg_data_a, true, out m_t, out m_v); break;
                case opcode_t.ope_subqi: dout.alu_data_c = (UInt16)sub(din.reg_data_b, din.reg_data_a, true, out m_t, out m_v); break;
                case opcode_t.ope_subtqi: dout.alu_data_c = (UInt16)sub(din.reg_data_b, din.reg_data_a, m_t, out m_t, out m_v); break;
                case opcode_t.ope_subt: dout.alu_data_c = (UInt16)sub(din.reg_data_b, din.reg_data_a, m_t, out m_t, out m_v); break;

                case opcode_t.ope_neg: dout.alu_data_c = (UInt16)(0 - (Int16)din.reg_data_a); break;
                case opcode_t.ope_negts: dout.alu_data_c = m_t ? (UInt16)(0 - (Int16)din.reg_data_a) : din.reg_data_a; break; 

                case opcode_t.ope_lsri      :
                case opcode_t.ope_lsr       :
                case opcode_t.ope_lsrqi     : dout.alu_data_c = (UInt16)(din.reg_data_b >> (din.reg_data_a & 0xf)); break;

                case opcode_t.ope_lsrts     : dout.alu_data_c = m_t ? (UInt16)(din.reg_data_b >> (din.reg_data_a & 0xf)) : din.reg_data_b; break;

                case opcode_t.ope_asri      :
                case opcode_t.ope_asr       :
                case opcode_t.ope_asrqi     : dout.alu_data_c = (UInt16)(((Int16)din.reg_data_b) >> (din.reg_data_a & 0xf)); break;

                case opcode_t.ope_asrts     : dout.alu_data_c = m_t ? (UInt16)(((Int16)din.reg_data_b) >> (din.reg_data_a & 0xf)) : din.reg_data_b; break;

                case opcode_t.ope_muli      :
                case opcode_t.ope_mulqi     :
                case opcode_t.ope_mul       : 
                                                {
                                                    UInt32 mulResult = ((UInt32)din.reg_data_b) * ((UInt32)din.reg_data_a);
                                                    dout.alu_data_c = (UInt16)mulResult;
                                                    m_mulHigh = (UInt16)(mulResult >> 16);
                                                }
                                                break;

                case opcode_t.ope_muls:
                                                {
                                                    Int32 a = (Int32)((Int16)din.reg_data_b);
                                                    Int32 b = (Int32)((Int16)din.reg_data_a);
                                                   // Int32 mulResult = ((Int32)din.reg_data_b) * ((Int32)din.reg_data_a); // UInt16 to Int32 have no signed extend ... wtf
                                                    Int32 mulResult = a * b;
                                                    dout.alu_data_c = (UInt16)mulResult;
                                                    m_mulHigh = (UInt16)(mulResult >> 16);
                                                }
                                                break;

                case opcode_t.ope_mults:        if (m_t)
                                                {
                                                    UInt32 mulResult = ((UInt32)din.reg_data_b) * ((UInt32)din.reg_data_a);
                                                    dout.alu_data_c = (UInt16)mulResult;
                                                    m_mulHigh = (UInt16)(mulResult >> 16);
                                                }
                                                else
                                                {
                                                    dout.alu_data_c = (UInt16)din.reg_data_b;
                                                }
                                                break;

                case opcode_t.ope_mulsts:        if (m_t)
                                                {
                                                    Int32 mulResult = ((Int32)din.reg_data_b) * ((Int32)din.reg_data_a);
                                                    dout.alu_data_c = (UInt16)mulResult;
                                                    m_mulHigh = (UInt16)(mulResult >> 16);
                                                }
                                                else
                                                {
                                                    dout.alu_data_c = (UInt16)din.reg_data_b;
                                                }
                                                break;

                case opcode_t.ope_gmulhi    : dout.alu_data_c = m_mulHigh; break;

                case opcode_t.ope_cmpeqi    :
                case opcode_t.ope_cmpeq     :   if (din.reg_data_b == din.reg_data_a)
                                                {
                                                    m_t = true;
                                                }
                                                else
                                                {
                                                    m_t = false;
                                                }
                                                break;

                case opcode_t.ope_cmploi:
                case opcode_t.ope_cmplo: if (din.reg_data_b < din.reg_data_a)
                                                {
                                                    m_t = true;
                                                }
                                                else
                                                {
                                                    m_t = false;
                                                }
                                                break;

                case opcode_t.ope_cmplosi:
                case opcode_t.ope_cmplos: if (((Int16)din.reg_data_b) < ((Int16)din.reg_data_a))
                                                {
                                                    m_t = true;
                                                }
                                                else
                                                {
                                                    m_t = false;
                                                }
                                                break;

                case opcode_t.ope_cmple: if (din.reg_data_b <= din.reg_data_a)
                                                {
                                                    m_t = true;
                                                }
                                                else
                                                {
                                                    m_t = false;
                                                }
                                                break;

                case opcode_t.ope_cmples: if (((Int16)din.reg_data_b) <= ((Int16)din.reg_data_a))
                                                {
                                                    m_t = true;
                                                }
                                                else
                                                {
                                                    m_t = false;
                                                }
                                                break;
                case opcode_t.ope_gpci: dout.alu_data_c = (UInt16) (m_pc + din.reg_data_a); break;
                case opcode_t.ope_extri: m_t = ((din.reg_data_b >> din.reg_data_a) & 1) == 1; break;
                case opcode_t.ope_movets: dout.alu_data_c = m_t==true? din.reg_data_b :din.reg_data_a ; break;
              
                case opcode_t.ope_and: dout.alu_data_c = (UInt16)(din.reg_data_b & din.reg_data_a); break;
                case opcode_t.ope_or: dout.alu_data_c = (UInt16)(din.reg_data_b | din.reg_data_a); break;
                case opcode_t.ope_xor: dout.alu_data_c = (UInt16)(din.reg_data_b ^ din.reg_data_a); break;
                case opcode_t.ope_bic: dout.alu_data_c = (UInt16)(din.reg_data_b & ~din.reg_data_a); break;

                case opcode_t.ope_andts: dout.alu_data_c = m_t ? (UInt16)(din.reg_data_b & din.reg_data_a) : din.reg_data_b; break;
                case opcode_t.ope_orts: if (m_t)
                    { 
                     dout.alu_data_c =  (UInt16)(din.reg_data_b | din.reg_data_a);
                }
                else
                {
                    dout.alu_data_c = din.reg_data_b;
                }
                    break;


                   // dout.alu_data_c =  m_t ?(UInt16)(din.reg_data_b | din.reg_data_a): din.reg_data_b; break;
                case opcode_t.ope_xorts: dout.alu_data_c =  m_t ?(UInt16)(din.reg_data_b ^ din.reg_data_a): din.reg_data_b; break;
                case opcode_t.ope_bicts: dout.alu_data_c = m_t ? (UInt16)(din.reg_data_b & ~din.reg_data_a) : din.reg_data_b; break;

                case opcode_t.ope_swptc:    if (m_t == false)
                                            {
                                                dout.alu_data_c = (UInt16)(((din.reg_data_a) << 8) | (din.reg_data_a >> 8));
                                            }
                                            else 
                                            {
                                                dout.alu_data_c = din.reg_data_a;
                                            }
                                            break;

                case opcode_t.ope_swp: dout.alu_data_c = (UInt16)(((din.reg_data_a) << 8) | (din.reg_data_a >> 8)); ;
                                            break;

                case opcode_t.ope_tnt:  m_t = !m_t;
                                        break;

                case opcode_t.ope_vtt: m_t = m_v;
                                        break;
              
                case opcode_t.ope_ld_pop: dout.alu_data_c = din.mem_q;break;

               case opcode_t.ope_ssto: m_storeOffset = din.reg_data_a; break;

                case opcode_t.ope_st :		dout.mem_data	  	= din.reg_data_a;
											dout.mem_address	= din.reg_data_b;
											dout.mem_we = true;
                                            break;

                case opcode_t.ope_stinc:    dout.mem_data = din.reg_data_a;
                                            dout.mem_address = din.reg_data_b;
                                            dout.mem_we = true;
                                            dout.alu_data_c = (UInt16)(din.reg_data_b + 1);
                                            break;

                case opcode_t.ope_stwo:     dout.mem_data = din.reg_data_a;
                                            dout.mem_address = (UInt16)(din.reg_data_b + m_storeOffset);
                                            dout.mem_we = true;
                                            break;

                case opcode_t.ope_extb:     if (extr(din.reg_data_a, 7) == 1)
                                            {
                                                dout.alu_data_c = (UInt16)(0xff00 | (din.reg_data_a & 0xff));
                                            }
                                            else
                                            {
                                                dout.alu_data_c = (UInt16)(din.reg_data_a & 0xff);
                                            }
                                            break;
        
                case opcode_t.ope_bffo:     {
                                                m_t = false;		// default
											    dout.alu_data_c	= 0x0010;	// default
													
											    for (int i = 15;i>=0;i--)
                                                {
                                                    if(extr(din.reg_data_a,i) == 1)
                                                    {
                                                        dout.alu_data_c	= (UInt16)i;
                                                        m_t = true;
                                                        break;
                                                    }
                                                
                                                }
                                            }
                                            break;

                case opcode_t.ope_bset:     dout.alu_data_c = (UInt16) (din.reg_data_b | (1 << (din.reg_data_a & 0xf)));
                                            break;

                case opcode_t.ope_bclr:     dout.alu_data_c = (UInt16)(din.reg_data_b & ~(1 << (din.reg_data_a & 0xf)));
                                            break;    

                case opcode_t.ope_push: 	dout.mem_data = din.reg_data_b;
											m_sp = (UInt16) (m_sp - 1);	
											dout.mem_address = m_sp;
											dout.mem_we = true;
                                            break;

                case opcode_t.ope_rqld  :
                case opcode_t.ope_rqldi	:	{
                                                bool m_dummy;
                                                dout.mem_address = add(din.reg_data_b, din.reg_data_a, false, out m_dummy, out m_dummy); 
                                            }
                                            break;
                case opcode_t.ope_rqpop	:	dout.mem_address	= m_sp;
											m_sp = (UInt16) (m_sp + 1);	
                                            break;

                case opcode_t.ope_poparqp:  dout.alu_data_c = din.mem_q;
                                            dout.mem_address	= m_sp;
											m_sp = (UInt16) (m_sp + 1);	
                                            break;

                case opcode_t.ope_rti	:   m_t	= m_stored_t;
                                            m_v = m_stored_v;
                                            break;


                case opcode_t.ope_cli: m_i = false; break;
                case opcode_t.ope_sei: m_i = true;break;

                case opcode_t.ope_getssr:	switch (din.reg_data_a & 0x7)
                                            {
                                                case 0: dout.alu_data_c = (UInt16)((m_stored_v ? 64 : 0) | (din.irq ? 4 : 0) | (m_stored_t ? 2 : 0) | (m_i ? 1 : 0)); break;
					                            case 1: dout.alu_data_c = m_stored_pc;break;
					                            case 2: dout.alu_data_c = m_irq_start_vector;break;
                                                case 3: dout.alu_data_c = 0;break;

	                                            // extension

                                                case 6: m_l2d.step(false, l2d_wsel, din.reg_data_b, l2d_rsel, out l2d_outdata, false);
                                                        dout.alu_data_c = l2d_outdata;
                                                        break;			                                          
                                                //

                                                default: throw new Exception("ope_getssr read from unknown special register");

                                            }
                                            break;

                case opcode_t.ope_setssr:	switch (din.reg_data_a & 0x7)
                                            {
                                                case 0: m_i = (din.reg_data_b & 1) == 1 ? true : false; break;
                                                case 1: m_stored_pc = din.reg_data_b; break;
                                                case 2: m_irq_start_vector = din.reg_data_b; break;
                                                case 3: m_emulatorSpecialSelect = din.reg_data_b;  break;

                                                // extension

												case 4:  l2d_rsel = (int) din.reg_data_b;break;
                                                case 5:  l2d_wsel = (int) din.reg_data_b;break;
                                                case 6:  m_l2d.step(true, l2d_wsel, din.reg_data_b, l2d_rsel, out l2d_outdata,false); break;

                                                case 7:  m_emulator_special.setData(m_emulatorSpecialSelect, din.reg_data_b); break;
                                                    
                                                //

                                                default: throw new Exception("ope_setssr write to unknown special register");
                                            }
		                                    break;
                // ucore1 extensions

                case opcode_t.ope_espage:   m_espage = ((UInt32)din.reg_data_b) << 16 | ((UInt32)din.reg_data_a);
                                            break;
              
                case opcode_t.ope_epushsadrl:   dout.mem_data = (UInt16) m_espage;
											    m_sp = (UInt16) (m_sp - 1);	
											    dout.mem_address = m_sp;
											    dout.mem_we = true;
                                                break;

                case opcode_t.ope_epushsadrh:   dout.mem_data = (UInt16)(m_espage>>16);
                                                m_sp = (UInt16)(m_sp - 1);
                                                dout.mem_address = m_sp;
                                                dout.mem_we = true;
                                                break;

                case opcode_t.ope_epopsadrl: m_espage = (m_espage & 0xffff0000) | ((UInt32)din.mem_q); break;
                case opcode_t.ope_epopsadrh: m_espage = (m_espage & 0xffff) | (((UInt32)din.mem_q) << 16); break;

                case opcode_t.ope_est:      externalBusWrite(extBus,m_espage + ((UInt32)din.reg_data_a), din.reg_data_b);                                          
                                            break;


                case opcode_t.ope_edrqldi:  {
                                               UInt16 rvalue; 
                                               UInt32 address = (((UInt32)din.reg_data_b) << 16) | ((UInt32)din.reg_data_a) + din.simm;
                                               extBus.read(address, out rvalue); // get value
                                               m_extLoadQueue.Enqueue(rvalue); // push in fifo
                                            }
                                            break;

                case opcode_t.ope_erqld:    // same behavior only different decoder
                case opcode_t.ope_erqldi:   { 
                                                UInt16 rvalue;
                                                uint address = (m_espage + ((UInt32)din.reg_data_a));
                                                extBus.read(address, out rvalue); // get value
                                                m_extLoadQueue.Enqueue(rvalue); // push in fifo
                                            }
                                            break;

                case opcode_t.ope_erqldb:
                                            {
                                                UInt16 rvalue;
                                                uint address = (m_espage + ((UInt32)din.reg_data_a >> 1));
                                                extBus.read(address, out rvalue); // get value                                                
                                                m_extLoadQueue.Enqueue(rvalue); // push in fifo
                                            }
                                            break;   
                      
                case opcode_t.ope_eld:      if (m_extLoadQueue.Count != 0)
                                            {
                                                dout.alu_data_c = m_extLoadQueue.Dequeue();                                               
                                            }
                                            else
                                            {
                                               // throw new Exception("ope_eld read  from empty fifo!");
                                            }
                                            break;

                // ucore_endlos extensions

                case opcode_t.ope_estts:    if (m_t == true)
                                            {
                                                externalBusWrite(extBus, m_espage + ((UInt32)din.reg_data_a), din.reg_data_b);                                              
                                            }
                                            break;

                case opcode_t.ope_getsp:    dout.alu_data_c = (UInt16)(m_sp + din.reg_data_a);
                                            break;

                case opcode_t.ope_setsp:    m_sp = din.reg_data_a;
                                            break;
             
                // udiv

                case opcode_t.ope_udivinit: udiv_qoutient = 0;
                                            udiv_rest = 0;
                                            udiv_counter = din.reg_data_a;
                                            break;

                case opcode_t.ope_udivgq:   dout.alu_data_c = (UInt16)udiv_qoutient;break;
                case opcode_t.ope_udivgr:   dout.alu_data_c = (UInt16)udiv_rest; break;
                case opcode_t.ope_udivgqh:  dout.alu_data_c = (UInt16)(udiv_qoutient>>16); break;
                case opcode_t.ope_udivgrh:  dout.alu_data_c = (UInt16)(udiv_rest>>16); break;	

                case opcode_t.ope_udivsdd:  udiv_divident = din.reg_data_a;break;
                case opcode_t.ope_udivsdv:  udiv_divisor  = din.reg_data_a;break;
                case opcode_t.ope_udivsddh: udiv_divident = (((UInt32)din.reg_data_a) << 16) | (udiv_divident& 0xffff); break;
                case opcode_t.ope_udivsdvh: udiv_divisor  = (((UInt32)din.reg_data_a) << 16) | (udiv_divisor & 0xffff); break;

                case opcode_t.ope_udivstep:     UInt32 udiv_temp = (udiv_rest*2) | extr(udiv_divident,(int)udiv_counter);
                                                if (udiv_temp >= udiv_divisor)
                                                {
                                                    udiv_temp -= udiv_divisor;
                                                    udiv_qoutient |= 1U << (int)udiv_counter;
                                                }
                                                udiv_rest = udiv_temp;
                                                udiv_counter--;                                             
                                                break;              

                case opcode_t.ope_illegal:
                                                throw new Exception("illegal instruction");

                // l2d extension

                case opcode_t.ope_exl2den       :   m_l2d.step(false, l2d_wsel, din.reg_data_b, l2d_rsel, out l2d_outdata,true);
                                                    break;
                case opcode_t.ope_exl2denerqld  :
                                                    {
                                                        UInt32 address = m_l2d.step(false, l2d_wsel, din.reg_data_b, l2d_rsel, out l2d_outdata,true);
                                                        //print(string.Format("{0:X8} ", address));
                                                        UInt16 rvalue;
                       
                                                        extBus.read(address, out rvalue); // get value                                                        
                                                        m_extLoadQueue.Enqueue(rvalue); // push in fifo                                                        
                                                    } break;
                case opcode_t.ope_exl2derqld:
                                                {
                                                    UInt32 address = m_l2d.step(false, l2d_wsel, din.reg_data_b, l2d_rsel, out l2d_outdata, false);
                                                    //print(string.Format("{0:X8} ", address));
                                                    UInt16 rvalue;

                                                    extBus.read(address, out rvalue); // get value                                                    
                                                    m_extLoadQueue.Enqueue(rvalue); // push in fifo                                                        
                                                } break;
                   
                default: break;
            }

            ///////////////// jump opcodes

            // default pc increment

			UInt16 next_pc = (UInt16)(m_pc + 1);											
															
			// exec
         
			switch( din.opcode)
            {

                case opcode_t.ope_jmpi 	:   next_pc = (UInt16) (din.reg_data_b + din.reg_data_a);break;
                case opcode_t.ope_jmpts:    if (m_t == true)
                                            {
                                                next_pc = (UInt16)(din.reg_data_a); 
                                            }
                                            break;
                case opcode_t.ope_jmptc:    if (m_t == false)
                                            {
                                                next_pc = (UInt16)(din.reg_data_a);
                                            }
                                            break;
                case opcode_t.ope_br: next_pc = (UInt16)(((din.reg_data_a << 20) >> 20) + m_pc); break;// (m_pc + exts(din.reg_data_a, 11)); break;
                case opcode_t.ope_brts  :
                                            if (m_t == true)
                                            {
                                                next_pc = (UInt16)(((din.reg_data_a << 20) >> 20) + m_pc); break;//(UInt16) (m_pc + exts(din.reg_data_a,11));
                                            }
                                            break;												
													
                case opcode_t.ope_brtc  :   if (m_t == false)
                                            {
                                                next_pc = (UInt16)(((din.reg_data_a << 20) >> 20) + m_pc); break;//(UInt16) (m_pc + exts(din.reg_data_a,11));
                                            }
                                            break;	
											
                case opcode_t.ope_rti	:   next_pc = m_stored_pc;break;
																										
				default: break;					
             }

             switch (din.opcode)
            {

                case opcode_t.ope_jmpi:
                case opcode_t.ope_jmpts:
                case opcode_t.ope_jmptc: 
                case opcode_t.ope_br: 
                case opcode_t.ope_brts:
                case opcode_t.ope_brtc: 
                case opcode_t.ope_rti: 

                    if (m_i == true && din.irq == true)
                    {
                        m_ucore.setIrq(false);

                        if (m_extLoadQueue.Count() > 0 && m_runtimeCheck) // there is someting in loadqueue this is very dangerous
                        {
                            print(string.Format("Critical runtime warning at pc ${0:X4}! Loadqueue count is {1} (>0) at current take interrupt! So every erqld.. to eld will have instable data!\n",m_pc,m_extLoadQueue.Count()));
                        }

                        m_stored_pc = next_pc;
                        m_stored_t = m_t;
                        m_stored_v = m_v;
                        next_pc = m_irq_start_vector;

                        // stats

                        m_irqCount++;
                    }

                break;

                default: break;
            }

             if (m_runtimeCheck)
             {
                 if (m_sp > 0x8000)
                 {
                     print(string.Format("Critical runtime warning at pc ${0:X4}! Stack pointer is too high {1:X4}!\n", m_pc, m_sp));
                 }
             }

			m_pc = next_pc;
            dout.pc = m_pc;
        }

        private UInt16 add(UInt16 a, UInt16 b, bool cin, out bool cout, out bool vtt)
        {
            UInt32 r = ((UInt32)a) + ((UInt32)b) + (UInt32)(cin ? 1 : 0);
            cout = r > 0xffff;

            vtt = (((a ^ b) & (b ^ r)) & 0x8000) != 0;  
            return (UInt16)r;
        }

        private UInt16 sub(UInt16 a, UInt16 b, bool cin, out bool cout, out bool vtt)
        {
            UInt32 r = ((UInt32)a) - ((UInt32)b) + (UInt32)(cin ? 1 : 0) - 1;
            cout = r <= 0xffff;

            vtt = (((a ^ b) & (b ^ r)) & 0x8000) !=0;  
            return (UInt16)r;
        }

        //

        void externalBusWrite(ext_bus extBus,UInt32 address, UInt16 data)
        {
            if (m_runtimeCheck == true)
            {
                foreach (KeyValuePair<UInt32, UInt32> space in m_emulator_special.m_readOnlyMap)
                {
                    if (address >= space.Key && address <= space.Value)
                    {
                        print(string.Format("Critical runtime warning at pc ${0:X4}! Access ${1:X4} in read only configurated space ${2:X4}:${3:X4}!\n", m_pc, address,space.Key, space.Value));
                    }
                }
            }

           extBus.write(address, data); // note: no fifo no stall in emulator implemented                       
        }

        //

        public void print(string txt)
        {
            m_iprint.print(txt);
        }

        public void dprint(string txt)
        {
            m_iprint.dprint(txt);
        }
    }
}
