﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_emulator.core.components.stages;

namespace ucore_emulator.core.components.stages_hcore
{
    // S0

    public class hc_s0_out_s1_in : IHStageSignals<hc_s0_out_s1_in>
    {
        public UInt32 address;
        public UInt32 data;

        public hc_s0_out_s1_in()
        {
            reset();
        }

        public UInt32 debug_pc
        {
            get { return address; }
        }

        public void copy(ref hc_s0_out_s1_in dout,bool stall)
        {
            if (stall == false)
            {
                dout.address = address;
                dout.data = data;
            }
        }

        public string trace()
        {
            return string.Format("-- S0:\nAddress: {0:X8} Data: {1:X8}\n", address,data);
        }

        public void reset()
        {
            address = 0xd0100000;
            data = 0xffffffff;
        }
    }

    // S1

    public class hc_s1_out_s2_in : IHStageSignals<hc_s1_out_s2_in>
    {
        public UInt32 data_in;

        // debug
        public UInt32 debug_pc;

        public hc_s1_out_s2_in()
        {
            reset();
        }

        public void copy(ref hc_s1_out_s2_in dout, bool stall)
        {
            if (stall == false)
            {
                dout.data_in = data_in;
                dout.debug_pc = debug_pc;
            }
        }

        public string trace()
        {
            return string.Format("-- S1:\nData: {0:X8}\n", data_in);
        }

        public void reset()
        {
            data_in = 0xffffffff;
            debug_pc = 0;
        }
    }

    // S2

    public class hc_s2_out_s2b_in : IHStageSignals<hc_s2_out_s2b_in>
    {
        public uint ra;
        public uint rb;
        public uint rc;
        public bool rc_we;
        public bool rc_lock;

        public uint imm;

        public bool a_as_reg;
        public bool b_as_reg;
        public bool c_as_reg;
        

        public hcondition_t condition;
        public hopcode_t opcode;

        // debug
        public UInt32 debug_pc;

        public hc_s2_out_s2b_in()
        {
            reset();
        }

        public void reset()
        {
            ra = 0;
            rb = 0;
            rc = 0;
            rc_we = false;
            rc_lock = false;
            imm = 0;

            a_as_reg = false;
            b_as_reg = false;
            c_as_reg = false;

            condition = hcondition_t.hcc_never;
            opcode = hopcode_t.hope_nop;

            debug_pc = 0;
        }

        public string trace()
        {
            return string.Format("-- S2:\nOpcode: {0}\n cc: {1} ra: {2}\nrb: {3}\nrc: {4}\nrc_we: {5}\nimm: {6}\n casr {7}\nbasr {8}\naasr {9}\n\n", opcode.ToString(), condition.ToString(), ra, rb, rc, rc_we, imm,c_as_reg,b_as_reg,a_as_reg);                
        }

        public void copy(ref hc_s2_out_s2b_in dout, bool stall)
        {
            if (stall == false)
            {
                dout.ra = ra;
                dout.rb = rb;
                dout.rc = rc;
                dout.rc_we = rc_we;
                dout.rc_lock = rc_lock;
                dout.imm = imm;
                dout.a_as_reg = a_as_reg;
                dout.b_as_reg = b_as_reg;
                dout.c_as_reg = c_as_reg;
                dout.condition = condition;
                dout.opcode = opcode;
                dout.debug_pc = debug_pc;
            }
        }
    }

    public class hc_s2b_out_s3_in : IHStageSignals<hc_s2b_out_s3_in>
    {
        public uint ra;
        public uint rb;
        public uint rc;
        public bool rc_we;
        public bool rc_lock;

        public uint imm;

        public bool a_as_reg;
        public bool b_as_reg;
        public bool c_as_reg;


        public hcondition_t condition;
        public hopcode_t opcode;

        // debug
        public UInt32 debug_pc;

        public hc_s2b_out_s3_in()
        {
            reset();
        }

        public void reset()
        {
            ra = 0;
            rb = 0;
            rc = 0;
            rc_we = false;
            rc_lock = false;
            imm = 0;

            a_as_reg = false;
            b_as_reg = false;
            c_as_reg = false;

            condition = hcondition_t.hcc_never;
            opcode = hopcode_t.hope_nop;

            debug_pc = 0;
        }

        public string trace()
        {
            return string.Format("-- S2b:\nOpcode: {0}\n cc: {1} ra: {2}\nrb: {3}\nrc: {4}\nrc_we: {5}\nimm: {6}\n casr {7}\nbasr {8}\naasr {9}\n\n", opcode.ToString(), condition.ToString(), ra, rb, rc, rc_we, imm, c_as_reg, b_as_reg, a_as_reg);
        }

        public void copy(ref hc_s2b_out_s3_in dout, bool stall)
        {
            if (stall == false)
            {
                dout.ra = ra;
                dout.rb = rb;
                dout.rc = rc;
                dout.rc_we = rc_we;
                dout.rc_lock = rc_lock;
                dout.imm = imm;
                dout.a_as_reg = a_as_reg;
                dout.b_as_reg = b_as_reg;
                dout.c_as_reg = c_as_reg;
                dout.condition = condition;
                dout.opcode = opcode;
                dout.debug_pc = debug_pc;
            }
        }
    }

    // S3

    public class hc_s3_out_s4_in : IHStageSignals<hc_s3_out_s4_in>
    {
        public uint rc_through_out;
        public bool rc_we_through_out;

        public uint imm_out;

        public bool a_as_reg_out;
        public bool b_as_reg_out;
        public bool c_as_reg_out;

        public hcondition_t condition_out;
        public hopcode_t opcode_out;

        public uint ra_data;
        public uint rb_data;

        // debug
        public UInt32 debug_pc;

        public hc_s3_out_s4_in()
        {
            reset();
        }

        public void reset()
        {
            opcode_out = hopcode_t.hope_nop;
            condition_out = hcondition_t.hcc_never;
        }

        public string trace()
        {
            return "-- S3:\n";
        }

        public void copy(ref hc_s3_out_s4_in dout, bool stall)
        {
            if (stall == false)
            {
                dout.rc_through_out = rc_through_out;
                dout.rc_we_through_out = rc_we_through_out;

                dout.imm_out = imm_out;

                dout.a_as_reg_out = a_as_reg_out;
                dout.b_as_reg_out = b_as_reg_out;
                dout.c_as_reg_out = c_as_reg_out;

                dout.condition_out = condition_out;
                dout.opcode_out = opcode_out;

                dout.ra_data = ra_data;
                dout.rb_data = rb_data;

                dout.debug_pc = debug_pc;
            }
           
        }
    }

    public class hc_s4_out_s5_in : IHStageSignals<hc_s4_out_s5_in>
    {
        public uint rc_data;
        public uint rc;
        public bool rc_we;
        public bool rc_unlock;
        
        // to decoder stage b
        public uint roffset;
        public uint en_roffset_num;
        public bool en_roffset_data;
        public bool en_roffset_en;	

        public uint pc_out;

        // debug
        public UInt32 debug_pc;

        public hc_s4_out_s5_in()
        {
            reset();
        }

        public void reset()
        {
            pc_out = 0xd0100000;
            roffset = 0;
            rc_we = false;
            rc_unlock = false;
        }

        public string trace()
        {
            return "-- S4:\n";
        }


        public void copy(ref hc_s4_out_s5_in dout, bool stall)
        {
           
            dout.rc_data = rc_data;
            dout.rc = rc;
            dout.rc_we = rc_we;
            dout.rc_unlock = rc_unlock;
            dout.roffset = roffset;
            dout.en_roffset_num = en_roffset_num;
            dout.en_roffset_data= en_roffset_data;
            dout.en_roffset_en = en_roffset_en;	

            if (stall == false)
            {
                dout.pc_out = pc_out;
                dout.debug_pc = debug_pc;
            }
        }
    }

    public class hc_s5_out_s6_in : IHStageSignals<hc_s5_out_s6_in>
    {
        public uint rc_data;
        public uint rc;
        public bool rc_we;
        public bool rc_unlock;

        public uint pc_out;

        // debug
        public UInt32 debug_pc;

        public hc_s5_out_s6_in()
        {
            reset();
        }

        public void reset()
        {
            pc_out = 0xd0100000;
        }

        public string trace()
        {
            return "-- S5:\n";
        }


        public void copy(ref hc_s5_out_s6_in dout, bool stall)
        {

            dout.rc_data = rc_data;
            dout.rc = rc;
            dout.rc_we = rc_we;
            dout.rc_unlock = rc_unlock;

            if (stall == false)
            {
                dout.pc_out = pc_out;
                dout.debug_pc = debug_pc;
            }
        }
    }

    public class hc_s6_out : IHStageSignals<hc_s6_out>
    {

        public hc_s6_out()
        {
            reset();
        }

        public void copy(ref hc_s6_out dout,bool stall)
        {
        }

        public string trace()
        {
            return "-- S6:\n";
        }

        public void reset()
        {           
        }
    }
}
