﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_emulator.core.components.stages;
using ucore_emulator.core.components.stages_hcore;
using xcore_tools.Common;

namespace ucore_emulator.core.stages_hcore
{
    class s2_decode : bitExtraction, IStage
    {
        public void reset()
        {

        }

        public hc_s2_out_s2b_in dataOut
        {
            get { return m_out; }
        }

        hc_s2_out_s2b_in m_out = new hc_s2_out_s2b_in();

        internal void process(components.stages_hcore.hc_s1_out_s2_in din, out components.stages_hcore.hc_s2_out_s2b_in dout)
        {
            dout = m_out;

            //debug

            dout.debug_pc = din.debug_pc;

            // register

            switch ((din.data_in >> 21) & 0x7)//extr(din.data_in,23,21))
            {
                case 0: 
                    dout.rc = (din.data_in >> 14) & 0x7f; 
                    dout.rb = (din.data_in >> 7) & 0x7f; 
                    dout.ra = (din.data_in) & 0x7f; 
                    dout.imm = 0; break;
                case 1: 
                    dout.rc = 0; 
                    dout.rb = (din.data_in >> 14) & 0x7f; 
                    dout.ra = (din.data_in >> 7) & 0x7f; 
                    dout.imm = (din.data_in) & 0x7f; break;
                case 2: 
                    dout.rc = (din.data_in >> 14) & 0x7f; 
                    dout.rb = 0; 
                    dout.ra = (din.data_in >> 7) & 0x7f; 
                    dout.imm = (din.data_in) & 0x7f; break;
                case 3: 
                    dout.rc = 0; 
                    dout.rb = 0; 
                    dout.ra = (din.data_in >> 14) & 0x7f; 
                    dout.imm = (din.data_in) & 0x3fff; break;
                case 4: 
                    dout.rc = (din.data_in >> 14) & 0x7f; 
                    dout.rb = (din.data_in >> 7) & 0x7f;
                    dout.ra = 0;
                    dout.imm = (din.data_in) & 0x7f; break;
                case 5: 
                    dout.rc = 0; 
                    dout.rb = (din.data_in >> 14) & 0x7f;
                    dout.ra = 0;
                    dout.imm = (din.data_in) & 0x3fff; break;
                case 6: 
                    dout.rc = (din.data_in >> 14) & 0x7f;
                    dout.rb = 0;
                    dout.ra = 0;
                    dout.imm = (din.data_in) & 0x3fff; break;
                case 7: 
                    dout.rc = 0; 
                    dout.rb = 0;
                    dout.ra = 0;
                    dout.imm = (uint)exts(extr(din.data_in, 20, 0), 20); break;
                default: break;
            }

            // abc as reg 

            switch ((din.data_in >> 21) & 0x7)//extr(din.data_in,23 , 21))
            {
                case 0: dout.a_as_reg = true; dout.b_as_reg = true; dout.c_as_reg = true; break;
                case 1: dout.a_as_reg = true; dout.b_as_reg = true; dout.c_as_reg = false; break;
                case 2: dout.a_as_reg = true; dout.b_as_reg = false; dout.c_as_reg = true; break;
                case 3: dout.a_as_reg = true; dout.b_as_reg = false; dout.c_as_reg = false; break;
                case 4: dout.a_as_reg = false; dout.b_as_reg = true; dout.c_as_reg = true; break;
                case 5: dout.a_as_reg = false; dout.b_as_reg = true; dout.c_as_reg = false; break;
                case 6: dout.a_as_reg = false; dout.b_as_reg = false; dout.c_as_reg = true; break;
                case 7: dout.a_as_reg = false; dout.b_as_reg = false; dout.c_as_reg = false; break;
                default: break;
            }

            // opcode / we

            dout.rc_we = false;
            dout.rc_lock = false;

            if (((din.data_in >> 30) & 3) != 3)
            {
                /* with cc */

                switch ((din.data_in >> 26) & 0x3f)//extr(din.data_in, 30, 26))
                {
                    case 0: dout.opcode = hopcode_t.hope_move; dout.rc_we = true; break;
                    case 1: dout.opcode = hopcode_t.hope_moveh; dout.rc_we = true; break;
                    case 2: dout.opcode = hopcode_t.hope_add; dout.rc_we = true; break;
                    case 3: dout.opcode = hopcode_t.hope_sub; dout.rc_we = true; break;
                    case 4: dout.opcode = hopcode_t.hope_cmplo; break;
                    case 5: dout.opcode = hopcode_t.hope_cmplos; break;
                    case 6: dout.opcode = hopcode_t.hope_cmple; break;
                    case 7: dout.opcode = hopcode_t.hope_cmples; break;
                    case 8: dout.opcode = hopcode_t.hope_cmpeq; break;
                    case 9: dout.opcode = hopcode_t.hope_jmp; break;
                    case 10: dout.opcode = hopcode_t.hope_br; break;
                    case 11: dout.opcode = hopcode_t.hope_gpc; dout.rc_we = true; break;
                    case 12: dout.opcode = hopcode_t.hope_and; dout.rc_we = true; break;
                    case 13: dout.opcode = hopcode_t.hope_or; dout.rc_we = true; break;
                    case 14: dout.opcode = hopcode_t.hope_xor; dout.rc_we = true; break;
                    case 15: dout.opcode = hopcode_t.hope_bic; dout.rc_we = true; break;
                    case 16: dout.opcode = hopcode_t.hope_mul; dout.rc_we = true; break;
                    case 17: dout.opcode = hopcode_t.hope_mulh; dout.rc_we = true; break;
                    case 18: dout.opcode = hopcode_t.hope_lsr; dout.rc_we = true; break;
                    case 19: dout.opcode = hopcode_t.hope_lsl; dout.rc_we = true; break;
                    case 20: dout.opcode = hopcode_t.hope_asr; dout.rc_we = true; break;

                    case 21: dout.opcode = hopcode_t.hope_st; break; // memory
                    case 22: dout.opcode = hopcode_t.hope_stw; break; // memory
                    case 23: dout.opcode = hopcode_t.hope_uexec; break; // uctrl/ucore
                    case 24: dout.opcode = hopcode_t.hope_rsub; dout.rc_we = true; break;
                    case 25: dout.opcode = hopcode_t.hope_stc; break; // memory / cache
                    case 26: dout.opcode = hopcode_t.hope_roe; break; 

                    case 45: dout.opcode = hopcode_t.hope_idle; break;
                    case 46: dout.opcode = hopcode_t.hope_dump; break;
                    case 47: dout.opcode = hopcode_t.hope_nop; break;

                    // 

                    default: dout.opcode = hopcode_t.hope_illegal; break;
                }

                 // conditions

                switch ((din.data_in >> 24) & 3)//extr(din.data_in, 25, 24))
                {
                    case 0: dout.condition = hcondition_t.hcc_ever; break;
                    case 1: dout.condition = hcondition_t.hcc_ts; break;
                    case 2: dout.condition = hcondition_t.hcc_tc; break;
                    case 3: dout.condition = hcondition_t.hcc_never; break;
                    default: break;
                }
            }
            else
            {
                // without cc
                switch ((din.data_in >> 24) & 0x3f)//extr(din.data_in, 29, 24))
                {


                    case 0: dout.opcode = hopcode_t.hope_ld; dout.rc_lock = true; break;  // memory
                    case 1: dout.opcode = hopcode_t.hope_ldw; dout.rc_lock = true; break;  // memory
                    case 2: dout.opcode = hopcode_t.hope_ldws; dout.rc_lock = true; break;  // memory

                    case 3: dout.opcode = hopcode_t.hope_gssr; dout.rc_we = true; break;
                    case 4: dout.opcode = hopcode_t.hope_sssr; break;

                    case 5: dout.opcode = hopcode_t.hope_uexecs; dout.rc_lock = true; break;  // uctrl/ucore
                    case 6: dout.opcode = hopcode_t.hope_ldc; dout.rc_lock = true; break;  // memory
                 

                    // 

                    default: dout.opcode = hopcode_t.hope_illegal; break;
                }

                // no selectable condition

                 dout.condition = hcondition_t.hcc_ever;
            }           
        }
    }
}
