﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_emulator.core.components.stages;
using ucore_emulator.core.components.stages_hcore;
using xcore_tools;

namespace ucore_emulator.core.stages_hcore
{
    class s3_readReg : IStage, IPrint
    {
        private uint[] m_regs;
        private bool[] m_regLocks;
        IPrint m_iprint;

        public s3_readReg(IPrint iprint,uint[] regs, bool[] regLocks)
        {
            m_iprint = iprint;
            m_regs = regs;
            m_regLocks = regLocks;
        }

        public hc_s3_out_s4_in dataOut
        {
            get { return m_out; }
        }

        hc_s3_out_s4_in m_out = new hc_s3_out_s4_in();

        public void reset()
        {
            for(int i=0;i<m_regLocks.Length;i++)
            {
                m_regLocks[i] = false;
            }
        }

        internal void process(components.stages_hcore.hc_s2b_out_s3_in din, out components.stages_hcore.hc_s3_out_s4_in dout, out bool regLockStall)
        {
            dout = m_out;

            // default

            regLockStall = false;

            // debug

            dout.debug_pc = din.debug_pc;

            // through

            dout.rc_through_out = din.rc;
            dout.rc_we_through_out = din.rc_we;

            dout.imm_out = din.imm;

            dout.a_as_reg_out = din.a_as_reg;
            dout.b_as_reg_out = din.b_as_reg;
            dout.c_as_reg_out = din.c_as_reg;

            dout.condition_out = din.condition;
            dout.opcode_out = din.opcode;

            // test lock

            if (din.rc_lock == true)
            {
                if (m_regLocks[din.rc] == true)
                {
                    regLockStall = true;
                    throw new Exception("s3_readReg: Error: Outoforder lock to and already locked register WriteToWrite Hazard");
                }
                else
                {
                  //  print(string.Format("Lock r{0} at pc {1:X8}\n",din.rc,din.debug_pc));
                    m_regLocks[din.rc] = true;
                }
            }

            // read regs

            dout.ra_data = m_regs[din.ra & 0x7f];
            dout.rb_data = m_regs[din.rb & 0x7f];

            // read locks and check

            if (
                (din.a_as_reg && m_regLocks[din.ra & 0x7f])||
                (din.b_as_reg && m_regLocks[din.rb & 0x7f])
                )

            {
              // print(string.Format("Run into Readstall at pc {0:X8}\n",din.debug_pc));

                // read register is locked ! stall stage before, skip stages after

                regLockStall = true;
                dout.condition_out = hcondition_t.hcc_ever;
                dout.opcode_out = hopcode_t.hope_skip;
                dout.rc_we_through_out = false;
            }

            
        }

        public void print(string txt)
        {
            m_iprint.print(txt);
        }

        public void dprint(string txt)
        {
            m_iprint.dprint(txt);
        }
    }
}
