﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_emulator.core.components.stages;
using ucore_assembler_library;
using System.IO;
using xcore_tools;

namespace ucore_emulator.core.components
{
    public class ucore : IClocked, IPrint
    {
        IPrint m_iprint;
        ucore_bus m_bus;
        ext_bus m_ext_bus;

        public UInt64 m_cycleCount = 0;
        public bool m_irq;
        int m_statistic = 0;
        UCoreAnalaystic m_analystic = new UCoreAnalaystic();
        double m_frequency = 106.484848E+6;

        s0_prepareFetch s0;
        s1_fetch s1;
        s2_decode s2;
        s3_readReg s3;
        s4_execute s4;
        s5_writeReg s5;

        m0_prepareLoad m0;
        m1_load m1;

        UInt16[] m_regs = new UInt16[8];

        public UCoreAnalaystic analystic
        {
            get { return m_analystic; }
        }

        public double frequency
        {
            get { return m_frequency; }
        }

        public int enableStatistic
        {
            set { m_statistic = value; }
            get { return m_statistic; }
        }

        public bool enableRuntimeCheck
        {
            set
            {
                if (s4 != null)
                {
                    s4.enableRuntimeCheck = value;
                }
            }

            get 
            {
                if (s4 != null)
                {
                    return s4.enableRuntimeCheck;
                }
                else
                {
                    return false;
                }
            }
        }

        public void setIrq(bool iirq)
        {
            m_irq = iirq;
        }

        #region debug_get_set

        public s0_prepareFetch stage_preParefetch
        {
            get { return s0; }
        }

        public s1_fetch stage_fetch
        {
            get { return s1;}
        }

        public s2_decode stage_decode
        {
            get { return s2;}
        }

        public s3_readReg stage_readReg
        {
            get { return s3;}
        }

        public s4_execute stage_execute
        {
            get { return s4; }
        }

        public s5_writeReg stage_writeReg
        {
            get { return s5; }
        }

        #endregion
     //   System.IO.StreamWriter file;

        public ucore(ucore_bus ibus,ext_bus iext_bus,IPrint iprint)
        {
            m_iprint = iprint;
            m_irq = false;
            m_bus = ibus;
            m_ext_bus = iext_bus;
            m_cycleCount = 0;

            //file = new System.IO.StreamWriter(@"f:\file.txt");

            //

            s0 = new s0_prepareFetch();
            s1 = new s1_fetch(m_bus);
            s2 = new s2_decode();
            s3 = new s3_readReg(m_regs);
            s4 = new s4_execute(iprint);
            s5 = new s5_writeReg(m_regs);

            m0 = new m0_prepareLoad();
            m1 = new m1_load(m_bus);
        }

        public ucore_bus bus
        {
            get { return m_bus; }
        }

        public ext_bus ext_bus
        {
            get { return m_ext_bus; }
        }

        public UInt16[] regs
        {
            get { return m_regs; }
        }

        //

       s0_out_s1_in s1_current = new s0_out_s1_in();
       s1_out_s2_in s2_current = new s1_out_s2_in();
       s2_out_s3_in s3_current = new s2_out_s3_in();
       s3_out_s4_in s4_current = new s3_out_s4_in();
       s4_out_s5_in s5_current = new s4_out_s5_in();
       m0_out_m1_in m1_current = new m0_out_m1_in();
        UInt16 m2_current = 0;

        public s0_out_s1_in s1_next = new s0_out_s1_in();
        public s1_out_s2_in s2_next = new s1_out_s2_in();
        public s2_out_s3_in s3_next = new s2_out_s3_in();
        public s3_out_s4_in s4_next = new s3_out_s4_in();
        public s4_out_s5_in s5_next = new s4_out_s5_in();
        public m0_out_m1_in m1_next = new m0_out_m1_in();
        public UInt16 m2_next;
       
      
        public void step()
        {
            bool dump;

            m0.process(s5_current, out m1_next);
            m1.process(m1_current, out m2_next, s5_current);

            s0.process(s5_current, out s1_next);
            s1.process(s1_current, out s2_next);
            s2.process(s2_current, out s3_next);
            s5.process(s5_current);
            s3.process(s3_current, out s4_next);

            s4_current.irq = m_irq;
            s4_current.mem_q = m2_current;
            s4.process(this, s4_current, out s5_next, m_ext_bus, out dump);

            if (dump) // emulator dump instruction
            {
                UInt16 edata = 0;
                //  m_ext_bus.read(s4.m_espage, out edata,true);
                string dumpMessage = string.Format("Dump: pc:{22:X4} pc_st:{23:X4} r0:{0:X4} r1:{1:X4} r2:{2:X4} r3:{3:X4} r4:{4:X4} r5:{5:X4} r6:{6:X4} r7:{7:X4} " +
                                                    "sp:{8:X4} esasdr:{9:X8}={10:X4} t:{11} i:{12} LoadQueueCnt:{13}    L2D U:{14:X8} V:{15:X8} DU:{16:X8} DV:{17:X8} XMSK:{18:X4} YMSK:{19:X4} WIDTH:{20:X4} SADR:{21:X4} IRQMEMORY:{24:X4} IRQCURRENT:{25:X4} IRQMSK:{26:X4}",
                                                                        m_regs[0], m_regs[1], m_regs[2], m_regs[3], m_regs[4], m_regs[5], m_regs[6], m_regs[7],
                                                                        s4.m_sp, s4.m_espage, edata, s4.m_t ? 1 : 0, s4.m_i ? 1 : 0, s4.m_extLoadQueue.Count(),
                                                                        s4.l2d.s0.iu,
                                                                        s4.l2d.s0.iv,
                                                                        s4.l2d.s0.idu,
                                                                        s4.l2d.s0.idv,
                                                                        s4.l2d.s0.ixmask,
                                                                        s4.l2d.s0.iymask,
                                                                        s4.l2d.s0.ixwidth,
                                                                        s4.l2d.s0.isadr,
                                                                        s4.m_pc,
                                                                        s4.m_stored_pc,
                                                                        m_bus.ctrlMemory.irqMemory,
                                                                        m_bus.ctrlMemory.irqCurrent,
                                                                        m_bus.ctrlMemory.irqMsk
                                                                        );
                print(dumpMessage + "\n");
            }

          /*  {
                string dumcpMessage = string.Format("r0:{0:X4} r1:{1:X4} r2:{2:X4} r3:{3:X4} r4:{4:X4} r5:{5:X4} r6:{6:X4} r7:{7:X4} sp:{8:X4} pc:{9:X4}",
                                                        m_regs[0], m_regs[1], m_regs[2], m_regs[3], m_regs[4], m_regs[5], m_regs[6], m_regs[7], s4.m_sp, s4.m_pc);
                file.WriteLine(dumcpMessage);
            }*/

            s1_next.copy(ref s1_current);
            s2_next.copy(ref s2_current);
            s3_next.copy(ref s3_current);
            s4_next.copy(ref s4_current);
            s5_next.copy(ref s5_current);
            m1_next.copy(ref m1_current);
            m2_current = m2_next;

            // statistic

            if (m_statistic != 0)
            {
                opcode_t decodedOpcode = s3_next.decode_opcode;

                m_analystic.increasOpcodeUsage(decodedOpcode);
             
                //

                if (decodedOpcode == opcode_t.ope_br ||
                    decodedOpcode == opcode_t.ope_brtc ||
                    decodedOpcode == opcode_t.ope_brts ||
                    decodedOpcode == opcode_t.ope_jmpts ||
                    decodedOpcode == opcode_t.ope_jmptc ||
                    decodedOpcode == opcode_t.ope_jmpi
                   )
                {
                    Int64 diff = (Int64) m_cycleCount;
                    diff -= (Int64) (m_analystic.lastBranchCycle == 0 ? m_cycleCount+4 : m_analystic.lastBranchCycle);

                    UInt64 diffAbs = (UInt64)( diff < 0 ? -diff : diff); 

                    m_analystic.branchCount++;

                    m_analystic.aprBranchDelay += diffAbs;
                    m_analystic.minBranchDelay = Math.Min(diffAbs, m_analystic.minBranchDelay);
                    m_analystic.maxBranchDelay = Math.Max(diffAbs, m_analystic.maxBranchDelay);
                    m_analystic.lastBranchCycle = m_cycleCount;
                }

                m_analystic.incrementLineUsage(s5_current.pc);
            }

            m_cycleCount++;
        }

        public void reset()
        {
            //m_cycleCount = 0;
            m_irq = false;

            s0.reset();
            s1.reset();
            s2.reset();
            s3.reset();
            s4.reset();
            s5.reset();

            m0.reset();
            m1.reset();

            s1_current.reset();
            s2_current.reset();
            s3_current.reset();
            s4_current.reset();
            s5_current.reset();
            m1_current.reset();

            s1_next.reset();
            s2_next.reset();
            s3_next.reset();
            s4_next.reset();
            s5_next.reset();
            m1_next.reset();

            m_cycleCount = 0;
            m_analystic.clear();
        }

        public void print(string txt)
        {
            m_iprint.print(txt);
        }

        public void dprint(string txt)
        {
            m_iprint.dprint(txt);
        }
    }
}
