﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace ucore_emulator.core.components.video
{
    public interface IVideoEvent
    {
        void verticalBlank();
    }

    public class videoPackage
    {
        // 1440x900x60 -- 720x450 -- 360x225 --

	    // video sync etc.
	
	    public const int  HSYNC_default      = 152-1;	
	    public const int  HSTART_default	= 384-1;
        public const int HMEMSTART_default = 382 - 1;
        public const int HSTOP_default = 1824 - 1;
        public const int HTOTAL_default = 1904 - 1;

        public const int VSYNC_default = 3 - 1;
        public const int VSTART_default = 31 - 1;
        public const int VSTOP_default = 931 - 1;
        public const int VTOTAL_default = 932 - 1;	
	
	    // external load

        public const int HLOADSTART_default = 8;		// external memory start
        public const int HLOADSTOP_default = 8 + 360;

        public const int VLOADSTART_default = 31 - 1;		// external memory start
        public const int VLOADSTOP_default = 931 - 1;
        public const int VLOADNEXT_default = 4;
	
	    // memory processing

        public const UInt32 MEM_STARTADR_default = 0xf0000000;		// external memory start
        public const UInt32 MEM_LINEOFFSET_default = 0;
        public const UInt16 MEM_MODE_default = 1;
	
	    // line cache

        public const UInt32 lc_add_default = 0x00400000;
        public const UInt32 lc_start_default = 0;
    }

    public class screenDefinitons
    {
        public screenDefinitons()
        {
            width = 1440;
            height = 900;
            scale = (float)(1.0 / 2.0);
        }

        public int width
        {
            get { return m_width; }
            set { m_width = value; }
        }

        public int height
        {
            get { return m_height; }
            set { m_height = value; }
        }

        public float scale
        {
            get { return m_scale; }
            set { m_scale = value; }
        }

        int m_width;
        int m_height;
        float m_scale;

    }

    public class VideoComponent : IClocked
    {
        screenDefinitons m_screenDefinitons = new screenDefinitons();
        IVideoEvent m_videoEvents = null;
        ext_bus m_extBus;
        bool m_videoEnabled = true;
        int m_vblank_v = 0;
        int m_vblank_h = 0;

        UInt32[] m_screenBuffer;
        UInt32[] m_lineCache = new UInt32[4096];


        public VideoComponent(IVideoEvent videoEvents,ext_bus extBus)
        { 
            m_videoEvents = videoEvents;
            m_screenBuffer = new UInt32[m_screenDefinitons.width * m_screenDefinitons.height];
            m_extBus = extBus;
            m_videoEnabled = true;
        }

        public void enableVideo(bool enable)
        {
            m_videoEnabled = enable;
        }

        public screenDefinitons screenDefinitons
        {
            get { return m_screenDefinitons; }
            set { m_screenDefinitons = value; }
        }
    
        public int visibleWidth()
        {
            return (HSTOP - HSTART) * (((int)lc_add) >> 16) / 256;
        }

        public int visibleHeight()
        {
            return (VSTOP - VSTART) / VLOADNEXT;
        }

       

        public int vblank_v
        {
            set { m_vblank_v = value; }
        }

        public int vblank_h
        {
            set { m_vblank_h = value; }
        }

        public void  reset()
        {

        }

        public void step2()
        {
          //  if (doBlank == true)
            // full speed blanking
            {
                /*doBlank = false;*/
                UInt32 startAdr = 0; ;
                UInt32 rvalue;
                UInt32 wvalue = 0;

                int xcount = visibleWidth();
                int ycount = visibleHeight();

                //

                if (m_videoEnabled == false)    // clear screen
                {
                    for (int y = 0; y < ycount; y++)
                    {
                        for (int x = 0; x < xcount; x++)
                        {
                            m_screenBuffer[x + y * xcount] = 0;
                        }
                    }
                }
                else
                {

                    // only copy visible data

                    for (int y = 0; y < ycount; y++)
                    {
                        for (int x = 0; x < xcount; )
                        {
                            m_extBus.read(startAdr + MEM_STARTADR, out rvalue);

                            switch (MEM_MODE)
                            {
                                case 0:
                                    m_screenBuffer[x + y * xcount] = rvalue; x++;
                                    break;
                                case 1:

                                    // 565
                                    //
                                    // 33322222222221111111111
                                    // 10987654321098765432109876543210
                                    // rrrrrggggggbbbbbrrrrrggggggbbbbb
                                    // 
                                    // 33322222222221111111111
                                    // 10987654321098765432109876543210
                                    // xxxxxxxxrrrrrrrrggggggggbbbbbbbb

                                    wvalue = 0;
                                    wvalue |= ((rvalue >> 27) & 0x1f) << (16 + 3);
                                    wvalue |= ((rvalue >> 21) & 0x3f) << (8 + 2);
                                    wvalue |= ((rvalue >> 16) & 0x1f) << (0 + 3);

                                    m_screenBuffer[x + y * xcount] = wvalue; x++;

                                    wvalue = 0;
                                    wvalue |= ((rvalue >> 11) & 0x1f) << (16 + 3);
                                    wvalue |= ((rvalue >> 5) & 0x3f) << (8 + 2);
                                    wvalue |= ((rvalue >> 0) & 0x1f) << (0 + 3);

                                    m_screenBuffer[x + y * xcount] = wvalue; x++;
                                    break;

                                case 2:
                                    // 555
                                    //
                                    //33322222222221111111111
                                    //10987654321098765432109876543210
                                    //-rrrrrgggggbbbbb-rrrrrgggggbbbbb

                                    wvalue = 0;
                                    wvalue |= ((rvalue >> 26) & 0x1f) << (16 + 3);
                                    wvalue |= ((rvalue >> 21) & 0x1f) << (8 + 3);
                                    wvalue |= ((rvalue >> 16) & 0x1f) << (0 + 3);

                                    m_screenBuffer[x + y * xcount] = wvalue; x++;

                                    wvalue = 0;
                                    wvalue |= ((rvalue >> 10) & 0x1f) << (16 + 3);
                                    wvalue |= ((rvalue >> 5) & 0x1f) << (8 + 3);
                                    wvalue |= ((rvalue >> 0) & 0x1f) << (0 + 3);

                                    m_screenBuffer[x + y * xcount] = wvalue; x++;
                                    break;

                                case 3:
                                    // 444
                                    //
                                    //33322222222221111111111
                                    //10987654321098765432109876543210
                                    //----rrrrggggbbbb----rrrrggggbbbb

                                    wvalue = 0;
                                    wvalue |= ((rvalue >> 24) & 0xf) << (16 + 4);
                                    wvalue |= ((rvalue >> 20) & 0xf) << (8 + 4);
                                    wvalue |= ((rvalue >> 16) & 0xf) << (0 + 4);

                                    m_screenBuffer[x + y * xcount] = wvalue; x++;

                                    wvalue = 0;
                                    wvalue |= ((rvalue >> 8) & 0xf) << (16 + 4);
                                    wvalue |= ((rvalue >> 4) & 0xf) << (8 + 4);
                                    wvalue |= ((rvalue >> 0) & 0xf) << (0 + 4);

                                    m_screenBuffer[x + y * xcount] = wvalue; x++;
                                    break;

                            }

                            startAdr += 2;
                        }
                        startAdr += MEM_LINEOFFSET;
                    }
                }
                m_videoEvents.verticalBlank();
           }
        }

        public bool pvcountHit = false;
        public UInt16 PVCOUNT
        {
            get
            {
                return m_pcvount;
            }

            set
            {
                m_pcvount = value;
                pvcountHit = false;
            }
        }

        UInt16 m_pcvount;

       /* bool doBlank = true;*/

        public void  step()
        {
            

            // vblank for gfx update

            if (m_hcount == m_vblank_h && m_vcount == m_vblank_v)
            {
                /*doBlank = true;   */           
                //m_videoEvents.verticalBlank();
            }

            // h/v count

            if (m_hcount < HTOTAL)
            {
				m_hcount = m_hcount + 1;
            }
			else
            {

                if (m_vcount < VTOTAL)
                {
                    m_vcount = m_vcount + 1;
                }
                else
                {
                    m_vcount = 0;
                }

                if (pvcountHit == false)
                {
                    if (m_pcvount == m_vcount)
                    {
                        pvcountHit = true;
                    }
                }

				m_hcount = 0;					
            }
        }

        internal UInt32[] getScreenBuffer()
        {
            return m_screenBuffer;
        }

        public int vcount
        {
            get { return m_vcount; }
        }

        public int hcount
        {
            get { return m_hcount; }
        }

        public int m_vcount = 0;
        public int m_hcount = 0;
        public UInt32 lc_q = 0, lc_q_next = 0;

        public int HSYNC = videoPackage.HSYNC_default;
        public int HSTART = videoPackage.HSTART_default;
        public int HMEMSTART = videoPackage.HMEMSTART_default;
        public int HSTOP = videoPackage.HSTOP_default;
        public int HTOTAL = videoPackage.HTOTAL_default;

        public int VSYNC = videoPackage.VSYNC_default;
        public int VSTART = videoPackage.VSTART_default;
        public int VSTOP = videoPackage.VSTOP_default;
        public int VTOTAL = videoPackage.VTOTAL_default;

        // external load

        public int HLOADSTART = videoPackage.HLOADSTART_default;		// external memory start
        public int HLOADSTOP = videoPackage.HLOADSTOP_default;

        public int VLOADSTARTt = videoPackage.VLOADSTART_default;		// external memory start
        public int VLOADSTOP = videoPackage.VLOADSTOP_default;
        public int VLOADNEXT = videoPackage.VLOADNEXT_default;

        // memory processing

        public UInt32 MEM_STARTADR = videoPackage.MEM_STARTADR_default;		// external memory start
        public UInt32 MEM_LINEOFFSET = videoPackage.MEM_LINEOFFSET_default;
        public UInt16 MEM_MODE = videoPackage.MEM_MODE_default;

        // line cache

        public UInt32 lc_add = videoPackage.lc_add_default;
        public UInt32 lc_start = videoPackage.lc_start_default;
    }
}
