﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ucore_emulator.core.components.memory.ctrl;
using ucore_emulator.common;
using System.Threading;
using ucore_emulator.core.components;
using ucore_emulator.core.components.video;
using ucore_assembler_library;
using xcore_tools;

namespace ucore_emulator.core
{
    class HCoreThreadParameter
    {
        public HCoreThreadParameter(uint istart, uint istop)
        {
            start = istart;
            stop = istop;
        }

        public uint start { get; set; }
        public uint stop { get; set; }
    }

    public class emulator : IPrint
    {
        BoardCtrl m_boardCtrl = null;
        IPrint m_iprint = null;
        Thread m_ucoreThread = null;
        Thread m_videoThread = null;
        Thread[] m_hcoreThreads = null;
        hcore_ctrl m_hcore_ctrl = null;
        hcore_thread_ctrl m_hcore_thread_ctrl;
     

        //Thread m_fcoreThread = null;
        Int64 m_runCount;
        bool m_stop;
        bool m_reset = false;
        bool m_hcoreReset = false;
        UInt64 m_clkCount = 0;
        UInt64 m_clkCountHcore = 0;

        ucore m_ucore;

        public hcore_ctrl hcore_ctrl
        {
            get { return m_hcore_ctrl; }
        }

        public uint hcoreCount 
        { 
            get 
            { 
                return m_hcoreCount; 
            }

            set
            {
                uint newCount = value;
                print("stop Emulator ... WAIT ...\n");
                try
                {
                    stop();
                }
                catch(Exception ex)
                {
                    print(ex.Message);
                }
                Thread.Sleep(1000);
                print("make " + newCount +" HCores\n");

                m_hcoreCount = newCount;
                initHCores();
            }
        }

        public hcore getHCore(uint pos)
        {
            return m_hcores[pos];  
        }

        uint m_hcoreCount = 2;
        hcore[] m_hcores;

        public bool m_hcoreSingleThread = true;

        ext_bus m_ext_bus;
        ucore_bus m_ucoreBus;
        ucore_form m_ucore_form;
        Timer m_ucoreFormUpdateTimer;
        VideoComponent m_videoComponent;

        public ext_bus getExtBus()
        {
            return m_ext_bus;
        }

        const uint m_runtDebugGUIRate = 250;
        const uint m_defaultDebugGUIRate = 2000;

        public emulator(IPrint iprint,IVideoEvent videoEvent,ScanCodeTranslator iScanCodeTranslator)
        {
            m_iprint = iprint;
            m_runCount = 0;

            m_ext_bus = new ext_bus(iprint); // SRAM / DRAM / FLASH
            m_boardCtrl = new BoardCtrl(iScanCodeTranslator);
            m_videoComponent = new VideoComponent(videoEvent, m_ext_bus);

            // ucore

            m_ucoreBus = new ucore_bus(m_boardCtrl, m_videoComponent, iprint,this);
            m_ucore = new ucore(m_ucoreBus, m_ext_bus, iprint);
            m_ucoreBus.setUCore(m_ucore);   // for irq

            m_ucore_form = new ucore_form(this,this);
            m_ucoreFormUpdateTimer = new Timer(new TimerCallback(ucoreFormUpdateCalback), null, 0, m_defaultDebugGUIRate); // 
            ucoreFormUpdateCalback(null);   
    
            // hcores

            m_hcore_ctrl = new hcore_ctrl();
            m_hcore_thread_ctrl = new hcore_thread_ctrl(this);
            initHCores();
           
        }

        void initHCores()
        {
            m_hcores = new hcore[m_hcoreCount];
            for (uint i = 0; i < m_hcoreCount; i++)
            {
                m_hcores[i] = new hcore(m_ext_bus, m_iprint, i, m_hcore_thread_ctrl, m_hcoreCount, m_hcore_ctrl);
            }

            m_ucoreBus.setHCores(m_hcores);   // for config
            m_hcoreThreads = new Thread[m_hcoreCount];
        }

        public void kill()
        {
            if (m_ucoreThread != null)
            {
                m_ucoreThread.Abort();
            }

            if (m_videoThread != null)
            {
                m_videoThread.Abort();
            }
            m_ucoreThread = null;
            m_videoThread = null;
        }

        public VideoComponent videoComponent
        {
            get { return m_videoComponent; }
        }


        void ucoreFormUpdateCalback(Object o)
        { 
            m_ucore_form.update();
        }


        internal void showUcoreForm()
        {
            m_ucore_form.Show();
        }
     
        public BoardCtrl boardCtrl
        {
            get { return m_boardCtrl;   }
            set { m_boardCtrl = value;  }
        }

        // start/stop/step/reset

        public void start(bool disableHCores)
        {           
            print("emulator start\n");
            threadAction(ref m_ucoreThread, internalThreadModes.I_RUN, ucore1MainFunction,null);
            threadAction(ref m_videoThread, internalThreadModes.I_RUN, videoMainFunction, null);
            if (disableHCores == false)
            {
                m_ucoreBus.setHCores(m_hcores);   // for config
                if (m_hcoreSingleThread)
                {

                    threadAction(ref m_hcoreThreads[0], internalThreadModes.I_RUN, hcoreMainFunction, new HCoreThreadParameter(0, m_hcoreCount - 1));
                }
                else
                {
                    for (uint i = 0; i < m_hcoreCount; i++)
                    {
                        threadAction(ref m_hcoreThreads[i], internalThreadModes.I_RUN, hcoreMainFunction, new HCoreThreadParameter(i, i));
                    }

                }
            }
            else
            {
                m_ucoreBus.setHCores(null);   // for config
            }
        }

        public void stop()
        {
            print("emulator stop\n");
            threadAction(ref m_ucoreThread, internalThreadModes.I_STOP, ucore1MainFunction, null);
            threadAction(ref m_videoThread, internalThreadModes.I_STOP, videoMainFunction, null);
            if (m_hcoreSingleThread)
            {
                threadAction(ref m_hcoreThreads[0], internalThreadModes.I_STOP, hcoreMainFunction, new HCoreThreadParameter(0, m_hcoreCount - 1));
            }
            else 
            {
                for (uint i = 0; i < m_hcoreCount; i++)
                {
                    threadAction(ref  m_hcoreThreads[i], internalThreadModes.I_STOP,hcoreMainFunction, new HCoreThreadParameter(i, i));
                }
             
            }
        }

        public void step()
        {
            print("emulator step\n");
            threadAction(ref m_ucoreThread, internalThreadModes.I_STEP, ucore1MainFunction, null);
            threadAction(ref m_videoThread, internalThreadModes.I_STEP, videoMainFunction, null);
            if (m_hcoreSingleThread)
            {
                threadAction(ref m_hcoreThreads[0], internalThreadModes.I_STEP, hcoreMainFunction, new HCoreThreadParameter(0, m_hcoreCount-1));
            }
            else 
            {
                for (uint i = 0; i < m_hcoreCount; i++)
                {
                    threadAction(ref  m_hcoreThreads[i], internalThreadModes.I_STEP,hcoreMainFunction, new HCoreThreadParameter(i, i));
                    m_hcoreThreads[i].Priority = ThreadPriority.BelowNormal;
                }
            }

            ucoreFormUpdateCalback(null);   
        }

        public void reset()
        {
            print("ucore reset\n");
            m_reset = true;
            m_hcoreReset = false;

            m_stop = false;
            m_runCount = 1;
            ucore1MainFunction(null);
            m_hcore_ctrl.reset();

            foreach(hcore h in m_hcores)
            {
                h.reset();
            }

        }

        // internal thread

        public UInt64 clkCountUCore
        {
            get { return m_clkCount; }
        }

        public UInt64 clkCountHCores
        {
            get { return m_clkCountHcore; }
        }

        public ucore ucore
        {
            get { return m_ucore; }
        }
       
        void videoMainFunction(object unused)
        {
            string message = "";
            try
            {
                Int64 opcodeCount = 0;
                for (m_stop = false; m_stop != true; )
                {
                    if (opcodeCount == m_runCount)  // runcount finish
                    {
                        break;
                    }

                    opcodeCount++;
                    m_videoComponent.step2();
                }
            }
            catch (Exception ex)
            {
                message = ex.Message;
            }
            // signal stop

            if (message.Length == 0) // no error 
            {
                message = string.Format("Video stop at V/H {0}/{1}\n", m_videoComponent.vcount,m_videoComponent.hcount);
            }
            else
            {
                message = string.Format("Video stop at V/H {0}/{1} at error '{1}'\n", m_videoComponent.vcount, m_videoComponent.hcount, message);
            }
            print(message);
        }

        void ucore1MainFunction(object unused)
        {
            m_ucoreFormUpdateTimer.Change(0, m_runtDebugGUIRate);

            Int64 opcodeCount = 0;
            string message = "";
            try
            {
                for (m_stop = false; m_stop != true; )
                {
                    if (opcodeCount == m_runCount)  // runcount finish
                    {
                        break;
                    }

                    opcodeCount++;

                    // here is the clock

                    if (m_reset)
                    {
                        m_ucore.reset();
                        m_ucoreBus.reset();
                        m_ext_bus.reset();
                        m_reset = false; 
                    }

                    m_ucore.step();
                    m_ucoreBus.step();
                    m_ext_bus.step();
                    m_videoComponent.step();

                    //

                    //m_boardCtrl.set_gled((uint)m_clkCount);
                    //m_boardCtrl.set_rled((uint)m_clkCount >> 8);

                    m_clkCount++;
                }
            }
            catch (Exception ex)
            {
                message = ex.Message;
            }
            // signal stop

            if (message.Length == 0) // no error 
            {
                message = string.Format("UCore stop (pc = ${0:X4})\n", ucore.stage_execute.m_pc);
            }
            else
            {
                message = string.Format("UCore stop (pc = ${0:X4}) at error '{1}'\n", ucore.stage_execute.m_pc, message);
            }

            print(message);

            m_ucoreFormUpdateTimer.Change(0, m_defaultDebugGUIRate);
        }

        void hcoreMainFunction(object parameter)
        {
            HCoreThreadParameter hcoreParameter = (HCoreThreadParameter) parameter;

            Int64 opcodeCount = 0;
            string message = "";
            try
            {
                for (m_stop = false; m_stop != true; )
                {
                    if (opcodeCount == m_runCount)  // runcount finish
                    {
                        break;
                    }

                    opcodeCount++;

                    for (uint hcoreId = hcoreParameter.start; hcoreId <= hcoreParameter.stop; hcoreId++)
                    {
                        hcore hcore = m_hcores[hcoreId];

                       if (m_hcoreReset)
                        {
                            hcore.reset();
                            m_hcoreReset = false;
                        }

                        hcore.step();
                        m_clkCountHcore ++;
                    }

                    //m_fcore.step();

                    
                }
            }
            catch (Exception ex)
            {
                message = ex.Message;
            }
            // signal stop

            if (message.Length == 0) // no error 
            {
                message = "";
                for (uint hcoreId = hcoreParameter.start; hcoreId <= hcoreParameter.stop; hcoreId++)
                {
                    hcore hcore = m_hcores[hcoreId];
                    message += string.Format("HCores ("+hcoreId+") stop at (pc = ${0:X8})\n", hcore.s5_next.pc_out);
                }                              
            }
            else
            {
                string error = message;
                message = "";
                for (uint hcoreId = hcoreParameter.start; hcoreId <= hcoreParameter.stop; hcoreId++)
                {
                    hcore hcore = m_hcores[hcoreId];
                    message += string.Format("HCores (" + hcoreId + ") stop at (pc = ${0:X8}) at error '{1}'\n", hcore.s5_next.pc_out, error);
                }  
            }

            print(message);
        }

        //

        public void print(string txt)
        {
           m_iprint.print(txt);
        }

        public void dprint(string txt)
        {
            m_iprint.dprint(txt);
        }

        #region thread_managment

        enum internalThreadModes
        {
            I_RUN,
            I_STEP,
            I_STOP
        };

        public void setRunCount(Int64 count)
        {
            m_runCount = count;
        }

        void threadAction(ref Thread ethread, internalThreadModes mode, ParameterizedThreadStart threadMainFunction, object parameter)
        {
            if (ethread == null)
            {
                ethread = new Thread(threadMainFunction);
            }

            // check state

            if (ethread.ThreadState == ThreadState.Running)
            {
                switch (mode)
                {
                    case internalThreadModes.I_STEP:
                    case internalThreadModes.I_RUN: throw new Exception("run/step is currently in progress\n");
                    case internalThreadModes.I_STOP: m_stop = true; break;
                }
            }
            else if (ethread.ThreadState == ThreadState.Unstarted ||
                     ethread.ThreadState == ThreadState.Stopped)
            {
                ethread = new Thread(threadMainFunction);


                switch (mode)
                {
                    case internalThreadModes.I_STEP: setRunCount(1); break;
                    case internalThreadModes.I_RUN: setRunCount(-1); break;
                    case internalThreadModes.I_STOP: throw new Exception("core is not running\n");
                }

                // signal start

                string message = string.Format("Start at ${0:X8}", 0/*m_cpu.pc*/);
               // m_emulatorEventManager.NotifySubscribers(emulatorMode_e.RUN, message);

                // start thread

                ethread.Start(parameter);
            }
            else
            {
                throw new Exception("Unknown Thread mode (" + ethread.ThreadState.ToString() + ")?!\n");
            }
        }

        #endregion        
    }
}
