﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using ucore_emulator.core;
using ucore_emulator.core.components.stages;
using ucore_emulator.core.components.extMemory;
using ucore_emulator.core.components.extMemory.SDCard;
using ucore_emulator.core.components.memory;
using ucore_emulator.core.components.memory.ctrl;
using ucore_emulator.core.components;
using ucore_assembler_library;
using System.Drawing.Imaging;
using xcore_tools;
using ucore_assembler_library.Data;
using hcore_assembler_library.Opcode;

namespace ucore_emulator
{
    public partial class ucore_form : Form, IPrint 
    {
        const int defaultRowHeight = 15;

        emulator m_emulator;
        IPrint m_iprint;

        int visibleWidth = 256;
        int visibleHeight = 16;
        int memViewMax = 65536;
        uint hcorePcOffset = 0xd013c680;//0xd0100000;//;

        object _lockRepaint = new object();
        Bitmap m_screenBitmap = null;

        public ucore_form(IPrint iprint,emulator emulator)
        {
            m_iprint = iprint;
            m_emulator = emulator;

            m_screenBitmap = new Bitmap(visibleWidth, visibleHeight);

            InitializeComponent();           

            comboBoxMemoryType.Items.Clear();
            comboBoxMemoryType.Items.Add("IRAM");
            comboBoxMemoryType.Items.Add("SRAM");
            comboBoxMemoryType.Items.Add("DRAM");
            comboBoxMemoryType.Items.Add("FLASH");
            comboBoxMemoryType.SelectedIndex = 0;

            numericUpDown_DacOutputSampleDepth.Value = m_emulator.ucore.bus.ctrlMemory.audio.audioMask;

            // hcore select
          
            uint hcoreCount = m_emulator.hcoreCount;

            numericUpDownHCcoreSelect.Minimum = 0;
            numericUpDownHCcoreSelect.Maximum = hcoreCount - 1;

            //

            InitUCoreAsmView();
            InitHCoreAsmView();
            InitUCoreMemView();

            update();
        }

        void InitUCoreMemView()
        {
            dataGridViewMemory.CellValueNeeded += new DataGridViewCellValueEventHandler(ucoreMemGridCellUpdate);
            dataGridViewMemory.VirtualMode = true;

            int lineCount = memViewMax; 

            dataGridViewMemory.RowCount = lineCount;         
        }

        void ucoreMemGridCellUpdate(object o, DataGridViewCellValueEventArgs e)
        {
            int x = e.ColumnIndex;
            int y = e.RowIndex;

            IMemIO mem = null;
            IExtMemory emem = null;
            int memSize = 0;
            int dataPerLine = 16;

            switch (comboBoxMemoryType.Items[comboBoxMemoryType.SelectedIndex].ToString())
            {
                case "IRAM":                   
                    mem = m_emulator.ucore.bus.ucore1MainMemory;
                     memSize = m_emulator.ucore.bus.ucore1MainMemory.size;  
                    break;
                case "SRAM":
                    emem = m_emulator.ucore.ext_bus.sram;
                     memSize = m_emulator.ucore.ext_bus.sram.size;
                    break;
                case "DRAM":
                    emem = m_emulator.ucore.ext_bus.dsram;
                    memSize = m_emulator.ucore.ext_bus.dsram.size;
                    break;
                case "FLASH":
                    emem = m_emulator.ucore.ext_bus.flash;
                     memSize = m_emulator.ucore.ext_bus.flash.size;
                    break;
            }

            int lineCount = ((memSize < memViewMax) ? memSize : memViewMax) / dataPerLine;
            int offset = (int)(numericUpDownMV64kOffset.Value * 65536);

            if (offset > memSize)
            {
                offset = 0;
            }
    
            String lineValue = String.Format("${0:X8} ", y * dataPerLine+ offset);

            UInt16[] memData = new UInt16[dataPerLine];
           
            // copy data

            for (int i = 0; i < dataPerLine; i++)
            {
                UInt16 rValue = 0;

                if (mem != null)
                {
                    rValue = mem.read((uint)((y * (dataPerLine)) + i + offset));
                }

                if (emem != null)
                {
                    emem.read((uint)((y * (dataPerLine)) + i + offset), out rValue);
                }
                memData[i] = rValue;
            }

            // print hex 

            for (int i = 0; i < dataPerLine; i++)
            {
                UInt16 rValue = memData[i];              
                lineValue += String.Format("${0:X4} ",rValue );
            }

            lineValue += " '";

            // print ascii 

            for (int i = 0; i < dataPerLine; i++)
            {
                UInt16 rValue = memData[i];
                lineValue += toAscii((UInt16)(rValue >> 8));
                lineValue += toAscii((UInt16)(rValue & 0xff));
            }

            lineValue += "'";

            dataGridViewMemory.Rows[y].Height = defaultRowHeight;
            e.Value = lineValue;
            
        }

        char toAscii(UInt16 v)
        {
            v &= 0xff; 
            char b = ' ';

            if (v >= 32 && v < 127)
            {
                b= (char)v;
            }
            return b;
        }

        void InitUCoreAsmView()
        {
            dataGridViewUcoreAsm.CellValueNeeded += new DataGridViewCellValueEventHandler(ucoreAsmGridCellUpdate);
            dataGridViewUcoreAsm.VirtualMode = true;

            int lineCount = 0x4000;

            dataGridViewUcoreAsm.RowCount = lineCount;
            dataGridViewUcoreAsm.CurrentCell = dataGridViewUcoreAsm.Rows[0].Cells[0];

            for (int y = 0; y < lineCount; y++)
            {
                dataGridViewUcoreAsm.Rows[y].Height = defaultRowHeight;
            }
        }

        void InitHCoreAsmView()
        {
            dataGridViewHCoreAsm.CellValueNeeded += new DataGridViewCellValueEventHandler(hcoreAsmGridCellUpdate);
            dataGridViewHCoreAsm.VirtualMode = true;

            int lineCount = 32768;

            dataGridViewHCoreAsm.RowCount = lineCount;
            dataGridViewHCoreAsm.CurrentCell = dataGridViewHCoreAsm.Rows[0].Cells[0];

            for (int y = 0; y < lineCount; y++)
            {
                dataGridViewHCoreAsm.Rows[y].Height = defaultRowHeight;
            }
        }

        void ucoreAsmGridCellUpdate(object o, DataGridViewCellValueEventArgs e)
        {
            int x = e.ColumnIndex;
            int y = e.RowIndex;

            UInt32 realAddress = ((uint)y);
           
            string cellValue = ""; 
            if (x == 0) // addressline
            {
                cellValue = String.Format("${0:X4}", realAddress);
            }
            else if (x == 1) // disasm
            {
                UInt16 value = m_emulator.ucore.bus.read(realAddress);

                Opcode tOpcode = new Opcode(null);
                tOpcode.decode(value);   // disassemble

                cellValue = String.Format("${0:X4} {1}", value, tOpcode.trace(false));
            }
            else if (x == 2) // stage
            {
                uint s0_pc = m_emulator.ucore.stage_preParefetch.dataOut.debug_pc;
                uint s1_pc = m_emulator.ucore.stage_fetch.dataOut.debug_pc;
                uint s2_pc = m_emulator.ucore.stage_decode.dataOut.debug_pc;
                uint s3_pc = m_emulator.ucore.stage_readReg.dataOut.debug_pc;
                uint s4_pc = m_emulator.ucore.stage_execute.dataOut.debug_pc;
                uint s5_pc = m_emulator.ucore.stage_writeReg.dataOut.debug_pc;

                if (y == s0_pc)
                {
                    cellValue += "prepare ";
                }
                if (y == s1_pc)
                {
                    cellValue += "fetch ";
                }
                if (y == s2_pc)
                {
                    cellValue += "decode ";
                }
                if (y == s3_pc)
                {
                    cellValue += "read reg ";
                }
                if (y == s4_pc)
                {
                    cellValue += "execute ";
                }
                if (y == s5_pc)
                {
                    cellValue += "write reg ";
                }
            }
            else if (x == 3) // analystic
            {
                uint usage = m_emulator.ucore.analystic.getLineUsage((uint)y);
                if (usage > 0)
                {
                    cellValue = "" + usage;
                }
            }

            e.Value = cellValue;         
        }


        void hcoreAsmGridCellUpdate(object o, DataGridViewCellValueEventArgs e)
        {
            int x = e.ColumnIndex;
            int y = e.RowIndex;

            uint hcoreId = (uint)numericUpDownHCcoreSelect.Value;
            hcorePcOffset = m_emulator.getHCore(hcoreId).s4_execute.pc_reset;
            UInt32 realAddress = ((uint)y*2) + hcorePcOffset;
           
            string cellValue = ""; 
            if (x == 0) // addressline
            {
                cellValue = String.Format("${0:X8}", realAddress);
            }
            else if (x == 1) // disasm
            {
                UInt32 value;
                m_emulator.getExtBus().read(realAddress, out value);

                HOpcode tOpcode = new HOpcode();
                tOpcode.decode(value);   // disassemble
                

                cellValue = String.Format("${0:X8} {1}", value, tOpcode.trace());
            }
            else if (x == 2) // stage
            {
              
            }
            else if (x == 3) // path
            {
                UInt32 value;
                m_emulator.getExtBus().read(realAddress, out value);

                HOpcode tOpcode = new HOpcode();
                tOpcode.decode(value);   // disassemble

                cellValue = tOpcode.getExecutionPathAsString();
            }

            e.Value = cellValue;         
        }

        public void update()
        {
            InvokeIfRequired(this, (MethodInvoker)delegate()
           {
               try
               {
                   ucoreViewUpdate();
                   videoViewUpdate();
                   boardViewUpdate();
                   sdCardViewUpdate();
                   memoryViewUpdate();
                   statisticViewUpdate();
                   audioViewUpdate();
                   hcoresUpdate();
               }
               catch (Exception)
               { }
           });
        }

        private void hcoresUpdate()
        {
            uint hcoreNum = (uint) numericUpDownHCcoreSelect.Value;
            hcore hcore = m_emulator.getHCore(hcoreNum);
          
            dataGridViewHCoreRegister.RowCount = 128;
            for(int y=0;y<128;y++)
            {
                DataGridViewRow row = dataGridViewHCoreRegister.Rows[y];

                row.Height = defaultRowHeight;
                row.Cells[0].Value = "" + y;

                UInt32 regValue = hcore.getRegValue((uint)y);
                bool   regLock  = hcore.getRegLock((uint)y);
                bool regOffsetEnable = hcore.getRegOffsetEnabled((uint)y);

                if (regLock)
                {
                    row.Cells[1].Style.BackColor = Color.Red;
                }
                else
                {
                    row.Cells[1].Style.BackColor = Color.White;
                }

                row.Cells[1].Value = string.Format("{0:X8}",regValue);
                row.Cells[2].Value = string.Format("{0}", regOffsetEnable?"E":"");

            }

            int pcPos = (int)hcore.s4_execute.pc;

            uint hcoreId = (uint)numericUpDownHCcoreSelect.Value;
            hcorePcOffset = m_emulator.getHCore(hcoreId).s4_execute.pc_reset;

            pcPos -= (int)hcorePcOffset;
            if (pcPos >= 0 && pcPos < dataGridViewHCoreAsm.Rows.Count)
            {
                dataGridViewHCoreAsm.CurrentCell = dataGridViewHCoreAsm.Rows[pcPos/2].Cells[0];
            }

            maskedTextBoxHCorePC.Text = string.Format("{0:X8}", hcore.s4_execute.pc);
            maskedTextBoxHCoreLast.Text = string.Format("{0:X8}", hcore.s4_execute.last);
            maskedTextBoxHCoreRegOffset.Text = string.Format("{0}", hcore.s5_next.roffset);
            maskedTextBoxHCoreCycleCount.Text = string.Format("{0}", hcore.s4_execute.cycleCount);

            maskedTextBoxDCacheHit.Text = string.Format("{0}", hcore.s4_execute.m_dcacheHit);
            maskedTextBoxDCacheMiss.Text = string.Format("{0}", hcore.s4_execute.m_dcacheMiss);

            radioButtonHCoreT.Checked = hcore.s4_execute.t;
            radioButtonHCoreIdle.Checked = hcore.s4_execute.idle;

        }



        private void audioViewUpdate()
        {
            Audio audio = m_emulator.ucore.bus.ctrlMemory.audio;

            maskedTextBox_AudioDacFifoUsedW.Text = string.Format("{0}",audio.fifoUseCount);
            maskedTextBox_AudioDacFifoUsedWMax.Text = string.Format("{0}", audio.fifoMaxUsedW);
            maskedTextBox_AudioDacFifoUsedWMin.Text = string.Format("{0}", audio.fifoMinUsedW);

            checkBox_AudioDacEnabled.Checked = audio.isDacEnabled;

            pictureBoxDACLeft.Refresh();
            pictureBoxDACRight.Refresh();
        }

        private void memoryViewUpdate()
        {
            int memSize = 0;
            int dataPerLine = 16;

            switch (comboBoxMemoryType.Items[comboBoxMemoryType.SelectedIndex].ToString())
            {
                case "IRAM":
                    memSize = m_emulator.ucore.bus.ucore1MainMemory.size;                   
                    break;
                case "SRAM":
                    memSize = m_emulator.ucore.ext_bus.sram.size;
                    break;
                case "DRAM":
                    memSize = m_emulator.ucore.ext_bus.dsram.size;
                    break;
                case "FLASH":
                    memSize = m_emulator.ucore.ext_bus.flash.size;
                    break;
            }

            int lineCount = ((memSize < memViewMax) ? memSize : memViewMax) / dataPerLine;

            int min64k = 0;
            int max64k = (memSize / memViewMax)-1;

            max64k = max64k >= 0 ? max64k : 0;

            numericUpDownMV64kOffset.Minimum = min64k;
            numericUpDownMV64kOffset.Maximum = max64k;
            dataGridViewMemory.RowCount = lineCount;

            dataGridViewMemory.Refresh();
        }

        private void sdCardViewUpdate()
        {
            sdCard_ctrl sdCard = m_emulator.ucore.ext_bus.sdCard;
            
            // clk

            sd_clk_msk_textBox.Text = String.Format("{0:X3}", sdCard.sdClk.prg_clk_msk);
            sd_clk_fall_textBox.Text = String.Format("{0:X3}", sdCard.sdClk.prg_clk_edge_cmp_down);
            sd_clk_rise_textBox.Text = String.Format("{0:X3}", sdCard.sdClk.prg_clk_edge_cmp_up);
            sd_clk_read_textBox.Text = String.Format("{0:X3}", sdCard.sdClk.prg_clk_read_edge_cmp);
            sd_clk_write_textBox.Text = String.Format("{0:X3}", sdCard.sdClk.prg_clk_write_edge_cmp);

            // io

            sd_burstaddress_textBox.Text = String.Format("{0:X8}", sdCard.burstAddress);

            SDCommand sdLastSendCmd = sdCard.sdCard_emulation.lastSendCmd;
            sd_send_cmd_textBox.Text = String.Format("SDCard CMD receive CMD{0:00} Parameter {1:X8} next receive is {2}", sdLastSendCmd.cmd, sdLastSendCmd.param, sdLastSendCmd.resive136 ? "136Bit" : "48Bit");

            SDCommandResponse sdCmdResponse = sdCard.sdCard_emulation.receiveCmd;          
            sd_receive_cmd_textBox.Text = String.Format("SDCard CMD result CMD{0:00} Parameter {1:X4}{2:X4}{3:X4}{4:X4}{5:X4}{6:X4}{7:X4}{8:X4} crc {9} excrc {10}", sdCmdResponse.cmd, sdCmdResponse.arg[0], sdCmdResponse.arg[1], sdCmdResponse.arg[2], sdCmdResponse.arg[3], sdCmdResponse.arg[4], sdCmdResponse.arg[5], sdCmdResponse.arg[6], sdCmdResponse.arg[7], sdCmdResponse.crc7, sdCmdResponse.crc7expected);

            sd_readFifo_usage_textBox.Text = String.Format("{0}", sdCard.sdCard_emulation.dataInReadQueueDataCount());
                
        }

        private void statisticViewUpdate()
        {
            string message = "";

            if ( m_emulator.ucore.enableStatistic == 0)
            {
                message = "Ucore analyse is disabled\n";
            }
            else
            { 
                message = "Ucore analyse is enabled\n";

                UCoreAnalaystic anly = m_emulator.ucore.analystic;

                if ((m_emulator.ucore.enableStatistic & 2) != 0)
                {
               
                    // branch

                    double minBrFreq = m_emulator.ucore.frequency / anly.maxBranchDelay;
                    message += String.Format("Branch difference min {0} max {1} apr {2} cycles (min irq frequency {3:F4}MHz) \n", anly.minBranchDelay, anly.maxBranchDelay, anly.aprBranchDelay / (double)anly.branchCount, minBrFreq / 1e+6);

                }


                if ((m_emulator.ucore.enableStatistic & 1) != 0)
                {

                    // ocpode usage

                    message += "\nOpcode usage:\n";

                    KeyValuePair<opcode_t, Int64>[] opcodeCountArray = anly.safeOpcodeUsage;

                    Int64 totalCount = 0;
                    SortedDictionary<opcode_t, Int64> opcodeCountDict = new SortedDictionary<opcode_t, Int64>();


                    foreach (KeyValuePair<opcode_t, Int64> oc in opcodeCountArray)
                    {
                        opcodeCountDict[oc.Key] = oc.Value;
                        totalCount += oc.Value;
                    }



                    foreach (opcode_t o in Enum.GetValues(typeof(opcode_t)))
                    {
                    
                        Int64 usage = opcodeCountDict.ContainsKey(o)?opcodeCountDict[o]:0;
                        float totalUsage = totalCount!=0?usage *100f / totalCount:0;
                        message += String.Format("opcode {0} is used {1} times {2}%\n", o.ToString().Remove(0, 4), usage, totalUsage);
                    }  
                }
            }

            textBox_UcoreAnalyse.Text = message;
        }    
  
        void boardViewUpdate()
        {
            irqCurrent.Text = String.Format("{0:X4}",m_emulator.ucore.bus.ctrlMemory.irqCurrent);
            irqMsk.Text = String.Format("{0:X4}",m_emulator.ucore.bus.ctrlMemory.irqMsk);
            irqMemory.Text = String.Format("{0:X4}", m_emulator.ucore.bus.ctrlMemory.irqMemory);

            // timer 1

            ucore_emulator.core.components.Timer32 t1 = m_emulator.ucore.bus.ctrlMemory.timer(0);
            t1_state.Text = t1.timerState.ToString();
            t1_loadValue.Text = String.Format("{0:X8}", t1.loadValue);
            t1_value.Text = String.Format("{0:X8}", t1.count);
            t1_start.Checked = t1.start;
            t1_stop.Checked = t1.stop;
            t1_autoreload.Checked = t1.autoReload;


            // timer 2

            ucore_emulator.core.components.Timer32 t2 = m_emulator.ucore.bus.ctrlMemory.timer(1);
            t2_state.Text = t2.timerState.ToString();
            t2_loadValue.Text = String.Format("{0:X8}", t2.loadValue);
            t2_value.Text = String.Format("{0:X8}", t2.count);
            t2_start.Checked = t2.start;
            t2_stop.Checked = t2.stop;
            t2_autoreload.Checked = t2.autoReload;
            
        }

        private void videoViewUpdate()
        {
            maskedTextBoxHCOUNT.Text = "" + m_emulator.videoComponent.m_hcount;
            maskedTextBoxHLOADSTART.Text = "" + m_emulator.videoComponent.HLOADSTART;
            maskedTextBoxHLOADSTOP.Text = "" + m_emulator.videoComponent.HLOADSTOP;
            maskedTextBoxHMEMSTART.Text = "" + m_emulator.videoComponent.HMEMSTART;
            maskedTextBoxHSTART.Text = "" + m_emulator.videoComponent.HSTART;
            maskedTextBoxHSTOP.Text = "" + m_emulator.videoComponent.HSTOP;
            maskedTextBoxHTOTAL.Text = "" + m_emulator.videoComponent.HTOTAL;
            maskedTextBoxLCGAIN.Text = String.Format("{0:X}", m_emulator.videoComponent.lc_add);
            maskedTextBoxLCSTART.Text = String.Format("{0:X}", m_emulator.videoComponent.lc_start);
            maskedTextBoxMEMLINEOFFSET.Text = "" + m_emulator.videoComponent.MEM_LINEOFFSET;
            maskedTextBoxMEMMODE.Text = "" + m_emulator.videoComponent.MEM_MODE;
            maskedTextBoxMEMSTARTADR.Text = String.Format("{0:X8}", m_emulator.videoComponent.MEM_STARTADR);
            maskedTextBoxVCOUNT.Text = "" + m_emulator.videoComponent.m_vcount;
            maskedTextBoxVLOADNEXT.Text = "" + m_emulator.videoComponent.VLOADNEXT;
            maskedTextBoxVLOADSTART.Text = "" + m_emulator.videoComponent.VLOADSTARTt;
            maskedTextBoxVLOADSTOP.Text = "" + m_emulator.videoComponent.VLOADSTOP;
            maskedTextBoxVSTART.Text = "" + m_emulator.videoComponent.VSTART;
            maskedTextBoxVSTOP.Text = "" + m_emulator.videoComponent.VSTOP;
            maskedTextBoxVSYNC.Text = "" + m_emulator.videoComponent.VSYNC;
            maskedTextBoxVTOTAL.Text = "" + m_emulator.videoComponent.VTOTAL;    
     
            int xres = m_emulator.videoComponent.HSTOP - m_emulator.videoComponent.HSTART;
            int yres = m_emulator.videoComponent.VSTOP - m_emulator.videoComponent.VSTART;


            int dxres = m_emulator.videoComponent.visibleWidth();
            int dyres = m_emulator.videoComponent.visibleHeight();
            
            //0x00400000


            string depth = "unknown";
            switch(m_emulator.videoComponent.MEM_MODE)
            {
                case 0: depth = "24 Bit [8X 8R 8G 8B] in 32 Bit word";break;
                case 1: depth = "16 Bit [5R 6G 5B] in 16 Bit word";break;
                case 2: depth = "15 Bit [1X 5R 5G 5B] in 16 Bit word";break;
                case 3: depth = "12 Bit [4X 4R 4G 4B] in 16 Bit word";break;
            }

            textBoxVideoSetup.Text = String.Format("{0} x {1} at {2}; drawed is {3} x {4}", xres, yres, depth, dxres  , dyres);
        }
       
        private void ucoreViewUpdate()
        {
            // s4 regs

            maskedTextBox_cycles.Text = String.Format("{0}", m_emulator.ucore.m_cycleCount);
            maskedTextBox_irqCount.Text = String.Format("{0}", m_emulator.ucore.stage_execute.m_irqCount);

            maskedTextBoxR0.Text = String.Format("{0:X4}", m_emulator.ucore.regs[0]);
            maskedTextBoxR1.Text = String.Format("{0:X4}", m_emulator.ucore.regs[1]);
            maskedTextBoxR2.Text = String.Format("{0:X4}", m_emulator.ucore.regs[2]);
            maskedTextBoxR3.Text = String.Format("{0:X4}", m_emulator.ucore.regs[3]);
            maskedTextBoxR4.Text = String.Format("{0:X4}", m_emulator.ucore.regs[4]);
            maskedTextBoxR5.Text = String.Format("{0:X4}", m_emulator.ucore.regs[5]);
            maskedTextBoxR6.Text = String.Format("{0:X4}", m_emulator.ucore.regs[6]);
            maskedTextBoxR7.Text = String.Format("{0:X4}", m_emulator.ucore.regs[7]);

            maskedTextBoxPC.Text = String.Format("{0:X4}", m_emulator.ucore.stage_execute.m_pc);
            maskedTextBoxPCStored.Text = String.Format("{0:X4}", m_emulator.ucore.stage_execute.m_stored_pc);
            maskedTextBoxSP.Text = String.Format("{0:X4}", m_emulator.ucore.stage_execute.m_sp);
            maskedTextBoxIRQVector.Text = String.Format("{0:X4}", m_emulator.ucore.stage_execute.m_irq_start_vector);

            maskedTextBoxEspage.Text = String.Format("{0:X8}", m_emulator.ucore.stage_execute.m_espage);

            radioButtonT.Checked = m_emulator.ucore.stage_execute.m_t;
            radioButtonTStored.Checked = m_emulator.ucore.stage_execute.m_stored_t;
            radioButtonI.Checked = m_emulator.ucore.stage_execute.m_i;

            textBox_memRequestQueue.Text = String.Format("{0}", m_emulator.ucore.stage_execute.m_extLoadQueue.Count);

            // udiv

            maskedTextBox_udivDivident.Text = String.Format("{0:X8}", m_emulator.ucore.stage_execute.udiv_divident);
            maskedTextBox_udivDivisor.Text = String.Format("{0:X8}", m_emulator.ucore.stage_execute.udiv_divisor);
            maskedTextBox_udivQoutient.Text = String.Format("{0:X8}", m_emulator.ucore.stage_execute.udiv_qoutient);
            maskedTextBox_udivRemainder.Text = String.Format("{0:X8}", m_emulator.ucore.stage_execute.udiv_rest);
            maskedTextBox_udivBitCounter.Text = String.Format("{0:X8}", m_emulator.ucore.stage_execute.udiv_counter);

            // l2d extension

            maskedTextBox_exl2d_u.Text = String.Format("{0:X8}", m_emulator.ucore.stage_execute.l2d.s0.iu);
            maskedTextBox_exl2d_du.Text = String.Format("{0:X8}", m_emulator.ucore.stage_execute.l2d.s0.idu);
            maskedTextBox_exl2d_v.Text = String.Format("{0:X8}", m_emulator.ucore.stage_execute.l2d.s0.iv);
            maskedTextBox_exl2d_dv.Text = String.Format("{0:X8}", m_emulator.ucore.stage_execute.l2d.s0.idv);
            maskedTextBox_exl2d_xwidth.Text = String.Format("{0:X4}", m_emulator.ucore.stage_execute.l2d.s0.ixwidth);
            maskedTextBox_exl2d_sadr.Text = String.Format("{0:X8}", m_emulator.ucore.stage_execute.l2d.s0.isadr);
            maskedTextBox_exl2d_xmask.Text = String.Format("{0:X4}", m_emulator.ucore.stage_execute.l2d.s0.ixmask);
            maskedTextBox_exl2d_ymask.Text = String.Format("{0:X4}", m_emulator.ucore.stage_execute.l2d.s0.iymask);


            // disasm

           

            // disasm cpu

            uint s0_pc = m_emulator.ucore.stage_preParefetch.dataOut.debug_pc;
            uint s1_pc = m_emulator.ucore.stage_fetch.dataOut.debug_pc;
            uint s2_pc = m_emulator.ucore.stage_decode.dataOut.debug_pc;
            uint s3_pc = m_emulator.ucore.stage_readReg.dataOut.debug_pc;
            uint s4_pc = m_emulator.ucore.stage_execute.dataOut.debug_pc;
            uint s5_pc = m_emulator.ucore.stage_writeReg.dataOut.debug_pc;

            for(int i=0,length = dataGridViewUcoreAsm.Rows.Count;i<length;i++)
            {
                DataGridViewRow row = dataGridViewUcoreAsm.Rows[i];
                if(row.Displayed)
                {
                    row.Cells[0].Value = "update";
                    row.Cells[1].Value = "update";
                    row.Cells[2].Value = "update";
                    row.Cells[3].Value = "update"; 
                }
            }
             
            int pcPos = (int)s4_pc;
            if(pcPos >= 0 && pcPos < dataGridViewUcoreAsm.Rows.Count)
            {
                dataGridViewUcoreAsm.CurrentCell = dataGridViewUcoreAsm.Rows[pcPos].Cells[0];          
            }

            // stage

            // prepare fetch

            textBox_prpareFetchAddress.Text = string.Format("{0:X4}", m_emulator.ucore.s1_next.address);

            // fetch

            textBox_fetchData.Text = string.Format("{0:X4}", m_emulator.ucore.s2_next.data_in);

            // decode

            textBox_decodeOpcode.Text = m_emulator.ucore.s3_next.decode_opcode.ToString();
            textBox_decodeRegA.Text = string.Format("{0}", m_emulator.ucore.s3_next.decode_ra);
            textBox_decodeRegB.Text = string.Format("{0}", m_emulator.ucore.s3_next.decode_rb);
            textBox_decodeRegC.Text = string.Format("{0}", m_emulator.ucore.s3_next.decode_rc);
            textBox_decodeRegCWe.Text = m_emulator.ucore.s3_next.decode_rc_we ? "1" : "0";
            textBox_decodeImmUse.Text = m_emulator.ucore.s3_next.decode_imm_a ? "1" : "0";
            textBox_decodeImm.Text = string.Format("{0:X4}", m_emulator.ucore.s3_next.decode_imm);

            // read reg

            textBox_rrOpcode.Text = m_emulator.ucore.s4_next.opcode.ToString();
            textBox_rrRA.Text = string.Format("{0:X4}", m_emulator.ucore.s4_next.reg_data_a);
            textBox_rrRB.Text = string.Format("{0:X4}", m_emulator.ucore.s4_next.reg_data_b);
            textBox_rrRIC.Text = string.Format("{0}", m_emulator.ucore.s4_next.reg_rc);
            textBox_rrRCWE.Text = m_emulator.ucore.s4_next.reg_rc_we ? "1" : "0";
            textBox_rrMemQ.Text = string.Format("{0:X4}", m_emulator.ucore.s4_next.mem_q);
            textBox_RRIRQ.Text = m_emulator.ucore.s4_next.irq ? "1" : "0";

            // execute

            textBox_executeOpcode.Text = m_emulator.ucore.s5_next.opcode.ToString();
            textBox_executePc.Text = string.Format("{0:X4}", m_emulator.ucore.s5_next.pc);
            textBox_executeRegDataC.Text = string.Format("{0:X4}", m_emulator.ucore.s5_next.alu_data_c);
            textBox_executeRegCWe.Text = m_emulator.ucore.s5_next.alu_rc_we ? "1" : "0";
            textBox_executeRegC.Text = string.Format("{0}", m_emulator.ucore.s5_next.alu_rc);
            textBox_executeMAddress.Text = string.Format("{0:X4}", m_emulator.ucore.s5_next.mem_address);
            textBox_executeMData.Text = string.Format("{0:X4}", m_emulator.ucore.s5_next.mem_data);
            textBox_executeMemWe.Text = m_emulator.ucore.s5_next.mem_we ? "1" : "0";

            // mem prepare

            textBox_memPrepareLoad.Text = string.Format("{0:X4}", m_emulator.ucore.m1_next.address);

            // mem read

            textBox_memReadData.Text = string.Format("{0:X4}", m_emulator.ucore.m2_next);
        }

        void InvokeIfRequired(Control target, Delegate methodToInvoke)
        {
            if (target.InvokeRequired)
            {
                target.Invoke(methodToInvoke);
            }
            else
            {
                methodToInvoke.DynamicInvoke();
            }
        }

        private void maskedTextBox_KeyPress(object sender, KeyPressEventArgs e)
        {
            char v = e.KeyChar;

            if (v >= '0' && v <= '9' ||
                v >= 'a' && v <= 'f' ||
                v >= 'A' && v <= 'F' ||
                v == '\b' || v == '\r')
            {

            }
            else
            {
                e.Handled = true;
            }
        }

        private void ucore_form_FormClosing(object sender, FormClosingEventArgs e)
        {
            Hide();
            e.Cancel = true;
        }

        #region toolbar_events

        private void toolStripButton_internal_reset_Click(object sender, EventArgs e)
        {
            try
            {
                m_emulator.reset();
            }
            catch (Exception ex)
            {
                m_emulator.print(ex.Message);
            }
        }

        private void toolStripButton_internal_run_Click(object sender, EventArgs e)
        {
            try
            {                
                m_emulator.start(false);
            }
            catch (Exception ex)
            {
                m_emulator.print(ex.Message);
            }
        }

        private void toolStripButton_internal_step_Click(object sender, EventArgs e)
        {
            try
            {m_emulator.step();}
            catch (Exception ex)
            {
                m_emulator.print(ex.Message);
            }
        }

        private void toolStripButton_internal_stop_Click(object sender, EventArgs e)
        {
            try
            { m_emulator.stop(); }
            catch (Exception ex)
            {
                m_emulator.print(ex.Message);
            }
        }

        #endregion

       

        private void comboBoxMemoryType_SelectedIndexChanged(object sender, EventArgs e)
        {
            InvokeIfRequired(this, (MethodInvoker)delegate()
            {
                try
                {
                    memoryViewUpdate();
                }
                catch (Exception)
                { }
            });
        }

        private void pictureBoxDACLeft_Paint(object sender, PaintEventArgs e)
        {        
            Graphics g = e.Graphics;

            int width  = pictureBoxDACLeft.Width;
            int height = pictureBoxDACLeft.Height;

            int channel = 0;

            drawDacWave(g, width, height, channel);

        }

        private void pictureBoxDacRight_Paint(object sender, PaintEventArgs e)
        {
            Graphics g = e.Graphics;
            

            int width = pictureBoxDACRight.Width;
            int height = pictureBoxDACRight.Height;

            int channel = 1;

            drawDacWave(g, width, height, channel);
        }

        private void drawDacWave(Graphics g, int width, int height, int channel)
        {
            Audio audio = m_emulator.ucore.bus.ctrlMemory.audio;
            int[,] audioData = audio.audioOut;
            int audioLength = audio.audioOutLength;

            double incX = ((double)audioLength) / ((double)width);
            double incY = ((double)height) / 16777216.0;

            Point[] p = new Point[width];
            for (int i = 0; i < width; i++)
            {
                int v = audioData[channel, (int)(i * incX)];
                double y = ((double)height / 2) - ((double)v) * incY;
                p[i] = new Point(i, (int)y);

            }

            Pen pen = new Pen(Color.Yellow);
            g.DrawLines(pen, p);
        }

        private void numericUpDown_DacOutputSampleDepth_ValueChanged(object sender, EventArgs e)
        {
            m_emulator.ucore.bus.ctrlMemory.audio.audioMask = (int)numericUpDown_DacOutputSampleDepth.Value;
        }

        private void buttonClearDacFifo_Click(object sender, EventArgs e)
        {
            m_emulator.ucore.bus.ctrlMemory.audio.clearDacFifo();
        }

        private void dataGridViewUcoreAsm_DoubleClick(object sender, EventArgs e)
        {
            // dump disassembly table     

            StringBuilder allInOne = new StringBuilder();

            for (int i = 0, length = dataGridViewUcoreAsm.Rows.Count; i < length; i++)
            {
                DataGridViewRow row = dataGridViewUcoreAsm.Rows[i];
                for (int j = 0; j < row.Cells.Count; j++)
                {
                    if (j != 2)  // not stage
                    {
                        allInOne.Append(row.Cells[j].Value + "; ");
                    }
                }
                allInOne.Append("\n");
            }

            print(allInOne.ToString());
        }

        public void print(string txt)
        {
            m_iprint.print(txt);
        }

        public void dprint(string txt)
        {
            m_iprint.dprint(txt);
        }
       
        void enableUcoreAnalyse(bool opcode,bool branch)
        {
            m_emulator.ucore.enableStatistic  = (opcode) ? 1 : 0;
            m_emulator.ucore.enableStatistic |= (branch) ? 2 : 0;
        }

        private void checkBoxEnableUcoreAnalyse_CheckedChanged(object sender, EventArgs e)
        {
            enableUcoreAnalyse(checkBoxEnableUcoreOpcodeAnalyse.Checked, checkBoxEnableUcoreBranchAnalyse.Checked);
        }

        private void numericUpDownMV64kOffset_ValueChanged(object sender, EventArgs e)
        {
            update();
        }

        private void buttonClearStats_Click(object sender, EventArgs e)
        {
            InvokeIfRequired(this, (MethodInvoker)delegate()
            {
                m_emulator.ucore.analystic.clear();
            });
           
        }
    }
}
