library ieee;
use work.bus_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity bus_arbiter_sd_card is
		port 
		(
				-- global

				clk			: in std_logic;
				reset			: in std_logic;
				
				-- sd card
				
				-- ctrl port
	
				-- send
				
				ctrl_send_data			: in std_logic_vector(15 downto 0);
				ctrl_send_address		: in std_logic_vector(31 downto 0);
				ctrl_send_we			: in std_logic;
				ctrl_send_busy			: out std_logic;
				ctrl_send_wrreq		: in std_logic;
				
				-- recive
				
				ctrl_recive_data		: out std_logic_vector(15 downto 0);
				ctrl_recive_address	: out std_logic_vector(31 downto 0);
				ctrl_recive_we			: out std_logic;
				ctrl_recive_busy		: out std_logic;				
				ctrl_recive_ack		: in std_logic;
				
				-- memory clks
				
				sram_clk					: in std_logic;				
				sdram_clk				: in std_logic;
				ctrl_clk					: in std_logic;
				
				-- data port
				
				-- send
				
				data_send_data			: in  std_logic_vector(15 downto 0);
				data_send_address		: in  std_logic_vector(31 downto 0);
				data_send_we			: in  std_logic;
				data_send_busy			: out std_logic;
				data_send_wrreq		: in  std_logic;
				
				-- external bus
				
				-- data
				
				-- to memory (DATA PORT)
				
				data_to_sram_data			: out std_logic_vector(71 downto 0);		
				data_to_sram_busy			: out std_logic;
				data_to_sram_ack			: in std_logic;
							
				data_to_sdram_data		: out std_logic_vector(71 downto 0);		
				data_to_sdram_busy		: out std_logic;
				data_to_sdram_ack			: in std_logic;
				
				
				-- ctrl
				
				-- to memory
				
				-- recive ports
				
				ctrl_to_sd_data			: in sd_ctrl_n_data_t;		
				ctrl_to_sd_busy			: in sd_ctrl_n_busy_t;
				ctrl_to_sd_ack				: out sd_ctrl_n_ack_t;	
							
				-- from memory fifos
				
				ctrl_from_sd_data			: out sd_ctrl_n_data_t;	
				ctrl_from_sd_busy			: out sd_ctrl_n_busy_t;	
				ctrl_from_sd_ack			: in sd_ctrl_n_ack_t	
							
																	
		);
end entity;

architecture rtl of bus_arbiter_sd_card is

-- external memory fifo

--component dcfifo_min IS
--	PORT
--	(
--		aclr		: IN STD_LOGIC  := '0';
--		data		: IN STD_LOGIC_VECTOR (71 DOWNTO 0);
--		rdclk		: IN STD_LOGIC ;
--		rdreq		: IN STD_LOGIC ;
--		wrclk		: IN STD_LOGIC ;
--		wrreq		: IN STD_LOGIC ;
--		q			: OUT STD_LOGIC_VECTOR (71 DOWNTO 0);
--		rdempty	: OUT STD_LOGIC ;
--		wrusedw	: OUT STD_LOGIC_VECTOR (7 DOWNTO 0)
--	);
--END component;

component dc_fifo IS
	
	GENERIC 
	(
		FIFO_DEPTH 		: natural;
		FIFO_WIDTH 		: natural;
		FIFO_SHOWAHEAD		: string;
		FIFO_USEDWWIDTH		: natural;
		FIFO_SYNCSTAGES		: natural		
	);
	PORT
	(
		aclr		: IN STD_LOGIC  := '0';
		data		: IN STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdclk		: IN STD_LOGIC ;
		rdreq		: IN STD_LOGIC ;
		wrclk		: IN STD_LOGIC ;
		wrreq		: IN STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdempty		: OUT STD_LOGIC ;
		wrusedw		: OUT STD_LOGIC_VECTOR (FIFO_USEDWWIDTH-1 DOWNTO 0)
	);
END component;

-- DATA

type data_send_xn_data_t is array(0 to mem_sd_card_data_reciver_count-1) of STD_LOGIC_VECTOR (71 DOWNTO 0);
type data_send_xn_rdreq_t is array(0 to mem_sd_card_data_reciver_count-1) of STD_LOGIC;
type data_send_xn_rclk_t is array(0 to mem_sd_card_data_reciver_count-1) of STD_LOGIC;
type data_send_xn_wrreq_t is array(0 to mem_sd_card_data_reciver_count-1) of STD_LOGIC;
type data_send_xn_q_t is array(0 to mem_sd_card_data_reciver_count-1) of STD_LOGIC_VECTOR (71 DOWNTO 0);
type data_send_xn_rdempty_t is array(0 to mem_sd_card_data_reciver_count-1) of STD_LOGIC;
type data_send_xn_wrfull_t is array(0 to mem_sd_card_data_reciver_count-1) of STD_LOGIC; 
type data_send_xn_wrusedw_t is array(0 to mem_sd_card_data_reciver_count-1) of STD_LOGIC_VECTOR (7 DOWNTO 0);

signal data_send_xn_data		: data_send_xn_data_t := (others => (others => '0'));
signal data_send_xn_rdreq		: data_send_xn_rdreq_t := (others => '0');
signal data_send_xn_rclk		: data_send_xn_rclk_t := (others => '0');
signal data_send_xn_wrreq		: data_send_xn_wrreq_t := (others => '0');
signal data_send_xn_q			: data_send_xn_q_t := (others => (others => '0'));
signal data_send_xn_rdempty	: data_send_xn_rdempty_t := (others => '0');
signal data_send_xn_wrfull	: data_send_xn_wrfull_t := (others => '0'); 
signal data_send_xn_wrusedw	: data_send_xn_wrusedw_t := (others => (others => '0'));

-- CTRL

type ctrl_send_xn_data_t is array(0 to mem_sd_card_ctrl_reciver_count-1) of STD_LOGIC_VECTOR (71 DOWNTO 0);
type ctrl_send_xn_rdreq_t is array(0 to mem_sd_card_ctrl_reciver_count-1) of STD_LOGIC;
type ctrl_send_xn_rclk_t is array(0 to mem_sd_card_ctrl_reciver_count-1) of STD_LOGIC;
type ctrl_send_xn_wrreq_t is array(0 to mem_sd_card_ctrl_reciver_count-1) of STD_LOGIC;
type ctrl_send_xn_q_t is array(0 to mem_sd_card_ctrl_reciver_count-1) of STD_LOGIC_VECTOR (71 DOWNTO 0);
type ctrl_send_xn_rdempty_t is array(0 to mem_sd_card_ctrl_reciver_count-1) of STD_LOGIC;
type ctrl_send_xn_wrfull_t is array(0 to mem_sd_card_ctrl_reciver_count-1) of STD_LOGIC; 
type ctrl_send_xn_wrusedw_t is array(0 to mem_sd_card_ctrl_reciver_count-1) of STD_LOGIC_VECTOR (7 DOWNTO 0);

signal ctrl_send_xn_data		: ctrl_send_xn_data_t := (others => (others => '0'));
signal ctrl_send_xn_rdreq		: ctrl_send_xn_rdreq_t := (others => '0');
signal ctrl_send_xn_rclk		: ctrl_send_xn_rclk_t := (others => '0');
signal ctrl_send_xn_wrreq		: ctrl_send_xn_wrreq_t := (others => '0');
signal ctrl_send_xn_q			: ctrl_send_xn_q_t := (others => (others => '0'));
signal ctrl_send_xn_rdempty	: ctrl_send_xn_rdempty_t := (others => '0');
signal ctrl_send_xn_wrfull		: ctrl_send_xn_wrfull_t := (others => '0'); 
signal ctrl_send_xn_wrusedw	: ctrl_send_xn_wrusedw_t := (others => (others => '0'));

-- fifo for recive 

--component fifo_max
--	PORT
--	(
--		clock		: IN STD_LOGIC ;
--		data		: IN STD_LOGIC_VECTOR (71 DOWNTO 0);
--		rdreq		: IN STD_LOGIC ;
--		sclr		: IN STD_LOGIC ;
--		wrreq		: IN STD_LOGIC ;
--		empty		: OUT STD_LOGIC ;
--		full		: OUT STD_LOGIC ;
--		q			: OUT STD_LOGIC_VECTOR (71 DOWNTO 0);	
--		usedw	: OUT STD_LOGIC_VECTOR (7 DOWNTO 0)	
--	);
--end component;

component sc_fifo_usedw
	GENERIC 
	(
		FIFO_DEPTH 			: natural;
		FIFO_WIDTH 			: natural;
		FIFO_SHOWAHEAD 	: string;
		FIFO_USERAM			: string;
		FIFO_USEDWWIDTH	: natural
		
	);
	PORT
	(
		clock		: IN STD_LOGIC ;
		data		: IN STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdreq		: IN STD_LOGIC ;
		sclr		: IN STD_LOGIC ;
		wrreq		: IN STD_LOGIC ;
		empty		: OUT STD_LOGIC ;
		full		: OUT STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		usedw		: OUT STD_LOGIC_VECTOR (FIFO_USEDWWIDTH-1 DOWNTO 0)
	);
end component;

signal buf_data	: STD_LOGIC_VECTOR (71 DOWNTO 0) := (others => '0');
signal buf_rdreq	: STD_LOGIC := '0';
signal buf_wrreq	: STD_LOGIC := '0';
signal buf_empty	: STD_LOGIC := '0';
signal buf_full	: STD_LOGIC := '0';
signal buf_full_dmy	: STD_LOGIC := '0';
signal buf_q		: STD_LOGIC_VECTOR (71 DOWNTO 0) := (others => '0');
signal buf_usedw		: STD_LOGIC_VECTOR (7 DOWNTO 0) := (others => '0');

--

begin

	-----------------------------------------------------------------------------------------------------------
	-----------------------------------------------------------------------------------------------------------
	---- CTRL
	-----------------------------------------------------------------------------------------------------------
	-----------------------------------------------------------------------------------------------------------

	-----------------------------------------------------------------------------------------------------------
	---- RECIVE  
	-----------------------------------------------------------------------------------------------------------
	
	-- recive fifo
	
	recive_buffer : sc_fifo_usedw
	generic map
	(
		FIFO_DEPTH 			=> 256,
		FIFO_WIDTH 			=> 72,
		FIFO_SHOWAHEAD 	=> "ON",
		FIFO_USERAM			=> "ON",
		FIFO_USEDWWIDTH	=> 8				
	) 
	port map
	(
		clk,
		buf_data,	
	   buf_rdreq,	
		reset,
	   buf_wrreq,	
	   buf_empty,	
	   buf_full_dmy,	
	   buf_q,
		buf_usedw
	);
	
	process(clk)
	begin
		if rising_edge(clk) then		
			if buf_usedw >=  x"fb" then --
					buf_full <= '1';
				else
					buf_full <= '0';-- or wor;
				end if;
			end if;		
	end process;
	
	-- recive ack
	
	process(all)
	begin
	
		for i in 0 to mem_sd_card_ctrl_reciver_count-1 loop
			ctrl_to_sd_ack(i) <= '0';
		end loop;
	
		if reset = '0' then
			for i in 0 to mem_sd_card_ctrl_reciver_count-1 loop
				if ctrl_to_sd_busy(i) = '0' and buf_full = '0' then						
					ctrl_to_sd_ack(i) <= '1';	
					exit; -- stop loop		
				end if;	
			end loop;
		end if;
	end process;
	
	process(clk)
		variable t_data : port_struct := fill_port('0');
	begin
		if rising_edge(clk) then
		
			-- default
		
			buf_wrreq 	<= '0';		
			
			--
			
			if reset = '1' then
				--
			else
				for i in 0 to mem_sd_card_ctrl_reciver_count-1 loop
				if ctrl_to_sd_busy(i) = '0' and buf_full = '0' then						
					
				
					t_data 		:= raw_to_port(ctrl_to_sd_data(i)); 	
					t_data.id	:= conv_std_logic_vector(i,6);	-- store id
			
					buf_data		<=  port_to_raw(t_data);
					buf_wrreq 	<= '1';
					
					exit; -- stop loop
				end if;	
			end loop;
			end if;
		end if;
	end process;
	
	ctrl_recive_data		<= raw_to_port(buf_q).data(15 downto 0); 
	ctrl_recive_address	<= raw_to_port(buf_q).address; 
	ctrl_recive_we			<= raw_to_port(buf_q).we;
	ctrl_recive_busy		<= buf_empty;
	buf_rdreq				<= ctrl_recive_ack;
	
	-----------------------------------------------------------------------------------------------------------
	---- SEND  
	-----------------------------------------------------------------------------------------------------------
	
	process(all)
		variable tport : port_struct := fill_port('0');
	begin
	
			for i in 0 to mem_sd_card_ctrl_reciver_count-1 loop
	
				
				tport.data 		:= x"0000" & ctrl_send_data;				-- no data
				tport.address 	:= ctrl_send_address;
				tport.we			:= ctrl_send_we;					
				tport.dmode		:= '1';								-- 16 bit mode
	
				-- data
				
	
				ctrl_send_xn_data(i) <=  port_to_raw(tport);
			
				-- wrreq logic 
		
				if ctrl_send_wrreq = '1' then					
					ctrl_send_xn_wrreq(i) <= '1'; 
				else
					ctrl_send_xn_wrreq(i) <= '0'; 
				end if;
											
				-- full generate
			
				if rising_edge(clk) then
					if ctrl_send_xn_wrusedw(i) >=  x"fb" then --x"1fb" at 9 bit?
						ctrl_send_xn_wrfull(i) <= '1';
					else
						ctrl_send_xn_wrfull(i) <= '0';
					end if;
				end if;				
				
			
			end loop;
	end process;
	
	ctrl_xf_loop : for i in 0 to mem_sd_card_ctrl_reciver_count-1 generate
         send_xf : dc_fifo 
					generic map
					(
						FIFO_DEPTH 			=> 256,
						FIFO_WIDTH 			=> 72,
						FIFO_SHOWAHEAD 	=> "ON",
						FIFO_USEDWWIDTH	=> 8,
						FIFO_SYNCSTAGES	=> 3
					)  
					port map
					(
						'0',
						ctrl_send_xn_data(i),
						ctrl_send_xn_rclk(i),		-- read with component clk
						ctrl_send_xn_rdreq(i),	
						clk,					-- 
						ctrl_send_xn_wrreq(i),	
						ctrl_send_xn_q(i),
						ctrl_send_xn_rdempty(i),
						ctrl_send_xn_wrusedw(i)
					);
   end generate;
		
	ctrl_send_busy <=  ctrl_send_xn_wrfull(0); 
	
	ctrl_from_sd_data(0) <= ctrl_send_xn_q(0);	
	ctrl_from_sd_busy(0) <= ctrl_send_xn_rdempty(0);	
	ctrl_send_xn_rdreq(0) <= ctrl_from_sd_ack(0);
	ctrl_send_xn_rclk(0) <= ctrl_clk;
		
	-----------------------------------------------------------------------------------------------------------
	-----------------------------------------------------------------------------------------------------------
	---- SEND DATA DATA 
	-----------------------------------------------------------------------------------------------------------
	-----------------------------------------------------------------------------------------------------------
	
	process(all)
		variable tport : port_struct := fill_port('0');
	begin
	
			for i in 0 to mem_sd_card_data_reciver_count-1 loop
	
				
				tport.data 		:= x"0000" & data_send_data;			-- no data
				tport.address 	:= data_send_address;
				tport.we			:= data_send_we;					
				tport.dmode		:= '1';								-- 16 bit mode
	
				-- data
				
	
				data_send_xn_data(i) <=  port_to_raw(tport);
			
				-- wrreq logic 
		
				if (I = 0 and data_send_address(31 downto 28) = x"f" and data_send_wrreq = '1') or
					(I = 1 and data_send_address(31 downto 28) = x"d" and data_send_wrreq = '1') then
					
					data_send_xn_wrreq(i) <= '1'; 
				else
					data_send_xn_wrreq(i) <= '0'; 
				end if;
											
				-- full generate
			
				if rising_edge(clk) then
					if data_send_xn_wrusedw(i) >=  x"fb" then --x"1fb" at 9 bit?
						data_send_xn_wrfull(i) <= '1';
					else
						data_send_xn_wrfull(i) <= '0';
					end if;
				end if;				
				
			
			end loop;
	end process;
	
	send_xf_loop : for i in 0 to mem_sd_card_data_reciver_count-1 generate
         send_xf : dc_fifo 
					generic map
					(
						FIFO_DEPTH 			=> 256,
						FIFO_WIDTH 			=> 72,
						FIFO_SHOWAHEAD 	=> "ON",
						FIFO_USEDWWIDTH	=> 8,
						FIFO_SYNCSTAGES	=> 3
					)   
					port map
					(
						'0',
						data_send_xn_data(i),
						data_send_xn_rclk(i),		-- read with component clk
						data_send_xn_rdreq(i),	
						clk,					-- write with ucore clk
						data_send_xn_wrreq(i),	
						data_send_xn_q(i),
						data_send_xn_rdempty(i),
						data_send_xn_wrusedw(i)
					);	
   end generate;
		
	data_send_busy <=  data_send_xn_wrfull(0) or data_send_xn_wrfull(1); 
	
	data_to_sram_data <= data_send_xn_q(0);	
	data_to_sram_busy <= data_send_xn_rdempty(0);	
	data_send_xn_rdreq(0) <= data_to_sram_ack;
	data_send_xn_rclk(0) <= sram_clk;
	
	data_to_sdram_data <= data_send_xn_q(1);	
	data_to_sdram_busy <= data_send_xn_rdempty(1);	
	data_send_xn_rdreq(1) <= data_to_sdram_ack;
	data_send_xn_rclk(1) <= sdram_clk;
	
	
	

end rtl;