library ieee;
use work.hcore_package.all;
use work.bus_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity bus_arbiter_hcore_data is
		port 
		(
				-- global

				clk			: in std_logic;
				reset			: in std_logic;
				
				-- hcores
				
				-- send
				
				hcore_send_data		: in  hcore_send_data_t;	
				hcore_send_address	: in  hcore_send_address_t;
				hcore_send_we			: in  hcore_send_we_t;	
				hcore_send_dmode		: in hcore_send_dmode_t;		
				hcore_send_busy		: out hcore_send_busy_t;	
				hcore_send_wrreq		: in  hcore_send_wrreq_t;	
				
				-- recive
				
				hcore_recive_data		: out hcore_recive_data_t;
				hcore_recive_busy		: out hcore_recive_busy_t;
				hcore_recive_ack		: in  hcore_recive_ack_t;	
								
				-- memory clks
				
				sram_clk					: in std_logic;
				
				-- to memory
				
				to_sdram_data			: out std_logic_vector(71 downto 0);		
				to_sdram_busy			: out std_logic;
				to_sdram_ack				: in std_logic;

				-- from memory fifos
				
				from_sdram_data			: in std_logic_vector(71 downto 0);	
				from_sdram_busy			: in std_logic;	
				from_sdram_ack			: out std_logic
				
		);
end entity;

architecture rtl of bus_arbiter_hcore_data is

-- ids

component hcore_data_id_gen is
		port 
		(
				-- global

				clk			: in std_logic;
				reset			: in std_logic;
				
				-- hcores
				
				ids			: out use_id_t			
		);
end component;

-- 

--component fifo_min
--	PORT
--	(
--		clock		: IN STD_LOGIC ;
--		data		: IN STD_LOGIC_VECTOR (71 DOWNTO 0);
--		rdreq		: IN STD_LOGIC ;
--		sclr		: IN STD_LOGIC ;
--		wrreq		: IN STD_LOGIC ;
--		empty		: OUT STD_LOGIC ;
--		full		: OUT STD_LOGIC ;
--		q		: OUT STD_LOGIC_VECTOR (71 DOWNTO 0);
--		usedw		: OUT STD_LOGIC_VECTOR (7 DOWNTO 0)
--	);
--end component;

component sc_fifo_usedw
	GENERIC 
	(
		FIFO_DEPTH 			: natural;
		FIFO_WIDTH 			: natural;
		FIFO_SHOWAHEAD 	: string;
		FIFO_USERAM			: string;
		FIFO_USEDWWIDTH	: natural
		
	);
	PORT
	(
		clock		: IN STD_LOGIC ;
		data		: IN STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdreq		: IN STD_LOGIC ;
		sclr		: IN STD_LOGIC ;
		wrreq		: IN STD_LOGIC ;
		empty		: OUT STD_LOGIC ;
		full		: OUT STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		usedw		: OUT STD_LOGIC_VECTOR (FIFO_USEDWWIDTH-1 DOWNTO 0)
	);
end component;


type buf_data_t 	is array(0 to hcore_count-1) of STD_LOGIC_VECTOR (71 DOWNTO 0);
type buf_rdreq_t 	is array(0 to hcore_count-1) of STD_LOGIC;
type buf_wrreq_t 	is array(0 to hcore_count-1) of STD_LOGIC;
type buf_empty_t 	is array(0 to hcore_count-1) of STD_LOGIC;
type buf_full_t 	is array(0 to hcore_count-1) of STD_LOGIC;
type buf_q_t 		is array(0 to hcore_count-1) of STD_LOGIC_VECTOR (71 DOWNTO 0);
type buf_usedw_t 	is array(0 to hcore_count-1) of STD_LOGIC_VECTOR (7 DOWNTO 0);

signal bufin_data		: buf_data_t  := (others => (others => '0'));
signal bufin_rdreq	: buf_rdreq_t := (others => '0');
signal bufin_empty	: buf_empty_t := (others => '0');
signal bufin_q			: buf_q_t 	 	:= (others => (others => '0'));
signal bufin_usedw	: buf_usedw_t	:= (others => (others => '0'));
signal bufin_full		: buf_full_t := (others => '0');

signal bufout_data	: STD_LOGIC_VECTOR (71 DOWNTO 0)  := (others => '0');
signal bufout_full	: buf_full_t := (others => '0');
signal bufout_wrreq	: buf_wrreq_t := (others => '0');
signal bufout_q		: buf_q_t := (others => (others => '0'));

-- external memory fifo

--component dcfifo_min IS
--	PORT
--	(
--		aclr		: IN STD_LOGIC  := '0';
--		data		: IN STD_LOGIC_VECTOR (71 DOWNTO 0);
--		rdclk		: IN STD_LOGIC ;
--		rdreq		: IN STD_LOGIC ;
--		wrclk		: IN STD_LOGIC ;
--		wrreq		: IN STD_LOGIC ;
--		q			: OUT STD_LOGIC_VECTOR (71 DOWNTO 0);
--		rdempty	: OUT STD_LOGIC ;
--		wrusedw	: OUT STD_LOGIC_VECTOR (7 DOWNTO 0)
--	);
--END component;

component dc_fifo IS
	
	GENERIC 
	(
		FIFO_DEPTH 		: natural;
		FIFO_WIDTH 		: natural;
		FIFO_SHOWAHEAD		: string;
		FIFO_USEDWWIDTH		: natural;
		FIFO_SYNCSTAGES		: natural		
	);
	PORT
	(
		aclr		: IN STD_LOGIC  := '0';
		data		: IN STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdclk		: IN STD_LOGIC ;
		rdreq		: IN STD_LOGIC ;
		wrclk		: IN STD_LOGIC ;
		wrreq		: IN STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdempty		: OUT STD_LOGIC ;
		wrusedw		: OUT STD_LOGIC_VECTOR (FIFO_USEDWWIDTH-1 DOWNTO 0)
	);
END component;

--

signal send_data		: STD_LOGIC_VECTOR (71 DOWNTO 0) := (others => '0');
signal send_rdreq		: STD_LOGIC := '0';
signal send_rclk		: STD_LOGIC := '0';
signal send_wrreq		: STD_LOGIC := '0';
signal send_q			: STD_LOGIC_VECTOR (71 DOWNTO 0) := (others => '0');
signal send_rdempty	: STD_LOGIC := '0';
signal send_wrfull	: STD_LOGIC := '0'; 
signal send_wrusedw	: STD_LOGIC_VECTOR (7 DOWNTO 0) := (others => '0');

--

signal use_id : use_id_t;

--

signal buf_do_read      : std_logic := '0';
signal buf_do_read_from : integer range 0 to hcore_count-1 := 0;

--

begin

	-----------------------------------------------------------------------------------------------------------
	---- id gen 
	-----------------------------------------------------------------------------------------------------------

	id_gen : hcore_data_id_gen port map
	(
		clk,
		reset,			
		use_id		
	);

	-----------------------------------------------------------------------------------------------------------
	---- SEND 
	-----------------------------------------------------------------------------------------------------------

	-- in buffer
	
	process(hcore_send_data,hcore_send_address,hcore_send_we,hcore_send_dmode)
		variable tport : port_struct := fill_port('0');
	begin
		for i in 0 to hcore_count-1 loop
		
			tport.data 		:= hcore_send_data(i);
			tport.address 	:= hcore_send_address(i);
			tport.we			:= hcore_send_we(i);
			tport.dmode		:= hcore_send_dmode(i);
		
			bufin_data(i) <= port_to_raw(tport);
		end loop;
	end process;

	bufin_loop : for i in 0 to hcore_count-1 generate
         bufin_n_loop : sc_fifo_usedw
				generic map
				(
					FIFO_DEPTH 			=> 256,
					FIFO_WIDTH 			=> 72,
					FIFO_SHOWAHEAD 	=> "ON",
					FIFO_USERAM			=> "ON",
					FIFO_USEDWWIDTH	=> 8				
				)
				port map
            (
					clk,
					bufin_data(i),
					bufin_rdreq(i),
					reset,
					hcore_send_wrreq(i),
					bufin_empty(i),	
					bufin_full(i),	-- not used
					bufin_q(i),
					bufin_usedw(i)
				);	
   end generate;

	process(clk)
	begin
		if rising_edge(clk) then
			for i in 0 to hcore_count-1 loop
				if bufin_usedw(i) >=  x"fc" then
					hcore_send_busy(i) <= '1';
				else
					hcore_send_busy(i) <= '0';
				end if;					
			end loop;
		end if;
	end process;
	
	-- to sdram
	
--	process(clk)
--		variable tport : port_struct := fill_port('0');
--		variable iu : integer range 0 to hcore_count-1;		
--	begin
--		if rising_edge(clk) then
--		
--			send_wrreq <= '0';
--		
--			if send_wrfull = '0' then	-- space in fifo
--				for i in 0 to hcore_count-1 loop
--				
--					iu := use_id(i);
--				
--					if bufin_empty(iu) = '0' then
--											
--						tport									:= raw_to_port(bufin_q(iu));
--						tport.address(31 downto 28) 	:= conv_std_logic_vector(iu,4); -- add id for recive dispatch
--						
--						send_wrreq		<= '1';
--						
--						exit; -- stop loop	
--					end if;
--				end loop;
--			end if;
--			
--			send_data <= port_to_raw(tport);
--			
--		end if;
--	end process;
--	
--	-- buf ack
--	
--	process(all)
--		variable iu : integer range 0 to hcore_count-1;	
--	begin
--		for i in 0 to hcore_count-1 loop
--			bufin_rdreq(i) <= '0';
--		end loop;
--			
--		if send_wrfull = '0' then	-- space in fifo
--			for i in 0 to hcore_count-1 loop
--			
--				iu := use_id(i);
--			
--				if bufin_empty(iu) = '0' then					
--					bufin_rdreq(iu) <= '1';						
--					exit; -- stop loop	
--				end if;
--			end loop;
--		end if;
--	end process;
	
		-- buf get data (after ack)
	
	process(clk)
		variable tport : port_struct := fill_port('0');
	begin
		if rising_edge(clk) then
		
			send_wrreq <= '0';
			
			if buf_do_read = '1' then
			
				tport									:= raw_to_port(bufin_q(buf_do_read_from));
				tport.address(31 downto 28) 	:= conv_std_logic_vector(buf_do_read_from,4); -- add id for recive dispatch
						
				send_wrreq		<= '1'; -- send 
				
			end if;

			send_data <= port_to_raw(tport);
			
		end if;
	end process;
	
	-- buf ack
	
	process(all)
		variable iu : integer range 0 to hcore_count-1;	
	begin
		if rising_edge(clk) then
		
			-- default
		
			buf_do_read 		<= '0';
			buf_do_read_from	<= 0;
			
			for i in 0 to hcore_count-1 loop
				bufin_rdreq(i) <= '0';
			end loop;
			
			--
			
			if reset = '1' then
			else		
				if send_wrfull = '0' then	-- space in fifo
					for i in 0 to hcore_count-1 loop
					
						iu := use_id(i);
					
						-- note: i have to set scfifo to type area, because wrreq / rdreq to usedw was to long 
					
						if bufin_empty(iu) = '1' then
						else
							if bufin_usedw(iu) = 1 and buf_do_read = '1' then
							else
																
								bufin_rdreq(iu) 	<= '1';
								buf_do_read			<= '1';
								buf_do_read_from	<= iu;
							
							   exit; -- stop loop	
						  end if;
						 
						 end if;
					end loop;
				end if;
			end if;
			
		end if;
	end process;
	
	-- out fifo full generate
	
	process(clk)
	begin
		if rising_edge(clk) then
			if send_wrusedw >=  x"fb" then
					send_wrfull <= '1';
				else
					send_wrfull <= '0';
				end if;
		end if;
	end process;
	
	send_fifo : dc_fifo
	generic map
	(
		FIFO_DEPTH 			=> 256,
		FIFO_WIDTH 			=> 72,
		FIFO_SHOWAHEAD 	=> "ON",
		FIFO_USEDWWIDTH	=> 8,
		FIFO_SYNCSTAGES	=> 2
	)
	port map
	(
		'0',
		send_data,
		send_rclk,		-- read with component clk
		send_rdreq,	
		clk,				-- write with ucore clk
		send_wrreq,	
		send_q,
		send_rdempty,
		send_wrusedw
	);
	
	to_sdram_data 	<= send_q;	
	to_sdram_busy 	<= send_rdempty;	
	send_rdreq 		<= to_sdram_ack;
	send_rclk 		<= sram_clk;
		
	-----------------------------------------------------------------------------------------------------------
	---- RECIVE DATA 
	-----------------------------------------------------------------------------------------------------------
	
	process(clk)
		variable id : integer range 0 to hcore_count-1;
	begin
		if rising_edge(clk) then
		
			-- default
			
			for i in 0 to hcore_count-1 loop
				bufout_wrreq(i) <= '0';
			end loop;
		
			if reset = '1' then
			else

				id := conv_integer(raw_to_port(from_sdram_data).address(31 downto 28));
				
				if from_sdram_busy = '0' and bufout_full(id) = '0' then -- there are data and hcore buffer is not full
					bufout_wrreq(id) <= '1';
				end if;

			end if;
			
			bufout_data <= from_sdram_data;
		end if;
	end process;
	
	-- sram ack
	
	process(all)
		variable id : integer range 0 to hcore_count-1;
	begin
	
		-- default
		
		from_sdram_ack <= '0';
	
		if reset = '1' then
		else
		
			id := conv_integer(raw_to_port(from_sdram_data).address(31 downto 28));
					
			if from_sdram_busy = '0' and bufout_full(id) = '0' then 
				from_sdram_ack <= '1';
			end if;
		end if;
	end process;
	
	-- to hcore fifos
	
	bufout_loop : for i in 0 to hcore_count-1 generate
         bufout_n_loop : sc_fifo_usedw
				generic map
				(
					FIFO_DEPTH 			=> 256,
					FIFO_WIDTH 			=> 72,
					FIFO_SHOWAHEAD 	=> "ON",
					FIFO_USERAM			=> "ON",
					FIFO_USEDWWIDTH	=> 8				
				)
				port map
            (
					clk,
					bufout_data,
					hcore_recive_ack(i),
					reset,
					bufout_wrreq(i),
					hcore_recive_busy(i),
					bufout_full(i),	
					bufout_q(i)
				);	
   end generate;
	
	process(bufout_q)
	begin
		for i in 0 to hcore_count-1 loop
			hcore_recive_data(i) <= raw_to_port(bufout_q(i)).data;
		end loop;
	end process;
	
end rtl;