library ieee;
use work.hcore_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity hcore_decode is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;
				
				-- in
				
				i_stall			: in std_logic;			
				i_data			: in std_logic_vector(31 downto 0);
				
				i_roffset		: in std_logic_vector(6 downto 0);
				
				i_en_roffset_num	: in std_logic_vector(6 downto 0);
				i_en_roffset_data	: in std_logic;	
				i_en_roffset_en	: in std_logic;	
				
				-- out
				
				o_ra				: out std_logic_vector(6 downto 0);	
				o_rb				: out std_logic_vector(6 downto 0);	
				o_rc				: out std_logic_vector(6 downto 0);	
				o_rc_we			: out std_logic;	
				o_rc_lock		: out std_logic;
				
				o_imm				: out std_logic_vector(31 downto 0);	
				
				o_a_as_reg		: out std_logic;
				o_b_as_reg		: out std_logic;
				o_c_as_reg		: out std_logic;
				
				o_condition		: out hcondition_t;
				o_opcode			: out hopcode_t;
				o_oo_opcode		: out hopcode_oo_t
			
		);
end entity;


architecture rtl of hcore_decode is

	type offset_enable_t is array(0 to 127) of std_logic;	
	signal offset_enable 	: offset_enable_t := (others => '0');

	--
	
signal s1_o_ra				: std_logic_vector(6 downto 0) := (others => '0');	
signal s1_o_rb				: std_logic_vector(6 downto 0) := (others => '0');
signal s1_o_rc				: std_logic_vector(6 downto 0) := (others => '0');	
signal s1_o_rc_we			: std_logic := '0';	
signal s1_o_rc_lock		: std_logic := '0';	
signal s1_o_imm			: std_logic_vector(31 downto 0) := (others => '0');	
signal s1_o_a_as_reg		: std_logic := '0';	
signal s1_o_b_as_reg		: std_logic := '0';	
signal s1_o_c_as_reg		: std_logic := '0';	
signal s1_o_condition	: hcondition_t := hcc_never;
signal s1_o_opcode		: hopcode_t := hope_nop;
signal s1_o_oo_opcode	: hopcode_oo_t := hope_oo_nop;	

begin

	------------------------------------------------------------
	-- decode stage 1
	------------------------------------------------------------

	process(clk)
	begin
		if rising_edge(clk) then
			if reset = '1' then

				s1_o_ra			<= (others => '0');		
				s1_o_rb			<= (others => '0');			
				s1_o_rc			<= (others => '0');		
				s1_o_rc_we		<= '0';	
				s1_o_rc_lock 	<= '0'; 
				
				s1_o_imm			<= (others => '0');	
				
				
				s1_o_a_as_reg	<= '0';				
				s1_o_b_as_reg	<= '0';	
				s1_o_c_as_reg	<= '0';	
				
				s1_o_condition	<= hcc_never;
				s1_o_opcode		<= hope_nop;
				s1_o_oo_opcode	<= hope_oo_nop;	
						
			else
				if i_stall = '0' then 
									
					-- default
					
					s1_o_rc_lock <= '0'; -- no o_rc lock is default

					-- s1_o_opcode
					
					s1_o_rc_we		<= '0';
					
					if i_data(31 downto 30) = "11" then
					
						-- without s1_o_condition
						
						s1_o_condition	<= hcc_ever;
						
						-- s1_o_opcode
						
						case i_data(29 downto 24)  is
							when "00" & x"0" => s1_o_oo_opcode <= hope_oo_ld; 		 s1_o_opcode <= hope_nop;		s1_o_rc_lock <= '1';							-- oo
							when "00" & x"1" => s1_o_oo_opcode <= hope_oo_ldw; 	 s1_o_opcode <= hope_nop;		s1_o_rc_lock <= '1';							-- oo
							when "00" & x"2" => s1_o_oo_opcode <= hope_oo_ldws; 	 s1_o_opcode <= hope_nop;		s1_o_rc_lock <= '1';							-- oo
							when "00" & x"3" => s1_o_oo_opcode <= hope_oo_nop; 	 s1_o_opcode <= hope_gssr;		s1_o_rc_we	 <= '1';
							when "00" & x"4" => s1_o_oo_opcode <= hope_oo_nop; 	 s1_o_opcode <= hope_sssr;
							when "00" & x"5" => s1_o_oo_opcode <= hope_oo_uexecs;  s1_o_opcode <= hope_nop;		s1_o_rc_we	 <= '1';	s1_o_rc_lock <= '1';	-- oo
							when "00" & x"6" => s1_o_oo_opcode <= hope_oo_ldc; 	 s1_o_opcode <= hope_nop;		s1_o_rc_lock <= '1';	
							when others 	  => s1_o_oo_opcode <= hope_oo_illegal; s1_o_opcode <= hope_illegal;							
						end case;
					
					else							
					
						-- with s1_o_condition
						
						case i_data(25 downto 24) is
							when "00" 	=> s1_o_condition	<= hcc_ever;
							when "01"	=> s1_o_condition	<= hcc_ts;
							when "10" 	=> s1_o_condition	<= hcc_tc;
							when "11" 	=> s1_o_condition	<= hcc_never;
							when others => s1_o_condition	<= hcc_never;
						end case;
						
						-- s1_o_opcode
					
						case i_data(31 downto 26)  is
							when "00" & x"0" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_move;		s1_o_rc_we		<= '1';
							when "00" & x"1" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_moveh;	s1_o_rc_we		<= '1';
							when "00" & x"2" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_add;		s1_o_rc_we		<= '1';
							when "00" & x"3" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_sub;		s1_o_rc_we		<= '1';
							when "00" & x"4" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_cmplo;
							when "00" & x"5" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_cmplos;
							when "00" & x"6" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_cmple;
							when "00" & x"7" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_cmples;
							when "00" & x"8" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_cmpeq;
							when "00" & x"9" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_jmp;
							when "00" & x"a" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_br;
							when "00" & x"b" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_gpc;		s1_o_rc_we		<= '1';
							when "00" & x"c" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_and;		s1_o_rc_we		<= '1';
							when "00" & x"d" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_or;		s1_o_rc_we		<= '1';
							when "00" & x"e" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_xor;		s1_o_rc_we		<= '1';
							when "00" & x"f" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_bic;		s1_o_rc_we		<= '1';
							when "01" & x"0" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_mul;		s1_o_rc_we		<= '1';
							when "01" & x"1" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_mulh;		s1_o_rc_we		<= '1';
							when "01" & x"2" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_lsr;		s1_o_rc_we		<= '1';		
							when "01" & x"3" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_lsl;		s1_o_rc_we		<= '1';	
							when "01" & x"4" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_asr;		s1_o_rc_we		<= '1';					
							when "01" & x"5" => s1_o_oo_opcode <= hope_oo_st;		 s1_o_opcode <= hope_nop;									-- oo	
							when "01" & x"6" => s1_o_oo_opcode <= hope_oo_stw;		 s1_o_opcode <= hope_nop;									-- oo
							when "01" & x"7" => s1_o_oo_opcode <= hope_oo_uexec;	 s1_o_opcode <= hope_nop;									-- oo
							when "01" & x"8" => s1_o_oo_opcode <= hope_oo_nop;	 	 s1_o_opcode <= hope_rsub;		s1_o_rc_we		<= '1';	
							when "01" & x"9" => s1_o_oo_opcode <= hope_oo_stc;		 s1_o_opcode <= hope_nop;									-- oo	
							when "01" & x"a" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_roe;									
							
							when "10" & x"e" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_nop;	-- dump -> nop
							when "10" & x"f" => s1_o_oo_opcode <= hope_oo_nop;		 s1_o_opcode <= hope_nop;	-- nop / fead

							when others 	  => s1_o_oo_opcode <= hope_oo_illegal; s1_o_opcode <= hope_illegal;							
						end case;
						
					end if;
					
					-- abc as reg 

					case i_data(23 downto 21) is
						when "000" 	=>	s1_o_a_as_reg	<= '1';	s1_o_b_as_reg	<= '1';	s1_o_c_as_reg <= '1';			
						when "001" 	=>	s1_o_a_as_reg	<= '1';	s1_o_b_as_reg	<= '1';	s1_o_c_as_reg <= '0';					
						when "010" 	=>	s1_o_a_as_reg	<= '1';	s1_o_b_as_reg	<= '0';	s1_o_c_as_reg <= '1';					
						when "011" 	=>	s1_o_a_as_reg	<= '1';	s1_o_b_as_reg	<= '0';	s1_o_c_as_reg <= '0';					
						when "100" 	=>	s1_o_a_as_reg	<= '0';	s1_o_b_as_reg	<= '1';	s1_o_c_as_reg <= '1';					
						when "101" 	=>	s1_o_a_as_reg	<= '0';	s1_o_b_as_reg	<= '1';	s1_o_c_as_reg <= '0';					
						when "110" 	=>	s1_o_a_as_reg	<= '0';	s1_o_b_as_reg	<= '0';	s1_o_c_as_reg <= '1';					
						when "111" 	=>	s1_o_a_as_reg	<= '0';	s1_o_b_as_reg	<= '0';	s1_o_c_as_reg <= '0';	
						when others => 
					end case;
					
					-- register
							
					case i_data(23 downto 21) is
						when "000" =>	s1_o_rc <= i_data(20 downto 14);		s1_o_rb <= i_data(13 downto 7);		s1_o_ra <= i_data(6 downto 0);		s1_o_imm <= (others => '0');			
						when "001" =>	s1_o_rc <= (others => '0');			s1_o_rb <= i_data(20 downto 14);		s1_o_ra <= i_data(13 downto 7);		s1_o_imm <= "0000000000000000000000000" & i_data(6 downto 0);					
						when "010" =>	s1_o_rc <= i_data(20 downto 14);		s1_o_rb <= (others => '0');			s1_o_ra <= i_data(13 downto 7);		s1_o_imm <= "0000000000000000000000000" & i_data(6 downto 0);				
						when "011" =>	s1_o_rc <= (others => '0');			s1_o_rb <= (others => '0');			s1_o_ra <= i_data(20 downto 14);		s1_o_imm <= "000000000000000000" & i_data(13 downto 0);		
						when "100" =>	s1_o_rc <= i_data(20 downto 14);		s1_o_rb <= i_data(13 downto 7);														s1_o_imm <= "0000000000000000000000000" & i_data(6 downto 0);			
						when "101" =>	s1_o_rc <= (others => '0');			s1_o_rb <= i_data(20 downto 14);														s1_o_imm <= "000000000000000000" & i_data(13 downto 0);
						when "110" =>	s1_o_rc <= i_data(20 downto 14);		s1_o_rb <= (others => '0');															s1_o_imm <= "000000000000000000" & i_data(13 downto 0);
						when "111" =>	s1_o_rc <= (others => '0');			s1_o_rb <= (others => '0');															s1_o_imm <= conv_std_logic_vector(signed(i_data(20 downto 0)),32);
						when others => 
					end case;
									
				end if;
			end if;
		end if;
	end process;
	
	------------------------------------------------------------
	-- decode stage 2
	------------------------------------------------------------

	process(clk)	
	begin
		if rising_edge(clk) then
			if reset = '1' then

				o_ra			<= (others => '0');		
				o_rb			<= (others => '0');			
				o_rc			<= (others => '0');		
				o_rc_we		<= '0';	
				o_rc_lock 	<= '0'; 
				
				o_imm			<= (others => '0');	
				
				
				o_a_as_reg	<= '0';				
				o_b_as_reg	<= '0';	
				o_c_as_reg	<= '0';	
				
				o_condition	<= hcc_never;
				o_opcode		<= hope_nop;
				o_oo_opcode	<= hope_oo_nop;	
				
				for i in 0 to 127 loop
					offset_enable(i) <= '0';
				end loop;
				
			else
				if i_stall = '0' then 						
			
					-- offset enable
											
					if i_en_roffset_en = '1' then		
						offset_enable(conv_integer(i_en_roffset_num)) <= i_en_roffset_data;
					end if;
					
					-- default reg setup
					
					o_ra <= s1_o_ra;
					o_rb <= s1_o_rb;
					o_rc <= s1_o_rc;
					
					--
			
					if offset_enable(conv_integer(s1_o_ra)) = '1' then
						o_ra <= s1_o_ra + i_roffset;
					end if;
					
					if offset_enable(conv_integer(s1_o_rb)) = '1' then
						o_rb <= s1_o_rb + i_roffset;
					end if;
					
					if offset_enable(conv_integer(s1_o_rc)) = '1' then
						o_rc <= s1_o_rc + i_roffset;
					end if;
			
					--
		
					o_rc_we			<= s1_o_rc_we;			
					o_rc_lock		<= s1_o_rc_lock;		
					o_imm				<= s1_o_imm;
					o_a_as_reg		<= s1_o_a_as_reg;	
					o_b_as_reg		<= s1_o_b_as_reg;	
					o_c_as_reg		<= s1_o_c_as_reg;		
					o_condition		<= s1_o_condition;	
					o_opcode			<= s1_o_opcode;		
					o_oo_opcode		<= s1_o_oo_opcode;
					
				end if;
			end if;	
		end if;
	end process;
					
					
					
								
	
					
	
end rtl;
