library ieee;
use work.hcore_package.all;
use work.bus_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity hcore_dispatch_merge is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;
				
				-------------------------------------------------------------------------------------------------
				-- dispatch
				-------------------------------------------------------------------------------------------------
			
				-- stall
				
				i_stall				: in std_logic;			
				o_stall				: out std_logic;
				
				-- from reg file
				
				i_hcore_id			: in std_logic_vector(3 downto 0);

				i_imm					: in std_logic_vector(31 downto 0);	
				i_ra_data			: in std_logic_vector(31 downto 0);		
				i_rb_data			: in std_logic_vector(31 downto 0);
			

				i_a_as_reg			: in std_logic;
				i_b_as_reg			: in std_logic;
				i_c_as_reg			: in std_logic;
			
				i_rc					: in std_logic_vector(6 downto 0);	
				i_rc_we				: in std_logic;	
			
				i_condition			: in hcondition_t;
				i_oo_opcode			: in hopcode_oo_t;
				
				i_t					: in std_logic;
				i_last				: in std_logic_vector(31 downto 0);
				
				i_dbuffer_clear	: in std_logic;
			
				-- to uctrl
			
				o_to_hctrl			: out hcore_to_uctrl_struct_t;
				o_to_hctrl_empty	: out std_logic;
				i_to_hctrl_rdreq	: in std_logic;
				
				-- from uctrl
				
				i_from_uctrl			: in uctrl_to_hcore_struct_t;
				o_from_uctrl_full		: out std_logic;
				i_from_uctrl_wrreq	: in std_logic;
				
				-------------------------------------------------------------------------------------------------
				-- merge
				-------------------------------------------------------------------------------------------------
			
				i_exec_rc_data			: in std_logic_vector(31 downto 0);
				i_exec_rc				: in std_logic_vector(6 downto 0);	
				i_exec_rc_we			: in std_logic;
				
				-- to reg file
				
				o_rc_data			: out std_logic_vector(31 downto 0);
				o_rc					: out std_logic_vector(6 downto 0);	
				o_rc_we				: out std_logic;
				o_rc_unlock			: out std_logic;
				
				-------------------------------------------------------------------------------------------------
				-- memory
				-------------------------------------------------------------------------------------------------
				
				-- recive
				
				data_recive_data	: in std_logic_vector(31 downto 0);
				data_recive_busy	: in std_logic;
				data_recive_ack	: out std_logic;	
				
				-- send port
				
				data_send_data		: out std_logic_vector(31 downto 0);
				data_send_address	: out std_logic_vector(31 downto 0);
				data_send_we		: out std_logic;
				data_send_mode		: out std_logic;
				data_send_busy		: in std_logic;
				data_send_wrreq	: out std_logic;
				
				-- preformance 
				
				o_dbuffer_hit		: out std_logic;
				o_dbuffer_miss		: out std_logic
				
		);
end entity;



architecture rtl of hcore_dispatch_merge is

	-- request/store load
	
	component hcore_requester is
		port 
		(
				-- 
				
				clk				: in std_logic;
				reset				: in std_logic;
				
				--
				
				request 			: in std_logic;
				address 			: in std_logic_vector(31 downto 0);
				data				: in std_logic_vector(31 downto 0);
				mode				: in std_logic;
				iswe				: in std_logic;
				
				ready				: out std_logic;
				
				-- send port
				
				data_send_data		: out std_logic_vector(31 downto 0);
				data_send_address	: out std_logic_vector(31 downto 0);
				data_send_we		: out std_logic;
				data_send_mode		: out std_logic;
				data_send_busy		: in std_logic;
				data_send_wrreq	: out std_logic
		);
	end component;
	
	component hcore_load is	
		port 
		(
				-- 
				
				clk				: in std_logic;
				reset				: in std_logic;
				
				-- 
				
				data				: out STD_LOGIC_VECTOR (31 DOWNTO 0);
				has_data 		: out std_logic;
				ack_data			: in std_logic;
				
				-- recive
				
				data_recive_data	: in std_logic_vector(31 downto 0);
				data_recive_busy	: in std_logic;
				data_recive_ack	: out std_logic											
	
		);
	end component;
	
	component hcore_dbuffer is
		generic
		(
			LINEWIDTH : integer := 8
		);
		port 
		(
				-- global

				clk											: in std_logic;
				reset											: in std_logic;
				clear											: in std_logic;	
				
				-- write a
				
				i_adr_a				: in std_logic_vector(30 downto 0);
				i_data_a				: in std_logic_vector(31 downto 0);
				i_data_a_we			: in std_logic;
				
				-- write b
				
				i_adr_b				: in std_logic_vector(30 downto 0);
				i_data_b				: in std_logic_vector(31 downto 0);
				i_data_b_we			: in std_logic;
				
				-- read
				
				i_read_adr				: in std_logic_vector(30 downto 0);
				o_read_data				: out std_logic_vector(31 downto 0);
				o_read_data_dok		: out std_logic
		
		);
	end component;
	
	-- sc_fifo
	
	component sc_fifo IS
	GENERIC 
	(
		FIFO_DEPTH 		: natural;
		FIFO_WIDTH 		: natural;
		FIFO_SHOWAHEAD : string;
		FIFO_USERAM		: string
		
	);
	PORT
	(
		clock		: IN STD_LOGIC ;
		data		: IN STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdreq		: IN STD_LOGIC ;
		sclr		: IN STD_LOGIC ;
		wrreq		: IN STD_LOGIC ;
		empty		: OUT STD_LOGIC ;
		full		: OUT STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0)
	);
	END component;
	
	--
	
	signal to_uctrl_fifo_data	: std_logic_vector(63 downto 0) := (others => '0');
	signal to_uctrl_fifo_rdreq	: std_logic := '0';
	signal to_uctrl_fifo_wrreq	: std_logic := '0';
	signal to_uctrl_fifo_empty	: std_logic := '0';
	signal to_uctrl_fifo_full	: std_logic := '0';
	signal to_uctrl_fifo_q		: std_logic_vector(63 downto 0) := (others => '0');

	--
		
	signal from_uctrl_fifo_data	: std_logic_vector(31 downto 0) := (others => '0');
	signal from_uctrl_fifo_rdreq	: std_logic := '0';
	signal from_uctrl_fifo_wrreq	: std_logic := '0';
	signal from_uctrl_fifo_empty	: std_logic := '0';
	signal from_uctrl_fifo_full	: std_logic := '0';
	signal from_uctrl_fifo_q		: std_logic_vector(31 downto 0) := (others => '0');
	
	--

	signal memory_loop_fifo_data	: std_logic_vector(47 downto 0) := (others => '0');
	signal memory_loop_fifo_rdreq	: std_logic := '0';
	signal memory_loop_fifo_wrreq	: std_logic := '0';
	signal memory_loop_fifo_empty	: std_logic := '0';
	signal memory_loop_fifo_full	: std_logic := '0';
	signal memory_loop_fifo_q		: std_logic_vector(47 downto 0) := (others => '0');
	
	--
	
	signal dbuffer_fifo_data	: std_logic_vector(39 downto 0) := (others => '0');
	signal dbuffer_fifo_rdreq	: std_logic := '0';
	signal dbuffer_fifo_wrreq	: std_logic := '0';
	signal dbuffer_fifo_empty	: std_logic := '0';
	signal dbuffer_fifo_full	: std_logic := '0';
	signal dbuffer_fifo_q		: std_logic_vector(39 downto 0) := (others => '0');
	
	--
	
	signal uctrl_loop_fifo_data	: std_logic_vector(8 downto 0) := (others => '0');
	signal uctrl_loop_fifo_rdreq	: std_logic := '0';
	signal uctrl_loop_fifo_wrreq	: std_logic := '0';
	signal uctrl_loop_fifo_empty	: std_logic := '0';
	signal uctrl_loop_fifo_full	: std_logic := '0';
	signal uctrl_loop_fifo_q		: std_logic_vector(8 downto 0) := (others => '0');
	
	-- dbuffer
	
	signal dbuffer_clear				: std_logic := '0';
	signal dbuffer_adr_a				: std_logic_vector(30 downto 0) := (others => '0');
	signal dbuffer_data_a			: std_logic_vector(31 downto 0) := (others => '0');
	signal dbuffer_data_a_we		: std_logic := '0';
	signal dbuffer_adr_b				: std_logic_vector(30 downto 0) := (others => '0');
	signal dbuffer_data_b			: std_logic_vector(31 downto 0) := (others => '0');
	signal dbuffer_data_b_we		: std_logic := '0';
	signal dbuffer_read_adr			: std_logic_vector(30 downto 0) := (others => '0');
	signal dbuffer_read_data		: std_logic_vector(31 downto 0) := (others => '0');
	signal dbuffer_read_data_dok	: std_logic := '0';
	
	--
	
	function signedExtended32(din : std_logic_vector(15 downto 0))
		return std_logic_vector is
	variable result : std_logic_vector(31 downto 0);
	begin 
	
		for i in 31 downto 16 loop
			result(i) := din(15);
		end loop;
		
		result(15 downto 0) := din;
		
		return  result;		
	end;
	
	-- request/store
	
	signal rs_request 		: std_logic := '0';
	signal rs_address 		: std_logic_vector(31 downto 0) := (others => '0');
	signal rs_data				: std_logic_vector(31 downto 0) := (others => '0');
	signal rs_mode				: std_logic := '0';
	signal rs_iswe				: std_logic := '0';
	signal rs_ready			: std_logic := '0';
	
	-- load
	
	signal ld_data					: std_logic_vector (31 DOWNTO 0) := (others => '0');
	signal ld_has_data 			: std_logic := '0';
	signal ld_ack_data			: std_logic := '0';
		
	--
	
	signal dispatch_execute_stall	: std_logic := '0';
	signal a 		: std_logic_vector(31 downto 0) := (others => '0');
	signal b 		: std_logic_vector(31 downto 0) := (others => '0');
	
	-- calc stage signals
	
	signal calc_address	: std_logic_vector(31 downto 0) := (others => '0');	
	signal calc_data 		: std_logic_vector(31 downto 0) := (others => '0');
	signal calc_opcode	: hopcode_oo_t := hope_oo_nop;
	signal calc_rc			: std_logic_vector(6 downto 0) := (others => '0');
	
begin

	-------------------------------------------------------------------------------------------------
	-------------------------------------------------------------------------------------------------
	-------------------------------------------------------------------------------------------------
	-------------------------------------------------------------------------------------------------
	-- dispatch
	-------------------------------------------------------------------------------------------------
	-------------------------------------------------------------------------------------------------
	-------------------------------------------------------------------------------------------------	
	-------------------------------------------------------------------------------------------------

	-------------------------------------------------------------------------------------------------
	-- dispatch calculate stage
	-------------------------------------------------------------------------------------------------
	
	o_stall 	<= dispatch_execute_stall;
		
	a <= i_ra_data when i_a_as_reg = '1' else i_imm;	-- a mux
	b <= i_rb_data when i_b_as_reg = '1' else i_last; 	-- b mux
	
	-- calculate process (will never stall)
	
	process(all)
	  variable calc_temp : std_logic_vector(31 downto 0);
	begin
	
		calc_temp := a + b;
		dbuffer_read_adr 	<= calc_temp(31 downto 1); -- ever long word access
	
		if rising_edge(clk) then
				
			-------------------------------------------------------------------------------------------------
			-- ever
			-------------------------------------------------------------------------------------------------			

			if reset = '1' then	
				calc_opcode <= hope_oo_nop;
			else
			
				-- default				
				
				calc_opcode <= hope_oo_nop;
							
				-------------------------------------------------------------------------------------------------
				-- start opcode processing
				-------------------------------------------------------------------------------------------------	
						
				if i_stall = '0' and dispatch_execute_stall = '0' then
							
					calc_rc		<= i_rc;
						
					-- test condition
					
					if i_condition = hcc_ever or
						(i_condition = hcc_ts and i_t = '1') or
						(i_condition = hcc_tc and i_t = '0') then
						
						calc_opcode <= i_oo_opcode; -- opcode becomes valid						
						
						case i_oo_opcode is
						
							-------------------------------------------------------------------------------------------------
							-- memory
							-------------------------------------------------------------------------------------------------							
							
							when hope_oo_st |
								  hope_oo_stc |
							     hope_oo_stw		=>	calc_address		<= a + conv_std_logic_vector(signed(i_imm(6 downto 0)),32); 													
															calc_data 			<= b; 														
															
							when hope_oo_ldc		=>	
															calc_address 		<= a + b;
							
							when hope_oo_ld 	|								
								  hope_oo_ldw	|
								  hope_oo_ldws		=>	calc_address <= a + b;
 								  

							-------------------------------------------------------------------------------------------------
							-- uctrl
							-------------------------------------------------------------------------------------------------	
			
							when hope_oo_uexec |
								  hope_oo_uexecs	=>	calc_address <= b;
															calc_data	 <= a;
								  
							when others =>
						end case;
						
					else
						-- condition not fit or never
					end if;
				end if;	
			end if;	
		end if;	
	end process;
	
	-------------------------------------------------------------------------------------------------
	-- dispatch execute stage
	-------------------------------------------------------------------------------------------------
	
	process(all)
		variable rc_we_temp			: std_logic := '0'; 
		variable mloop_temp		   : hcore_memory_loop_struct_t := raw_to_memory_loop((others => '0'));
		variable uloop_temp			: hcore_uctrl_loop_struct_t := raw_to_uctrl_loop((others => '0'));
		variable dqueue_temp			: hcore_memory_dbuffer_queue_struct_t := raw_to_dbuffer_queue((others => '0'));
		variable to_uctrl_temp		: hcore_to_uctrl_struct_t := raw_to_to_uctrl((others => '0'));
	begin
	
		
	
		if rising_edge(clk) then
				
			-------------------------------------------------------------------------------------------------
			-- ever
			-------------------------------------------------------------------------------------------------			

			-- rs defaults
			
			rs_address 	<= calc_address; 
			rs_data 		<= calc_data; 
			
			-- memory loop defaults
			
			mloop_temp.rc 			:= calc_rc;
			mloop_temp.address	:= calc_address; -- important for ldc
			
			-- uloop defaults
			
			uloop_temp.rc := calc_rc; -- optimize me
			
			-- dqueue defaults
			
			dqueue_temp.data := dbuffer_read_data;
			dqueue_temp.rc	  := calc_rc;
			
			-- uctrl defaults
														
			to_uctrl_temp.do_return := '0';
			to_uctrl_temp.hcore_id  := "0000" & i_hcore_id;
			to_uctrl_temp.data		:= calc_data;
			to_uctrl_temp.address	:= calc_address(15 downto 8);
			to_uctrl_temp.code		:= calc_address(7 downto 0);
			
			-- performance
			
			o_dbuffer_hit				<= '0';	
			o_dbuffer_miss				<= '0';	
			
			--
		
			if reset = '1' then	
			
				rs_request 					<= '0';
				rs_mode  					<= '0';					
				rs_iswe  					<= '0';	
				memory_loop_fifo_wrreq 	<= '0';
				uctrl_loop_fifo_wrreq	<= '0';
				to_uctrl_fifo_wrreq		<= '0';
				dbuffer_fifo_wrreq 		<= '0';
				
				dbuffer_adr_a				<= (others => '0');				
				dbuffer_data_a				<= (others => '0');
				dbuffer_data_a_we 		<= '0';	
				
				
				
			else
			
				-- default
										
				rs_mode  					<= '0';					
				rs_iswe  					<= '0';				
				rs_request 					<= '0'; 
				memory_loop_fifo_wrreq 	<= '0';
				uctrl_loop_fifo_wrreq	<= '0';
				to_uctrl_fifo_wrreq		<= '0';
				dbuffer_fifo_wrreq 		<= '0';
				
				-- dbuffer
				
				dbuffer_adr_a		<= calc_address(31 downto 1); 				
				dbuffer_data_a		<= calc_data;	
				dbuffer_data_a_we <= '0';	
				
				-------------------------------------------------------------------------------------------------
				-- start opcode processing
				-------------------------------------------------------------------------------------------------	
						
			--	if i_stall = '0' then
					case calc_opcode is
					
						when hope_oo_nop		=> -- nothing to do
					
						-------------------------------------------------------------------------------------------------
						-- memory
						-------------------------------------------------------------------------------------------------							
						
						when hope_oo_st |
							  hope_oo_stc |
							  hope_oo_stw		=>	--rs_address 	<= calc_address; 
														rs_iswe 		<= '1'; 
														--rs_data 		<= calc_data; 	
														
														if calc_opcode = hope_oo_st or calc_opcode = hope_oo_stc then
															rs_mode <= '0'; 	-- 32 bit write
														else
															rs_mode <= '1'; 	-- 16 bit write
														end if;

														if rs_ready = '1' then
															rs_request <= '1'; 
														end if;
														
														-- stc will write through (also into dbuffer)
														
														if calc_opcode = hope_oo_stc then		
															dbuffer_data_a_we <= '1';	
														end if;
						
						when hope_oo_ld 	|
							  hope_oo_ldc 	|
							  hope_oo_ldw	|
							  hope_oo_ldws		=>	if calc_opcode = hope_oo_ldc and dbuffer_read_data_dok = '1' then -- cach hit
														
															if dbuffer_fifo_full = '0' then
																dbuffer_fifo_wrreq <= '1';
															end if;
														
														else
							  
															--rs_address <= calc_address; 
								  
															if calc_opcode = hope_oo_ld or calc_opcode = hope_oo_ldc then
																rs_mode <= '0';		-- 32 bit read 
															else
																rs_mode <= '1'; 		-- 16 bit read
															end if;
															
															rs_iswe <= '0'; 
														
															-- prepare memory loop fifo (push to top ...)

															case calc_opcode is
																when hope_oo_ld 	=> mloop_temp.code := hmc_ld;
																when hope_oo_ldc 	=> mloop_temp.code := hmc_ldc;
																when hope_oo_ldw 	=> mloop_temp.code := hmc_ldw;
																when hope_oo_ldws => mloop_temp.code := hmc_ldws;
																when others 	=> -- model sim
															end case;
															
															if rs_ready = '1' and memory_loop_fifo_full = '0' then
																rs_request 					<= '1'; -- push to out
																memory_loop_fifo_wrreq	<= '1'; -- push to loop		
															end if;	
															
														end if;
														
														-- preformance measurement
							  
														if calc_opcode = hope_oo_ldc then	
															if dbuffer_read_data_dok = '1' and dbuffer_fifo_full = '0' then
																o_dbuffer_hit				<= '1';	
															else
																o_dbuffer_miss				<= '1';	
															end if;
														end if;
						
						-------------------------------------------------------------------------------------------------
						-- uctrl
						-------------------------------------------------------------------------------------------------	
		
						when hope_oo_uexec |
							  hope_oo_uexecs	=>
							  
													if to_uctrl_fifo_full = '0' then -- to uctrl fifo not full
														-- loop back setup
								  
														if calc_opcode = hope_oo_uexecs then	-- setup loop buffer
															if uctrl_loop_fifo_full = '0' then
																to_uctrl_temp.do_return := '1'; -- we want data back
																uctrl_loop_fifo_wrreq	<= '1'; -- push to loop		
																to_uctrl_fifo_wrreq 		<= '1'; -- push to uctrl	
															end if;
														else -- only send to uctrl without response
															to_uctrl_fifo_wrreq <= '1'; -- push to uctrl
														end if;
													end if;
													
							  
							  
						when others =>
					end case;

				end if;	
		
				memory_loop_fifo_data 	<= memory_loop_to_raw(mloop_temp);
				uctrl_loop_fifo_data 	<= uctrl_loop_to_raw(uloop_temp);
				to_uctrl_fifo_data		<= to_uctrl_to_raw(to_uctrl_temp);
				dbuffer_fifo_data			<= dbuffer_queue_to_raw(dqueue_temp);
				
		--	end if;
		end if;
	end process;
	
	-- ack/stall generator
	
	process(all)
	begin
	
		--default
		
		dispatch_execute_stall 		<= '0';
		
		--
		
		--if i_stall = '0' then	
				case calc_opcode is
				
					-- request/store
				
					when hope_oo_st	 |
						  hope_oo_stc	 |
						  hope_oo_stw	 	=>	if rs_ready = '0' then	-- request fifo is full so stall
													dispatch_execute_stall <= '1';										
												end if;

					when hope_oo_ld	 |
						  hope_oo_ldc	 |
						  hope_oo_ldw	 |
						  hope_oo_ldws		
												=>	if calc_opcode = hope_oo_ldc and dbuffer_read_data_dok = '1' then -- cach hit
														if dbuffer_fifo_full = '1' then
															dispatch_execute_stall <= '1';
														end if;
													else			
														if rs_ready = '0' or memory_loop_fifo_full = '1' then	-- request fifo / loop fifo is full so stall
															dispatch_execute_stall <= '1';										
														end if;
													end if;
												
					when hope_oo_uexec |
						  hope_oo_uexecs	=>	
						 
												if to_uctrl_fifo_full = '1' then -- to uctrl fifo full
													dispatch_execute_stall <= '1';										
												else
													if calc_opcode = hope_oo_uexecs then	
														if uctrl_loop_fifo_full = '1' then -- loop buffer full 
															dispatch_execute_stall <= '1';
														end if;
													end if;
												end if;																									
					--
					
					when others =>
						
				end case;
		--end if;
	end process;
	
	-------------------------------------------------------------------------------------------------
	-------------------------------------------------------------------------------------------------
	-------------------------------------------------------------------------------------------------
	-------------------------------------------------------------------------------------------------
	-- merge
	-------------------------------------------------------------------------------------------------
	-------------------------------------------------------------------------------------------------
	-------------------------------------------------------------------------------------------------	
	-------------------------------------------------------------------------------------------------
	
	process(all)
		variable rc_we_temp			: std_logic := '0'; 
	begin
		if rising_edge(clk) then
				
			if reset = '1' then

				o_rc_data			<= (others => '0');
				o_rc					<= (others => '0');
				o_rc_we				<= '0';
				o_rc_unlock			<= '0';
				
				dbuffer_adr_b		<= (others => '0');		
				dbuffer_data_b		<= (others => '0');
				dbuffer_data_b_we <= '0';	
			
			else	
				-- default	
					
				o_rc_data		<= i_exec_rc_data;
				o_rc				<= i_exec_rc;
				o_rc_we			<= '0';
				o_rc_unlock		<= '0';
				
				-- dbuffer
				
				dbuffer_adr_b		<= raw_to_memory_loop(memory_loop_fifo_q).address(31 downto 1); 				
				dbuffer_data_b		<= ld_data;	
				dbuffer_data_b_we <= '0';	
				
				if i_exec_rc_we	= '1' then	-- exec has priority so dont merge anything here
					
					o_rc_we <= '1'; 				-- we through 
					
				else									-- check from dispatch side
					
					-- data from load
					
					if dbuffer_fifo_empty = '0' then -- dbuffer has data
					
						o_rc 			<= raw_to_dbuffer_queue(dbuffer_fifo_q).rc;
						o_rc_data 	<= raw_to_dbuffer_queue(dbuffer_fifo_q).data;
						o_rc_we  	<= '1';
						o_rc_unlock	<= '1';
					
					elsif ld_has_data = '1' and memory_loop_fifo_empty = '0' then -- there are data 
						case raw_to_memory_loop(memory_loop_fifo_q).code is
							when hmc_ld		=>	o_rc_data 		<= ld_data;
							
							when hmc_ldc	=> o_rc_data		<= ld_data; 
																			dbuffer_data_b_we <= '1'; -- write into dbuffer (try)
																			
							when hmc_ldw	=>	o_rc_data 		<= x"0000" & ld_data(15 downto 0);
							when hmc_ldws	=>	o_rc_data 		<= signedExtended32(ld_data(15 downto 0));
							when others 	=>	o_rc_data 		<= ld_data;
						end case;
						
						o_rc 			<= raw_to_memory_loop(memory_loop_fifo_q).rc;
						o_rc_we  	<= '1';
						o_rc_unlock	<= '1';
					elsif from_uctrl_fifo_empty = '0' and uctrl_loop_fifo_empty = '0' then -- datas from uctrl
					
						o_rc_data	<= from_uctrl_fifo_q;
						o_rc 			<= raw_to_uctrl_loop(uctrl_loop_fifo_q).rc;
						o_rc_we  	<= '1';
						o_rc_unlock	<= '1';
						
					end if;
					
				end if;
			end if;
		end if;
		
		-- memory ack

		memory_loop_fifo_rdreq	<= '0';
		uctrl_loop_fifo_rdreq	<= '0';
		from_uctrl_fifo_rdreq	<= '0';
		dbuffer_fifo_rdreq		<= '0';
		
		ld_ack_data 				<= '0';
		if i_exec_rc_we = '0' and reset = '0' then	-- only if no rc_we is not used
			if dbuffer_fifo_empty = '0' then -- dbuffer has data
				dbuffer_fifo_rdreq		<= '1';
			elsif ld_has_data = '1' and memory_loop_fifo_empty = '0' then -- there are data 
				ld_ack_data 				<= '1';
				memory_loop_fifo_rdreq	<= '1';
			elsif from_uctrl_fifo_empty = '0' and uctrl_loop_fifo_empty = '0' then -- datas from uctrl
				uctrl_loop_fifo_rdreq	<= '1';
				from_uctrl_fifo_rdreq	<= '1';
			end if;
		end if;
		
	end process;
	
	-- request/store
	
	request_store : hcore_requester port map
	(
		clk,
		reset,
		rs_request, 				
		rs_address, 				
		rs_data,						
		rs_mode,						
		rs_iswe,						
		rs_ready,					
		data_send_data,		
		data_send_address,	
		data_send_we,		
		data_send_mode,		
		data_send_busy,		
		data_send_wrreq	
	);
	
	-- load
	
	load : hcore_load port map
	(
		clk,
		reset,
		ld_data,			
		ld_has_data,
		ld_ack_data,
		data_recive_data,	
	   data_recive_busy,	
      data_recive_ack
	);
	
	-- memory loop fifo
	
	memory_loop : sc_fifo 
	generic map 
	(
		FIFO_DEPTH 		=> 256,
		FIFO_WIDTH 		=> 48,
		FIFO_SHOWAHEAD => "ON",
		FIFO_USERAM		=> "ON"
	)
	port map
	(
		clk,
		memory_loop_fifo_data,
		memory_loop_fifo_rdreq,
		reset,
		memory_loop_fifo_wrreq,
		memory_loop_fifo_empty,
		memory_loop_fifo_full,
		memory_loop_fifo_q
	);
	
	-- memory loop fifo
	
	dbuffer_queue : sc_fifo 
	generic map 
	(
		FIFO_DEPTH 		=> 256,
		FIFO_WIDTH 		=> 40,
		FIFO_SHOWAHEAD => "ON",
		FIFO_USERAM		=> "ON"
	)
	port map
	(
		clk,
		dbuffer_fifo_data,
		dbuffer_fifo_rdreq,
		reset,
		dbuffer_fifo_wrreq,
		dbuffer_fifo_empty,
		dbuffer_fifo_full,
		dbuffer_fifo_q
	);
	
	-------------------------------------------------------------------------------------------------
	-- uctrl
	-------------------------------------------------------------------------------------------------
	
	-- uctrl loop fifo
	
	uctrl_loop : sc_fifo 
	generic map 
	(
		FIFO_DEPTH 		=> 256,
		FIFO_WIDTH 		=> 9,
		FIFO_SHOWAHEAD => "ON",
		FIFO_USERAM		=> "ON"
	)
	port map
	(
		clk,
		uctrl_loop_fifo_data,
		uctrl_loop_fifo_rdreq,
		reset,
		uctrl_loop_fifo_wrreq,
		uctrl_loop_fifo_empty,
		uctrl_loop_fifo_full,
		uctrl_loop_fifo_q
	);
	
	-- to uctrl fifo
	
	o_to_hctrl				<= raw_to_to_uctrl(to_uctrl_fifo_q);	
	o_to_hctrl_empty		<= to_uctrl_fifo_empty;
	to_uctrl_fifo_rdreq 	<= i_to_hctrl_rdreq;
	
	to_uctrl : sc_fifo 
	generic map 
	(
		FIFO_DEPTH 		=> 256,
		FIFO_WIDTH 		=> 64,
		FIFO_SHOWAHEAD => "ON",
		FIFO_USERAM		=> "ON"
	)
	port map
	(
		clk,
		to_uctrl_fifo_data,
		to_uctrl_fifo_rdreq,
		reset,
		to_uctrl_fifo_wrreq,
		to_uctrl_fifo_empty,
		to_uctrl_fifo_full,
		to_uctrl_fifo_q
	);
	
	-- from uctrl
	
	from_uctrl_fifo_data		<= i_from_uctrl.data;
	from_uctrl_fifo_wrreq	<= i_from_uctrl_wrreq;
	o_from_uctrl_full			<= from_uctrl_fifo_full;
	
	from_uctrl : sc_fifo 
	generic map 
	(
		FIFO_DEPTH 		=> 256,
		FIFO_WIDTH 		=> 32,
		FIFO_SHOWAHEAD => "ON",
		FIFO_USERAM		=> "ON"
	)
	port map
	(
		clk,						--
		from_uctrl_fifo_data,	--
		from_uctrl_fifo_rdreq,
		reset,
		from_uctrl_fifo_wrreq,	--
		from_uctrl_fifo_empty,	
		from_uctrl_fifo_full,	--
		from_uctrl_fifo_q		

	);
	
	-- dbuffer
	
	dbuffer_clear <= i_dbuffer_clear;
	
	dbuffer : hcore_dbuffer 
	generic map 
	(
		LINEWIDTH => 4 -- 64 entrys
	)
	port map
	(
		clk,
		reset,
		dbuffer_clear,
		dbuffer_adr_a,				
		dbuffer_data_a,			
		dbuffer_data_a_we,		
		dbuffer_adr_b,				
		dbuffer_data_b,			
		dbuffer_data_b_we,		
		dbuffer_read_adr,			
		dbuffer_read_data,		
		dbuffer_read_data_dok
	);	

end rtl;
