library ieee;
use work.hcore_package.all;
use work.bus_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity hcore_execute is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;
				
				--
			
				i_hcore_id					: in std_logic_vector(3 downto 0);
				
				-- stall
				
				i_stall				: in std_logic;			
				o_stall				: out std_logic;
				
				--
				
				i_hcore_data		: in hcore_ctrl_struct_t;
				i_hcore_we			: in std_logic;
				
				-- from reg file

				i_imm					: in std_logic_vector(31 downto 0);	
				i_ra_data			: in std_logic_vector(31 downto 0);		
				i_rb_data			: in std_logic_vector(31 downto 0);
			

				i_a_as_reg			: in std_logic;
				i_b_as_reg			: in std_logic;
				i_c_as_reg			: in std_logic;
			
				i_rc					: in std_logic_vector(6 downto 0);	
				i_rc_we				: in std_logic;	
			
				i_condition			: in hcondition_t;
				i_opcode				: in hopcode_t;
				
				-- to merge stage
				
				o_rc_data			: out std_logic_vector(31 downto 0);
				o_rc					: out std_logic_vector(6 downto 0);	
				o_rc_we				: out std_logic;
				
				-- to fetch
				
				o_pc					: out std_logic_vector(31 downto 0);
				
				-- to dispatch
				
				o_t					: out std_logic;
				o_last				: out std_logic_vector(31 downto 0);
				o_dbuffer_clear	: out std_logic;
				
				-- to decoder
				
				o_roffset			: out std_logic_vector(6 downto 0);
				
				o_en_roffset_num 	: out std_logic_vector(6 downto 0);
				o_en_roffset_data	: out std_logic;	
				o_en_roffset_en	: out std_logic;	
				
				-- performance
				
				o_update_cnt_max							: out std_logic_vector(31 downto 0);
				o_update_cnt_max_we						: out std_logic;
				
				o_preformance_clear_cnt					: out std_logic;
				o_preformance_enable						: out std_logic;
				
				o_performance_soft_trigger				: out std_logic;
				
				i_stored_icache_hit_cnt 				: in std_logic_vector(31 downto 0);
				i_stored_fetch_reload_cnt 			   : in std_logic_vector(31 downto 0);
				i_stored_score_stall_cnt 			   : in std_logic_vector(31 downto 0);
				i_stored_execute_stall_cnt 			: in std_logic_vector(31 downto 0);
				i_stored_dispatch_merge_stall_cnt   : in std_logic_vector(31 downto 0);
				i_stored_dbuffer_hit_cnt				: in std_logic_vector(31 downto 0);
				i_stored_dbuffer_miss_cnt				: in std_logic_vector(31 downto 0);
				i_stored_soft_cnt							: in std_logic_vector(31 downto 0)
				
				
		);
end entity;

architecture rtl of hcore_execute is

	-- op components with reslut
	
	component move is port(clk : in std_logic;	reset : in std_logic; a : in std_logic_vector(31 downto 0); b : in std_logic_vector(31 downto 0); res : out std_logic_vector(31 downto 0)	); end component;
	component moveh is port(clk : in std_logic;	reset : in std_logic; a : in std_logic_vector(31 downto 0); b : in std_logic_vector(31 downto 0); res : out std_logic_vector(31 downto 0)	); end component;
	component add is port(clk : in std_logic;	reset : in std_logic; a : in std_logic_vector(31 downto 0); b : in std_logic_vector(31 downto 0); res : out std_logic_vector(31 downto 0)	); end component;	
	component sub is port(clk : in std_logic;	reset : in std_logic; a : in std_logic_vector(31 downto 0); b : in std_logic_vector(31 downto 0); res : out std_logic_vector(31 downto 0)	); end component;
	component oand is port(clk : in std_logic;	reset : in std_logic; a : in std_logic_vector(31 downto 0); b : in std_logic_vector(31 downto 0); res : out std_logic_vector(31 downto 0)	); end component;
	component oor is port(clk : in std_logic;	reset : in std_logic; a : in std_logic_vector(31 downto 0); b : in std_logic_vector(31 downto 0); res : out std_logic_vector(31 downto 0)	); end component;
	component oxor is port(clk : in std_logic;	reset : in std_logic; a : in std_logic_vector(31 downto 0); b : in std_logic_vector(31 downto 0); res : out std_logic_vector(31 downto 0)	); end component;	
	component obic is port(clk : in std_logic;	reset : in std_logic; a : in std_logic_vector(31 downto 0); b : in std_logic_vector(31 downto 0); res : out std_logic_vector(31 downto 0)	); end component;	
	component asr is port(clk : in std_logic;	reset : in std_logic; a : in std_logic_vector(31 downto 0); b : in std_logic_vector(31 downto 0); res : out std_logic_vector(31 downto 0)	); end component;
	component lsr is port(clk : in std_logic;	reset : in std_logic; a : in std_logic_vector(31 downto 0); b : in std_logic_vector(31 downto 0); res : out std_logic_vector(31 downto 0)	); end component;
	component lsl is port(clk : in std_logic;	reset : in std_logic; a : in std_logic_vector(31 downto 0); b : in std_logic_vector(31 downto 0); res : out std_logic_vector(31 downto 0)	); end component;
	
	-- clocked multicycle
	
	component mul is port(clk : in std_logic;	reset : in std_logic; stall : in std_logic; a : in std_logic_vector(31 downto 0); b : in std_logic_vector(31 downto 0); res : out std_logic_vector(63 downto 0)	); end component;
	
	-- op components with t
	
	component cmplo is port(clk : in std_logic;	reset : in std_logic; a : in std_logic_vector(31 downto 0); b : in std_logic_vector(31 downto 0); t : out std_logic ); end component;
	component cmplos is port(clk : in std_logic;	reset : in std_logic; a : in std_logic_vector(31 downto 0); b : in std_logic_vector(31 downto 0); t : out std_logic ); end component;
	component cmple is port(clk : in std_logic;	reset : in std_logic; a : in std_logic_vector(31 downto 0); b : in std_logic_vector(31 downto 0); t : out std_logic ); end component;
	component cmples is port(clk : in std_logic;	reset : in std_logic; a : in std_logic_vector(31 downto 0); b : in std_logic_vector(31 downto 0); t : out std_logic ); end component;
	component cmpeq is port(clk : in std_logic;	reset : in std_logic; a : in std_logic_vector(31 downto 0); b : in std_logic_vector(31 downto 0); t : out std_logic ); end component;

	-- op results
	
	signal res_move  	: std_logic_vector(31 downto 0) := (others => '0');
	signal res_moveh 	: std_logic_vector(31 downto 0) := (others => '0');
	signal res_add	  	: std_logic_vector(31 downto 0) := (others => '0');
	signal res_sub	  	: std_logic_vector(31 downto 0) := (others => '0');
	signal res_rsub	: std_logic_vector(31 downto 0) := (others => '0');
	signal res_and  	: std_logic_vector(31 downto 0) := (others => '0');
	signal res_or 		: std_logic_vector(31 downto 0) := (others => '0');
	signal res_xor		: std_logic_vector(31 downto 0) := (others => '0');
	signal res_bic		: std_logic_vector(31 downto 0) := (others => '0');
	signal res_mul		: std_logic_vector(63 downto 0) := (others => '0');
	signal res_asr 	: std_logic_vector(31 downto 0) := (others => '0');
	signal res_lsr		: std_logic_vector(31 downto 0) := (others => '0');
	signal res_lsl		: std_logic_vector(31 downto 0) := (others => '0');
	
	signal res_cmplo  : std_logic := '0';
	signal res_cmplos : std_logic := '0';
	signal res_cmple  : std_logic := '0';
	signal res_cmples	: std_logic := '0';
	signal res_cmpeq	: std_logic := '0';
	
	-- udiv steper
	
	signal udiv_remainder: std_logic_vector(31 downto 0) := (others => '0');
	signal udiv_qoutient : std_logic_vector(31 downto 0) := (others => '0');
	signal udiv_divisor  : std_logic_vector(31 downto 0) := (others => '0');
	signal udiv_divident : std_logic_vector(31 downto 0) := (others => '0');
	signal udiv_counter	: std_logic_vector(4 downto 0)  := (others => '0');
	
	-- constant
	
	constant RESET_PC 			: std_logic_vector(31 downto 0) := x"d0100000";
	
	-- internal

	signal t 				: std_logic := '0';
	signal a 				: std_logic_vector(31 downto 0) := (others => '0');
	signal b 				: std_logic_vector(31 downto 0) := (others => '0');
	signal last 			: std_logic_vector(31 downto 0) := (others => '0');
	signal pc				: std_logic_vector(31 downto 0) := RESET_PC;
	signal stall			: std_logic := '0';
	signal stall_total	: std_logic := '0';	
	signal roffset 		: std_logic_vector(6 downto 0) := (others => '0');
	
	--

begin

	--

	stall			<= '0'; -- no internal stall yet
	
	o_stall 		<= stall;
	stall_total <= stall or i_stall;
	
	o_pc 			<= pc;
	o_t			<= t;
	o_last		<= last;
	o_rc_data 	<= last;	
	o_roffset	<= roffset;
	
	a <= i_ra_data when i_a_as_reg = '1' else i_imm;	-- a mux
	b <= i_rb_data when i_b_as_reg = '1' else last; 	-- b mux
		
	process(all)
		variable ssradr_temp			: std_logic_vector(31 downto 0) := (others => '0');
	begin
	
		
	
		if rising_edge(clk) then
				
			-------------------------------------------------------------------------------------------------
			-- ever (also on reset)
			-------------------------------------------------------------------------------------------------			
			
			if i_hcore_we = '1' then
				case i_hcore_data.address is
					when x"0000" 	=>	pc(31 downto 16) 				<= i_hcore_data.data;	-- set pc high
					when x"0001" 	=>	pc(15 downto 0) 				<= i_hcore_data.data;	-- set pc low
					when others =>
				end case;
			end if;
			
			--
		
			if reset = '1' then
				t 									<= '0';	
				o_rc_we 							<= '0';
				o_rc 								<= (others => '0');
				last								<= (others => '0');
				o_update_cnt_max_we			<= '0';
				o_preformance_clear_cnt		<= '1';				
				o_preformance_enable			<= '1';	
				o_dbuffer_clear				<= '0';
				o_performance_soft_trigger <= '0';
				roffset							<= (others => '0');
				o_en_roffset_en				<= '0';
				
			else
			
				-- default			
				
				o_rc_we 							<= '0';	
				o_rc 								<= i_rc;
				o_update_cnt_max_we			<= '0';
				o_preformance_clear_cnt		<= '0';	-- 
				o_dbuffer_clear				<= '0';
				o_performance_soft_trigger <= '0';
				o_en_roffset_en				<= '0';
	
				-------------------------------------------------------------------------------------------------
				-- start opcode processing
				-------------------------------------------------------------------------------------------------	
						
				if i_stall = '0' and stall = '0' then
				
					-- pc
					
					pc <= pc + 2;
					
					-- reg we
					
					if (i_condition = hcc_ever or
						(i_condition = hcc_ts and t = '1') or
						(i_condition = hcc_tc and t = '0')) and
						i_c_as_reg = '1' then
					
							if i_opcode = hope_nop then
								-- nop will never enable rc we (bypass from oo)
							else
								o_rc_we <= i_rc_we;	-- if condition is true and c is a reg then push rc_we through
							end if;
							
					end if;
				
					-- test condition
					
					if i_condition = hcc_ever or
						(i_condition = hcc_ts and t = '1') or
						(i_condition = hcc_tc and t = '0') then
						
						case i_opcode is
						
							-------------------------------------------------------------------------------------------------
							-- normal in order
							-------------------------------------------------------------------------------------------------							
													
							when hope_move			=> last <= res_move;
							when hope_moveh		=> last <= res_moveh;
							when hope_add			=> last <= res_add;
							when hope_sub			=> last <= res_sub;
							when hope_rsub			=> last <= res_rsub;
							when hope_cmplo		=> t <= res_cmplo;
							when hope_cmplos		=> t <= res_cmplos;
							when hope_cmple		=> t <= res_cmple;
							when hope_cmples		=> t <= res_cmples;
							when hope_cmpeq		=> t <= res_cmpeq;
							when hope_jmp			=> pc <= res_add(31 downto 1) & '0';
							when hope_br			=> pc <= pc + (a(31 downto 1) & '0');
							when hope_gpc			=> last <= pc + (a(31 downto 1) & '0');							
							when hope_and			=> last <= res_and;
							when hope_or			=> last <= res_or;
							when hope_xor			=> last <= res_xor;
							when hope_bic			=> last <= res_bic;
							when hope_mul			=> last <= res_mul(31 downto 0);
							when hope_mulh			=> last <= res_mul(63 downto 32);
							when hope_asr			=> last <= res_asr;
							when hope_lsr			=> last <= res_lsr;
							when hope_lsl			=> last <= res_lsl;
							
							when hope_gssr			=> case a(3 downto 0) is
																when x"0" => 	last <=  x"0000000" & i_hcore_id;						-- core num
																when x"1" => 	last <=  conv_std_logic_vector(hcore_count,32);	-- total core count
																when x"2" => 	last <=  (others => '0');	
																when x"3" => 	last <= x"000000" & '0' & roffset;	
																when x"4" =>	last <= i_stored_icache_hit_cnt; 			
																when x"5" =>	last <= i_stored_fetch_reload_cnt; 			
																when x"6" =>	last <= i_stored_score_stall_cnt; 			
																when x"7" =>	last <= i_stored_execute_stall_cnt; 		
																when x"8" =>	last <= i_stored_dispatch_merge_stall_cnt;
																when x"9" =>	last <= i_stored_dbuffer_hit_cnt;
																when x"A" =>	last <= i_stored_dbuffer_miss_cnt;
																when x"B" => 	last <= i_stored_soft_cnt;
																
																when others => last <=  (others => '0');
															end case;
														
							when hope_sssr			=> case a(3 downto 0) is	
															
																when x"3" =>	roffset <= b(6 downto 0); 
																when x"4" =>	if b(0) = '1' then
																						o_preformance_clear_cnt <= '1';
																					end if;
																					
																					if b(2) = '1' then
																						o_preformance_enable <= b(1);
																					end if;
																					
																					if b(3) = '1' then
																						o_dbuffer_clear <= '1';
																					end if;
																					
																					if b(4) = '1' then
																						o_performance_soft_trigger <= '1';
																					end if;
																					
																when x"5" =>	o_update_cnt_max_we		<= '1';
																					o_update_cnt_max			<= b;
																
																					
																when others =>
															end case;
							
							when hope_roe			=> o_en_roffset_num 	<= b(6 downto 0);
															o_en_roffset_data	<= a(0);
															o_en_roffset_en	<= '1';
							
--							when hope_udivinit 	=>	udiv_remainder <= (others => '0');
--															udiv_qoutient 	<= (others => '0');
--															udiv_divisor 	<= a;
--															udiv_divident 	<= b;
--															udiv_counter	<= i_imm(4 downto 0);
--																						
--							when hope_udivstep	=>	udiv_qtemp := udiv_qoutient;
--															udiv_temp  := udiv_remainder(30 downto 0) & udiv_divident(conv_integer(udiv_counter));
--															if udiv_temp >= udiv_divisor then
--																udiv_temp := udiv_temp - udiv_divisor;
--																udiv_qtemp(conv_integer(udiv_counter)) := '1';
--															end if;
--															udiv_remainder <= udiv_temp;
--															udiv_counter <= udiv_counter - 1;
--															last <=  udiv_qtemp;
--															udiv_qoutient <= udiv_qtemp;
--							
--							when hope_udivgrem	=>	last <=  udiv_remainder;
																
							when others =>
						end case;
						
					else
						-- condition not fit or never
					end if;
				end if;			
			end if;					
		end if;	
	end process;
	
	-- operations (0 cycle delay)
	
	op_move 		: move 	port map ( clk,reset,a,b,res_move	);
	op_moveh 	: moveh	port map ( clk,reset,a,b,res_moveh	);
	op_add 		: add 	port map ( clk,reset,a,b,res_add		);
	op_sub 		: sub 	port map ( clk,reset,a,b,res_sub		);
	op_rsub 		: sub 	port map ( clk,reset,b,a,res_rsub	);
	op_and 		: oand 	port map ( clk,reset,a,b,res_and		);
	op_or 		: oor		port map ( clk,reset,a,b,res_or		);
	op_xor		: oxor 	port map ( clk,reset,a,b,res_xor		);
	op_bic 		: obic 	port map ( clk,reset,a,b,res_bic		);
	op_asr 		: asr 	port map ( clk,reset,a,b,res_asr		);
	op_lsr 		: lsr 	port map ( clk,reset,a,b,res_lsr		);
	op_lsl 		: lsl 	port map ( clk,reset,a,b,res_lsl		);
	op_cmplo 	: cmplo 	port map ( clk,reset,a,b,res_cmplo  );
	op_cmplos 	: cmplos	port map ( clk,reset,a,b,res_cmplos );
	op_cmple 	: cmple 	port map ( clk,reset,a,b,res_cmple  );
	op_cmples 	: cmples port map ( clk,reset,a,b,res_cmples	);
	op_cmpeq 	: cmpeq  port map ( clk,reset,a,b,res_cmpeq	);
	
	-- operation (1 cycle delay)
	
	op_mul 		: mul  	port map ( clk,reset,stall_total,a,b,res_mul		);
			
end rtl;
