library ieee;
use work.hcore_package.all;
use work.bus_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity hcore_ssregs is
		generic 
		(
			reg_cnt : natural := 8
		);

		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;
				
				-- p1
				
				p1_adr			: in integer range 0 to reg_cnt-1;
				p1_q				: out std_logic_vector(31 downto 0);
				p1_data			: in std_logic_vector(31 downto 0);
				p1_we				: in std_logic;
				
				-- p2
				
				p2_adr			: in integer range 0 to reg_cnt-1;
				p2_q				: out std_logic_vector(31 downto 0);
				p2_data			: in std_logic_vector(31 downto 0);
				p2_we				: in std_logic
	
		);
end entity;

architecture rtl of hcore_ssregs is

type ss_regs_t is array(0 to reg_cnt-1) of std_logic_vector(31 downto 0);		
signal ss_regs : ss_regs_t := (others => (others => '0'));	

begin

	process(clk)
	begin
		if rising_edge(clk) then
			if reset = '1' then
			else
				if p1_we = '1' then
					ss_regs(p1_adr) <= p1_data;	
				end if;
				
				if p2_we = '1' then
					ss_regs(p2_adr) <= p2_data;	
				end if;
			end if;
			
			p1_q <= ss_regs(p1_adr);
			p2_q <= ss_regs(p2_adr);
		end if;
	end process;

end rtl;