library ieee;
use work.bus_package.all;
use work.hcore_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity hcore is
		generic 
		(
			has_performance_counter : std_logic := '0'			
		);
		port 
		(
				-- global

				clk					: in std_logic;
				reset					: in std_logic;
				
				-- ids
				
				i_hcore_id			: in std_logic_vector(3 downto 0);
				
				-- ctrl
				
				i_ctrl_hcore_data	: in hcore_ctrl_struct_t;
				i_ctrl_hcore_we	: in std_logic;
				
				-- uto ctrl
				
				o_to_hctrl			: out hcore_to_uctrl_struct_t;
				o_to_hctrl_empty	: out std_logic;
				i_to_hctrl_rdreq	: in std_logic;
				
				-- from uctrl
				
				i_from_uctrl			: in uctrl_to_hcore_struct_t;
				o_from_uctrl_full		: out std_logic;
				i_from_uctrl_wrreq	: in std_logic;
				
				--
				
				o_stall			: out std_logic;
				
				-- inst
				
				-- send
				
				inst_send_data		: out std_logic_vector(31 downto 0);
				inst_send_address	: out std_logic_vector(31 downto 0);
				inst_send_we		: out std_logic;
				inst_send_mode		: out std_logic;
				inst_send_busy		: in std_logic;
				inst_send_wrreq	: out std_logic;
				
				-- recive
				
				inst_recive_data	: in std_logic_vector(31 downto 0);
				inst_recive_busy	: in std_logic;
				inst_recive_ack	: out std_logic;
				
				-- data 
				
				-- send port
				
				data_send_data		: out std_logic_vector(31 downto 0);
				data_send_address	: out std_logic_vector(31 downto 0);
				data_send_we		: out std_logic;
				data_send_mode		: out std_logic;
				data_send_busy		: in std_logic;
				data_send_wrreq	: out std_logic;
				
				-- recive
				
				data_recive_data	: in std_logic_vector(31 downto 0);
				data_recive_busy	: in std_logic;
				data_recive_ack	: out std_logic
				
			
		);
end entity;



architecture rtl of hcore is

------------------------------------------------------------
-- s0 fetch
------------------------------------------------------------

component ifetch is
		generic
		(
			CACHE_LINES			: natural := 8;	
			CACHE_LINE_DEPTH	: natural := 6		-- 2^6 == 64
		);
		port 
		(
				-- 
				
				clk					: in std_logic;
				reset					: in std_logic;
				
				-- in
				
				i_address			: in std_logic_vector(31 downto 0);	-- not address is 16 bit (lowest bit is ignored)
				i_stall				: in std_logic;
				o_stall				: out std_logic;
				
				--
				
				i_clear				: in std_logic;
			
				-- out
				
				o_data					: out std_logic_vector(31 downto 0);
				
				-- debug
				
				o_hit						: out std_logic; -- hit for debug
				o_valid_line_found	: out std_logic; 
				o_reload					: out std_logic;
				
				-- send
				
				inst_send_data		: out std_logic_vector(31 downto 0);
				inst_send_address	: out std_logic_vector(31 downto 0);
				inst_send_we		: out std_logic;
				inst_send_busy		: in std_logic;
				inst_send_wrreq	: out std_logic;
				
				-- recive
				
				inst_recive_data	: in std_logic_vector(31 downto 0);
				inst_recive_busy	: in std_logic;
				inst_recive_ack	: out std_logic	
				
		);
end component;

	
signal s0_fetch_i_address			 : std_logic_vector(31 downto 0) := (others => '0');
signal s0_fetch_i_stall				 : std_logic := '0';
signal s0_fetch_o_stall				 : std_logic := '0';
signal s0_fetch_i_clear				 : std_logic := '0';
signal s0_fetch_o_data				 : std_logic_vector(31 downto 0) := (others => '0');
signal s0_fetch_o_hit				 : std_logic := '0';
signal s0_fetch_valid_line_found : std_logic := '0';
signal s0_fetch_o_reload			 : std_logic := '0';

------------------------------------------------------------
-- s1 decode
------------------------------------------------------------

component hcore_decode is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;
				
				-- in
				
				i_stall			: in std_logic;			
				i_data			: in std_logic_vector(31 downto 0);
				
				i_roffset		: in std_logic_vector(6 downto 0);
				
				i_en_roffset_num	: in std_logic_vector(6 downto 0);
				i_en_roffset_data	: in std_logic;	
				i_en_roffset_en	: in std_logic;
				
				-- out
				
				o_ra				: out std_logic_vector(6 downto 0);	
				o_rb				: out std_logic_vector(6 downto 0);	
				o_rc				: out std_logic_vector(6 downto 0);	
				o_rc_we			: out std_logic;	
				o_rc_lock		: out std_logic;
				
				o_imm				: out std_logic_vector(31 downto 0);	
				
				o_a_as_reg		: out std_logic;
				o_b_as_reg		: out std_logic;
				o_c_as_reg		: out std_logic;
				
				o_condition		: out hcondition_t;
				o_opcode			: out hopcode_t;
				o_oo_opcode		: out hopcode_oo_t
		);
end component;

signal s1_s2_decode_i_stall		: std_logic := '0';			
signal s1_s2_decode_i_data			: std_logic_vector(31 downto 0) := (others => '0');
signal s1_s2_decode_o_ra			: std_logic_vector(6 downto 0) := (others => '0');	
signal s1_s2_decode_o_rb			: std_logic_vector(6 downto 0) := (others => '0');	
signal s1_s2_decode_o_rc			: std_logic_vector(6 downto 0) := (others => '0');	
signal s1_s2_decode_o_rc_we		: std_logic  := '0';	
signal s1_s2_decode_o_rc_lock		: std_logic  := '0';
signal s1_s2_decode_o_imm			: std_logic_vector(31 downto 0) := (others => '0');	
signal s1_s2_decode_o_a_as_reg	: std_logic  := '0';
signal s1_s2_decode_o_b_as_reg	: std_logic  := '0';
signal s1_s2_decode_o_c_as_reg	: std_logic  := '0';
signal s1_s2_decode_o_condition	: hcondition_t  := hcc_never;
signal s1_s2_decode_o_opcode		: hopcode_t := hope_nop;
signal s1_s2_decode_o_oo_opcode	: hopcode_oo_t := hope_oo_nop;

------------------------------------------------------------
-- s2_s6 score
------------------------------------------------------------

component hcore_score is
		port 
		(
				-- global

				clk						: in std_logic;
				reset						: in std_logic;
				
				-- 
				
				i_stall					: in std_logic;
				i_clear_locks			: in std_logic;
				o_reg_lock				: out std_logic;
						
				--	s3 / s5
						
				i_ra						: in std_logic_vector(6 downto 0);		
				i_rb						: in std_logic_vector(6 downto 0);		
				o_ra						: out std_logic_vector(6 downto 0);
				o_rb						: out std_logic_vector(6 downto 0);
								
				i_rc_lock				: in std_logic;
				i_rc_unlock				: in std_logic;
				i_rc_num_unlock		: in std_logic_vector(6 downto 0);
						
				o_lock_exception		: out std_logic;
				o_unlock_exception	: out std_logic;
		
				i_rc				: in std_logic_vector(6 downto 0);		-- write
				i_rc_we  		: in std_logic;								--			enable
				o_rc				: out std_logic_vector(6 downto 0);		-- write
				o_rc_we 			: out std_logic;								--			enable
						
				i_imm				: in std_logic_vector(31 downto 0);	
				o_imm				: out std_logic_vector(31 downto 0);	
				
				i_a_as_reg		: in std_logic;
				i_b_as_reg		: in std_logic;
				i_c_as_reg		: in std_logic;
				
				o_a_as_reg		: out std_logic;
				o_b_as_reg		: out std_logic;
				o_c_as_reg		: out std_logic;

				i_condition		: in hcondition_t;
				i_opcode			: in hopcode_t;
				i_oo_opcode		: in hopcode_oo_t;
				o_condition		: out hcondition_t;
				o_opcode			: out hopcode_t;
				o_oo_opcode		: out hopcode_oo_t				
		);
end component;

------------------------------------------------------------
-- s2_s6 score
------------------------------------------------------------

signal s3_s7_score_i_stall					: std_logic := '0';	
signal s3_s7_score_i_clear_locks			: std_logic := '0';	
signal s3_s7_score_o_reg_lock				: std_logic := '0';		
signal s3_s7_score_i_ra						: std_logic_vector(6 downto 0) := (others => '0');			
signal s3_s7_score_i_rb						: std_logic_vector(6 downto 0) := (others => '0');			
signal s3_s7_score_o_ra						: std_logic_vector(6 downto 0) := (others => '0');	
signal s3_s7_score_o_rb						: std_logic_vector(6 downto 0) := (others => '0');	
signal s3_s7_score_i_rc_lock				: std_logic := '0';	
signal s3_s7_score_i_rc_unlock			: std_logic := '0';	
signal s3_s7_score_i_rc_num_unlock				: std_logic_vector(6 downto 0) := (others => '0');	
signal s3_s7_score_o_lock_exception		: std_logic := '0';	-- not used yet FIXME
signal s3_s7_score_o_unlock_exception	: std_logic := '0';	-- not used yet FIXME
signal s3_s7_score_i_rc						: std_logic_vector(6 downto 0) := (others => '0');			
signal s3_s7_score_i_rc_we  				: std_logic := '0';									
signal s3_s7_score_o_rc						: std_logic_vector(6 downto 0) := (others => '0');			
signal s3_s7_score_o_rc_we 				: std_logic := '0';									
signal s3_s7_score_i_imm					: std_logic_vector(31 downto 0) := (others => '0');		
signal s3_s7_score_o_imm					: std_logic_vector(31 downto 0) := (others => '0');		
signal s3_s7_score_i_a_as_reg				: std_logic := '0';
signal s3_s7_score_i_b_as_reg				: std_logic := '0';
signal s3_s7_score_i_c_as_reg				: std_logic := '0';
signal s3_s7_score_o_a_as_reg				: std_logic := '0';
signal s3_s7_score_o_b_as_reg				: std_logic := '0';
signal s3_s7_score_o_c_as_reg				: std_logic := '0';
signal s3_s7_score_i_condition			: hcondition_t := hcc_never;
signal s3_s7_score_i_opcode				: hopcode_t := hope_nop;
signal s3_s7_score_i_oo_opcode			: hopcode_oo_t := hope_oo_nop;
signal s3_s7_score_o_condition			: hcondition_t := hcc_never;
signal s3_s7_score_o_opcode				: hopcode_t := hope_nop;	
signal s3_s7_score_o_oo_opcode			: hopcode_oo_t := hope_oo_nop;

------------------------------------------------------------
-- s3_s6 reg io
------------------------------------------------------------

component hcore_regio is
		port 
		(
				-- global

				clk						: in std_logic;
				reset						: in std_logic;
				
				-- in
				
				i_stall					: in std_logic;
						
				--	s3 / s5
						
				i_ra						: in std_logic_vector(6 downto 0);		-- read
				i_rb						: in std_logic_vector(6 downto 0);		-- read
					
				i_rc						: in std_logic_vector(6 downto 0);		-- write
				i_rc_we					: in std_logic;								--	enable			
				i_rc_data				: in std_logic_vector(31 downto 0);

				o_ra_data				: out std_logic_vector(31 downto 0);
			   o_rb_data				: out std_logic_vector(31 downto 0);
		
				-- hold / through	
	
				i_hold_rc				: in std_logic_vector(6 downto 0);		-- write
				i_hold_rc_we  			: in std_logic;								--			enable
				o_hold_rc				: out std_logic_vector(6 downto 0);		-- write
				o_hold_rc_we 			: out std_logic;								--			enable
								
				i_hold_imm				: in std_logic_vector(31 downto 0);	
				o_hold_imm				: out std_logic_vector(31 downto 0);	
				
				i_hold_a_as_reg		: in std_logic;
				i_hold_b_as_reg		: in std_logic;
				i_hold_c_as_reg		: in std_logic;
				
				o_hold_a_as_reg		: out std_logic;
				o_hold_b_as_reg		: out std_logic;
				o_hold_c_as_reg		: out std_logic;

				i_hold_condition		: in hcondition_t;
				i_hold_opcode			: in hopcode_t;
				i_hold_oo_opcode		: in hopcode_oo_t;
				o_hold_condition		: out hcondition_t;
				o_hold_opcode			: out hopcode_t;	
				o_hold_oo_opcode		: out hopcode_oo_t												
		);
end component;

signal s4_s7_regio_i_stall					: std_logic := '0';
signal s4_s7_regio_i_ra						: std_logic_vector(6 downto 0) := (others => '0');	
signal s4_s7_regio_i_rb						: std_logic_vector(6 downto 0):= (others => '0');	
signal s4_s7_regio_i_rc						: std_logic_vector(6 downto 0):= (others => '0');	
signal s4_s7_regio_i_rc_we					: std_logic := '0';	
signal s4_s7_regio_i_rc_data				: std_logic_vector(31 downto 0) := (others => '0');	
signal s4_s7_regio_o_ra_data				: std_logic_vector(31 downto 0) := (others => '0');	
signal s4_s7_regio_o_rb_data				: std_logic_vector(31 downto 0) := (others => '0');	
signal s4_s7_regio_i_hold_rc				: std_logic_vector(6 downto 0) := (others => '0');
signal s4_s7_regio_i_hold_rc_we  		: std_logic := '0';
signal s4_s7_regio_o_hold_rc				: std_logic_vector(6 downto 0) := (others => '0');	
signal s4_s7_regio_o_hold_rc_we 			: std_logic := '0';
signal s4_s7_regio_i_hold_imm				: std_logic_vector(31 downto 0) := (others => '0');	
signal s4_s7_regio_o_hold_imm				: std_logic_vector(31 downto 0) := (others => '0');	
signal s4_s7_regio_i_hold_a_as_reg		: std_logic := '0';
signal s4_s7_regio_i_hold_b_as_reg		: std_logic := '0';
signal s4_s7_regio_i_hold_c_as_reg		: std_logic := '0';
signal s4_s7_regio_o_hold_a_as_reg		: std_logic := '0';
signal s4_s7_regio_o_hold_b_as_reg		: std_logic := '0';
signal s4_s7_regio_o_hold_c_as_reg		: std_logic := '0';
signal s4_s7_regio_i_hold_condition		: hcondition_t := hcc_never;
signal s4_s7_regio_i_hold_opcode			: hopcode_t := hope_nop;
signal s4_s7_regio_i_hold_oo_opcode		: hopcode_oo_t := hope_oo_nop;
signal s4_s7_regio_o_hold_condition		: hcondition_t := hcc_never;
signal s4_s7_regio_o_hold_opcode			: hopcode_t := hope_nop;	
signal s4_s7_regio_o_hold_oo_opcode		: hopcode_oo_t := hope_oo_nop;

------------------------------------------------------------
-- s4 execute
------------------------------------------------------------

component hcore_execute is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;
				
				--
			
				i_hcore_id					: in std_logic_vector(3 downto 0);
				
				-- stall
				
				i_stall				: in std_logic;			
				o_stall				: out std_logic;
				
				--
				
				i_hcore_data		: in hcore_ctrl_struct_t;
				i_hcore_we			: in std_logic;
				
				-- from reg file

				i_imm					: in std_logic_vector(31 downto 0);	
				i_ra_data			: in std_logic_vector(31 downto 0);		
				i_rb_data			: in std_logic_vector(31 downto 0);
			

				i_a_as_reg			: in std_logic;
				i_b_as_reg			: in std_logic;
				i_c_as_reg			: in std_logic;
			
				i_rc					: in std_logic_vector(6 downto 0);	
				i_rc_we				: in std_logic;	
			
				i_condition			: in hcondition_t;
				i_opcode				: in hopcode_t;
				
				-- to merge stage
				
				o_rc_data			: out std_logic_vector(31 downto 0);
				o_rc					: out std_logic_vector(6 downto 0);	
				o_rc_we				: out std_logic;
				
				-- to fetch
				
				o_pc					: out std_logic_vector(31 downto 0);
				
				-- to dispatch
				
				o_t					: out std_logic;
				o_last				: out std_logic_vector(31 downto 0);
				o_dbuffer_clear	: out std_logic;
				
				-- to decoder
				
				o_roffset			: out std_logic_vector(6 downto 0);
				
				o_en_roffset_num 	: out std_logic_vector(6 downto 0);
				o_en_roffset_data	: out std_logic;	
				o_en_roffset_en	: out std_logic;	
				
				-- performance
				
				o_update_cnt_max							: out std_logic_vector(31 downto 0);
				o_update_cnt_max_we						: out std_logic;
				
				o_preformance_clear_cnt					: out std_logic;
				o_preformance_enable						: out std_logic;
				
				o_performance_soft_trigger				: out std_logic;
				
				i_stored_icache_hit_cnt 				: in std_logic_vector(31 downto 0);
				i_stored_fetch_reload_cnt 			   : in std_logic_vector(31 downto 0);
				i_stored_score_stall_cnt 			   : in std_logic_vector(31 downto 0);
				i_stored_execute_stall_cnt 			: in std_logic_vector(31 downto 0);
				i_stored_dispatch_merge_stall_cnt   : in std_logic_vector(31 downto 0);
				i_stored_dbuffer_hit_cnt				: in std_logic_vector(31 downto 0);
				i_stored_dbuffer_miss_cnt				: in std_logic_vector(31 downto 0);
				i_stored_soft_cnt							: in std_logic_vector(31 downto 0)
				
		);
end component;

signal s5_execute_i_hcore_id			: std_logic_vector(3 downto 0) := (others => '0');	
signal s5_execute_i_stall				: std_logic := '0';			
signal s5_execute_o_stall				: std_logic := '0';
signal s5_execute_i_hcore_data		: hcore_ctrl_struct_t := raw_to_hcore_ctrl((others => '0'));
signal s5_execute_i_hcore_we			: std_logic := '0';
signal s5_execute_i_imm					: std_logic_vector(31 downto 0) := (others => '0');		
signal s5_execute_i_ra_data			: std_logic_vector(31 downto 0) := (others => '0');			
signal s5_execute_i_rb_data			: std_logic_vector(31 downto 0) := (others => '0');	
signal s5_execute_i_a_as_reg			: std_logic := '0';
signal s5_execute_i_b_as_reg			: std_logic := '0';
signal s5_execute_i_c_as_reg			: std_logic := '0';
signal s5_execute_i_rc					: std_logic_vector(6 downto 0) := (others => '0');		
signal s5_execute_i_rc_we				: std_logic := '0';	
signal s5_execute_i_condition			: hcondition_t := hcc_never;
signal s5_execute_i_opcode				: hopcode_t := hope_nop; 
signal s5_execute_o_rc_data			: std_logic_vector(31 downto 0) := (others => '0');	
signal s5_execute_o_rc					: std_logic_vector(6 downto 0) := (others => '0');		
signal s5_execute_o_rc_we				: std_logic := '0'; 
signal s5_execute_o_pc					: std_logic_vector(31 downto 0) := (others => '0');	
signal s5_execute_o_t					: std_logic := '0';
signal s5_execute_o_last				: std_logic_vector(31 downto 0) := (others => '0');	
signal s5_execute_o_dbuffer_clear	: std_logic := '0';
signal s4_performance_soft_trigger	: std_logic := '0'; 	
signal s4_o_roffset						: std_logic_vector(6 downto 0) := (others => '0');	
signal s4_o_en_roffset_num 			: std_logic_vector(6 downto 0) := (others => '0');	
signal s4_o_en_roffset_data			: std_logic := '0'; 		
signal s4_o_en_roffset_en				: std_logic := '0'; 	


------------------------------------------------------------
-- s4_s5 dispatch / merge
------------------------------------------------------------

component hcore_dispatch_merge is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;
				
				-------------------------------------------------------------------------------------------------
				-- dispatch
				-------------------------------------------------------------------------------------------------
			
				-- stall
				
				i_stall				: in std_logic;			
				o_stall				: out std_logic;
				
				-- from reg file
				
				i_hcore_id			: in std_logic_vector(3 downto 0);

				i_imm					: in std_logic_vector(31 downto 0);	
				i_ra_data			: in std_logic_vector(31 downto 0);		
				i_rb_data			: in std_logic_vector(31 downto 0);
			

				i_a_as_reg			: in std_logic;
				i_b_as_reg			: in std_logic;
				i_c_as_reg			: in std_logic;
			
				i_rc					: in std_logic_vector(6 downto 0);	
				i_rc_we				: in std_logic;	
			
				i_condition			: in hcondition_t;
				i_oo_opcode			: in hopcode_oo_t;
				
				i_t					: in std_logic;
				i_last				: in std_logic_vector(31 downto 0);
				
				i_dbuffer_clear	: in std_logic;
				
				-- uctrl
				
				o_to_hctrl			: out hcore_to_uctrl_struct_t;
				o_to_hctrl_empty	: out std_logic;
				i_to_hctrl_rdreq	: in std_logic;
				
				-- from uctrl
				
				i_from_uctrl			: in uctrl_to_hcore_struct_t;
				o_from_uctrl_full		: out std_logic;
				i_from_uctrl_wrreq	: in std_logic;
								
				-------------------------------------------------------------------------------------------------
				-- merge
				-------------------------------------------------------------------------------------------------
			
				i_exec_rc_data			: in std_logic_vector(31 downto 0);
				i_exec_rc				: in std_logic_vector(6 downto 0);	
				i_exec_rc_we			: in std_logic;
				
						
				-- to reg file
				
				o_rc_data			: out std_logic_vector(31 downto 0);
				o_rc					: out std_logic_vector(6 downto 0);	
				o_rc_we				: out std_logic;
				o_rc_unlock			: out std_logic;
				
				-------------------------------------------------------------------------------------------------
				-- memory
				-------------------------------------------------------------------------------------------------
				
				-- recive
				
				data_recive_data	: in std_logic_vector(31 downto 0);
				data_recive_busy	: in std_logic;
				data_recive_ack	: out std_logic;	
				
				-- send port
				
				data_send_data		: out std_logic_vector(31 downto 0);
				data_send_address	: out std_logic_vector(31 downto 0);
				data_send_we		: out std_logic;
				data_send_mode		: out std_logic;
				data_send_busy		: in std_logic;
				data_send_wrreq	: out std_logic;
				
				-- preformance 
				
				o_dbuffer_hit		: out std_logic;
				o_dbuffer_miss		: out std_logic
				
		);
end component;

signal s5_s6_dispatch_merge_i_stall					: std_logic := '0';			
signal s5_s6_dispatch_merge_o_stall					: std_logic := '0';
signal s5_s6_dispatch_merge_i_imm					: std_logic_vector(31 downto 0) := (others => '0');	
signal s5_s6_dispatch_merge_i_ra_data				: std_logic_vector(31 downto 0) := (others => '0');		
signal s5_s6_dispatch_merge_i_rb_data				: std_logic_vector(31 downto 0) := (others => '0');
signal s5_s6_dispatch_merge_i_a_as_reg				: std_logic := '0';
signal s5_s6_dispatch_merge_i_b_as_reg				: std_logic := '0';
signal s5_s6_dispatch_merge_i_c_as_reg				: std_logic := '0';
signal s5_s6_dispatch_merge_i_rc						: std_logic_vector(6 downto 0) := (others => '0');	
signal s5_s6_dispatch_merge_i_rc_we					: std_logic := '0';	
signal s5_s6_dispatch_merge_i_condition			: hcondition_t := hcc_never;
signal s5_s6_dispatch_merge_i_oo_opcode			: hopcode_oo_t := hope_oo_nop;
signal s5_s6_dispatch_merge_i_t						: std_logic := '0';	
signal s5_s6_dispatch_merge_i_last					: std_logic_vector(31 downto 0) := (others => '0');
signal s5_s6_dispatch_merge_i_dbuffer_clear		: std_logic := '0';	
signal s5_s6_dispatch_merge_i_exec_rc_data		: std_logic_vector(31 downto 0) := (others => '0');
signal s5_s6_dispatch_merge_i_exec_rc				: std_logic_vector(6 downto 0) := (others => '0');	
signal s5_s6_dispatch_merge_i_exec_rc_we			: std_logic := '0';
signal s5_s6_dispatch_merge_o_rc_data				: std_logic_vector(31 downto 0) := (others => '0');
signal s5_s6_dispatch_merge_o_rc						: std_logic_vector(6 downto 0) := (others => '0');	
signal s5_s6_dispatch_merge_o_rc_we					: std_logic := '0';
signal s5_s6_dispatch_merge_o_rc_unlock			: std_logic := '0';
signal s5_s6_dispatch_merge_o_dbuffer_hit			: std_logic := '0';
signal s5_s6_dispatch_merge_o_dbuffer_miss		: std_logic := '0';
	
------------------------------------------------------------
-- performance counter
------------------------------------------------------------

component performance_cnt is
		port 
		(
							-- global

				clk											: in std_logic;
				reset											: in std_logic;
				
				-- config
				
				i_update_cnt_max							: in std_logic_vector(31 downto 0);
				i_update_cnt_max_we						: in std_logic;
				
				i_clear_cnt									: in std_logic; -- clear counters (not update max)
				i_enable										: in std_logic;
				
				--	in
				
				i_icache_hit								: in std_logic;
				i_fetch_reload								: in std_logic;
				i_score_stall								: in std_logic;
				i_execute_stall							: in std_logic;
				i_dispatch_merge_stall					: in std_logic;
				i_dbuffer_hit								: in std_logic;
				i_dbuffer_miss								: in std_logic;
				i_soft_trigger								: in std_logic;
				
				-- out
				
				o_stored_icache_hit_cnt 				: out std_logic_vector(31 downto 0);
				o_stored_fetch_reload_cnt 			   : out std_logic_vector(31 downto 0);
				o_stored_score_stall_cnt 			   : out std_logic_vector(31 downto 0);
				o_stored_execute_stall_cnt 			: out std_logic_vector(31 downto 0);
				o_stored_dispatch_merge_stall_cnt   : out std_logic_vector(31 downto 0);
				o_stored_dbuffer_hit_cnt 				: out std_logic_vector(31 downto 0);
				o_stored_dbuffer_miss_cnt   			: out std_logic_vector(31 downto 0);
				o_stored_soft_cnt   						: out std_logic_vector(31 downto 0)
		);
end component;


signal performance_clear_cnt									: std_logic := '0'; 
signal performance_enable										: std_logic := '1'; -- enabled by default
signal performance_enable_generic							: std_logic := '1'; -- enabled by default

signal performance_update_cnt_max							: std_logic_vector(31 downto 0)  := (others => '0');
signal performance_update_cnt_max_we						: std_logic := '0'; 

signal performance_stored_icache_hit_cnt 					: std_logic_vector(31 downto 0) := (others => '0');
signal performance_stored_fetch_reload_cnt 			   : std_logic_vector(31 downto 0) := (others => '0');
signal performance_stored_score_stall_cnt 			   : std_logic_vector(31 downto 0) := (others => '0');
signal performance_stored_execute_stall_cnt 				: std_logic_vector(31 downto 0) := (others => '0');
signal performance_stored_dispatch_merge_stall_cnt   	: std_logic_vector(31 downto 0) := (others => '0');
signal performance_stored_dbuffer_hit_cnt					: std_logic_vector(31 downto 0) := (others => '0'); 
signal performance_stored_dbuffer_miss_cnt				: std_logic_vector(31 downto 0) := (others => '0');
signal performance_stored_soft_cnt							: std_logic_vector(31 downto 0) := (others => '0');

signal debug_stall : std_logic := '0'; 
	
begin

	------------------------------------------------------------
	-- debug / default
	------------------------------------------------------------

	process(clk)
	begin
		if rising_edge(clk) then
			debug_stall <=  s0_fetch_o_stall or s3_s7_score_o_reg_lock or s5_execute_o_stall or s5_s6_dispatch_merge_o_stall; 
			o_stall		 <= debug_stall;
		end if;
	end process;

	------------------------------------------------------------
	-- s0 fetch
	------------------------------------------------------------
	
	inst_send_mode	<= '0'; -- ever 32 bit reads from instruction cache

	s0_fetch_i_address	<= s5_execute_o_pc;	
	s0_fetch_i_stall 		<= s3_s7_score_o_reg_lock or s5_execute_o_stall or s5_s6_dispatch_merge_o_stall;					
	s0_fetch_i_clear 		<= '0'; -- FIXME 			
	
	s0_fetch : ifetch 
	generic map
	(
		CACHE_LINES			=> 8,	
		CACHE_LINE_DEPTH	=> 6		-- 2^6 == 64
	)
	port map
	(
		clk,
		reset,
		s0_fetch_i_address,			
		s0_fetch_i_stall,				
		s0_fetch_o_stall,				
		s0_fetch_i_clear,				
		s0_fetch_o_data,				
		s0_fetch_o_hit,	
		s0_fetch_valid_line_found,
		s0_fetch_o_reload,		
		inst_send_data,
		inst_send_address,
		inst_send_we,
		inst_send_busy,
		inst_send_wrreq,
		inst_recive_data,
		inst_recive_busy,
		inst_recive_ack
	);
	
	------------------------------------------------------------
	-- s1_s2 decode
	------------------------------------------------------------

	s1_s2_decode_i_stall	<= s0_fetch_o_stall or s3_s7_score_o_reg_lock or s5_execute_o_stall or s5_s6_dispatch_merge_o_stall;					
	s1_s2_decode_i_data 	<=	s0_fetch_o_data;
	
	s1_s2_decode : hcore_decode port map	
	(
		clk,
		reset,
		s1_s2_decode_i_stall,					
		s1_s2_decode_i_data,	
		s4_o_roffset,			
		s4_o_en_roffset_num, 
		s4_o_en_roffset_data,
		s4_o_en_roffset_en,			
		s1_s2_decode_o_ra,						
		s1_s2_decode_o_rb,						
		s1_s2_decode_o_rc,						
		s1_s2_decode_o_rc_we,					
		s1_s2_decode_o_rc_lock,				
		s1_s2_decode_o_imm,					
		s1_s2_decode_o_a_as_reg,				
		s1_s2_decode_o_b_as_reg,				
		s1_s2_decode_o_c_as_reg,				
		s1_s2_decode_o_condition,			
		s1_s2_decode_o_opcode,
		s1_s2_decode_o_oo_opcode		
	);

	------------------------------------------------------------
	-- s3_s6 score
	------------------------------------------------------------

	s3_s7_score_i_stall			<= s0_fetch_o_stall or s5_execute_o_stall or s5_s6_dispatch_merge_o_stall;							
	s3_s7_score_i_clear_locks	<= '0'; -- FIXME
	s3_s7_score_i_ra				<= s1_s2_decode_o_ra;
	s3_s7_score_i_rb				<= s1_s2_decode_o_rb;
	s3_s7_score_i_rc_lock		<= s1_s2_decode_o_rc_lock;	
	s3_s7_score_i_rc_unlock		<= s5_s6_dispatch_merge_o_rc_unlock; -- from merge
	s3_s7_score_i_rc_num_unlock<= s5_s6_dispatch_merge_o_rc;			 -- from merge
	s3_s7_score_i_rc				<= s1_s2_decode_o_rc;
	s3_s7_score_i_rc_we			<= s1_s2_decode_o_rc_we;
	s3_s7_score_i_imm				<= s1_s2_decode_o_imm;			
	s3_s7_score_i_a_as_reg		<= s1_s2_decode_o_a_as_reg;	
	s3_s7_score_i_b_as_reg		<= s1_s2_decode_o_b_as_reg;	
	s3_s7_score_i_c_as_reg		<= s1_s2_decode_o_c_as_reg;	
	s3_s7_score_i_condition		<= s1_s2_decode_o_condition;	
	s3_s7_score_i_opcode			<= s1_s2_decode_o_opcode;		
	s3_s7_score_i_oo_opcode		<= s1_s2_decode_o_oo_opcode;
	
	s3_s7_score : hcore_score port map	
	(			
		clk,
		reset,
		s3_s7_score_i_stall,					
		s3_s7_score_i_clear_locks,
		s3_s7_score_o_reg_lock,
		s3_s7_score_i_ra,
		s3_s7_score_i_rb,
		s3_s7_score_o_ra,
		s3_s7_score_o_rb,
		s3_s7_score_i_rc_lock,
		s3_s7_score_i_rc_unlock,
		s3_s7_score_i_rc_num_unlock,
		s3_s7_score_o_lock_exception,	
		s3_s7_score_o_unlock_exception,	
		s3_s7_score_i_rc,	
		s3_s7_score_i_rc_we,	
		s3_s7_score_o_rc,	
		s3_s7_score_o_rc_we,
		s3_s7_score_i_imm,
		s3_s7_score_o_imm,
		s3_s7_score_i_a_as_reg,	
		s3_s7_score_i_b_as_reg,	
		s3_s7_score_i_c_as_reg,	
		s3_s7_score_o_a_as_reg,	
		s3_s7_score_o_b_as_reg,	
		s3_s7_score_o_c_as_reg,	
		s3_s7_score_i_condition,	
	   s3_s7_score_i_opcode,	
      s3_s7_score_i_oo_opcode,
      s3_s7_score_o_condition,
	   s3_s7_score_o_opcode,	
		s3_s7_score_o_oo_opcode
	); 

	------------------------------------------------------------
	-- s4_s7 regio
	------------------------------------------------------------
		
	s4_s7_regio_i_stall				<=	s0_fetch_o_stall or s3_s7_score_o_reg_lock or s5_execute_o_stall or s5_s6_dispatch_merge_o_stall;
	s4_s7_regio_i_ra					<= s3_s7_score_o_ra;
	s4_s7_regio_i_rb					<= s3_s7_score_o_rb;
	s4_s7_regio_i_rc					<= s5_s6_dispatch_merge_o_rc; 		-- from merge
	s4_s7_regio_i_rc_we				<= s5_s6_dispatch_merge_o_rc_we;	 	-- from merge
	s4_s7_regio_i_rc_data			<= s5_s6_dispatch_merge_o_rc_data; 	-- from merge
	s4_s7_regio_i_hold_rc			<= s3_s7_score_o_rc;
	s4_s7_regio_i_hold_rc_we		<= s3_s7_score_o_rc_we;
	s4_s7_regio_i_hold_imm			<= s3_s7_score_o_imm;
	s4_s7_regio_i_hold_a_as_reg	<= s3_s7_score_o_a_as_reg;
	s4_s7_regio_i_hold_b_as_reg	<= s3_s7_score_o_b_as_reg;
	s4_s7_regio_i_hold_c_as_reg	<= s3_s7_score_o_c_as_reg;
	s4_s7_regio_i_hold_condition	<=	s3_s7_score_o_condition;
	s4_s7_regio_i_hold_opcode		<= s3_s7_score_o_opcode;
	s4_s7_regio_i_hold_oo_opcode	<= s3_s7_score_o_oo_opcode;
	
	s4_s7_regio : hcore_regio port map
	(		
		clk,
		reset,
		s4_s7_regio_i_stall,
		s4_s7_regio_i_ra,
		s4_s7_regio_i_rb,
		s4_s7_regio_i_rc,
		s4_s7_regio_i_rc_we,
		s4_s7_regio_i_rc_data,
		s4_s7_regio_o_ra_data,
		s4_s7_regio_o_rb_data,
		s4_s7_regio_i_hold_rc,
		s4_s7_regio_i_hold_rc_we,
		s4_s7_regio_o_hold_rc,
		s4_s7_regio_o_hold_rc_we,
		s4_s7_regio_i_hold_imm,
		s4_s7_regio_o_hold_imm,
		s4_s7_regio_i_hold_a_as_reg,
		s4_s7_regio_i_hold_b_as_reg,
		s4_s7_regio_i_hold_c_as_reg,
		s4_s7_regio_o_hold_a_as_reg,
		s4_s7_regio_o_hold_b_as_reg,
		s4_s7_regio_o_hold_c_as_reg,
		s4_s7_regio_i_hold_condition,	
		s4_s7_regio_i_hold_opcode,		
		s4_s7_regio_i_hold_oo_opcode,	
		s4_s7_regio_o_hold_condition,	
		s4_s7_regio_o_hold_opcode,		
		s4_s7_regio_o_hold_oo_opcode	
	);
	
	------------------------------------------------------------
	-- s5 execute
	------------------------------------------------------------

	s5_execute_i_stall		<=	s0_fetch_o_stall or s3_s7_score_o_reg_lock or s5_s6_dispatch_merge_o_stall;	
	s5_execute_i_hcore_id	<= i_hcore_id;		
	s5_execute_i_hcore_data	<= i_ctrl_hcore_data;	
	s5_execute_i_hcore_we	<= i_ctrl_hcore_we;		
	s5_execute_i_imm			<= s4_s7_regio_o_hold_imm;
	s5_execute_i_ra_data		<= s4_s7_regio_o_ra_data;
	s5_execute_i_rb_data		<= s4_s7_regio_o_rb_data;
	s5_execute_i_a_as_reg	<= s4_s7_regio_o_hold_a_as_reg;	
	s5_execute_i_b_as_reg	<= s4_s7_regio_o_hold_b_as_reg;	
	s5_execute_i_c_as_reg	<= s4_s7_regio_o_hold_c_as_reg;	
	s5_execute_i_rc			<= s4_s7_regio_o_hold_rc;
	s5_execute_i_rc_we		<= s4_s7_regio_o_hold_rc_we;
	s5_execute_i_condition	<= s4_s7_regio_o_hold_condition;
	s5_execute_i_opcode		<= s4_s7_regio_o_hold_opcode;	

	s5_execute : hcore_execute port map
	(
		clk,
		reset,
		s5_execute_i_hcore_id,
		s5_execute_i_stall,	
		s5_execute_o_stall,	
		s5_execute_i_hcore_data,
		s5_execute_i_hcore_we,	
		s5_execute_i_imm,	
		s5_execute_i_ra_data,
		s5_execute_i_rb_data,
		s5_execute_i_a_as_reg,	
		s5_execute_i_b_as_reg,	
		s5_execute_i_c_as_reg,	
		s5_execute_i_rc,	
		s5_execute_i_rc_we,	
		s5_execute_i_condition,	
		s5_execute_i_opcode,	
		s5_execute_o_rc_data,
		s5_execute_o_rc,
		s5_execute_o_rc_we,
		s5_execute_o_pc,
		s5_execute_o_t,		
		s5_execute_o_last,
		s5_execute_o_dbuffer_clear,
		s4_o_roffset,			
		s4_o_en_roffset_num, 
		s4_o_en_roffset_data,
		s4_o_en_roffset_en,	
		performance_update_cnt_max,						
		performance_update_cnt_max_we,	
		performance_clear_cnt,	
		performance_enable,
		s4_performance_soft_trigger,
		performance_stored_icache_hit_cnt, 				
		performance_stored_fetch_reload_cnt, 			
		performance_stored_score_stall_cnt, 			
		performance_stored_execute_stall_cnt, 			
		performance_stored_dispatch_merge_stall_cnt,
		performance_stored_dbuffer_hit_cnt,	
		performance_stored_dbuffer_miss_cnt,
		performance_stored_soft_cnt
	);
	
	------------------------------------------------------------
	-- s5_s6 dispatch merge
	------------------------------------------------------------

	-- dispatch
	
	s5_s6_dispatch_merge_i_stall						<= s0_fetch_o_stall or s3_s7_score_o_reg_lock;								
	s5_s6_dispatch_merge_i_imm							<= s4_s7_regio_o_hold_imm;		
	s5_s6_dispatch_merge_i_ra_data					<= s4_s7_regio_o_ra_data;			
	s5_s6_dispatch_merge_i_rb_data					<= s4_s7_regio_o_rb_data;			
	s5_s6_dispatch_merge_i_a_as_reg					<= s4_s7_regio_o_hold_a_as_reg;				
	s5_s6_dispatch_merge_i_b_as_reg					<= s4_s7_regio_o_hold_b_as_reg;				
	s5_s6_dispatch_merge_i_c_as_reg					<= s4_s7_regio_o_hold_c_as_reg;				
	s5_s6_dispatch_merge_i_rc							<= s4_s7_regio_o_hold_rc;			
	s5_s6_dispatch_merge_i_rc_we						<= s4_s7_regio_o_hold_rc_we;			
	s5_s6_dispatch_merge_i_condition					<= s4_s7_regio_o_hold_condition;		
	s5_s6_dispatch_merge_i_oo_opcode					<= s4_s7_regio_o_hold_oo_opcode;		
	s5_s6_dispatch_merge_i_t							<= s5_execute_o_t;										-- from s4 execute	
	s5_s6_dispatch_merge_i_last                  <= s5_execute_o_last;									-- from s4 execute		
	s5_s6_dispatch_merge_i_dbuffer_clear			<= s5_execute_o_dbuffer_clear;						-- from s4 execute		
	
	-- merge
	
	s5_s6_dispatch_merge_i_exec_rc_data				<= s5_execute_o_rc_data; 								-- from s4 execute	
	s5_s6_dispatch_merge_i_exec_rc				   <= s5_execute_o_rc; 		 								-- from s4 execute 
	s5_s6_dispatch_merge_i_exec_rc_we			   <= s5_execute_o_rc_we;	 								-- from s4 execute
	
	s5_s6_dispatch_merge : hcore_dispatch_merge port map
	(
		clk,
		reset,
		s5_s6_dispatch_merge_i_stall,					
		s5_s6_dispatch_merge_o_stall,	
		i_hcore_id,	
		s5_s6_dispatch_merge_i_imm,					
		s5_s6_dispatch_merge_i_ra_data,				
		s5_s6_dispatch_merge_i_rb_data,				
		s5_s6_dispatch_merge_i_a_as_reg,				
		s5_s6_dispatch_merge_i_b_as_reg,				
		s5_s6_dispatch_merge_i_c_as_reg,				
		s5_s6_dispatch_merge_i_rc,						
		s5_s6_dispatch_merge_i_rc_we,					
		s5_s6_dispatch_merge_i_condition,			
		s5_s6_dispatch_merge_i_oo_opcode,		
		s5_s6_dispatch_merge_i_t,	
		s5_s6_dispatch_merge_i_last,
		s5_s6_dispatch_merge_i_dbuffer_clear,
		o_to_hctrl,		
		o_to_hctrl_empty,
		i_to_hctrl_rdreq,		
		i_from_uctrl,		
		o_from_uctrl_full,	
		i_from_uctrl_wrreq,
		s5_s6_dispatch_merge_i_exec_rc_data,		
		s5_s6_dispatch_merge_i_exec_rc,				
		s5_s6_dispatch_merge_i_exec_rc_we,			
		s5_s6_dispatch_merge_o_rc_data,				
		s5_s6_dispatch_merge_o_rc,						
		s5_s6_dispatch_merge_o_rc_we,					
		s5_s6_dispatch_merge_o_rc_unlock,			
		data_recive_data,
		data_recive_busy,
		data_recive_ack,
		data_send_data,
		data_send_address,
		data_send_we,
		data_send_mode,
		data_send_busy,
		data_send_wrreq,
		s5_s6_dispatch_merge_o_dbuffer_hit,	
		s5_s6_dispatch_merge_o_dbuffer_miss		
	);
	
------------------------------------------------------------
-- performance counter
------------------------------------------------------------

	performance_enable_generic <= performance_enable and has_performance_counter;

	performace_counter : performance_cnt port map
	(
		clk,
		reset,
		performance_update_cnt_max,	
		performance_update_cnt_max_we,
		performance_clear_cnt,	
		performance_enable_generic,		
		s0_fetch_o_hit,
		s0_fetch_o_reload,
		s3_s7_score_o_reg_lock,
		s5_execute_o_stall,
		s5_s6_dispatch_merge_o_stall,
		s5_s6_dispatch_merge_o_dbuffer_hit,	
		s5_s6_dispatch_merge_o_dbuffer_miss,	
		s4_performance_soft_trigger,
		performance_stored_icache_hit_cnt,
		performance_stored_fetch_reload_cnt,
		performance_stored_score_stall_cnt,
		performance_stored_execute_stall_cnt,
		performance_stored_dispatch_merge_stall_cnt,
		performance_stored_dbuffer_hit_cnt, 
		performance_stored_dbuffer_miss_cnt,
		performance_stored_soft_cnt		
	);
		
end rtl;
