library ieee;
use work.hcore_package.all;
use work.bus_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity hcore_ctrl is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;
								
				-- in
				
				hcore_data				: in hcore_ctrl_struct_t;
				hcore_busy				: in std_logic;
				hcore_ack				: out std_logic;
				
				-- out
				
				hcore_ctrls				: out hcore_ctrls_t;
				hcore_wes				: out hcore_wes_t;	
				
				hcores_reset			: out std_logic
				
		);
end entity;


architecture rtl of hcore_ctrl is	

signal ihcores_reset : std_logic := '1';

begin 		

	hcores_reset <= ihcores_reset;

	process(clk)
	begin
		if rising_edge(clk) then
		
			-- default
			
			for i in 0 to hcore_count-1 loop
				hcore_wes(i) <= '0';
			end loop;
		
			if reset = '1' then
				ihcores_reset <= '1';
			else
				if hcore_busy = '0' then	
					if hcore_data.hcore_sel = x"ff" then -- special
						ihcores_reset <= hcore_data.data(0);
					else									 -- to given hcore
						for i in 0 to hcore_count-1 loop
							if i = conv_integer(hcore_data.hcore_sel) then
								hcore_wes(i) <= '1';
							end if;
						end loop;
					end if;
				end if;
			end if;
			
			for i in 0 to hcore_count-1 loop
				hcore_ctrls(i) <= hcore_data;
			end loop;
			
		end if;
	end process;
	
	process(reset,hcore_busy)
	begin
		hcore_ack <= '0';
	
		if reset = '1' then
		else
			if hcore_busy = '0' then	-- note we accept ever, also if hcore select has no valid hcore			
				hcore_ack <= '1';
			end if;
		end if;
	end process;



end rtl;