 library ieee;
use work.hcore_package.all;
use work.bus_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity hcore_uctrl_arb is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;
								
				-- from hcores
				
				i_hcore_to_hctrl			: in hcore_to_uctrl_structs_t;	
				i_hcore_to_hctrl_empty	: in hctrl_emptys;			
				o_hcore_to_hctrl_rdreq	: out hctrl_rdreqs;	
				
				-- to hcores
				
				o_hcore_from_uctrl			: out uctrl_to_hcore_structs_t;
				i_hcore_from_uctrl_full		: in uctrl_fulls;					
				o_hcore_from_uctrl_wrreq	: out uctrl_wrreqs;	
				
				-- to uctrl
				
				i_to_hctrl_clk				: in std_logic;
				o_to_hctrl_data			: out hcore_to_uctrl_struct_t;
				o_to_hctrl_empty			: out std_logic;
				i_to_hctrl_rdreq			: in std_logic;
	
				-- from uctrl
				
				i_from_uctrl_data		: in uctrl_to_hcore_struct_t;
				i_from_uctrl_empty	: in std_logic;
				o_from_uctrl_rdreq	: out std_logic
				
		);
end entity;


architecture rtl of hcore_uctrl_arb is	

-- scfifo to real uctrl

--component from_hcores_to_uctrl_fifo
--	PORT
--	(
--		data		: IN STD_LOGIC_VECTOR (63 DOWNTO 0);
--		rdclk		: IN STD_LOGIC ;
--		rdreq		: IN STD_LOGIC ;
--		wrclk		: IN STD_LOGIC ;
--		wrreq		: IN STD_LOGIC ;
--		q		: OUT STD_LOGIC_VECTOR (63 DOWNTO 0);
--		rdempty		: OUT STD_LOGIC ;
--		wrfull		: OUT STD_LOGIC ;
--		wrusedw		: OUT STD_LOGIC_VECTOR (7 DOWNTO 0)
--	);
--end component;

component dc_fifo IS
	
	GENERIC 
	(
		FIFO_DEPTH 		: natural;
		FIFO_WIDTH 		: natural;
		FIFO_SHOWAHEAD		: string;
		FIFO_USEDWWIDTH		: natural;
		FIFO_SYNCSTAGES		: natural		
	);
	PORT
	(
		aclr		: IN STD_LOGIC  := '0';
		data		: IN STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdclk		: IN STD_LOGIC ;
		rdreq		: IN STD_LOGIC ;
		wrclk		: IN STD_LOGIC ;
		wrreq		: IN STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdempty		: OUT STD_LOGIC ;
		wrusedw		: OUT STD_LOGIC_VECTOR (FIFO_USEDWWIDTH-1 DOWNTO 0)
	);
END component;

signal to_uctrl_fifo_data		: STD_LOGIC_VECTOR (63 DOWNTO 0) := (others => '0');
signal to_uctrl_fifo_rdclk		: STD_LOGIC := '0';
signal to_uctrl_fifo_rdreq		: STD_LOGIC := '0';
signal to_uctrl_fifo_wrclk		: STD_LOGIC := '0';
signal to_uctrl_fifo_wrreq		: STD_LOGIC := '0';
signal to_uctrl_fifo_q			: STD_LOGIC_VECTOR (63 DOWNTO 0) := (others => '0');
signal to_uctrl_fifo_rdempty	: STD_LOGIC := '0';
signal to_uctrl_fifo_wrfull	: STD_LOGIC := '0';
signal to_uctrl_fifo_wrusedw	: STD_LOGIC_VECTOR (7 DOWNTO 0)  := (others => '0');

signal to_uctrl_fifo_full	: STD_LOGIC := '0';

begin 		

	-----------------------------------------------------------------------------------------------------------
	---- to uctrl
	-----------------------------------------------------------------------------------------------------------

	process(clk)
	begin
		if rising_edge(clk) then
			if reset = '1' then
				to_uctrl_fifo_wrreq <= '0';
			else		
				to_uctrl_fifo_wrreq <= '0';
				
				if to_uctrl_fifo_full = '0' then -- out not full
					for i in 0 to hcore_count-1 loop
						if i_hcore_to_hctrl_empty(i) = '0' then -- there are datas
							to_uctrl_fifo_data 	<= to_uctrl_to_raw(i_hcore_to_hctrl(i));
							to_uctrl_fifo_wrreq 	<= '1'; -- write out
							exit; -- stop loop	
						end if;
					end loop;		
				end if;
			end if;
		end if;
	end process;
	
	-- rdreq 
	
	process(all)
	begin
	
	-- default 
		
		for i in 0 to hcore_count-1 loop
			o_hcore_to_hctrl_rdreq(i) <= '0';
		end loop;
	
		if reset = '1' then
		else		
			if to_uctrl_fifo_full = '0' then -- out not full
				for i in 0 to hcore_count-1 loop
					if i_hcore_to_hctrl_empty(i) = '0' then -- there are datas
						o_hcore_to_hctrl_rdreq(i) <= '1';	 -- request						
						exit; -- stop loop	
					end if;
				end loop;	
			end if;
		end if;
	end process;

	-- real fifo to uctrl
	
	to_uctrl_fifo_wrclk 	<= clk;
	to_uctrl_fifo_rdclk	<= i_to_hctrl_clk;
					
	o_to_hctrl_data		<= raw_to_to_uctrl(to_uctrl_fifo_q);
	o_to_hctrl_empty		<= to_uctrl_fifo_rdempty;
	to_uctrl_fifo_rdreq	<= i_to_hctrl_rdreq;

	
	to_uctrl : dc_fifo
	generic map
	(
		FIFO_DEPTH 			=> 256,
		FIFO_WIDTH 			=> 64,
		FIFO_SHOWAHEAD 	=> "OFF",
		FIFO_USEDWWIDTH	=> 8,
		FIFO_SYNCSTAGES	=> 3
	) 
	port map
	(
		'0',
		to_uctrl_fifo_data,		--	
	   to_uctrl_fifo_rdclk,		--
	   to_uctrl_fifo_rdreq,		--
	   to_uctrl_fifo_wrclk,		--	
	   to_uctrl_fifo_wrreq,		-- 	
	   to_uctrl_fifo_q,			--
	   to_uctrl_fifo_rdempty,	--
	   to_uctrl_fifo_wrusedw	--
	);
	
	-- out fifo full generate
	
	process(clk)
	begin
		if rising_edge(clk) then
			if to_uctrl_fifo_wrusedw >=  x"fc" then
					to_uctrl_fifo_full <= '1';
				else
					to_uctrl_fifo_full <= '0';
				end if;
		end if;
	end process;
	
	-----------------------------------------------------------------------------------------------------------
	---- from uctrl
	-----------------------------------------------------------------------------------------------------------
	
	process(clk)
	begin
	
		if rising_edge(clk) then
		
			-- default
				
			for i in 0 to hcore_count-1 loop
				o_hcore_from_uctrl_wrreq(i) 	<= '0';
				o_hcore_from_uctrl(i) 			<= i_from_uctrl_data;
			end loop;
		
			if reset = '1' then
			else
						
				--
			
				if i_from_uctrl_empty = '0' then
					if i_hcore_from_uctrl_full(conv_integer(i_from_uctrl_data.hcore_id)) = '0' then
						o_hcore_from_uctrl_wrreq(conv_integer(i_from_uctrl_data.hcore_id)) <= '1';
					end if;
				end if;
			end if;
			
			
			
		end if;
	end process;
	
	-- rdreq 
	
	process(all)
	begin
	
		o_from_uctrl_rdreq <= '0';
		if i_from_uctrl_empty = '0' then
			if i_hcore_from_uctrl_full(conv_integer(i_from_uctrl_data.hcore_id)) = '0' then
				o_from_uctrl_rdreq <= '1';
			end if;
		end if;
		
	end process;
	
end rtl;