library ieee;
use work.ucore_package.all;
use work.hcore_package.all;
use work.bus_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity hcores is
		port 
		(
				-- global

				clk		: in std_logic;
				reset		: in std_logic;			
				
				-- uctrl in
				
				hcore_data				: in hcore_ctrl_struct_t;
				hcore_busy				: in std_logic;
				hcore_ack				: out std_logic;
				
				-- to uctrl
				
				i_to_hctrl_clk			: in std_logic;
				o_to_hctrl_data		: out hcore_to_uctrl_struct_t;
				o_to_hctrl_empty		: out std_logic;
				i_to_hctrl_rdreq		: in std_logic;	
				
				-- from uctrl
				
				i_from_uctrl_data		: in uctrl_to_hcore_struct_t;
				i_from_uctrl_empty	: in std_logic;
				o_from_uctrl_rdreq	: out std_logic;
			
				-- stall out

				stalls					: out std_logic_vector(15 downto 0);
								
				-- memory clks
							
				sdram_clk				: in std_logic;
				
				-- INST
				
				-- to memory
							
				inst_to_sdram_data		: out std_logic_vector(71 downto 0);		
				inst_to_sdram_busy		: out std_logic;
				inst_to_sdram_ack			: in std_logic;
				
				-- from memory fifos
							
				inst_from_sdram_data		: in std_logic_vector(71 downto 0);	
				inst_from_sdram_busy		: in std_logic;	
				inst_from_sdram_ack		: out std_logic;
				
				-- DATA
				
				-- to memory
							
				data_to_sdram_data		: out std_logic_vector(71 downto 0);		
				data_to_sdram_busy		: out std_logic;
				data_to_sdram_ack			: in std_logic;
				
				-- from memory fifos
							
				data_from_sdram_data		: in std_logic_vector(71 downto 0);	
				data_from_sdram_busy		: in std_logic;	
				data_from_sdram_ack		: out std_logic	
				
														
		);
end entity;

architecture rtl of hcores is

-- ctrl

component hcore_ctrl is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;
								
				-- in
				
				hcore_data				: in hcore_ctrl_struct_t;
				hcore_busy				: in std_logic;
				hcore_ack				: out std_logic;
				
				-- out
				
				hcore_ctrls				: out hcore_ctrls_t;
				hcore_wes				: out hcore_wes_t;
		
				hcores_reset			: out std_logic
				
		);
end component;

signal hcore_ctrls				: hcore_ctrls_t := (others => (others => (others => '0')));		
signal hcore_wes					: hcore_wes_t := (others => '0');		
signal hcores_reset				: std_logic := '0';

-- bus arb

component bus_arbiter_hcore_inst is
		port 
		(
				-- global

				clk			: in std_logic;
				reset			: in std_logic;
				
				-- hcores
				
				-- send
				
				hcore_send_data		: in  hcore_send_data_t;	
				hcore_send_address	: in  hcore_send_address_t;
				hcore_send_we			: in  hcore_send_we_t;	
				hcore_send_dmode		: in hcore_send_dmode_t;		
				hcore_send_busy		: out hcore_send_busy_t;	
				hcore_send_wrreq		: in  hcore_send_wrreq_t;	
				
				-- recive
				
				hcore_recive_data		: out hcore_recive_data_t;
				hcore_recive_busy		: out hcore_recive_busy_t;
				hcore_recive_ack		: in  hcore_recive_ack_t;	
								
				-- memory clks
				
				sram_clk					: in std_logic;
				
				-- to memory
				
				to_sdram_data			: out std_logic_vector(71 downto 0);		
				to_sdram_busy			: out std_logic;
				to_sdram_ack				: in std_logic;

				-- from memory fifos
				
				from_sdram_data			: in std_logic_vector(71 downto 0);	
				from_sdram_busy			: in std_logic;	
				from_sdram_ack			: out std_logic
				
		);
end component;

component bus_arbiter_hcore_data is
		port 
		(
				-- global

				clk			: in std_logic;
				reset			: in std_logic;
				
				-- hcores
				
				-- send
				
				hcore_send_data		: in  hcore_send_data_t;	
				hcore_send_address	: in  hcore_send_address_t;
				hcore_send_we			: in  hcore_send_we_t;	
				hcore_send_dmode		: in hcore_send_dmode_t;		
				hcore_send_busy		: out hcore_send_busy_t;	
				hcore_send_wrreq		: in  hcore_send_wrreq_t;	
				
				-- recive
				
				hcore_recive_data		: out hcore_recive_data_t;
				hcore_recive_busy		: out hcore_recive_busy_t;
				hcore_recive_ack		: in  hcore_recive_ack_t;	
								
				-- memory clks
				
				sram_clk					: in std_logic;
				
				-- to memory
				
				to_sdram_data			: out std_logic_vector(71 downto 0);		
				to_sdram_busy			: out std_logic;
				to_sdram_ack				: in std_logic;

				-- from memory fifos
				
				from_sdram_data			: in std_logic_vector(71 downto 0);	
				from_sdram_busy			: in std_logic;	
				from_sdram_ack			: out std_logic
				
		);
end component;

--- inst

signal inst_send_data		: hcore_send_data_t		:= (others => (others => '0'));	
signal inst_send_address	: hcore_send_address_t 	:= (others => (others => '0'));
signal inst_send_we			: hcore_send_we_t  		:= (others => '0');	
signal inst_send_dmode		: hcore_send_dmode_t 		:= (others => '0');		
signal inst_send_busy		: hcore_send_busy_t 		:= (others => '0');	
signal inst_send_wrreq		: hcore_send_wrreq_t 	:= (others => '0');	

signal inst_recive_data		: hcore_recive_data_t	:= (others => (others => '0'));
signal inst_recive_busy		: hcore_recive_busy_t	:= (others => '0');	
signal inst_recive_ack		: hcore_recive_ack_t		:= (others => '0');	

--- data

signal data_send_data		: hcore_send_data_t		:= (others => (others => '0'));	
signal data_send_address	: hcore_send_address_t 	:= (others => (others => '0'));
signal data_send_we			: hcore_send_we_t  		:= (others => '0');	
signal data_send_dmode		: hcore_send_dmode_t 		:= (others => '0');		
signal data_send_busy		: hcore_send_busy_t 		:= (others => '0');	
signal data_send_wrreq		: hcore_send_wrreq_t 	:= (others => '0');

signal data_recive_data		: hcore_recive_data_t	:= (others => (others => '0'));
signal data_recive_busy		: hcore_recive_busy_t	:= (others => '0');	
signal data_recive_ack		: hcore_recive_ack_t		:= (others => '0');	

--

component hcore is
		generic 
		(
			has_performance_counter : std_logic := '0'			
		);
		port 
		(
				-- global

				clk					: in std_logic;
				reset					: in std_logic;
				
				-- ids
				
				i_hcore_id			: in std_logic_vector(3 downto 0);
				
				-- ctrl
				
				i_ctrl_hcore_data	: in hcore_ctrl_struct_t;
				i_ctrl_hcore_we	: in std_logic;
				
				-- to uctrl
				
				o_to_hctrl			: out hcore_to_uctrl_struct_t;
				o_to_hctrl_empty	: out std_logic;
				i_to_hctrl_rdreq	: in std_logic;
				
				-- from uctrl
				
				i_from_uctrl			: in uctrl_to_hcore_struct_t;
				o_from_uctrl_full		: out std_logic;
				i_from_uctrl_wrreq	: in std_logic;
				
				o_stall			: out std_logic;
				
				-- inst
				
				-- send
				
				inst_send_data		: out std_logic_vector(31 downto 0);
				inst_send_address	: out std_logic_vector(31 downto 0);
				inst_send_we		: out std_logic;
				inst_send_mode		: out std_logic;
				inst_send_busy		: in std_logic;
				inst_send_wrreq	: out std_logic;
				
				-- recive
				
				inst_recive_data	: in std_logic_vector(31 downto 0);
				inst_recive_busy	: in std_logic;
				inst_recive_ack	: out std_logic;
				
				-- data 
				
				-- send port
				
				data_send_data		: out std_logic_vector(31 downto 0);
				data_send_address	: out std_logic_vector(31 downto 0);
				data_send_we		: out std_logic;
				data_send_mode		: out std_logic;
				data_send_busy		: in std_logic;
				data_send_wrreq	: out std_logic;
				
				-- recive
				
				data_recive_data	: in std_logic_vector(31 downto 0);
				data_recive_busy	: in std_logic;
				data_recive_ack	: out std_logic
	
		);
end component;

signal hcore_to_hctrl			: hcore_to_uctrl_structs_t;	
signal hcore_to_hctrl_empty	: hctrl_emptys	:= (others => '0');				
signal hcore_to_hctrl_rdreq	: hctrl_rdreqs	:= (others => '0');	

signal hcore_from_uctrl			: uctrl_to_hcore_structs_t;	
signal hcore_from_uctrl_full	: uctrl_fulls	:= (others => '0');											
signal hcore_from_uctrl_wrreq	: uctrl_wrreqs	:= (others => '0');											

signal hcore_ids		: hcore_ids_t := (others => (others => '0'));
signal hcore_stalls	: hcore_stalls_t := (others => '0');

-- uctrl arb

component hcore_uctrl_arb is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;
								
				-- from hcores
				
				i_hcore_to_hctrl			: in hcore_to_uctrl_structs_t;	
				i_hcore_to_hctrl_empty	: in hctrl_emptys;				
				o_hcore_to_hctrl_rdreq	: out hctrl_rdreqs;	
				
				-- to hcores
				
				o_hcore_from_uctrl			: out uctrl_to_hcore_structs_t;
				i_hcore_from_uctrl_full		: in uctrl_fulls;					
				o_hcore_from_uctrl_wrreq	: out uctrl_wrreqs;	
				
				-- to uctrl
				
				i_to_hctrl_clk				: in std_logic;
				o_to_hctrl_data			: out hcore_to_uctrl_struct_t;
				o_to_hctrl_empty			: out std_logic;
				i_to_hctrl_rdreq			: in std_logic;
			
				-- from uctrl
				
				i_from_uctrl_data		: in uctrl_to_hcore_struct_t;
				i_from_uctrl_empty	: in std_logic;
				o_from_uctrl_rdreq	: out std_logic
				
		);
end component;

--

begin

	-----------------------------------------------------------------------------------------------------------
	---- ctrl (from uctrl)  
	-----------------------------------------------------------------------------------------------------------
	
	ctrl : hcore_ctrl port map
	(
		clk,
		reset,
		hcore_data,	
		hcore_busy,	
		hcore_ack,	
		hcore_ctrls,
		hcore_wes,
		hcores_reset
	);	

	-----------------------------------------------------------------------------------------------------------
	---- arbiter 
	-----------------------------------------------------------------------------------------------------------
	
	-- inst
	
	arb_inst : bus_arbiter_hcore_inst port map 
	(
			clk,
			reset,

			inst_send_data,
			inst_send_address,
			inst_send_we,
			inst_send_dmode,	
			inst_send_busy,
			inst_send_wrreq,	

			inst_recive_data,
			inst_recive_busy,
			inst_recive_ack,
	
			sdram_clk,
				
			inst_to_sdram_data,	
			inst_to_sdram_busy,
			inst_to_sdram_ack,

			inst_from_sdram_data,
			inst_from_sdram_busy,	
			inst_from_sdram_ack
	);
	
	-- data
	
	arb_data : bus_arbiter_hcore_data port map 
	(
			clk,
			reset,

			data_send_data,
			data_send_address,
			data_send_we,
			data_send_dmode,	
			data_send_busy,
			data_send_wrreq,	

			data_recive_data,
			data_recive_busy,
			data_recive_ack,
	
			sdram_clk,
				
			data_to_sdram_data,	
			data_to_sdram_busy,
			data_to_sdram_ack,

			data_from_sdram_data,
			data_from_sdram_busy,	
			data_from_sdram_ack
	);
	
	-----------------------------------------------------------------------------------------------------------
	---- hcores 
	-----------------------------------------------------------------------------------------------------------
	
	hids: for i in 0 to hcore_count-1 generate
		hcore_ids(i) <= conv_std_logic_vector(i,4);
	end generate;
	
	hfor : for i in 0 to hcore_count-1 generate
		hcore_n : hcore 
			generic map 
			(
				has_performance_counter => '0'
			)
			port map
         (
				clk,
				hcores_reset,
				
				hcore_ids(i),
				hcore_ctrls(i),
				hcore_wes(i),
				
				hcore_to_hctrl(i),			
				hcore_to_hctrl_empty(i),	
				hcore_to_hctrl_rdreq(i),	
				
				hcore_from_uctrl(i),				
				hcore_from_uctrl_full(i),		
				hcore_from_uctrl_wrreq(i),		
				
				hcore_stalls(i),
				
				inst_send_data(i),
				inst_send_address(i),
				inst_send_we(i),
				inst_send_dmode(i),
				inst_send_busy(i),
				inst_send_wrreq(i),
				
				inst_recive_data(i),
				inst_recive_busy(i),
				inst_recive_ack(i),
				
				data_send_data(i),
				data_send_address(i),
				data_send_we(i),
				data_send_dmode(i),
				data_send_busy(i),
				data_send_wrreq(i),
				
				data_recive_data(i),
				data_recive_busy(i),
				data_recive_ack(i)
				
			);	
   end generate;
	
	uctrl_arb : hcore_uctrl_arb port map 
	(
			clk,
			reset,
			hcore_to_hctrl,			
			hcore_to_hctrl_empty,	
			hcore_to_hctrl_rdreq,
			hcore_from_uctrl,				
			hcore_from_uctrl_full,		
			hcore_from_uctrl_wrreq,		
			i_to_hctrl_clk,	
			o_to_hctrl_data,
			o_to_hctrl_empty,
			i_to_hctrl_rdreq,
			i_from_uctrl_data,	
			i_from_uctrl_empty,
			o_from_uctrl_rdreq
	);	
	
	-----------------------------------------------------------------------------------------------------------
	---- stall out (debug/info) 
	-----------------------------------------------------------------------------------------------------------
	
	process(clk)
	begin
		if rising_edge(clk) then
		
			stalls <= (others => '0');
		
			for i in 0 to hcore_count-1 loop
				stalls(i) <= hcore_stalls(i);
			end loop;
		end if;
	end process;
	

end rtl;
