library ieee;
use work.bus_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity ifetch_requester is
		generic
		(			
			LINE_DEPTH	: natural := 6		-- 2^6 == 64
		);
		port 
		(
				-- 
				
				clk					: in std_logic;
				reset					: in std_logic;
				
				--
				
				load_request 		: in std_logic;
				load_address 		: std_logic_vector((31-LINE_DEPTH) downto 0);
				
				-- send
				
				inst_send_data		: out std_logic_vector(31 downto 0);
				inst_send_address	: out std_logic_vector(31 downto 0);
				inst_send_we		: out std_logic;
				inst_send_busy		: in std_logic;
				inst_send_wrreq	: out std_logic
				
	
		);
end entity;

architecture rtl of ifetch_requester is


type state_t is 
( 
	IFR_WAIT,
	IFR_PUSH
);

signal state 			: state_t := IFR_WAIT; 

begin
	
	process(clk)
		variable iaddress		: std_logic_vector((31-LINE_DEPTH) downto 0) := (others => '0');
		variable wreq			: std_logic := '0'; 
		variable icounter		: integer; 
	begin
		if rising_edge(clk) then
		
			-- default
		
			wreq	:= '0';
			
			if reset = '1' then
				state <= IFR_WAIT; 
			else
				if state = IFR_WAIT then
					if load_request = '1' then
						state 				<= IFR_PUSH;
						icounter				:= 0;
						iaddress 			:= load_address;
						wreq					:= '1';
					end if;
				else
					if icounter = ((2**LINE_DEPTH)-1) then
						state 				<= IFR_WAIT;
					else
						icounter				:= icounter + 1;
						wreq					:= '1';
					end if;
				end if;
			end if;
			
			inst_send_address 	<= iaddress((31-LINE_DEPTH) downto 1) & conv_std_logic_vector(icounter,LINE_DEPTH) & '0'; -- even 32 bit
    		inst_send_we			<= '0';
    		inst_send_data			<= (others => '0');
			inst_send_wrreq		<= wreq;
		end if;
	end process;

end rtl;
