library ieee;
use work.bus_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity mem_sram is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;	
				
				-- SRAM 
				SRAM_ADDR			: out std_logic_vector(19 downto 0);
				SRAM_CE_N			: out std_logic;
				SRAM_DQ				: inout std_logic_vector(15 downto 0);
				SRAM_LB_N			: out std_logic;
				SRAM_OE_N			: out std_logic;
				SRAM_UB_N			: out std_logic;
				SRAM_WE_N			: out std_logic;

				-- recive ports
				
				to_sram_data		: in sram_n_data_t;		
				to_sram_busy		: in sram_n_busy_t;
				to_sram_ack			: out sram_n_ack_t;	
			
				-- send ports
				
				from_sram_rclk		: in sram_rclk_t;
				from_sram_data		: out sram_n_data_t;
				from_sram_busy		: out sram_n_busy_t;
				from_sram_ack		: in sram_n_ack_t	
				
		);
end entity;

architecture rtl of mem_sram is

-- dcfifo (for send)

--component dcfifo_min IS
--	PORT
--	(
--		aclr		: IN STD_LOGIC  := '0';
--		data		: IN STD_LOGIC_VECTOR (71 DOWNTO 0);
--		rdclk		: IN STD_LOGIC ;
--		rdreq		: IN STD_LOGIC ;
--		wrclk		: IN STD_LOGIC ;
--		wrreq		: IN STD_LOGIC ;
--		q			: OUT STD_LOGIC_VECTOR (71 DOWNTO 0);
--		rdempty	: OUT STD_LOGIC ;
--		wrusedw	: OUT STD_LOGIC_VECTOR (7 DOWNTO 0)
--	);
--END component;

component dc_fifo IS
	
	GENERIC 
	(
		FIFO_DEPTH 		: natural;
		FIFO_WIDTH 		: natural;
		FIFO_SHOWAHEAD		: string;
		FIFO_USEDWWIDTH		: natural;
		FIFO_SYNCSTAGES		: natural		
	);
	PORT
	(
		aclr		: IN STD_LOGIC  := '0';
		data		: IN STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdclk		: IN STD_LOGIC ;
		rdreq		: IN STD_LOGIC ;
		wrclk		: IN STD_LOGIC ;
		wrreq		: IN STD_LOGIC ;
		q		: OUT STD_LOGIC_VECTOR (FIFO_WIDTH-1 DOWNTO 0);
		rdempty		: OUT STD_LOGIC ;
		wrusedw		: OUT STD_LOGIC_VECTOR (FIFO_USEDWWIDTH-1 DOWNTO 0)
	);
END component;

--

type send_xn_data_t is array(0 to mem_sram_reciver_count-1) of STD_LOGIC_VECTOR (71 DOWNTO 0);
type send_xn_wrreq_t is array(0 to mem_sram_reciver_count-1) of STD_LOGIC;
type send_xn_wrfull_t is array(0 to mem_sram_reciver_count-1) of STD_LOGIC; 
type send_xn_wrusedw_t is array(0 to mem_sram_reciver_count-1) of STD_LOGIC_VECTOR (7 DOWNTO 0);

--

signal send_xn_data		: send_xn_data_t := (others => (others => '0'));
signal send_xn_wrreq		: send_xn_wrreq_t := (others => '0');
signal send_xn_wrfull	: send_xn_wrfull_t := (others => '0'); 
signal send_xn_wrusedw	: send_xn_wrusedw_t := (others => (others => '0'));

-- stats

type transferState_t is
	(
		WAIT_TRANSFER_START,PSUH_NEXT_READ,PSUH_NEXT_WRITE,STALL,
		DEBUG1,DEBUG2,DEBUG3
	);
	
signal transferState : transferState_t := WAIT_TRANSFER_START;

-- stage signals

signal sram_ready			: std_logic := '0';
	
signal ctrl_push_out			: std_logic := '0';
signal ctrl_data				: port_struct := fill_port('0');
signal ctrl_id					: integer range 0 to mem_sram_reciver_count-1 := 0;
signal ctrl_we					: std_logic := '0';	
signal ctrl_todo				: std_logic_vector(1 downto 0); 
signal ctrl_last_read		: std_logic := '0';

--
-- 00 -- read high store				, write high
-- 01 -- read low, push high low		, write low
-- 10 -- read push 
	
signal setadr_push_out		: std_logic := '0';
signal setadr_data			: port_struct := fill_port('0');
signal setadr_todo			: std_logic_vector(1 downto 0) := (others => '0'); 
signal setadr_id				: integer range 0 to mem_sram_reciver_count-1 := 0;

signal waitread_push_out  	: std_logic := '0';
signal waitread_data			: port_struct := fill_port('0');
signal waitread_todo 		: std_logic_vector(1 downto 0) := (others => '0'); 
signal waitread_id			: integer range 0 to mem_sram_reciver_count-1 := 0;

constant MODELSIM : boolean := false
-- synthesis translate_off
     	or true
-- synthesis translate_on
	;
	
begin
	
	-- sram 
	
	SRAM_CE_N		<= '0';	-- ever enable 
	SRAM_LB_N		<= '0';	-- ever enable 
	SRAM_UB_N		<= '0';	-- ever enable 	
	SRAM_OE_N		<= '0';	-- ever enable 
		
	-- ack
	
	process(all)
	begin
	
		for i in 0 to mem_sram_reciver_count-1 loop
			to_sram_ack(i) <= '0';
		end loop;
	
		if reset = '0' then
			for i in 0 to mem_sram_reciver_count-1 loop
				if to_sram_busy(i) = '0' and sram_ready = '1' and send_xn_wrfull(i) = '0' then						
					if raw_to_port(to_sram_data(i)).we = '1' and ctrl_last_read = '1' then -- add stall between write after read
					else
						to_sram_ack(i) <= '1';	
					end if;
				exit; -- stop loop
				end if;	
			end loop;
		end if;
	end process;
	
	sram_ready <= '1' when transferState = WAIT_TRANSFER_START else '0';
		
	-----------------------------------------------------------------------------------------------------------	
	-- ctrl stage
	-----------------------------------------------------------------------------------------------------------
	
	process(clk)
		variable data_in : port_struct:= fill_port('0');
	begin
		if rising_edge(clk) then
			if reset = '1' then
				transferState <= WAIT_TRANSFER_START;
				ctrl_push_out	<= '0';
				ctrl_we			<= '0';
				ctrl_todo 		<= "00";
				ctrl_last_read	<= '0';
				ctrl_id			<= 0;
			else
				--
				
				ctrl_push_out	<= '0';
				ctrl_we			<= '0';
				ctrl_todo 		<= "00";
				ctrl_last_read	<= '0';
				
				--
			
				case transferState is
					when WAIT_TRANSFER_START => for I in 0 to mem_sram_reciver_count-1 loop
															if to_sram_busy(i) = '0' and 		-- data in is ready
																send_xn_wrfull(i) = '0' and	-- out fifo is not full
																sram_ready = '1' then			-- internal ready
					
																data_in 	:= raw_to_port(to_sram_data(i)); -- data
																ctrl_id  <= i;	-- store usedreciver id 														
				
																if data_in.we = '0' then -- read mode	
																	if data_in.dmode = '1'then -- 16 bit mode
																	
																		ctrl_we			<= '0';  -- read
																		ctrl_todo 		<= "10"; -- read push 
																		ctrl_push_out	<= '1';	-- start pipe
																		ctrl_last_read	<= '1';	-- mark as last read
																		
																		transferState <= WAIT_TRANSFER_START;
																		
																	else -- 32 bit
																	
																		ctrl_we			<= '0';  -- read
																		ctrl_todo 		<= "00"; -- read high store
																		ctrl_push_out	<= '1';	-- start pipe
																		ctrl_last_read	<= '1';	-- mark as last read
																		
																		transferState <= PSUH_NEXT_READ;
																		
																	end if;
																else -- write mode
																	if ctrl_last_read = '1' then	-- read in queue stall one cycles
																		transferState <= STALL;
																	else
																		if data_in.dmode = '1'then -- 16 bit mode
																		
																			ctrl_we			<= '1';  -- write
																			ctrl_todo 		<= "01"; -- write low
																			ctrl_push_out	<= '1';	-- start pipe
																			
																			transferState <= WAIT_TRANSFER_START;
																			
																		else -- 32 bit
																		
																			ctrl_we			<= '1';  -- write
																			ctrl_todo 		<= "00"; -- write high
																			ctrl_push_out	<= '1';	-- start pipe
			
																			transferState <= PSUH_NEXT_WRITE;
																			
																		end if;
																	end if;
																end if;															
															exit;	-- stop loop
														end if;
													end loop;	
															
					when PSUH_NEXT_READ =>		data_in.address := data_in.address + 1; -- next address 
					
														ctrl_we			<= '0';  -- read
														ctrl_todo 		<= "01"; -- read low, push high low
														ctrl_push_out	<= '1';	-- start pipe
														ctrl_last_read	<= '1';	-- mark as last read
														
														transferState <= WAIT_TRANSFER_START;	
														
					when PSUH_NEXT_WRITE =>		data_in.address := data_in.address + 1; -- next address 
					
														ctrl_we			<= '1';  -- write
														ctrl_todo 		<= "01"; -- write low
														ctrl_push_out	<= '1';	-- start pipe
														
														transferState <= WAIT_TRANSFER_START;
														
					when DEBUG1 => transferState <= DEBUG2;
					when DEBUG2 => transferState <= DEBUG3;
					when DEBUG3 => transferState <= STALL;
					when STALL	=> transferState <= WAIT_TRANSFER_START;
				end case;
				
				ctrl_data <= data_in;
			end if;			
		end if;		
	end process;
	
	-----------------------------------------------------------------------------------------------------------	
	-- set address / data (at write stage) stage
	-----------------------------------------------------------------------------------------------------------
	
	process(clk)
	begin
		if rising_edge(clk) then
			if reset = '1' then
				setadr_push_out 	<= '0';
				SRAM_WE_N 			<= '1';
			else
			
				-- default
			
				setadr_push_out 	<= '0';
				setadr_data			<= ctrl_data;
				setadr_todo 		<= ctrl_todo;
				setadr_id			<= ctrl_id;
				
				SRAM_WE_N 			<= '1';
				SRAM_DQ 				<= (others => 'Z');
				
				--
				
				if ctrl_push_out = '1' then
					SRAM_ADDR <= ctrl_data.address(19 downto 0);
					
					if ctrl_we = '1' then
						SRAM_WE_N <= '0'; -- write
					
						if ctrl_todo = "00" then -- write high
							SRAM_DQ <= ctrl_data.data(31 downto 16);	-- one write only write and do nothing after this stage
						else
							SRAM_DQ <= ctrl_data.data(15 downto 0);	
						end if;
					else	-- read								
						setadr_push_out 	<= '1';												
					end if;					
				end if;
			end if;			
		end if;		
	end process;
	
	-----------------------------------------------------------------------------------------------------------	
	-- wait read stage
	-----------------------------------------------------------------------------------------------------------	
	
	process(clk)
	begin
		if rising_edge(clk) then
			if reset = '1' then
				waitread_push_out <= '0';
			else
				-- default
			
				waitread_push_out <= '0';
				waitread_data		<= setadr_data;
				waitread_todo 		<= setadr_todo;
				waitread_id			<= setadr_id;
				
				-- 
				if setadr_push_out = '1' then
					
					waitread_push_out <= '1';
				end if;
			end if;			
		end if;		
	end process;
	
	-- capture data
	
	process(clk)
		variable tmp_data : std_logic_vector(31 downto 0);
		variable data_in : port_struct := fill_port('0');
	begin
		if rising_edge(clk) then
		
			-- default
			
			for i in 0 to mem_sram_reciver_count-1 loop
				send_xn_wrreq(i) <= '0';
			end loop;
			
			--
		
			if reset = '1' then		
				-- nothing to do
			else				
				if waitread_push_out = '1' then
				
					data_in := waitread_data;
					
--					if MODELSIM then
--						tmp_data(31 downto 0) := (others => '0');	
--					end if;
--					
					case waitread_todo is
						when "00" => tmp_data(31 downto 16) := SRAM_DQ; 
						
						when "01" | 
							  "10" => tmp_data(15 downto 0)  := SRAM_DQ; 
							  send_xn_wrreq(waitread_id)	<= '1'; -- push
						when others =>
					end case;
					
				--	tmp_data := x"89abcdef";
					data_in.data := tmp_data;
	
				end if;
			end if;	
	
			-- set to all fifo inputs (send_xn_wrreq arbit write)
			
			for i in 0 to mem_sram_reciver_count-1 loop
				send_xn_data(i) <= port_to_raw(data_in);				
			end loop;
				
		end if;		
	end process;
	
	-- full generate
		
	process(clk)
	begin
		for i in 0 to mem_sram_reciver_count-1 loop
			if rising_edge(clk) then
					if send_xn_wrusedw(i) >=  x"fc" then --x"1fb" at 9 bit?
						send_xn_wrfull(i) <= '1';
					else
						send_xn_wrfull(i) <= '0';
					end if;
				end if;		
			end loop;
	end process;
	
	-----------------------------------------------------------------------------------------------------------
	---- SEND DATA 
	-----------------------------------------------------------------------------------------------------------
	
	 dcfifo_loop : for i in 0 to mem_sram_reciver_count-1 generate
         dcfifo : dc_fifo 
					generic map
					(
						FIFO_DEPTH 			=> 256,
						FIFO_WIDTH 			=> 72,
						FIFO_SHOWAHEAD 	=> "ON",
						FIFO_USEDWWIDTH	=> 8,
						FIFO_SYNCSTAGES	=> 3
					) 
					port map
					(
						'0',
						send_xn_data(i),
						from_sram_rclk(i),		-- read with component clk
						from_sram_ack(i),	
						clk,							-- write with flash clk
						send_xn_wrreq(i),	
						from_sram_data(i),
						from_sram_busy(i),
						send_xn_wrusedw(i)
					);
   end generate;
	
	
end rtl;
