library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity sdClk_io is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;	

				-- prg
				
				prg_clk						: in std_logic;	
				
				prg_clk_msk					: in std_logic_vector(8 downto 0);
				prg_clk_edge_cmp_down	: in std_logic_vector(8 downto 0);
				prg_clk_edge_cmp_up		: in std_logic_vector(8 downto 0);
				prg_clk_write_edge_cmp	: in std_logic_vector(8 downto 0);
				prg_clk_read_edge_cmp	: in std_logic_vector(8 downto 0);
				
				--
				
				SD_CLK			: out std_logic;
				
				read_edge		: out std_logic;
				write_edge		: out std_logic			
		);
end entity;

architecture rtl of sdClk_io is

-- clk / ctrl

constant clk_msk_default				: std_logic_vector(8 downto 0) := "111111111";	--	  := "011111111";

constant clk_read_edge_cmp_default	: std_logic_vector(8 downto 0) := "110111111";  --   := "001011111";
constant clk_write_edge_cmp_default	: std_logic_vector(8 downto 0) := "010000000";  --   := "001000000";

constant clk_edge_cmp_down_default	: std_logic_vector(8 downto 0) := "000000000";  --   := "000000000";                                                                    
constant clk_edge_cmp_up_default		: std_logic_vector(8 downto 0) := "100000000";  --   := "010000000";

--

signal clk_cnt 				: std_logic_vector(8 downto 0) := (others => '0');
signal clk_out					: std_logic := '0';

signal clk_msk					: std_logic_vector(8 downto 0) := clk_msk_default;

signal clk_edge_cmp_down	: std_logic_vector(8 downto 0) := clk_edge_cmp_down_default;
signal clk_edge_cmp_up		: std_logic_vector(8 downto 0) := clk_edge_cmp_up_default;

signal clk_write_edge_cmp	: std_logic_vector(8 downto 0) := clk_write_edge_cmp_default;
signal clk_read_edge_cmp	: std_logic_vector(8 downto 0) := clk_read_edge_cmp_default;

signal clk_write_edge		: std_logic := '0';
signal clk_read_edge			: std_logic := '0';

begin
	
	--
	
	read_edge	<=	clk_read_edge;
	write_edge	<= clk_write_edge;
	
	-----------------------------------------------------------------------------------------------------------
	-- clk counter / ctrl
	-----------------------------------------------------------------------------------------------------------

	process(clk)
	begin
		if rising_edge(clk) then
			if reset = '1' then
				clk_cnt 				<= (others => '0');
				
				clk_write_edge			<= '0';
				clk_read_edge			<= '0';
				
				clk_msk					<= clk_msk_default;
				
				clk_write_edge_cmp	<= clk_write_edge_cmp_default;
				clk_read_edge_cmp		<= clk_read_edge_cmp_default;
				
				clk_out					<= '0';
				
				clk_edge_cmp_up 		<= clk_edge_cmp_up_default;
				clk_edge_cmp_down 	<= clk_edge_cmp_down_default;						
			else
			
				-- default
				
				clk_write_edge		<= '0';
				clk_read_edge		<= '0';
				
				-- prg
				
				if prg_clk = '1' then				
					clk_msk				 	<= prg_clk_msk;				
					
					clk_edge_cmp_down 	<= prg_clk_edge_cmp_down;					
					clk_edge_cmp_up 		<= prg_clk_edge_cmp_up;				
					clk_write_edge_cmp 	<= prg_clk_write_edge_cmp;				
					clk_read_edge_cmp 	<= prg_clk_read_edge_cmp;				
				end if;
				
				-- io edges
			
				if (clk_cnt and  clk_msk) = clk_write_edge_cmp then
					clk_write_edge <= '1'; 
				end if;
				
				if (clk_cnt and clk_msk) = clk_read_edge_cmp then
					clk_read_edge	<= '1';
				end if;
				
				if (clk_cnt and clk_msk) = clk_edge_cmp_down then
					clk_out	<= '0';
				end if;
				
				if (clk_cnt and clk_msk) = clk_edge_cmp_up then
					clk_out	<= '1';
				end if;
				
				SD_CLK <= clk_out; 
				clk_cnt <= clk_cnt + 1;
			end if;
		end if;
	end process;
	
		
end rtl;
