library ieee;
use work.bus_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity sdCtrl is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;	

				-- SDCARD 
				
				SD_CLK			: out std_logic;
				SD_CMD			: inout std_logic;
				SD_DAT			: inout std_logic_vector(3 downto 0);
				SD_WP_N			: in std_logic;
				
				-- external
				
				-- ctrl port
	
				-- send
				
				ctrl_send_data			: out std_logic_vector(15 downto 0);
				ctrl_send_address		: out std_logic_vector(31 downto 0);
				ctrl_send_we			: out std_logic;
				ctrl_send_busy			: in std_logic;
				ctrl_send_wrreq		: out std_logic;
				
				-- recive
				
				ctrl_recive_data		: in std_logic_vector(15 downto 0);
				ctrl_recive_address	: in std_logic_vector(31 downto 0);
				ctrl_recive_we			: in std_logic;
				ctrl_recive_busy		: in std_logic;				
				ctrl_recive_ack		: out std_logic;
				
				-- data port
				
				-- send
				
				data_send_data			: out std_logic_vector(15 downto 0);
				data_send_address		: out std_logic_vector(31 downto 0);
				data_send_we			: out std_logic;
				data_send_busy			: in std_logic;
				data_send_wrreq		: out std_logic
		);
end entity;

architecture rtl of sdCtrl is

-- CLK ctrl

component sdClk_io is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;	

				-- prg
				
				prg_clk						: in std_logic;	
				
				prg_clk_msk					: in std_logic_vector(8 downto 0);
				prg_clk_edge_cmp_down	: in std_logic_vector(8 downto 0);
				prg_clk_edge_cmp_up		: in std_logic_vector(8 downto 0);
				prg_clk_write_edge_cmp	: in std_logic_vector(8 downto 0);
				prg_clk_read_edge_cmp	: in std_logic_vector(8 downto 0);
				
				--
				
				SD_CLK			: out std_logic;
				
				read_edge		: out std_logic;
				write_edge		: out std_logic			
		);
end component;

-- CMD IO

component sdCtrl_io is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;	

				-- SDCARD 
				
				SD_CMD			: inout std_logic;
				SD_WP_N			: in std_logic;
				
				-- clk in
				
				clk_write_edge		: in std_logic;
				clk_read_edge		: in std_logic;
				
				-- send
				
				send_cmd_arg		: in std_logic_vector(37 downto 0);
				send_start			: in std_logic;
				send_in_progress	: out std_logic;
				
				-- recive
				
				recive_mode			: in std_logic; -- 0 = 48 / 1 = 136 bit
				recive_reset		: in std_logic;
				recive_error		: out std_logic;
				recive_ok			: out std_logic;
				recive_cmd			: out std_logic_vector(5 downto 0);
				recive_crc7			: out std_logic_vector(6 downto 0);
				recive_exp_crc7	: out std_logic_vector(6 downto 0);
				recive_data			: out std_logic_vector(126 downto 0)
				
		);
end component;

signal io_send_cmd_arg		: std_logic_vector(37 downto 0) := (others => '0');
signal io_send_start			: std_logic := '0';
signal io_send_in_progress	: std_logic := '0';

signal io_recive_mode		: std_logic := '0'; -- 0 = 48 / 1 = 136 bit
signal io_recive_reset		: std_logic := '0';
signal io_recive_error		: std_logic := '0';
signal io_recive_ok			: std_logic := '0';
signal io_recive_cmd			: std_logic_vector(5 downto 0)	:= (others => '0');
signal io_recive_crc7		: std_logic_vector(6 downto 0)	:= (others => '0');
signal io_recive_exp_crc7	: std_logic_vector(6 downto 0)	:= (others => '0');
signal io_recive_data		: std_logic_vector(126 downto 0)	:= (others => '0');

-- sd data

component sdData_io_channel is
		port 
		(
				-- global

				clk						: in std_logic;
				reset						: in std_logic;	

				-- SDCARD 
				
				sd_data					: in std_logic;
				
				-- clk in
				
				clk_read_edge			: in std_logic;
				
				-- recive
				
				recive_reset			: in std_logic;
				recive_fifo_usage		: out std_logic_vector (12 DOWNTO 0);				
				recive_data_in_cnt	: in integer range 0 to 8191; -- cnt + 1			
				recive_crc_pass		: out std_logic;
				recive_finish			: out std_logic;
				recive_error			: out std_logic;
				recive_blocks_readed	: out integer range 0 to 65535;
								
				recive_data					: out std_logic;
				recive_data_available	: out std_logic;
				recive_data_ack			: in std_logic				
				
		);
end component;

signal sdData_io_channel1_recive_reset				: std_logic := '0';
signal sdData_io_channel1_recive_fifo_usage		: std_logic_vector (12 DOWNTO 0) := (others => '0');				
signal sdData_io_channel1_recive_data_in_cnt		: integer range 0 to 8191 := 1023;	
signal sdData_io_channel1_recive_crc_pass			: std_logic := '0';
signal sdData_io_channel1_recive_finish			: std_logic := '0';
signal sdData_io_channel1_recive_error				: std_logic := '0';
signal sdData_io_channel1_recive_blocks_readed	: integer range 0 to 65535 := 0;	
signal sdData_io_channel1_recive_data				: std_logic := '0';
signal sdData_io_channel1_recive_data_available				: std_logic := '0';
signal sdData_io_channel1_recive_data_ack			: std_logic	:= '0';	
		
signal sdData_io_channel2_recive_reset				: std_logic := '0';
signal sdData_io_channel2_recive_fifo_usage		: std_logic_vector (12 DOWNTO 0) := (others => '0');				
signal sdData_io_channel2_recive_data_in_cnt		: integer range 0 to 8191 := 1023;	
signal sdData_io_channel2_recive_crc_pass			: std_logic := '0';
signal sdData_io_channel2_recive_finish			: std_logic := '0';
signal sdData_io_channel2_recive_error				: std_logic := '0';
signal sdData_io_channel2_recive_blocks_readed	: integer range 0 to 65535 := 0;	
signal sdData_io_channel2_recive_data				: std_logic := '0';
signal sdData_io_channel2_recive_data_available				: std_logic := '0';
signal sdData_io_channel2_recive_data_ack			: std_logic	:= '0';	

signal sdData_io_channel3_recive_reset				: std_logic := '0';
signal sdData_io_channel3_recive_fifo_usage		: std_logic_vector (12 DOWNTO 0) := (others => '0');				
signal sdData_io_channel3_recive_data_in_cnt		: integer range 0 to 8191 := 1023;	
signal sdData_io_channel3_recive_crc_pass			: std_logic := '0';
signal sdData_io_channel3_recive_finish			: std_logic := '0';
signal sdData_io_channel3_recive_error				: std_logic := '0';
signal sdData_io_channel3_recive_blocks_readed	: integer range 0 to 65535 := 0;	
signal sdData_io_channel3_recive_data				: std_logic := '0';
signal sdData_io_channel3_recive_data_available				: std_logic := '0';
signal sdData_io_channel3_recive_data_ack			: std_logic	:= '0';	

signal sdData_io_channel4_recive_reset				: std_logic := '0';
signal sdData_io_channel4_recive_fifo_usage		: std_logic_vector (12 DOWNTO 0) := (others => '0');				
signal sdData_io_channel4_recive_data_in_cnt		: integer range 0 to 8191 := 1023;	
signal sdData_io_channel4_recive_crc_pass			: std_logic := '0';
signal sdData_io_channel4_recive_finish			: std_logic := '0';
signal sdData_io_channel4_recive_error				: std_logic := '0';
signal sdData_io_channel4_recive_blocks_readed	: integer range 0 to 65535 := 0;	
signal sdData_io_channel4_recive_data				: std_logic := '0';
signal sdData_io_channel4_recive_data_available				: std_logic := '0';
signal sdData_io_channel4_recive_data_ack			: std_logic	:= '0';			
		
-- stage 1

signal s1_push_out	: std_logic := '0';
signal s1_data			: STD_LOGIC_VECTOR (71 DOWNTO 0) := (others => '0');

-- clk ctrl

signal clk_write_edge	: std_logic := '0';
signal clk_read_edge		: std_logic := '0';

signal prg_clk						: std_logic := '0';

signal prg_clk_msk				: std_logic_vector(8 downto 0)  := (others =>'0');
signal prg_clk_edge_cmp_down	: std_logic_vector(8 downto 0)  := (others =>'0');
signal prg_clk_edge_cmp_up		: std_logic_vector(8 downto 0)  := (others =>'0');
signal prg_clk_write_edge_cmp	: std_logic_vector(8 downto 0)  := (others =>'0');
signal prg_clk_read_edge_cmp	: std_logic_vector(8 downto 0)  := (others =>'0');

-- data

signal burst_address_start_data_out	: STD_LOGIC_VECTOR (31 DOWNTO 0) := (others =>'0');
signal burst_address_start_set		: std_logic := '0';

signal burst_data_out_en				: std_logic := '0';

signal fifo_data_push_out 				: std_logic := '0';
signal fifo_data_value	  				: std_logic_vector(3 downto 0) := (others => '0');
signal burst_address_real_data_out	: STD_LOGIC_VECTOR(31 downto 0)  := (others =>'0');

signal  test_register					: std_logic_vector(15 downto 0)  := (others =>'0');

--

begin
	
	-----------------------------------------------------------------------------------------------------------	
	-- sd clk (programmable)
	-----------------------------------------------------------------------------------------------------------
	
	clk_io : sdClk_io port map 
	(
		clk,
		reset,
		prg_clk,					
		prg_clk_msk,
		prg_clk_edge_cmp_down,		
		prg_clk_edge_cmp_up,		
		prg_clk_write_edge_cmp,		
		prg_clk_read_edge_cmp,		
		SD_CLK,
		clk_read_edge,
		clk_write_edge	
	);
	
		
	-----------------------------------------------------------------------------------------------------------	
	-- sd cmd io
	-----------------------------------------------------------------------------------------------------------
	
	cmd_io : sdCtrl_io port map 
	(		
		clk,
		reset,		
		SD_CMD,
		SD_WP_N,
		clk_write_edge,
		clk_read_edge,
		io_send_cmd_arg,	
		io_send_start,
		io_send_in_progress,
		io_recive_mode,
		io_recive_reset,		
		io_recive_error,		
		io_recive_ok,		
		io_recive_cmd,			
		io_recive_crc7,		
		io_recive_exp_crc7,
		io_recive_data
	);
	
	-----------------------------------------------------------------------------------------------------------
	-- sd data (4 bit)
	-----------------------------------------------------------------------------------------------------------
	
	data_io_1 : sdData_io_channel port map 
	(
		clk,
		reset,
		SD_DAT(0),
		clk_read_edge,
		sdData_io_channel1_recive_reset,			
		sdData_io_channel1_recive_fifo_usage,		
		sdData_io_channel1_recive_data_in_cnt,
		sdData_io_channel1_recive_crc_pass,		
		sdData_io_channel1_recive_finish,			
		sdData_io_channel1_recive_error,			
		sdData_io_channel1_recive_blocks_readed,	
		sdData_io_channel1_recive_data,		
		sdData_io_channel1_recive_data_available,
		sdData_io_channel1_recive_data_ack			
	);
	
	data_io_2 : sdData_io_channel port map 
	(
		clk,
		reset,
		SD_DAT(1),
		clk_read_edge,
		sdData_io_channel2_recive_reset,			
		sdData_io_channel2_recive_fifo_usage,		
		sdData_io_channel2_recive_data_in_cnt,
		sdData_io_channel2_recive_crc_pass,		
		sdData_io_channel2_recive_finish,			
		sdData_io_channel2_recive_error,			
		sdData_io_channel2_recive_blocks_readed,	
		sdData_io_channel2_recive_data,		
		sdData_io_channel2_recive_data_available,	
		sdData_io_channel2_recive_data_ack			
	);
	
	data_io_3 : sdData_io_channel port map 
	(
		clk,
		reset,
		SD_DAT(2),
		clk_read_edge,
		sdData_io_channel3_recive_reset,			
		sdData_io_channel3_recive_fifo_usage,		
		sdData_io_channel3_recive_data_in_cnt,
		sdData_io_channel3_recive_crc_pass,		
		sdData_io_channel3_recive_finish,			
		sdData_io_channel3_recive_error,			
		sdData_io_channel3_recive_blocks_readed,	
		sdData_io_channel3_recive_data,	
		sdData_io_channel3_recive_data_available,		
		sdData_io_channel3_recive_data_ack			
	);
	
	data_io_4 : sdData_io_channel port map 
	(
		clk,
		reset,
		SD_DAT(3),
		clk_read_edge,
		sdData_io_channel4_recive_reset,			
		sdData_io_channel4_recive_fifo_usage,		
		sdData_io_channel4_recive_data_in_cnt,
		sdData_io_channel4_recive_crc_pass,		
		sdData_io_channel4_recive_finish,			
		sdData_io_channel4_recive_error,			
		sdData_io_channel4_recive_blocks_readed,	
		sdData_io_channel4_recive_data,	
		sdData_io_channel4_recive_data_available,		
		sdData_io_channel4_recive_data_ack			
	);

			
	-----------------------------------------------------------------------------------------------------------
	-- ctrl in/output
	-----------------------------------------------------------------------------------------------------------

	process(all)
	begin
	
		ctrl_recive_ack <= '0'; 
	
		if reset = '1' then		
		else
			if ctrl_recive_busy = '0' and  ctrl_send_busy = '0' then											
					ctrl_recive_ack <= '1';
			end if;
		end if;
	end process;
	
	process(clk)
		variable tmp_data : std_logic_vector(15 downto 0) := (others => '0');
	begin
		if rising_edge(clk) then
		
			if reset = '1' then
				burst_data_out_en <= '0';
				
				ctrl_send_wrreq 	<= '0';
				ctrl_send_we		<= '0';
				ctrl_send_data 	<= (others => '0');
				ctrl_send_address	<= ctrl_recive_address;
				
				io_send_start 		<= '0';
				io_recive_reset	<= '0';
				
				sdData_io_channel1_recive_reset <= '0';
				sdData_io_channel2_recive_reset <= '0';
				sdData_io_channel3_recive_reset <= '0';
				sdData_io_channel4_recive_reset <= '0';
				
				burst_address_start_set			  <= '0';
				prg_clk	<= '0';
			else
				-- default
				
				ctrl_send_wrreq 	<= '0';
				ctrl_send_we		<= '0';
				ctrl_send_data 	<= (others => '0');
				ctrl_send_address	<= ctrl_recive_address;
				
				io_send_start 		<= '0';
				io_recive_reset	<= '0';
				
				sdData_io_channel1_recive_reset <= '0';
				sdData_io_channel2_recive_reset <= '0';
				sdData_io_channel3_recive_reset <= '0';
				sdData_io_channel4_recive_reset <= '0';
				
				burst_address_start_set			  <= '0';
				
				prg_clk	<= '0';
				--
			
				if ctrl_recive_busy = '0' and  ctrl_send_busy = '0' then							
					if ctrl_recive_we = '0' then -- read
						case ctrl_recive_address(4 downto 0) is

							when "00000"  =>		tmp_data :=  "00" & io_recive_cmd & "00000" & io_recive_ok & io_recive_error & io_send_in_progress;
							when "00001" 	=>		tmp_data := '0' & io_recive_exp_crc7 & '0' & io_recive_crc7;
							
							when "00010" 	=>		tmp_data := io_recive_data(126 downto 111);	
							when "00011" 	=>		tmp_data := io_recive_data(110 downto 95);	
							when "00100" 	=>		tmp_data := io_recive_data(94 downto 79);	
							when "00101" 	=>		tmp_data := io_recive_data(78 downto 63);
							when "00110" 	=>		tmp_data := io_recive_data(62 downto 47);	
							when "00111" 	=>		tmp_data := io_recive_data(46 downto 31);
							when "01000" 	=>		tmp_data := io_recive_data(30 downto 15);	
							when "01001" 	=>		tmp_data := io_recive_data(14 downto 0) & '0';
								
									--
									
							when "10000" 	=>	tmp_data(0)  := sdData_io_channel4_recive_finish;
													tmp_data(1)  := sdData_io_channel3_recive_finish;
													tmp_data(2)  := sdData_io_channel2_recive_finish;
													tmp_data(3)  := sdData_io_channel1_recive_finish;
													
													tmp_data(4)  := sdData_io_channel4_recive_error;
													tmp_data(5)  := sdData_io_channel3_recive_error;
													tmp_data(6)  := sdData_io_channel2_recive_error;
													tmp_data(7)  := sdData_io_channel1_recive_error;
													
													tmp_data(8)  := not sdData_io_channel4_recive_crc_pass;
													tmp_data(9)  := not sdData_io_channel3_recive_crc_pass;
													tmp_data(10) := not sdData_io_channel2_recive_crc_pass;
													tmp_data(11) := not sdData_io_channel1_recive_crc_pass;
							
													tmp_data(15 downto 12) := (others => '0');
							
							when "10001" 	=>	tmp_data := conv_std_logic_vector(sdData_io_channel1_recive_blocks_readed,16);
							when "10010" 	=>	tmp_data := conv_std_logic_vector(sdData_io_channel2_recive_blocks_readed,16);
							when "10011" 	=>	tmp_data := conv_std_logic_vector(sdData_io_channel3_recive_blocks_readed,16);
							when "10100" 	=>	tmp_data := conv_std_logic_vector(sdData_io_channel4_recive_blocks_readed,16);
							
							when "10101"	=>	tmp_data := "000" & sdData_io_channel1_recive_fifo_usage;
							when "10110"	=>	tmp_data := "000" & sdData_io_channel2_recive_fifo_usage;
							when "10111"	=>	tmp_data := "000" & sdData_io_channel3_recive_fifo_usage;
							when "11000"	=>	tmp_data := "000" & sdData_io_channel4_recive_fifo_usage;
							when "11001"	=> tmp_data := test_register;
									
							when others =>		tmp_data := x"bada";					
						end case;
						
						-- write out
												
						ctrl_send_data 	<= tmp_data;
						ctrl_send_we		<= '1';
						ctrl_send_wrreq 	<= '1';
						
					else	-- write
						case ctrl_recive_address(4 downto 0) is							
							when "00000" =>  io_send_cmd_arg(37 downto 32) <= ctrl_recive_data(5 downto 0);
											
							
												if ctrl_recive_data(8) = '1' then
													io_send_start <= '1';
												end if;
												
												if ctrl_recive_data(9) = '1' then
													io_recive_reset <= '1';
												end if;
												
												io_recive_mode <= ctrl_recive_data(10);
												prg_clk	<= ctrl_recive_data(11);
												
							when "00001" 	=>
							
							when "00010" 	=>	io_send_cmd_arg(31 downto 16) <=  ctrl_recive_data(15 downto 0);
							when "00011" 	=>	io_send_cmd_arg(15 downto 0) 	<=  ctrl_recive_data(15 downto 0);
							
							
							-- clk
							
							when "01000" 	=>	prg_clk_msk					<=  ctrl_recive_data(8 downto 0);							
							when "01001" 	=>	prg_clk_edge_cmp_down	<=  ctrl_recive_data(8 downto 0);									
							when "01010" 	=>	prg_clk_edge_cmp_up		<=  ctrl_recive_data(8 downto 0);		
							when "01011" 	=>	prg_clk_write_edge_cmp	<=  ctrl_recive_data(8 downto 0);	
							when "01100" 	=>	prg_clk_read_edge_cmp	<=  ctrl_recive_data(8 downto 0);		

							--
							
							when "10000" 	=>	if ctrl_recive_data(0) = '1' then
														sdData_io_channel1_recive_reset <= '1';
														sdData_io_channel2_recive_reset <= '1';
														sdData_io_channel3_recive_reset <= '1';
														sdData_io_channel4_recive_reset <= '1';
													end if;
													burst_data_out_en <= ctrl_recive_data(1);
													
							when "10010"	=>	burst_address_start_data_out(15 downto 0) <= ctrl_recive_data(15 downto 0);	burst_address_start_set <= '1';
							when "10011"	=>	burst_address_start_data_out(31 downto 16) <= ctrl_recive_data(15 downto 0);	burst_address_start_set <= '1';
							when "11001"	=> test_register <= ctrl_recive_data;
							
							when others =>
						end case;
					end if;
				end if;
			end if;				
		end if;
	end process;
	
	-----------------------------------------------------------------------------------------------------------
	-- data out (fifo)
	-----------------------------------------------------------------------------------------------------------
	
	process(clk)
	begin
		if rising_edge(clk) then		
			if reset = '1' then
				fifo_data_push_out <= '0';
			else
			
				-- default
				
				fifo_data_push_out <= '0';
				
				--
			
				if burst_data_out_en = '1' and data_send_busy = '0' then
					if sdData_io_channel1_recive_data_available = '1' and
						sdData_io_channel2_recive_data_available = '1' and
						sdData_io_channel3_recive_data_available = '1' and
						sdData_io_channel4_recive_data_available = '1' then
						
							fifo_data_value 	 <= sdData_io_channel4_recive_data & sdData_io_channel3_recive_data & sdData_io_channel2_recive_data & sdData_io_channel1_recive_data;
							fifo_data_push_out <= '1';
						
					end if;
				end if;			
			end if;
		end if;
	end process;
	
	process(all)
	begin
		if reset = '1' then
			sdData_io_channel1_recive_data_ack <= '0';
			sdData_io_channel2_recive_data_ack <= '0';
			sdData_io_channel3_recive_data_ack <= '0';
			sdData_io_channel4_recive_data_ack <= '0';
		else
		
			sdData_io_channel1_recive_data_ack <= '0';
			sdData_io_channel2_recive_data_ack <= '0';
			sdData_io_channel3_recive_data_ack <= '0';
			sdData_io_channel4_recive_data_ack <= '0';
		
			if burst_data_out_en = '1' and data_send_busy = '0' then
					if sdData_io_channel1_recive_data_available = '1' and
						sdData_io_channel2_recive_data_available = '1' and
						sdData_io_channel3_recive_data_available = '1' and
						sdData_io_channel4_recive_data_available = '1' then
						
							
							sdData_io_channel1_recive_data_ack <= '1';
							sdData_io_channel2_recive_data_ack <= '1';
							sdData_io_channel3_recive_data_ack <= '1';
							sdData_io_channel4_recive_data_ack <= '1';
			
					end if;
			end if;			
		end if;
	end process;
	
	process(clk)		
		variable fo_state : integer range 0 to 3 := 0;
		variable fo_data  : std_logic_vector(15 downto 0) := (others => '0');
	begin
		if rising_edge(clk) then		
			if reset = '1' then
				fo_state := 0;
				data_send_wrreq 	<= '0';
			else
			
				-- default
										
				if burst_address_start_set = '1' then
					burst_address_real_data_out <= burst_address_start_data_out;
				end if;
						
				data_send_wrreq 	<= '0';		
										
				--
			
				if fifo_data_push_out = '1' then
					case fo_state is
						when 0 =>	fo_data(15 downto 12)	:= fifo_data_value;	
										fo_state := 1;
										
						when 1 =>	fo_data(11 downto 8)  	:= fifo_data_value;	
										fo_state := 2;
										
						when 2 =>	fo_data(7 downto 4) 		:= fifo_data_value;	
										fo_state := 3;
										
						when 3 =>	fo_data(3 downto 0) 		:= fifo_data_value;	
										fo_state := 0;
						
										burst_address_real_data_out 	<= burst_address_real_data_out + 1; -- next
										data_send_wrreq		 			<= '1';
					end case;
				end if;
							
			end if;	
					
			data_send_address	<= burst_address_real_data_out(31 downto 0);		
			data_send_we	 	<= '1';		
			data_send_data		<= fo_data;		
	
		end if;
	end process;
end rtl;
