library ieee;
use work.ucore_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity utility_ctrl_audio is
		port 
		(
			-- global

				clk				: in std_logic;
				reset				: in std_logic;
				
				-- audio 
				
				clk_audio			: in std_logic;
				reset_audio			: in std_logic;
				reset_fall_audio	: in std_logic;
				
				-- dac fifo
				
				dac_fifo_data		: in std_logic_vector(63 downto 0); 
				dac_fifo_wreq		: in std_logic;
				dac_used_low_irq	: in std_logic_vector(7 downto 0);
				dac_irq				: out std_logic;
				
				-- adc fifo
				
				adc_fifo_data		: out std_logic_vector(47 downto 0);
				adc_fifo_rdreq		: in std_logic;
				adc_used_high_irq	: in std_logic_vector(7 downto 0);
				adc_irq				: out std_logic;
				
				-- error
				
				error_dac_no_data			: out std_logic;
				error_dac_full				: out std_logic;
				
				error_adc_no_data			: out std_logic;
				error_adc_full				: out std_logic;
				
				-- wm8731
				
				AUD_ADCDAT		: in std_logic;
				AUD_ADCLRCK		: inout std_logic;
				AUD_DACDAT		: out std_logic;
				AUD_DACLRCK		: inout std_logic												
		);
end entity;

architecture rtl of utility_ctrl_audio is

-- fifo adc

component utility_ctrl_audio_fifo_adc
	PORT
	(
		data		: IN STD_LOGIC_VECTOR (47 DOWNTO 0);
		rdclk		: IN STD_LOGIC ;
		rdreq		: IN STD_LOGIC ;
		wrclk		: IN STD_LOGIC ;
		wrreq		: IN STD_LOGIC ;
		q			: OUT STD_LOGIC_VECTOR (47 DOWNTO 0);
		rdempty		: OUT STD_LOGIC ;
		rdusedw		: OUT STD_LOGIC_VECTOR (7 DOWNTO 0);
		wrfull		: OUT STD_LOGIC
	);
end component;

signal fifo_adc_data			: STD_LOGIC_VECTOR (47 DOWNTO 0) := (others => '0');
signal fifo_adc_rdclk		: STD_LOGIC := '0';
signal fifo_adc_rdreq		: STD_LOGIC := '0';
signal fifo_adc_wrclk		: STD_LOGIC := '0';
signal fifo_adc_wrreq		: STD_LOGIC := '0';
signal fifo_adc_q				: STD_LOGIC_VECTOR (47 DOWNTO 0) := (others => '0');
signal fifo_adc_rdempty		: STD_LOGIC := '0';
signal fifo_adc_rdusedw		: STD_LOGIC_VECTOR (7 DOWNTO 0) := (others => '0');
signal fifo_adc_wrfull		: STD_LOGIC := '0';

-- fifo dac

component utility_ctrl_audio_fifo_dac
	PORT
	(
		data			: IN STD_LOGIC_VECTOR (63 DOWNTO 0);
		rdclk			: IN STD_LOGIC ;
		rdreq			: IN STD_LOGIC ;
		wrclk			: IN STD_LOGIC ;
		wrreq			: IN STD_LOGIC ;
		q				: OUT STD_LOGIC_VECTOR (63 DOWNTO 0);
		rdempty		: OUT STD_LOGIC ;
		wrfull		: OUT STD_LOGIC ;
		wrusedw		: OUT STD_LOGIC_VECTOR (7 DOWNTO 0)
	);
end component;

signal fifo_dac_data			: STD_LOGIC_VECTOR (63 DOWNTO 0) := (others => '0');
signal fifo_dac_rdclk		: STD_LOGIC := '0';
signal fifo_dac_rdreq		: STD_LOGIC := '0';
signal fifo_dac_wrclk		: STD_LOGIC := '0';
signal fifo_dac_wrreq		: STD_LOGIC := '0';
signal fifo_dac_q				: STD_LOGIC_VECTOR (63 DOWNTO 0) := (others => '0');
signal fifo_dac_rdempty		: STD_LOGIC := '0';
signal fifo_dac_wrfull		: STD_LOGIC := '0';
signal fifo_dac_wrusedw		: STD_LOGIC_VECTOR (7 DOWNTO 0) := (others => '0');

--
	
constant frame_length 		: integer := 192; 	
constant dac_bit_width		: integer := 24; 	

--
	
signal ser_out_state : integer range 0 to (frame_length) := 0;

signal dac_counter		: std_logic_vector(3 downto 0) := (others => '0');
signal dac_counter_max	: std_logic_vector(3 downto 0) := (others => '0');
	
signal dac_left  	: std_logic_vector(23 downto 0) := (others => '0');
signal dac_right 	: std_logic_vector(23 downto 0) := (others => '0');
signal dac_enable	: std_logic := '0';	

signal adc_counter		: std_logic_vector(3 downto 0) := (others => '0');
signal adc_counter_max	: std_logic_vector(3 downto 0) := (others => '0');

signal adc_enable	: std_logic := '0';
	
--

signal iAUD_ADCDAT		: std_logic := '0';
signal iAUD_ADCLRCK		: std_logic := '0';
signal iAUD_DACDAT		: std_logic := '0';
signal iAUD_DACLRCK		: std_logic	:= '0'; 	
	
--	
	
begin
	
	process(clk_audio)
	begin
		if falling_edge(clk_audio) then
			if reset_fall_audio = '1' then
				AUD_DACDAT				<= '0';
				AUD_DACLRCK 			<= '0';
				AUD_ADCLRCK 			<= '0';				
			else
				AUD_DACDAT  <= iAUD_DACDAT;
				AUD_DACLRCK <= iAUD_DACLRCK;
				AUD_ADCLRCK <= iAUD_ADCLRCK;
				iAUD_ADCDAT <= AUD_ADCDAT;
			end if;
		end if;
	end process;
	
	-----------------------------------------------------------------------------------------------------------
	-- fifo dac
	-----------------------------------------------------------------------------------------------------------
	
	fifo_dac_data 	<= dac_fifo_data;
	fifo_dac_wrreq	<=	dac_fifo_wreq;
	
	fifo_dac_rdclk <= clk_audio;
	fifo_dac_wrclk <= clk;
	
	fifo_dac : utility_ctrl_audio_fifo_dac port map
	(
		fifo_dac_data,			
		fifo_dac_rdclk,	-- audio clk	
		fifo_dac_rdreq,	
		fifo_dac_wrclk,	-- ucore clk	
		fifo_dac_wrreq,		
		fifo_dac_q,				
		fifo_dac_rdempty,
		fifo_dac_wrfull,
		fifo_dac_wrusedw
	);
	
	error_dac_full 		<= fifo_dac_wrfull;
	error_dac_no_data		<= fifo_dac_rdempty;
	
	-----------------------------------------------------------------------------------------------------------
	-- fifo adc
	-----------------------------------------------------------------------------------------------------------
	
	adc_fifo_data		<= fifo_adc_q;
	fifo_adc_rdreq		<= adc_fifo_rdreq;
	
	fifo_adc_rdclk <= clk;
	fifo_adc_wrclk <= clk_audio;
	
	fifo_adc : utility_ctrl_audio_fifo_adc port map
	(
		fifo_adc_data,		
		fifo_adc_rdclk,	-- ucore clk			
		fifo_adc_rdreq,	
		fifo_adc_wrclk,	-- audio clk		
		fifo_adc_wrreq,		
		fifo_adc_q,			
		fifo_adc_rdempty,	
		fifo_adc_rdusedw,
		fifo_adc_wrfull
	);	
	
	error_adc_full			<= fifo_adc_wrfull;
	error_adc_no_data		<=	fifo_adc_rdempty;
	
	-----------------------------------------------------------------------------------------------------------
	-- irq dac (generate irq if fifo run low)
	-----------------------------------------------------------------------------------------------------------
	
	process(clk)
	begin
		if rising_edge(clk) then
			if reset = '1' then
				dac_irq <= '0';
			else
				if fifo_dac_wrusedw < dac_used_low_irq then
					dac_irq <= '1';
				else
					dac_irq <= '0';
				end if;
			end if;
		end if;
	end process;
	
	-----------------------------------------------------------------------------------------------------------
	-- irq adc (generate irq if fifo run high)
	-----------------------------------------------------------------------------------------------------------
	
	process(clk)
	begin
		if rising_edge(clk) then
			if reset = '1' then
				adc_irq <= '0';
			else
				if fifo_adc_rdusedw > adc_used_high_irq then
					adc_irq <= '1';
				else
					adc_irq <= '0';
				end if;
			end if;
		end if;
	end process;
	
	-----------------------------------------------------------------------------------------------------------
	-- serial out
	-----------------------------------------------------------------------------------------------------------
	
	process(clk_audio)	
	begin
		if rising_edge(clk_audio) then
			if reset_audio = '1' then
				ser_out_state 			<= 0;
				
				iAUD_DACDAT				<= '0';
				
				iAUD_DACLRCK 			<= '0';
				iAUD_ADCLRCK 			<= '0';
				
				fifo_dac_rdreq			<= '0';
				fifo_adc_wrreq 		<= '0';
				
				dac_enable				<= '0';
				
				dac_counter				<= (others => '0');
				dac_counter_max		<= (others => '0');
				
				adc_counter				<= (others => '0');
				adc_counter_max		<= (others => '0');
				
				adc_enable				<= '0';
			else
			
				-- defaults
				
				fifo_dac_rdreq		<= '0';
				fifo_adc_wrreq 	<= '0';
			
				-------------------------------------------------------------------------
				-- n bit left justified
				-------------------------------------------------------------------------
				
				-- DAC data
				
				if dac_enable = '1' then
					case ser_out_state is
						when 0 to (dac_bit_width-1) 												=>	iAUD_DACDAT	<= dac_left((dac_bit_width-1) - ser_out_state);
						when (frame_length/2) to ((frame_length/2)+(dac_bit_width-1))	=> iAUD_DACDAT 	<= dac_right(((frame_length/2)+(dac_bit_width-1)) - ser_out_state);						
						when others =>
					end case;
				end if;
				
				-- ADC data
				
				if adc_enable = '1' then
					case ser_out_state is
						when 1 to (dac_bit_width) 													=> fifo_adc_data((dac_bit_width) - ser_out_state) <= iAUD_ADCDAT;							
						when (frame_length/2)+1 to ((frame_length/2)+(dac_bit_width))	=> fifo_adc_data(((frame_length/2)+(dac_bit_width)) - ser_out_state) 	<= iAUD_ADCDAT;							
						when others =>
					end case;
				end if;
				
				-- ADC / DAC LRCK
				
				if ser_out_state < (frame_length/2) then
					iAUD_DACLRCK 	<= '1';
					iAUD_ADCLRCK	<= '1';
				else
					iAUD_DACLRCK 	<= '0';
					iAUD_ADCLRCK	<= '0';
				end if;
				
				-- DAC / ADC fifo check / read
				
				if ser_out_state = (frame_length/2) + dac_bit_width then
					
					-- ADC
					
					if adc_counter = adc_counter_max then	-- divider
					
						adc_counter <= (others => '0');
					
						if adc_enable = '1' then
							if fifo_adc_wrfull = '0' then
								fifo_adc_wrreq <= '1';
							else
								-- adc fifo full
							end if;
						end if;
					else
						adc_counter <= adc_counter + 1;
					end if;
						
					-- DAC
					
					
					if dac_counter = dac_counter_max then	-- divider
					
						dac_counter <= (others => '0');
						
						if fifo_dac_rdempty = '0' then
							fifo_dac_rdreq <= '1';	-- send ack
							
							dac_left  <= fifo_dac_q(63 downto 40);
							dac_right <= fifo_dac_q(39 downto 16);
							
							case fifo_dac_q (15 downto 8) is
								when x"01" => dac_enable 		<= fifo_dac_q(0);
								when x"02" => dac_counter_max <= fifo_dac_q(3 downto 0);
								when x"03" => adc_enable 		<= fifo_dac_q(0);
								when x"04" => adc_counter_max <= fifo_dac_q(3 downto 0);
								when others =>
							end case;
						else
							-- no data to send
						end if;
					else
						dac_counter <= dac_counter + 1;
					end if;
					
				end if;
				
				-- state counter
				
				ser_out_state <= ser_out_state + 1;
				if ser_out_state = (frame_length-1) then
					ser_out_state <= 0;
				end if;
			
			end if;
		end if;
	end process;
	
end rtl;
