library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity utility_ctrl_lcd is
		port 
		(
				-- 
				
				clk				: in std_logic;
				reset				: in std_logic;
				
				--
				
				lcd_data_write  : in std_logic := '0';
				lcd_data_read	 : in std_logic := '0';
				lcd_data_rs     : in std_logic;
				lcd_data_data   : in std_logic_vector(7 downto 0);
				lcd_data_q	    : out std_logic_vector(7 downto 0);
				lcd_wait			 : out std_logic;
				
				-- LCD 
				
				LCD_BLON			: out std_logic;
				LCD_DATA			: inout std_logic_vector(7 downto 0);
				LCD_EN			: out std_logic;
				LCD_ON			: out std_logic;
				LCD_RS			: out std_logic;
				LCD_RW			: out std_logic
				
				
		
		);
end entity;

architecture rtl of utility_ctrl_lcd is

	type lcd_state_t is 
	( 
		SLCD_WAIT,
		SLCD_WRITE,
		SLCD_READ
	);
	
	signal lcd_state : lcd_state_t := SLCD_WAIT;


begin
	
	-- lcd blon is not connected on my de2-115 :-(
	
	LCD_BLON <= '1';
	LCD_ON 	<= '1';
	
	-- lcd
	
	lcd_wait <= '1' when lcd_state = SLCD_WAIT else '0';
	
	process ( clk )
		variable delay : integer range 0 to 127 := 0;
		
		constant	SETUP_TAS 		:  integer :=	5;
		constant	SETUP_PWEH 		:  integer :=	29;
		constant	SETUP_TCYR 		:  integer :=	34;
		constant	SETUP_CAPTURE	:  integer :=	1;
		
	begin
		if rising_edge(clk) then	
			if reset = '1' then
				LCD_DATA			<= (others => 'Z');
				LCD_EN			<= '0';
				LCD_RS			<= '0';
				LCD_RW			<= '0';
				
				lcd_state		<= SLCD_WAIT;
				lcd_data_q		<= (others => '0');

			else
			
				-- default
				
				LCD_DATA			<= (others => 'Z');
				LCD_EN			<= '0';
				LCD_RS			<= '0';
				LCD_RW			<= '0';
				
				--
			
				case lcd_state is
					when SLCD_WAIT	=>		if lcd_data_write = '1' then								
													lcd_state <= SLCD_WRITE;													
												elsif lcd_data_read = '1' then
													lcd_state <= SLCD_READ;	
												end if;	
												
												delay	:= 0;
												
					when SLCD_WRITE =>	
					
--	TCYR					     270		ns	(34)					
--	PWEH			    230				ns (29)	
--	TAS			40						ns	(5)
--				   |-|-----|------|
--	RS				XXX****************XXX
--	RW				XXX________________XXX
--	E				______------__________
--	DB0..6		XXXXXXXX*******XXXXXXX		
			
												if delay <= SETUP_TAS then
																				LCD_RS		<= lcd_data_rs;
																				LCD_RW		<= '0';	
																				LCD_EN		<= '0';
																				LCD_DATA		<= (others => 'Z');
																				
												elsif delay <= SETUP_PWEH + SETUP_TAS then
																				
																				LCD_RS		<= lcd_data_rs;
																				LCD_RW		<= '0';	
																				LCD_EN		<= '1';
																				LCD_DATA		<= lcd_data_data;
													
												elsif delay <= SETUP_PWEH + SETUP_TAS + SETUP_TCYR then
													
																				LCD_RS		<= lcd_data_rs;
																				LCD_RW		<= '0';	
																				LCD_EN		<= '0';
																				LCD_DATA		<= lcd_data_data;
																				
												else 							
																				lcd_state 	<= SLCD_WAIT;
												end if;											
--												
																																			
												delay	:= delay	+ 1;		
												
					when SLCD_READ	=>
					
--	TCYR					     270		ns	(34)	
-- CAPTURE				   8			ns (1)				
--	PWEH			    230				ns (29)	
--	TAS			40						ns	(5)
--				   |-|-----|------|
--	RS				XXX****************XXX
--	RW				XXX________________XXX
--	E				______------__________
--	DB0..6		XXXXXXXX*******XXXXXXX						
					
													if delay <= SETUP_TAS then
													
																				LCD_RS		<= lcd_data_rs;
																				LCD_RW		<= '1';	
																				LCD_EN		<= '0';
																				LCD_DATA		<= (others => 'Z');
																				
													elsif delay <= SETUP_PWEH + SETUP_TAS then
																				
																				LCD_RS		<= lcd_data_rs;
																				LCD_RW		<= '1';	
																				LCD_EN		<= '1';
																				LCD_DATA		<= (others => 'Z');
																				
													elsif delay <= SETUP_PWEH + SETUP_TAS + SETUP_CAPTURE then
													
																				LCD_RS		<= lcd_data_rs;
																				LCD_RW		<= '1';	
																				LCD_EN		<= '1';
																				lcd_data_q	<= LCD_DATA;
													
													
													elsif delay <= SETUP_PWEH + SETUP_TAS + SETUP_CAPTURE + SETUP_TCYR then
													
																				LCD_RS		<= lcd_data_rs;
																				LCD_RW		<= '1';	
																				LCD_EN		<= '0';
																				LCD_DATA		<= (others => 'Z');
																				
													else 
																				lcd_state 	<= SLCD_WAIT;
												end if;
												
												delay	:= delay	+ 1;
				end case;
			end if;
		end if;
	end process;	

end rtl;
