library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity utility_ctrl_timer is
		port 
		(
				-- 
				
				clk				: in std_logic;
				reset				: in std_logic;
				
				--
				
				timer_load_value 		: in std_logic_vector(31 downto 0) := (others => '0');
				timer_start 			: in std_logic := '0';
				timer_stop  			: in std_logic := '0';
				timer_auto_reload  	: in std_logic := '0';
				timer_hit  				: out std_logic := '0';
				
				timer_current_value	: out std_logic_vector(31 downto 0)
				
				
		
		);
end entity;

architecture rtl of utility_ctrl_timer is

	-- timer
	
	signal timer_value : std_logic_vector(31 downto 0) := (others => '0');
	
	type timer_state_t is 
	( 
		STIMER_WAIT,
		STIMER_RUN
	);
	
	signal timer_state : timer_state_t := STIMER_WAIT;
	
begin
	
	timer_current_value  <= timer_value;
	
	-- timer
	
	process ( clk )
	begin
		if rising_edge(clk) then	
			if reset = '1' then
				timer_state <= STIMER_WAIT;
				timer_hit <= '0';				
			else
			
				timer_hit <= '0';
			
				case timer_state is 
					when STIMER_WAIT 		=>	if timer_start = '1' then	
														timer_state <= STIMER_RUN;
														timer_value <= timer_load_value;
													else
														timer_state <= STIMER_WAIT;
													end if;				 																									
					
					when STIMER_RUN		=>	timer_value <= timer_value - 1;
													timer_state <= STIMER_RUN;
																		
													if timer_stop = '1' then	
														timer_state <= STIMER_WAIT;
													else					
														if timer_value = x"00000000" then																									
															timer_hit <= '1';
															
															if timer_auto_reload <= '0' then 
																timer_state <= STIMER_WAIT;
															else
																timer_value <= timer_load_value;
															end if;
														end if;
													end if;
				end case;
			end if;	
		end if;
	end process;	

	

end rtl;
