library ieee;
use work.bus_package.all;
use work.video_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity utility_ctrl_vga is
		port 
		(
				-- 
				
				clk				: in std_logic;
				reset				: in std_logic;
							
				-- VGA 
				
				VGA_B				: out std_logic_vector(7 downto 0);
				VGA_BLANK_N		: out std_logic;
				VGA_CLK			: out std_logic;
				VGA_G				: out std_logic_vector(7 downto 0);
				VGA_HS			: out std_logic;
				VGA_R				: out std_logic_vector(7 downto 0);
				VGA_SYNC_N		: out std_logic;
				VGA_VS			: out std_logic;
				
				-- 
				
				HSYNC				: in integer range 0 to 4095;
				HSTART			: in integer range 0 to 4095;
				HMEMSTART		: in integer range 0 to 4095;
				HSTOP				: in integer range 0 to 4095;
				HTOTAL			: in integer range 0 to 4095;

				HLOADSTART		: in integer range 0 to 4095;
				HLOADSTOP		: in integer range 0 to 4095;

				VSYNC				: in integer range 0 to 2047;
				VSTART			: in integer range 0 to 2047;
				VSTOP				: in integer range 0 to 2047;
				VTOTAL			: in integer range 0 to 2047;

				VLOADSTART		: in integer range 0 to 4095;
				VLOADSTOP		: in integer range 0 to 4095;
				VLOADNEXT		: in integer range 0 to 4095;

				MEM_STARTADR	: in std_logic_vector(31 downto 0);
				MEM_LINEOFFSET	: in std_logic_vector(31 downto 0);	
				MEM_MODE			: in std_logic_vector(15 downto 0);

				video_on			: in std_logic;
				lc_add			: in std_logic_vector(31 DOWNTO 0);
				lc_start			: in std_logic_vector(10 DOWNTO 0);
				
				video_hcount	: out integer range 0 to 4095;
				video_vcount	: out integer range 0 to 2047;
				
				PVCOUNT			: in integer range 0 to 2047;
				PVCOUNT_WE		: in std_logic;				
				PVCOUNT_HIT		: out std_logic;
				
				-- external
				
				-- send
				
				ext_send_address	: out std_logic_vector(31 downto 0);
				ext_send_busy		: in std_logic;
				ext_send_wrreq		: out std_logic;
				
				-- recive
				
				ext_recive_data	: in std_logic_vector(31 downto 0);
				ext_recive_busy	: in std_logic;
				ext_recive_ack		: out std_logic						
		);
end entity;



architecture rtl of utility_ctrl_vga is

	component oddr is
		port
		(
				datain_h		: IN STD_LOGIC_VECTOR (0 DOWNTO 0);
				datain_l		: IN STD_LOGIC_VECTOR (0 DOWNTO 0);
				outclock		: IN STD_LOGIC ;
				dataout		: OUT STD_LOGIC_VECTOR (0 DOWNTO 0)
		);
end component;

signal video_clk_dataout	: STD_LOGIC_VECTOR (0 DOWNTO 0) := (others => '0');

-- line cache

component utility_ctrl_vga_rgb_line_cache is
		port 
		(
				-- 
				
				clk				: in std_logic;
				reset				: in std_logic;
				
	
				-- read 
							
				rdaddress		: in std_logic_vector (10 downto 0);
				q					: out std_logic_vector (23 downto 0);
				
				-- push
				
				setwaddress		: in std_logic;
				wdaddress		: in std_logic_vector (10 downto 0);
				
				pushin			: in std_logic;
				pushdata			: in std_logic_vector (31 downto 0);
				pushmode			: in std_logic_vector(1 downto 0);
				pushready		: out std_logic
		
		);
end component;


	signal lc_rdaddress		: std_logic_vector (10 DOWNTO 0) := (others => '0');
	signal lc_q					: std_logic_vector (23 DOWNTO 0) := (others => '0');		
	
	signal lc_setwaddress	: std_logic := '0';
	signal lc_wdaddress		: std_logic_vector (10 downto 0) := (others => '0');			
	signal lc_pushin			: std_logic := '0';
	signal lc_pushdata		: std_logic_vector (31 downto 0) := (others => '0');
	signal lc_pushmode		: std_logic_vector(1 downto 0) 	:= (others => '0');
	signal lc_pushready		: std_logic := '0';
	
	--
	
	signal hcount		: integer range 0 to 4095 := 0;
	signal vcount		: integer range 0 to 2047 := 0;
	
	--

	signal mem_rpos		: std_logic_vector(34 DOWNTO 0) := (others => '0');
	signal rgb 				: std_logic_vector(23 downto 0) := (others => '0');
	
	--
	
	type loadStage_t is 
	( 
		WAIT_VLOADSTART,
		WAIT_HLOADSTART,
		WAIT_HLOADSTOP,
		PROCESS_LINE,
		WAIT_NEXTLINE
	);
	
	signal loadStage : loadStage_t := WAIT_VLOADSTART;
	
	signal transferBlankStart : std_logic := '0';
	signal transferLineStart  : std_logic := '0';
	
	type transferStage_t is
	(
		WAIT_TRANSFER_START,
		PUSH_DATA_OUT
	);
	
	signal transferStage : transferStage_t := WAIT_TRANSFER_START;	
			
	signal internal_PVCOUNT_HIT :  std_logic := '0';
	
-- buffers to out
	
begin

	-- out

	video_hcount <= hcount;
	video_vcount <= vcount;
	PVCOUNT_HIT  <= internal_PVCOUNT_HIT;
	
	-----------------------------------------------------------------------------------------------------------
	-- external memory wait / start
	-----------------------------------------------------------------------------------------------------------
	
	process ( clk )	
		variable next_vpos : integer range 0 to 4095  := 0;
	begin
		if rising_edge(clk) then
			if reset = '1' then
				loadStage 			<= WAIT_VLOADSTART;
				
				transferBlankStart	<= '0';
				transferLineStart		<= '0';
			else
			
				-- default
				
				transferBlankStart	<= '0';
				transferLineStart		<= '0';
			
				--
			
				case loadStage is
					when WAIT_VLOADSTART	=>	if vcount = VLOADSTART then
														transferBlankStart <= '1'; -- reset emem adr
														loadStage <= WAIT_HLOADSTART;
													else
														loadStage <= WAIT_VLOADSTART;
													end if;
													
													next_vpos := vcount;
													
					when WAIT_HLOADSTART =>	if hcount = HLOADSTART then
														loadStage <= PROCESS_LINE;
														next_vpos := next_vpos + VLOADNEXT;
													else
														loadStage <= WAIT_HLOADSTART;
													end if;
					
					when PROCESS_LINE		=>	-- start transfer
													transferLineStart	<= '1';
													loadStage <= WAIT_HLOADSTOP;
					
					when WAIT_HLOADSTOP	=>	
													if hcount = HLOADSTOP then
														loadStage <= WAIT_NEXTLINE;
													else
														loadStage <= WAIT_HLOADSTOP;
													end if;
													
					when WAIT_NEXTLINE	=>	if vcount >= VLOADSTOP then
														loadStage <= WAIT_VLOADSTART;
													else
														if vcount = next_vpos then
															loadStage <= WAIT_HLOADSTART;
														else
															loadStage <= WAIT_NEXTLINE;
														end if;
													end if;
					
				end case;
			end if;
		end if;
	end process;
	
	-----------------------------------------------------------------------------------------------------------
	-- external memory address generate and push out
	-----------------------------------------------------------------------------------------------------------
	
	process ( clk )	
		variable current_memadr		: std_logic_vector(31 downto 0) := (others => '0');
		variable load_memadr			: std_logic_vector(31 downto 0) := (others => '0');
		variable current_datacount	: integer range 0 to 4095 := 0;
	begin
		if rising_edge(clk) then
			if reset = '1' then	
				transferStage 	<= WAIT_TRANSFER_START;
				current_memadr	:= (others => '0');
				ext_send_wrreq <= '0';
			else
			
				-- default
				
				load_memadr := MEM_STARTADR + current_memadr;
				
				ext_send_wrreq		<= '0';
				ext_send_address 	<= load_memadr;	-- load address (external memory)
																		
				--
			
				case transferStage is
					when WAIT_TRANSFER_START	=>	if transferBlankStart = '1' then
																current_memadr := (others => '0');	-- reset at blank
															end if;
					
															current_datacount := HLOADSTART_default;	-- update every line
															
															if transferLineStart = '1' then																
																transferStage <= PUSH_DATA_OUT;
															else
																transferStage <= WAIT_TRANSFER_START;
															end if;
															
					when PUSH_DATA_OUT			=>	if current_datacount = HLOADSTOP then -- FIXME
																
																-- next line
																transferStage <= WAIT_TRANSFER_START;
																current_memadr := current_memadr + MEM_LINEOFFSET;
															else
																-- into fifo if not full
																
																if ext_send_busy = '0' then -- send full (this will corrupt video, but should never occure, video have ever memory prio)
																
																	ext_send_wrreq	<= '1';
																
																	current_memadr := current_memadr + 2;
																	current_datacount := current_datacount + 1;
																end if;
															
																
																transferStage <= PUSH_DATA_OUT;
															end if;					
				end case;
			end if;
		end if;
	end process;
	
	-----------------------------------------------------------------------------------------------------------
	-- external memory data to lc cache
	-----------------------------------------------------------------------------------------------------------
	
	process ( clk )	
	begin
		if rising_edge(clk) then
			if reset = '1' then	
				lc_wdaddress <= (others => '0');
				lc_pushin	 <= '0';
			else
				
				-- default
				
				lc_pushin		<= '0';
				lc_pushdata		<= ext_recive_data;
				
				lc_pushmode 	<= MEM_MODE(1 downto 0);
				lc_setwaddress	<= '0';
				lc_wdaddress 	<= (others => '0');

				--
																	
				if transferLineStart = '1' then	
					lc_wdaddress 	<= (others => '0'); -- is set t0 zero as default not programmable yet (who need)?
					lc_setwaddress	<= '1';
				end if;
				
				if ext_recive_busy = '0' and lc_pushready = '1' then -- data available										
					lc_pushin		<= '1';
				end if;
																
			end if;
		end if;
	end process;
	
	process(ext_recive_busy,reset,lc_pushready)
	begin
		ext_recive_ack	<= '0';	
		
		if ext_recive_busy = '0' and lc_pushready = '1' and  reset = '0' then -- data available
				ext_recive_ack	<= '1';	-- ack
		end if;	
	end process;
			
	-----------------------------------------------------------------------------------------------------------		
	-- line cache
	-----------------------------------------------------------------------------------------------------------
	
	line_cache : utility_ctrl_vga_rgb_line_cache port map 
	(	
		clk,
		reset,
		lc_rdaddress,
		lc_q,						
		lc_setwaddress,
		lc_wdaddress,		
		lc_pushin,
		lc_pushdata,
		lc_pushmode,
		lc_pushready
	);
	
	
	 	
	lc_rdaddress 	<= mem_rpos(34 downto 24);

	-----------------------------------------------------------------------------------------------------------
	-- line cache processing
	-----------------------------------------------------------------------------------------------------------
	
	process ( clk )	
		variable rpos		: std_logic_vector(34 DOWNTO 0) := (others => '0');
	begin
		if rising_edge(clk) then
		
			if hcount < HMEMSTART or  hcount > HSTOP then
				rpos := lc_start & x"000000";
			else
				rpos := rpos + ("00" & lc_add); 
			end if;
			
			
			mem_rpos <= rpos;			
		end if;
	end process;
	
	-----------------------------------------------------------------------------------------------------------
	-- programmable vcount 
	-----------------------------------------------------------------------------------------------------------
	
	process ( clk, internal_PVCOUNT_HIT, PVCOUNT, PVCOUNT_WE, vcount )	
	begin
		if rising_edge(clk) then
		
			-- we
			
			if PVCOUNT_WE = '1' then
				internal_PVCOUNT_HIT <= '0';
			end if;
		
			-- trap
		
			if internal_PVCOUNT_HIT = '0' then
				if PVCOUNT = vcount then
					internal_PVCOUNT_HIT <= '1';
				end if;
			end if;
		end if;
	end process;
	
	-----------------------------------------------------------------------------------------------------------
	-- blank / vh / vs
	-----------------------------------------------------------------------------------------------------------
		
	process (clk)	
	begin
		if rising_edge(clk) then
		
			-- default 
			
			
			if reset = '1' then
				VGA_BLANK_N <= '0';
				VGA_VS 		<= '0'; 
				VGA_HS 		<= '0';
			else
				if hcount >= HSTART and hcount < HSTOP and
					vcount >= VSTART and vcount < VSTOP and video_on = '1' then
				
					VGA_BLANK_N <= '1';
				else
					VGA_BLANK_N <= '0';
				end if;
				
				if hcount >= HSYNC then
					VGA_HS <= '1';
				else
					VGA_HS <= '0';
				end if;
				
				if vcount >= VSYNC then
					VGA_VS <= '1';
				else
					VGA_VS <= '0';
				end if;
			end if;
			
		end if;
	end process;
	
	
	-----------------------------------------------------------------------------------------------------------
	-- clk / sync
	-----------------------------------------------------------------------------------------------------------
	
--	ext_video_clk : oddr port map
--	(
--			"1",
--			"0",
--			clk,
--			video_clk_dataout
--	);
	
	VGA_CLK		<= clk;--video_clk_dataout(0);	-- use system clk (106MHz)
	VGA_SYNC_N 	<= '1';									-- no sync through green
	
	-----------------------------------------------------------------------------------------------------------
	-- rgb
	-----------------------------------------------------------------------------------------------------------
	
	process ( clk )	
	begin
		if rising_edge(clk) then		
			if reset = '1' then
				VGA_R <= (others => '0');
				VGA_G <= (others => '0');
				VGA_B <= (others => '0');
			else
				VGA_R		   <= lc_q(23 downto 16);
				VGA_G	 		<= lc_q(15 downto 8);
				VGA_B	 	   <= lc_q(7 downto 0);
			end if;
		end if;
	end process;
	
	-----------------------------------------------------------------------------------------------------------
	-- h/vcount
	-----------------------------------------------------------------------------------------------------------
	
	process ( clk )
	begin
		if rising_edge(clk) then
			if reset = '1' then
				hcount <= 0;
				vcount <= 0;
			else
				if hcount < HTOTAL then
					hcount <= hcount + 1;
				else
					
					if vcount < VTOTAL then						
						vcount <= vcount + 1;
					else
						vcount <= 0;
					end if;
					
					hcount <= 0;					
				end if;
			end if;
		end if;
	end process;


	
end rtl;
