library ieee;
use work.ucore_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity ld2d is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;
				stall_in			: in std_logic;
				
				-- param 
				
				rsel				: in integer range 0 to 15;
				wsel				: in integer range 0 to 15;
				indata			: in std_logic_vector(15 downto 0);	
				inwe				: in std_logic;
				outdata			: out std_logic_vector(15 downto 0);	
									
				-- activate
				
				en 				: in std_logic;
				
				--
								
				dadr				: out std_logic_vector(31 downto 0)
																		
		);
end entity;



architecture rtl of ld2d is

signal iu			: std_logic_vector(31 downto 0) := (others => '0');
signal idu			: std_logic_vector(31 downto 0) := (others => '0');	
signal iv			: std_logic_vector(31 downto 0) := (others => '0');
signal idv			: std_logic_vector(31 downto 0) := (others => '0');
signal ixmask		: std_logic_vector(15 downto 0) := (others => '0');
signal iymask		: std_logic_vector(15 downto 0) := (others => '0');
signal ixwidth		: std_logic_vector(15 downto 0) := (others => '0');
signal isadr		: std_logic_vector(31 downto 0) := (others => '0');
	
signal s1_x			: std_logic_vector(15 downto 0)  := (others => '0');
signal s1_y			: std_logic_vector(15 downto 0) := (others => '0');
	
signal s2_x			: std_logic_vector(15 downto 0)  := (others => '0');
signal s2_y			: std_logic_vector(31 downto 0) := (others => '0');	
	
begin
	
	-- param get (not clocked (use in clk domain at read))
	
	process(rsel,iu,iv)	
	begin
		case rsel is
			when eld_u_l 			=>		outdata <= iu(15 downto 0);
			when eld_u_h			=>		outdata <= iu(31 downto 16);							
			when eld_v_l			=>		outdata <= iv(15 downto 0);					
			when eld_v_h			=>		outdata <= iv(31 downto 16);					
			when others			=>		outdata <= (others => '0');
		end case;
	end process;
	
	-- s1

	process(clk,stall_in,reset,en)
		variable x_tmp : std_logic_vector(15 downto 0) := (others => '0');
		variable y_tmp : std_logic_vector(15 downto 0) := (others => '0');
	begin 
		if rising_edge(clk) then
			if reset = '1' then				
			else		
				if stall_in = '0' then
			
					-- param set
					
					if inwe = '1' then
						case wsel is
							when eld_u_l 		=>		iu(15 downto 0) <= indata;
							when eld_u_h 		=>		iu(31 downto 16) <= indata;
							when eld_du_l		=>		idu(15 downto 0) <= indata;		
							when eld_du_h		=>		idu(31 downto 16) <= indata;							
							when eld_v_l		=>		iv(15 downto 0) <= indata;		
							when eld_v_h		=>		iv(31 downto 16) <= indata;				
							when eld_dv_l		=>		idv(15 downto 0) <= indata;	
							when eld_dv_h		=>		idv(31 downto 16) <= indata;
							when eld_xmask		=>		ixmask <= indata;
							when eld_ymask		=>		iymask <= indata;
							when eld_xwidth	=>		ixwidth <= indata;			 
							when eld_sadr_l	=>		isadr(15 downto 0) <= indata;	
							when eld_sadr_h	=>		isadr(31 downto 16) <= indata;
							when others			=>		
						end case;													
					end if;
				
					--
					
					if en = '1' then	
						x_tmp  := iu(31 downto 16) and ixmask;
						y_tmp  := iv(31 downto 16) and iymask;
		
						--
						
						iu <= iu + idu;
						iv <= iv + idv;
						
		
						s1_x  <= x_tmp;
						s1_y  <= y_tmp;
						
					end if;								
				end if;
				
			end if;
		end if;
	end process;
	
	-- s2
	
	process(clk,stall_in,en)
	begin 
		if rising_edge(clk) then
			if reset = '1' then				
			else
				if stall_in = '0' and en = '1' then
					s2_x <= s1_x;
					s2_y <= s1_y * ixwidth;
				end if;
			end if;
		end if;
	end process;
	
	-- s3
	
	process(clk,stall_in,en)
	begin 
		if rising_edge(clk) then
			if reset = '1' then				
			else
				if stall_in = '0' and en = '1'then
					dadr 			<= s2_x + s2_y + isadr;
				end if;
			end if;
		end if;
	end process;
	
end rtl;
