library ieee;
use work.ucore_package.all;
use work.bus_package.all;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity ucore is
		port 
		(
				-- global
			
				clk			: in std_logic;
				reset			: in std_logic;
				
				--
				
				irq			: in std_logic;
								
				-- ctrl port
				
				ctrl_port_a_adress   : out std_logic_vector(6 downto 0);
				ctrl_port_a_q		   : in std_logic_vector(15 downto 0);
				ctrl_port_a_data 		: out std_logic_vector(15 downto 0);
				ctrl_port_a_we			: out std_logic;
				
				-- external
				
				-- send
				
				ext_send_data		: out std_logic_vector(15 downto 0);
				ext_send_address	: out std_logic_vector(31 downto 0);
				ext_send_we			: out std_logic;
				ext_send_busy		: in std_logic;
				ext_send_wrreq		: out std_logic;
				
				-- recive
				
				ext_recive_data	: in std_logic_vector(15 downto 0);
				ext_recive_busy	: in std_logic;
				ext_recive_ack		: out std_logic;
	
				-- diag
				
				stall_out		: out std_logic
															
		);
end entity;



architecture rtl of ucore is

-- internal memory

component ucore_icmem
	PORT
	(
		address_a		: IN STD_LOGIC_VECTOR (13 DOWNTO 0);
		address_b		: IN STD_LOGIC_VECTOR (13 DOWNTO 0);
		clock		: IN STD_LOGIC  := '1';
		data_a		: IN STD_LOGIC_VECTOR (15 DOWNTO 0);
		data_b		: IN STD_LOGIC_VECTOR (15 DOWNTO 0);
		enable		: IN STD_LOGIC  := '1';
		rden_a		: IN STD_LOGIC  := '1';
		rden_b		: IN STD_LOGIC  := '1';
		wren_a		: IN STD_LOGIC  := '0';
		wren_b		: IN STD_LOGIC  := '0';
		q_a		: OUT STD_LOGIC_VECTOR (15 DOWNTO 0);
		q_b		: OUT STD_LOGIC_VECTOR (15 DOWNTO 0)
	);
end component;

-- decode
				
component ucore_decode is
		generic 
		(
			enable_udiv 	: std_logic := '1';
			enable_exl2d	: std_logic := '1'
		);
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;
				
				-- in
				
				stall				: in std_logic;
			
				data_in			: in std_logic_vector(15 downto 0);
				
				-- out
					
				decode_opcode	: out opcode_t;				  
				decode_rc		: out std_logic_vector(2 downto 0);	
				decode_rb		: out std_logic_vector(2 downto 0);	
				decode_ra		: out std_logic_vector(2 downto 0);	
				decode_imm		: out std_logic_vector(15 downto 0);	
				decode_rc_we	: out std_logic;
				decode_imm_a	: out std_logic
															
		);
end component;	
	
signal decode_opcode	: opcode_t := ope_nop;				  
signal decode_rc		: std_logic_vector(2 downto 0)  := (others => '0');	
signal decode_rb		: std_logic_vector(2 downto 0)  := (others => '0');		
signal decode_ra		: std_logic_vector(2 downto 0)  := (others => '0');		
signal decode_imm		: std_logic_vector(15 downto 0) := (others => '0');		
signal decode_rc_we	: std_logic := '0';
signal decode_imm_a	: std_logic := '0';

-- reg read/write

component ucore_regio is
		port 
		(
				-- global

				clk				: in std_logic;
				reset				: in std_logic;
				
				-- in
				
				stall_read		: in std_logic;
				stall_write		: in std_logic;
				
				decode_opcode	: in opcode_t;				  
				decode_rc		: in std_logic_vector(2 downto 0);	
				decode_rb		: in std_logic_vector(2 downto 0);	
				decode_ra		: in std_logic_vector(2 downto 0);	
				decode_imm		: in std_logic_vector(15 downto 0);	
				decode_rc_we	: in std_logic;
				decode_imm_a	: in std_logic;
						
				alu_rc_we		: in std_logic;	
				alu_rc			: in std_logic_vector(2 downto 0);	
				alu_data_c		: in std_logic_vector(15 downto 0);
				
				-- out
				
				reg_data_b		: out std_logic_vector(15 downto 0);	
				reg_data_a		: out std_logic_vector(15 downto 0);
				reg_rc			: out std_logic_vector(2 downto 0);	
				reg_rc_we		: out std_logic;
				reg_opcode		: out opcode_t;
				reg_simm			: out std_logic_vector(15 downto 0)
			
															
		);
end component;

signal reg_data_b		: std_logic_vector(15 downto 0)  := (others => '0');		
signal reg_data_a		: std_logic_vector(15 downto 0)  := (others => '0');
signal reg_rc			: std_logic_vector(2 downto 0)   := (others => '0');	
signal reg_rc_we		: std_logic := '0';
signal reg_opcode		: opcode_t := ope_nop;
signal reg_simm		: std_logic_vector(15 downto 0)  := (others => '0');
-- alu

component ucore_execute is
		port 
		(
				-- global

				clk				: in std_logic;
				reset		: in std_logic;
				
				--
				
				irq				: in std_logic;
				
				-- in
				
				reg_data_b		: in std_logic_vector(15 downto 0);	
				reg_data_a		: in std_logic_vector(15 downto 0);
				reg_rc			: in std_logic_vector(2 downto 0);	
				reg_rc_we		: in std_logic;
				reg_opcode		: in opcode_t;
				reg_simm			: in std_logic_vector(15 downto 0);
				
				-- out
				
				alu_pc 			: out std_logic_vector(15 DOWNTO 0);
				alu_rc_we		: out std_logic;	
				alu_rc			: out std_logic_vector(2 downto 0);	
				alu_data_c		: out std_logic_vector(15 downto 0);
				alu_stall		: out std_logic;
			
				-- memory
				
				mem_address		: out std_logic_vector(15 downto 0);
				mem_we			: out std_logic;
				mem_data			: out std_logic_vector(15 downto 0);
				mem_q				: in std_logic_vector(15 downto 0);
				
				-- external
				
				-- send
				
				ext_send_data		: out std_logic_vector(15 downto 0);
				ext_send_address	: out std_logic_vector(31 downto 0);
				ext_send_we			: out std_logic;
				ext_send_busy		: in std_logic;
				ext_send_wrreq		: out std_logic;
				
				-- recive
				
				ext_recive_data	: in std_logic_vector(15 downto 0);
				ext_recive_busy	: in std_logic;
				ext_recive_ack		: out std_logic

		);
end component;

signal alu_pc 					: std_logic_vector(15 DOWNTO 0) := (others => '0');
signal alu_rc_we				: std_logic := '0';	
signal alu_rc					: std_logic_vector(2 downto 0)  := (others => '0');		
signal alu_data_c				: std_logic_vector(15 downto 0)  := (others => '0');
signal alu_stall				: std_logic := '0';	
signal alu_mem_address		: std_logic_vector(15 downto 0)  := (others => '0');	
signal alu_mem_we				: std_logic := '0';
signal alu_mem_data			: std_logic_vector(15 downto 0)  := (others => '0');	
signal alu_mem_q				: std_logic_vector(15 downto 0)  := (others => '0');	

-- mem ports

signal code_port_a_adress  : std_logic_vector(13 downto 0) := (others => '0');
signal code_port_a_q		   : std_logic_vector(15 downto 0) := (others => '0');
signal code_port_a_data 	: std_logic_vector(15 downto 0) := (others => '0');
signal code_port_a_we		: std_logic := '0';
signal code_port_rden_a		: std_logic  := '1';
		
signal code_port_b_adress  : std_logic_vector(13 downto 0) := (others => '0');
signal code_port_b_q		   : std_logic_vector(15 downto 0) := (others => '0');
signal code_port_b_data 	: std_logic_vector(15 downto 0) := (others => '0');
signal code_port_b_we		: std_logic := '0';
signal code_port_rden_b		: std_logic  := '1';

signal code_port_enable		: std_logic  := '1';

begin

	stall_out <= alu_stall;
	
	-- decode
	
		decode : ucore_decode port map 
		(		
				-- global

				clk,
				reset,
				
				-- in
				
				alu_stall,			
				code_port_a_q,
				
				-- out
					
				decode_opcode,
				decode_rc,	
				decode_rb,	
				decode_ra,	
				decode_imm,		
				decode_rc_we,
				decode_imm_a
															
		);

		-- regio component
	
		regio : ucore_regio port map 
		(		
				clk,
				reset,
				
				-- in
				
				alu_stall,
				'0',	-- write is not stalled (because no stage behind)
				
				decode_opcode,				  
				decode_rc,
				decode_rb,
				decode_ra,
				decode_imm,
				decode_rc_we,
				decode_imm_a,
						
				alu_rc_we,
				alu_rc,
				alu_data_c,
				
				-- out
				
				reg_data_b,
				reg_data_a,
				reg_rc,
				reg_rc_we,
				reg_opcode,
				reg_simm
		);
		
	-- execute
		
		execute : ucore_execute port map 
		(
				-- global

				clk,
				reset,
				
				--
				
				irq,
				
				-- in
				
				reg_data_b,
				reg_data_a,	
				reg_rc,		
				reg_rc_we,	
				reg_opcode,	
				reg_simm,
				
				-- out
				
				alu_pc, 		
				alu_rc_we,	
				alu_rc,		
				alu_data_c,	
				alu_stall,	
			
				-- memory
				
				alu_mem_address,	
				alu_mem_we,		
				alu_mem_data,		
				alu_mem_q,
	
				-- external memory
				
				ext_send_data,
				ext_send_address,
				ext_send_we,
				ext_send_busy,
				ext_send_wrreq,		
				ext_recive_data,
				ext_recive_busy,
				ext_recive_ack
				
		);
		
		-- alu mem port a
		
		code_port_a_adress 	<= alu_pc(13 downto 0);	-- fetch / code memory
		code_port_a_data   	<= (others => '0');
		code_port_a_we			<= '0';
		
		-- alu mem port b
		
		code_port_b_adress 	<= alu_mem_address(13 downto 0);
		code_port_b_data	 	<= alu_mem_data;
		
		ctrl_port_a_adress	<= alu_mem_address(6 downto 0);
		ctrl_port_a_data 		<= alu_mem_data;
		
		code_port_rden_a		<= '1'; --not alu_stall;	-- fetch port can stall
		code_port_rden_b		<= '1';
		
		alu_mem_q			 	<= code_port_b_q when alu_mem_address(15) = '0' else
									   ctrl_port_a_q;
									 
		code_port_b_we		 <= '1' when alu_mem_we = '1' and alu_mem_address(15) = '0' else '0';
		ctrl_port_a_we		 <= '1' when alu_mem_we = '1' and alu_mem_address(15) = '1' else '0';
		
	-- data mem / code mem
	
	--
	-- 111111
	-- 5432109876543210
	--    iiiiiiiiiiiii
	--
	
   -- address
	-- 0000  --------------------
	-- 		- cmem     pa (r)  - <-> code_port_a (adress/q)  		<-> execute code
	-- 		- 16K*16  pb (rw)  - <-> code_port_b (address/data/q)	<-> execute data
	-- 		-                  -
	-- 		-                  -
	-- 3fff  --------------------
	-- 4000  --------------------
	-- 		- cmem mirror      - 
	-- 		-                  -
	-- 		-                  -
	-- 7fff  --------------------
	-- 8000 --------------------
	--       - devices  pa (rw) - <->  ctrl_port_a_we (address/data/q) <-> execute data
	-- 		- 		    			 -
	-- 		-                  -
	-- 		-                  -
	-- 0...  --------------------
	
	code_port_enable <= not alu_stall;
	
	icode_mem : ucore_icmem port map 
	(
		code_port_a_adress,
		code_port_b_adress,
		clk,
		code_port_a_data,
		code_port_b_data,
		code_port_enable,
		code_port_rden_a,
		code_port_rden_b,
		code_port_a_we,
		code_port_b_we,
		code_port_a_q,
		code_port_b_q
	);
	
end rtl;
